# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'blockbridge/api/internal-client'
require 'blockbridge/api/resources'

module Blockbridge
  module Api
    class Client
      def self.resource(*names)
        names.each do |name|
          define_method name.to_sym do |*args, &block|
            unless @client
              raise Blockbridge::Api::Error.new("client not connected")
            end

            resource = @cache[name]
            if resource.nil?
              resource_klass = Blockbridge::Api::Resources.for_type(name)
              resource = @cache[name] = resource_klass.new(@client, @opts)
            end

            resource
          end
        end
      end

      def self.defaults
        # pass through to Excon -- for tweaking SSL params
        Excon.defaults
      end

      def self.new_basic(password = nil, opts = nil, client_opts = {})
        Client.new(client_opts).tap do |c|
          c.connect_basic(password, opts)
        end
      end

      def self.new_oauth(token = nil, opts = nil, client_opts = {})
        Client.new(client_opts).tap do |c|
          c.connect_oauth(token, opts)
        end
      end

      # instance

      attr_reader :opts

      def initialize(opts = {})
        @opts = opts
        reset
      end

      def async?
        @opts[:async]
      end

      def reset
        @client = nil
        @cache = {}
      end

      def connect_basic(password = nil, options = nil)
        reset
        @client = BlockbridgeInternalClient.connect(password, options)
        nil
      end

      def connect_oauth(token = nil, options = nil)
        reset
        @client = BlockbridgeInternalClient.connect_oauth(token, options)
        nil
      end

      def obj_read(oid)
        o = Blockbridge::Api::ObjectId.new(oid)
        self.send(o.type.to_sym).info(oid)
      end

      #
      # define resources
      #

      # api status
      resource :status

      # authentication, et al.
      resource :oauth2_token
      resource :authorization

      # ca cert
      resource :ca_certificate

      # generic
      resource :xmd
      resource :objects

      # storage
      resource :vss
      resource :vdisk
      resource :initiator_profile
      resource :target
      resource :rule
      resource :snapshot
      resource :product
      resource :obj_store

      # infrastructure
      resource :site
      resource :node
      resource :sys_ds
      resource :iscsi_dev
      resource :os_dev
      resource :dev
      resource :net

      # account
      resource :account
      resource :account_usage
      resource :user_profile
    end
  end
end
