# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

module Blockbridge
  module Api
    # base BB error type
    class Error < StandardError; end

    class PathNotFound < Error; end
    class NotFound < Error; end
    class ResourceNotFound < Error; end
    class Locked < Error; end
    class RequestFailed < Error; end
    class Timeout < Error; end
    class InvalidObjectId < Error; end

    # bad requests (HTTP 4XX)
    class ClientError < Error
      attr_writer :default_message
      attr_reader :status
      attr_reader :data
      attr_writer :errors
      attr_reader :type

      def initialize(message = nil, data = nil)
        @default_message ||= "Bad request"

        @data     = data || {}
        @message  = @data[:message] || message
        @status   = @data[:status] || 400
        @type     = (@data[:type] || 'unspecified').to_sym

        if @data.has_key? :errors
          @errors = @data[:errors]
        elsif @data.has_key? :error
          @errors = [ @data[:error] ]
        else
          @errors = []
        end

        @errors.map! do |err|
          next err if !err.is_a? String

          {
            type:    err,
            message: err,
          }
        end

        unless @errors.empty?
          @message.concat ": #{error}"
        end
      end

      def errors
        @errors
      end

      def error
        errors.map { |err| err[:message] }.join(', ')
      end

      def to_s
        @message || @default_message
      end
    end

    class UnauthorizedError < ClientError
      def initialize(message = nil, data = nil)
        @default_message = 'Unauthorized'
        super(message, data)
      end

      def otp_required?
        @data[:otp_required]
      end
    end

    class ForbiddenError < ClientError
      def initialize(message = nil, data = nil)
        @default_message = 'Forbidden'
        super(message, data)
      end
    end

    class NotFoundError < ClientError
      def initialize(message = nil, data = nil)
        @default_message = 'Not Found'
        super(message, data)
      end
    end

    class ValidationError < ClientError
      def initialize(message = nil, data = nil)
        @default_message = 'Parameter validation failed'
        super(message, data)
      end
    end

    class ExecutionError < ClientError
      def initialize(message = nil, data = nil)
        @default_message = 'Task execution failed'
        super(message, data)
      end
    end

    class ConflictError < ClientError
      attr_reader :conflict

      def initialize(message = nil, data = nil)
        @default_message = 'Task execution failed'
        super(message, data)

        @conflict = @data[:conflict]
      end
    end

    ERRORS = {
      'exec'       => [ Blockbridge::Api::ExecutionError, 'Task execution failed' ],
      'validation' => [ Blockbridge::Api::ValidationError, 'Parameter validation failed' ],
    }
    DEFAULT_ERROR = [Blockbridge::Api::ClientError, "Unspecified failure"]

    def self.wrap_error(e)
      case e
      when Excon::Errors::BadRequest
        data = MultiJson.load(e.response.body, :symbolize_keys => true)
        error_class, error_message = ERRORS[data[:type]] || DEFAULT_ERROR
        error_class.new(error_message, data)
      when Excon::Errors::Conflict
        data = MultiJson.load(e.response.body, :symbolize_keys => true)
        ConflictError.new('Conflict', data)
      when Excon::Errors::Unauthorized
        data = MultiJson.load(e.response.body, :symbolize_keys => true)
        UnauthorizedError.new('Unauthorized', data)
      when Excon::Errors::Forbidden
        data = MultiJson.load(e.response.body, :symbolize_keys => true)
        ForbiddenError.new('Forbidden', data)
      when Excon::Errors::NotFound
        data = MultiJson.load(e.response.body, :symbolize_keys => true)
        NotFoundError.new('Not Found', data)
      end
    end

  end
end
