# frozen_string_literal: true

# Copyright (c) 2015-2017, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'blockbridge/api/xmd_data'
require 'blockbridge/api/hashie_mash'

begin
  require 'em-synchrony'
rescue LoadError
end

require 'jsonpath'

module Blockbridge
  module Api
    # wrapped API object with a variety of helper methods
    class Model < ::Blockbridge::HashieMash
      include Blockbridge::Api::CamelSnake

      def self.for_type(type)
        base_klass = Blockbridge::Api::CamelSnake.camel_case(type.to_s)
        full_klass = base_klass.to_s + 'Model'
        Blockbridge::Api.const_get(full_klass.to_sym) rescue Blockbridge::Api::Model
      end

      def self.build(object)
        oid = ObjectId.new(object['id'])
        klass = for_type(oid.type)
        klass.new(object)
      end

      attr_accessor :obj_type
      attr_accessor :friendly_type
      attr_accessor :resource_instance

      def obj_type
        @obj_type ||= id.type || @resource_instance.oid_type
      end

      def friendly_type
        @friendly_type ||= obj_type
      end

      def to_s
        begin
          label ? "#{label} (#{serial})" : serial
        rescue
          "<#{self.class.name}:#{id}>"
        end
      end

      def id
        @id ||= ObjectId.new(super)
      end

      # lookup by jsonpath
      def find_all(path)
        JsonPath.new(path).on(self)
      end

      def find_first(path, default)
        matches = find_all(path)
        if matches.length == 0
          if default
            default
          else
            raise Blockbridge::Api::PathNotFound.new("failed to match json path '#{path}'")
          end
        else
          matches.first
        end
      end

      def xmd
        @xmd ||= XmdDataDecorator.new(super)
      end

      private

      def method_missing(method, *args, &blk)
        ref_key = "#{method}_id"

        if !key?(method) && key?(ref_key)
          # look up the referred object
          ref_obj = resource_instance.fetch_ref self.send(ref_key)
          define_method(method) { ref_obj }
          ref_obj
        else
          super
        end
      end

      def respond_to_missing?(method, *)
        !key?(method) && key?("#{method}_id") || super
      end
    end

    class AccountModel < Model
      def label
        name
      end
    end

    class UserProfileModel < Model
      def initialize(data)
        super
        @friendly_type = 'user'
      end

      def label
        auth_user || login
      end
    end

    class VdiskModel < Model
      def initialize(data, dfl = nil)
        super
        @friendly_type = 'disk'
      end
    end

    class InitiatorProfileModel < Model
      def initialize(data, dfl = nil)
        super
        @friendly_type = 'profile'
      end
    end

    class SysDsModel < Model
      def initialize(data, dfl = nil)
        super
        @friendly_type = 'datastore'
      end
    end

    class IscsiDevModel < Model
      def initialize(data, dfl = nil)
        super
        @friendly_type = 'iscsi dev'
      end
    end

    class OsDevModel < Model
      def initialize(data, dfl = nil)
        super
        @friendly_type = 'local dev'
      end

      def label
        super || display_name
      end
    end

    module Resources
      include Blockbridge::Api::CamelSnake

      def self.for_type(type)
        const_get(Blockbridge::Api::CamelSnake.camel_case(type.to_s))
      end

      class Base
        # class methods

        def send_request(method, *args, &block)
          if async?
            args = args.dup
            res = EM::Synchrony.defer do
              begin
                @res_client.send(method, *args)
              rescue => ex
                ex
              end
            end
            raise res if res.is_a?(Exception)
            res
          else
            @res_client.send(method, *args, &block)
          end
        rescue StandardError => e
          raise (Blockbridge::Api::wrap_error(e) || e)
        end

        def self.model_response(*methods)
          methods.each do |method|
            define_method method.to_sym do |*args, &block|
              # retrieve object(s) from the API
              rsp = send_request(method, *args, &block)

              # wrap the response object(s)
              modelize(rsp)
            end
          end
        end

        def self.model(klass)
          @model_klass = klass
        end

        # default methods that return wrapped objects
        model_response :info, :create, :update, :list

        # instance methods
        def initialize(client, opts = {})
          # look up per-resource heroics client
          @res_client  = client.send(oid_type)
          @base_client = client
          @opts = opts
        end

        def async?
          @opts[:async]
        end

        def fetch_ref(ref_str)
          return if ref_str.nil? # null reference -> null object
          begin
            # use generic object fetch API -- this supports by id or serial.
            modelize_by_oid @base_client.objects.info(ref_str)
          rescue Excon::Errors::NotFound, Excon::Errors::Gone => e
            # translate not-found into nil; everything else raises.
            nil
          end
        end

        def bare_resource_klass
          @bare_resource_klass ||= self.class.name.split('::').last
        end

        def oid_type
          @type_str ||= Blockbridge::Api::CamelSnake.snake_case(bare_resource_klass)
        end

        def default_model_klass
          Blockbridge::Api::Model
        end

        def model_klass
          @@model_klass ||= default_model_klass
        end

        def modelize_by_oid(data)
          # XXX: placeholder for handling local client model specialization in
          # the future. at the moment, all responses are just wrapped by
          # a Hashie::Mash derived class. (see above)
          Blockbridge::Api::Model.build(data).tap do |m|
            m.resource_instance = self
          end
        end

        def modelize(data)
          if data.is_a?(Array)
            data.map { |item| modelize_by_oid(item) }
          else
            modelize_by_oid(data)
          end
        end

        def method_missing(method, *args, &block)
          return super unless @res_client.respond_to?(method)

          # other methods proxy through to Heroics client directly.
          rsp = send_request(method, *args, &block)

          # wrap the response object(s)
          if rsp.is_a?(Array)
            rsp.map do |item|
              model_klass.new(item).tap { |m| m.resource_instance = self }
            end
          else
            model_klass.new(rsp).tap do |m|
              m.resource_instance = self
            end
          end
        end

        def respond_to_missing?(method, include_private = false)
          @res_client.respond_to?(method) || super
        end
      end

      # api status
      class Status < Base; end

      # authn/authz
      class Oauth2Token < Base; end
      class Authorization < Base
        def oid_type
          'authz'
        end
      end

      # ca cert
      class CaCertificate < Base
        def oid_type
          'ca_cert'
        end
      end

      # generic resources
      class Xmd < Base; end

      # bulk object api
      class Objects < Base
        model_response :info, :list, :list_by_xref
      end

      # account/domain
      class Account < Base
        def model_klass
          Blockbridge::Api::AccountModel
        end
      end
      class AccountUsage < Base; end
      class UserProfile < Base
        def model_klass
          Blockbridge::Api::UserProfileModel
        end
      end

      # storage
      class Vdisk < Base
        def model_klass
          Blockbridge::Api::VdiskModel
        end
      end
      class InitiatorProfile < Base
        def model_klass
          Blockbridge::Api::InitiatorProfileModel
        end
      end

      class Vss < Base;      end
      class Target < Base;   end
      class Rule < Base;     end
      class Product < Base;  end
      class Snapshot < Base; end

      # infrastructure
      class SysDs < Base
        def model_klass
          Blockbridge::Api::SysDsModel
        end
      end
      class OsDev < Base
        def model_klass
          Blockbridge::Api::OsDevModel
        end
      end
      class IscsiDev < Base
        def model_klass
          Blockbridge::Api::IscsiDevModel
        end
      end

      class Node < Base;     end
      class Site < Base;     end
      class Dev < Base;      end
      class Net < Base;      end

      class ObjStore < Base; end

    end
  end
end
