# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'retriable'
require 'securerandom'
require 'socket'
require 'ipaddr'

require 'blockbridge/host/executor'

module BlockbridgeHost
  # Generic host information and controll class. This contains functionality
  # that could reasonably be used to control any (or most) platforms, and thus
  # should contain no platform specific details.
  class GenericControl
    def self.util(name, alts: nil, env_var: nil, paths: nil, missing: nil,
                  run: :spawn, privileged: true, from_package: nil, &blk)
      env_var ||= "BLOCKBRIDGE_#{name.to_s.upcase}_COMMAND"

      exec_name_attr = "#{name}_exec_name"
      has_exec_attr = "has_#{name}?"
      exec_name = nil

      define_method(exec_name_attr) do
        exec_name ||= determine_exec_name(name, alts, env_var, paths)
        return exec_name if exec_name
        fail "Failed to find \`#{name}\` command: #{missing_exec_name(missing, from_package)}."
      end

      define_method(has_exec_attr) do
        begin
          self.send(exec_name_attr)
          true
        rescue
          false
        end
      end

      define_method(name) do |*argv, **kwargs|
        xn = self.send(exec_name_attr)
        blk.call(xn, *argv, **kwargs) unless blk.nil?
        execute(xn.to_s, *argv, run: run, privileged: privileged, **kwargs)
      end
    end

    def runcmd(name, *argv, **kwargs)
      res = send(name, *argv, **kwargs, noraise: true)
      return res.success?, res.out, res.err
    end

    attr_accessor :options
    attr_reader :executor

    def initialize(**kwargs)
      @options ||= kwargs
      @executor ||= @options.fetch(:executor, GenericExecutor.new(@options))
      @options[:executor] = @executor
    end

    def warn_puts(*a)
      STDERR.puts(*a) unless options[:quiet]
    end

    def debug_puts(*a)
      puts(*a) if options[:debug]
    end

    def verbose_puts(*a)
      puts(*a) if options[:verbose]
    end

    def os_desc
      'generic'
    end

    def iscsi
      fail "iSCSI control is unsupported on #{os_desc} hosts"
    end

    def ipsec
      fail "IPsec control is unsupported on #{os_desc} hosts"
    end

    def sysfs
      fail "Sysfs control is unsupported on #{os_desc} hosts"
    end

    def md
      fail "MD control is unsupported on #{os_desc} hosts"
    end

    def stunnel
      fail "Stunnel control is unsupported on #{os_desc} hosts"
    end

    def machine_id
      return @machine_id if defined? @machine_id

      nifs = network_interfaces.dup

      # remove 'pan*', 'lo' and interfaces that have no hardware address
      nifs.reject! do |e|
        e[:name] =~ /^pan\d+$|^lo$/ || e[:hwaddr].nil?
      end

      primary_if = nifs.find { |e| e[:hwaddr] }
      primary_if[:hwaddr] if primary_if
    end

    def hostname
      Socket.gethostname
    end

    def ip_addr?(str)
      return false if str.nil?

      IPAddr.new(str)
      true
    rescue IPAddr::InvalidAddressError
      false
    end

    def local_ifaddrs
      @local_ifaddrs ||= Socket.getifaddrs.reject do |ifa|
        !ifa.addr || !ifa.addr.ip? || ifa.name.start_with?("docker")
      end
    end

    def local_ipaddrs(&blk)
      ifaddrs = local_ifaddrs
      ifaddrs = ifaddrs.select(&blk) unless blk.nil?
      ifaddrs.map { |ifa| ifa.addr.ip_address }.uniq
    end

    def local_ifnames(&blk)
      ifaddrs = local_ifaddrs
      ifaddrs = ifaddrs.select(&blk) unless blk.nil?
      ifaddrs.map { |ifa| ifa.name }.uniq
    end

    def udp_port_open?(host, port)
      UDPSocket.open do |s|
        s.connect(host, port)
      end
      true
    rescue
      false
    end

    # Open a TCP connection to the specified destination +host+ and +port+. If a
    # +local+ bind is specified, the source address is bound. +local+ may be an
    # ip address or an interface name -- if an interface name is specified,
    # returns true if binding to *any* of the IP addresses configured on the
    # interface results in an established connection.
    def tcp_port_open?(host, port, local: nil, local_port: nil, wait: 5)
      if local
        local_ip = ip_addr?(local)
        locals = local_ipaddrs do |ifa|
          if local_ip
            ifa.addr.ip_address == local
          else
            ifa.name == local
          end
        end
      else
        locals = [ nil ]
      end

      locals.any? do |lip|
        begin
          Socket.tcp(host.to_s, port, lip, local_port,
                     connect_timeout: wait) do |sock|
            # pass a block to have Socket auto-close on completion.
          end
          true
        rescue Errno::ECONNREFUSED,
               Errno::ENETUNREACH,
               Errno::EHOSTUNREACH,
               Errno::ETIMEDOUT,
               Errno::EINVAL,
               SocketError => e
          debug_puts "connect failed to #{host.to_s}:#{port} #{e.message}"
          false
        end
      end
    end

    def exist?(path)
      File.exist?(path)
    end

    def file?(path)
      File.file?(path)
    end

    def executable?(path)
      File.exist?(path)
    end

    def compute_target_paths(target, multipath: false, ipsec: false, ssl: false,
                             local_bind: true, locals: nil, remotes: nil,
                             af: Socket::AF_UNSPEC, **_extras)
      # remote interfaces
      remotes = verify_remotes(target, remotes)

      # target portals
      portals = target.portals.select { |p|
        next false if ipsec && p.interface != "IPsec"
        next false if ssl && p.interface != "SSL"
        remotes.include?(p.interface)
      }
      portal_ipaddrs = portals.flat_map { |p| p.ipaddrs }

      case af ||= Socket::AF_UNSPEC
      when Socket::AF_INET
        debug_puts("only considering IPv4 portal addresses")
      when Socket::AF_INET6
        debug_puts("only considering IPv6 portal addresses")
      when Socket::AF_UNSPEC
        debug_puts("considering both IPv4 and IPv6 portal addresses")
      else
        fail "Unsupported address family '#{af}' specified."
      end

      # build set of paths
      paths = []
      portal_ipaddrs.each do |rip|
        rip = rip.dup
        begin
          rip.addr = IPAddr.new(rip.addr)
        rescue => e
          debug_puts("skipping portal address '#{rip.addr}': #{e.message}")
          next
        end
        if af != Socket::AF_UNSPEC && rip.addr.family != af
          debug_puts("skipping portal address '#{rip.addr}': "\
                     "address family does not match.")
          next
        end

        if !local_bind || !multipath
          # probe each target portal ip without binding the source
          # address. multipath mode binds source unless otherwise instructed.
          if tcp_port_open?(rip.addr, rip.port)
            debug_puts("portal address reachable #{rip.addr}, port #{rip.port}")
            paths.push(remote: rip)
          else
            debug_puts("portal address NOT reachable #{rip.addr}, "\
                       "port #{rip.port}")
          end
        else
          # multipath local bind -- probe each portal ip from *each* local
          # interface/address.
          locals.each do |ip_or_ifname|
            if tcp_port_open?(rip.addr, rip.port, local: ip_or_ifname)
              debug_puts("portal address reachable #{rip.addr}, "\
                         "port #{rip.port} (via #{ip_or_ifname})")
              paths.push(local: ip_or_ifname, remote: rip)
            else
              debug_puts("portal address NOT reachable #{rip.addr}, "\
                         "port #{rip.port} (via #{ip_or_ifname})")
            end
          end
        end

        # keep looking for all (multipath) or 1 (otherwise) paths.
        break unless multipath || paths.empty?
      end

      # ensure at least one valid node
      if paths.empty?
        fail "No target portals found that are reachable from " \
             "local interfaces."
      end

      paths
    end

    protected

    def retriable(**kwargs, &blk)
      # set up some defaults
      kwargs[:max_elapsed_time] ||= 30
      kwargs[:base_interval] ||= 1.0
      kwargs[:multiplier] ||= 1.0
      kwargs[:rand_factor] ||= 0.0
      kwargs[:tries] ||= 60

      # delegate to Retriable gem
      Retriable.retriable(kwargs, &blk)
    end

    def execute(*args, **kwargs, &blk)
      @executor.execute(*args, **kwargs, &blk)
    end

    def execute_batch(*args, **kwargs, &blk)
      @executor.execute_batch(*args, **kwargs, &blk)
    end

    def exec_name_from_env(cmd, var)
      return unless (val = ENV[var])
      unless exist?(val)
        fail "#{cmd} command from environment (#{var}=#{val}) does not exist"
      end
      unless file?(val)
        fail "#{cmd} command from environment (#{var}=#{val}) " \
             "is not a regular file"
      end
      unless executable?(val)
        fail "#{cmd} command from environment (#{var}=#{val}) is not executable"
      end
      val
    end

    def exec_name_in_paths(cmd, fs_paths)
      return cmd if fs_paths.nil?
      dir = fs_paths.find do |path|
        fp = File.join(path, cmd.to_s)
        exist?(fp) && file?(fp) && executable?(fp)
      end
      return File.join(dir, cmd.to_s) if dir
    end

    def determine_exec_name(cmd, alts, env_var, fs_paths)
      exec_name_from_env(cmd, env_var) || exec_name_in_paths(cmd, fs_paths) ||
        begin
          alts = [ alts ] if alts.is_a?(String)
          alts&.find do |alt|
            exec_name_from_env(alt, env_var) || exec_name_in_paths(alt, fs_paths)
          end
        end
    end

    def missing_exec_name(missing, from_package)
      return missing if missing
      return from_package if from_package&.is_a? String
      from_package&.each do |cmd, pkg|
        next unless which cmd
        return "please ensure the '#{pkg}' package is installed and try again"
      end
      return "please ensure it is installed and try again"
    end

    def which_exts
      ENV['PATHEXT']&.split(';')
    end

    def which_paths
      ENV['PATH']&.split(File::PATH_SEPARATOR)
    end

    def which_cmds(cmd)
      cmds = []
      which_paths&.each do |path|
        cmds.push File.join(path, "#{cmd.to_s}")
        which_exts&.each do |ext|
          cmds.push File.join(path, "#{cmd.to_s}#{ext}")
        end
      end
      cmds
    end

    def which(cmd)
      which_cmds(cmd).find do |exe|
        exist?(exe) && file?(exe) && executable?(exe)
      end
    end
  end
end
