# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'blockbridge/host/control/generic'
require 'blockbridge/host/control/libc'
require 'blockbridge/host/control/udev'

module BlockbridgeHost
  # Specialization for controlling a local Linux host.
  class LinuxControl < GenericControl
    util :wipefs,    run: :system, paths: ['/usr/sbin', '/sbin']
    util :dd,        run: :system, paths: ['/usr/bin', '/bin']
    util :partprobe, run: :system, paths: ['/usr/sbin', '/sbin']
    util :ip,        run: :system, paths: ['/usr/sbin', '/sbin']
    util :luabb,     paths: ['/bb/bin']
    util :amtool,    paths: ['/bb/bin/prometheus']
    util :systemctl, paths: ['/bin', '/usr/bin']
    util :service,   paths: ['/sbin', '/usr/sbin']
    util :mv, paths: ['/bin', '/usr/bin'], run: :system, from_package: 'coreutils'
    util :rm, paths: ['/bin', '/usr/bin'], run: :system, from_package: 'coreutils'
    util :cat, paths: ['/bin', '/usr/bin'], run: :system, from_package: 'coreutils'
    util :modprobe, paths: ['/usr/sbin', '/sbin']
    util :chkconfig, paths: ['/sbin']
    util :rdma_probe, paths: ['/bb/libexec']
    util :pgrep, paths: ['/usr/bin']

    def initialize(**kwargs)
      @options ||= kwargs
      @executor ||= UnixExecutor.new(@options)
      super
    end

    def linux?
      true
    end

    def windows?
      false
    end

    def os_desc
      'Linux'
    end

    def verify_remotes(target, remotes)
      target_ifnames = target.portals.map { |p| p.interface }
      return target_ifnames unless remotes
      remotes.each do |ifa|
        unless target_ifnames.include?(ifa)
          fail "Remote interface '#{ifa} not found."
        end
      end
      remotes
    end

    def verify_locals(locals)
      return local_ifnames unless locals
      locals.each do |n|
        unless local_ifnames.include?(n)
          fail "Local interface '#{n}' not found."
        end
      end
      locals
    end

    def create_target_devs(target, mknod: false, mknod_prefix: nil,
                           mknod_name: nil, multipath: false, **)
      # lookup devs
      devs = iscsi.lookup_target_devs(target.iqn)
      fail 'Unable to locate device after login.' unless devs

      # manually create dev nodes if requested
      if mknod
        debug_puts "dev: make device node(s) in /dev"
        iscsi.makedevs(devs, prefix: mknod_prefix, name: mknod_name)
      end

      # wait for devs
      debug_puts "iscsi: waiting for mapped lun to show up as device"
      fail 'Unable to locate device after login.' if !iscsi.wait_for_devs(devs)

      # configure multipath
      if multipath
        debug_puts "multipath: looking up scsi id for device"
        scsi_id = iscsi.lookup_scsi_id(devs.first[:dev])
        fail 'Unable to determine SCSI ID for multipath device.' unless scsi_id

        # add wwid as multipath device
        debug_puts "multipath: tracking device #{devs.first[:dev]} with wwid #{scsi_id}"
        mpath.add_wwid(scsi_id)

        # wait for multipath device
        debug_puts "looking for multipath device (wwid = #{scsi_id})"
        mpath_dev = mpath.wait_for_dev(devs.map { |d| d[:dev] }, scsi_id)
        fail 'Unable to locate multipath device after login.' unless mpath_dev

        # multipath -- return the mp device as the 'dev', and also include the
        # constituent devices.
        {
          dev: mpath_dev,
          dev_path: mpath_dev[:dev],
          hostname: hostname,
          multipath_devs: devs
        }
      else
        # single path should only have a single device
        dev = devs.first
        {
          dev: dev,
          dev_path: dev[:dev],
          hostname: hostname,
        }
      end
    end

    # configure the linux initiator to connect to a remote blockbridge target.
    def target_attach(target, **conn_config)
      # verify locals selection, if any. in linux, we only support binding to
      # local interfaces by name.
      conn_config[:locals] = verify_locals(conn_config[:locals])

      # determine the path(s) to use to access the target
      paths = compute_target_paths(target, **conn_config)

      # create iscsi node for paths
      paths.each do |path|
        if path[:local]
          # path specifies a local interface binding -- use it.
          debug_puts "iscsi: creating interface #{path[:local]}"
          iscsi.create_iface(path[:local])
        end

        debug_puts "iscsi: creating target #{path[:remote].addr}, #{target.iqn}"
        iscsi.create_target(target.iqn, path[:remote].addr, path[:remote].port,
                            iface: path[:local],
                            automatic: conn_config[:automatic])
      end

      # set auth on target
      iscsi.set_target_auth(target.iqn, **conn_config[:chap])

      if conn_config[:multipath]
        # set timeout for multipath
        iscsi.set_target_timeout(target.iqn, 10)
      end

      # attempt to login to the target
      debug_puts "iscsi: logging in to #{target.iqn}"
      iscsi.login_target(target.iqn)

      # create/lookup local devs
      dev_info = create_target_devs(target, **conn_config)
      dev_info.merge!(paths: paths, multipath: conn_config[:multipath])
    rescue
      # something went wrong -- try to tear everything down.
      iscsi.remove_target(target.iqn) rescue nil
      raise
    end

    def lookup_target_paths(targ_name, dev_path)
      mpath.lookup_active_paths(dev_path)
    end

    def tune_settings(path, key, val)
      if val.is_a? Hash
        val.each_pair do |_key, _val|
          tune_settings("#{path}/#{key}", _key, _val)
        end
      else
        debug_puts "tuning device: #{path}/#{key}: setting to #{val}"
        begin
          sysfs.write(path, key, val)
        rescue Errno::EINVAL
          warn_puts "WARN: failed to set #{path}/#{key} to #{val}: Invalid argument"
        end
      end
    end

    def device_tune(dev, settings)
      case dev
      when String
        path = "/sys/block/#{File.basename(dev)}"
        desc = dev
      when Hash
        path = "/sys/dev/block/#{dev[:major]}:#{dev[:minor]}"
        desc = dev[:dev]
      else
        fail "device has invalid format (#{dev.class})"
      end

      debug_puts "tuning device: #{path} (#{desc})"
      settings.each_pair do |key, val|
        tune_settings(path, key, val)
      end
    end

    def device_settle
      LibUdev.settle
    end

    def device_scsi_serial(dev)
      LibUdev::Device.new_from_devnode(dev)&.id_scsi_serial
    end

    def device_from_uuid(uuid)
      LibUdev::Devices.each do |device|
        return device.devnode if device.id_scsi_serial == uuid
      end
    end

    def device_wipe(dev)
      wipefs('-a', dev)
    end

    def device_zero(dev, length, offset = 0)
      return unless length && length > 0
      offset = 0 unless offset && offset > 0
      if length < 4096
        count = 1
        bs    = length
        over  = 0
      else
        count = length / 4096
        bs    = 4096
        over  = length % 4096
      end

      # dd for length at offset
      args = ['if=/dev/zero', "of=#{dev}", "count=#{count}", "bs=#{bs}", "seek=#{offset}", 'oflag=seek_bytes']
      dd(*args)

      # dd remainder for any overage
      if over > 0
        args = ['if=/dev/zero', "of=#{dev}", "seek=#{offset + (count * bs)}", "count=1", "bs=#{over}", "oflag=seek_bytes"]
        dd(*args)
      end
    end

    def route_get(match)
      cmd = ['route', 'get', match]
      ip(*cmd)&.out&.split("\n")&.first&.strip
    end

    def route_get_gw(match)
      route_get(match)[/via ([^\s]+)\s/,1]
    end

    def route_add(route, gw, src, table)
      cmd = ['route', 'add', route, 'via', gw]
      cmd.concat ['src', src] if src
      cmd.concat ['table', table] if table
      ip(*cmd)
    end

    def route_list(table)
      cmd = ['route', 'list']
      cmd.concat ['table', table] if table
      ip(*cmd)&.out&.split("\n").map { |l| l.strip }
    rescue ExecError => e
      return [] if e.message =~ /FIB table does not exist./

      raise e
    end

    def route_del(route, table)
      cmd = ['route', 'del', route]
      cmd.concat ['table', table] if table
      ip(*cmd)
    end

    def device_partprobe(dev)
      partprobe(dev)
    end

    def iscsi
      @iscsi ||= LinuxISCSIControl.new(@options).tap do |ctl|
        ctl.iscsi_init
      end
    end

    def mpath
      @mpath ||= LinuxMultipathControl.new(@options).tap do |ctl|
        ctl.enable_multipath_service
      end
    end

    def ipsec
      @ipsec ||= LinuxIPsecControl.new(@options).tap do |ctl|
        ctl.ipsec_init
      end
    end

    def stunnel 
      @stunnel ||= LinuxStunnelControl.new(@options)
    end

    def sysfs
      @sysfs ||= LinuxSysfsControl.new(@options)
    end

    def md
      @md ||= LinuxMDControl.new(@options).tap do |ctl|
      end
    end

    def ipmi
      @ipmi ||= LinuxIPMIControl.new(@options)
    end

    def dmi
      @dmi ||= LinuxDMIControl.new(@options)
    end
  end
end

require 'blockbridge/host/control/linux_iscsi'
require 'blockbridge/host/control/linux_multipath'
require 'blockbridge/host/control/linux_ipsec'
require 'blockbridge/host/control/linux_stunnel'
require 'blockbridge/host/control/linux_sysfs'
require 'blockbridge/host/control/linux_md'
require 'blockbridge/host/control/linux_ipmi'
require 'blockbridge/host/control/linux_dmi'
