# frozen_string_literal: true

# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require_relative 'vici'
require 'pp'

module BlockbridgeHost
  class LinuxIPsecControl < LinuxControl
    util :ipsec,
         alts: 'strongswan',
         run: :system,
         paths: ['/usr/sbin/'],
         from_package: { rpm: 'epel-release, strongswan', dpkg: 'strongswan' }

    IPSEC_ROUTE_TABLE = '220'
    DUPLICHECK_DISABLE_CONF =
      '/etc/strongswan/strongswan.d/charon/duplicheck_disable.conf'
    VICI_SOCK_PATHS = [
      '/var/run/charon.vici',
      '/run/strongswan/charon.vici'
    ].freeze

    def ipsec_init
      plugin_configure
      ipsec('version')
      ipsec('start') rescue nil
      plugin_verify
    end

    LOADED_PLUGINS_RX = /loaded plugins:(.*)$/.freeze

    def loaded_plugins
      return [] if (sts_out = ipsec('statusall')&.out&.to_s).nil?
      return [] unless (md = LOADED_PLUGINS_RX.match(sts_out))

      md[1].strip.split(/\s+/)
    end

    def plugin_loaded?(name)
      loaded_plugins.include?(name)
    end

    def plugin_verify
      return unless plugin_loaded?('duplicheck')

      puts <<~WARN
        WARNING: Detected strongswan running with 'duplicheck' plugin enabled.
                 Multiple IPsec attached disks from the same host may not operate correctly.
                 Restart strongswan and re-attach disks to resolve.
      WARN
    end

    def plugin_configure
      return if File.exist? DUPLICHECK_DISABLE_CONF

      Tempfile.open('duplicheck_disable.conf') do |f|
        f.write(duplicheck_disable_text)
        mv(f.path, DUPLICHECK_DISABLE_CONF)
      end
    end

    def duplicheck_disable_text
      "duplicheck {\n" \
      "    # Disable duplicheck plugin\n" \
      "    enable = no\n" \
      "\n" \
      "    # Don't load the plugin\n" \
      "    load = no\n" \
      "}\n"
    end

    def vici_sock_path
      path = VICI_SOCK_PATHS.find { |p| File.exist?(p) }
      return path unless path.nil?

      raise 'Unable to communicate with strongswan; please ensure strongswan'\
            ' is installed, running and vici plugin is enabled.'
    end

    def vici_connect(wait: 5)
      vici_e = nil
      vici = retriable max_elapsed_time: wait do
        begin
          Vici::Connection.new(UNIXSocket.new(vici_sock_path))
        rescue SystemCallError => e
          raise 'Unable to communicate with strongswan; please ensure '\
                'strongswan is installed, running and vici plugin '\
                "is enabled: #{e.message}"
        rescue StandardError => e
          if e&.message =~ /No such file or directory.*charon.vici/
            raise 'Unable to communicate with strongswan; please ensure'\
                  'strongswan is installed, running and vici plugin is'\
                  " enabled: #{e.message}"
          end
          vici_e = e
          nil
        end
      end

      raise vici_e if vici_e

      vici
    end

    def vici
      @vici ||= vici_connect
    end

    def hostname_ike_avail?(hostname)
      udp_port_open?(hostname, 500) || udp_port_open?(hostname, 4500)
    end

    # set ca
    def load_cert(cert)
      vici.load_cert('type' => 'X509', 'data' => cert)
    rescue Vici::CommandExecError => e
      pp e if options[:debug]
      raise 'Unable to load CA certificate'
    end

    def load_cred(type, user, pass)
      vici.load_shared('type' => type, 'data' => pass, 'owners' => [user])
    rescue Vici::CommandExecError => e
      pp e if options[:debug]
      raise 'Unable to load IPsec credentials'
    end

    def load_conn(conn)
      vici.load_conn(conn)
    rescue Vici::CommandExecError => e
      pp e if options[:debug]
      raise 'Unable to load IPsec connection definition'
    end

    def unload_conn(name)
      begin
        vici.unload_conn('name' => name)
      rescue Vici::CommandExecError => e
        raise unless e.message =~ /not found/
      end
    rescue Vici::CommandExecError => e
      pp e if options[:debug]
      raise 'Unable to unload IPsec connection definition'
    end

    def start_conn(hostname, name, wait: 30)
      unless hostname_ike_avail?(hostname)
        raise "Unable to connect to host '#{hostname}'. Check configuration."
      end

      vici.initiate('child' => name, 'timeout' => 10000)
      retriable max_elapsed_time: wait do
        unless get_sa(name)
          raise 'timed out waiting for IPsec Security Association '\
                'after starting connection'
        end
      end
    rescue Vici::CommandExecError => e
      pp e if options[:debug]
      raise "Unable to start IPsec Security Association (#{e.message})."\
            ' Check syslog for more information.'
    end

    def get_sa(name)
      vici.list_sas(name)&.first&.dig(name)
    end

    def add_src_routes(name)
      sa = get_sa(name)
      unless sa
        raise "could not retrieve IPsec Security Association by name #{name}"
      end
      dst_ip    = sa['remote-host']
      src_ip    = sa['local-vips'].first

      child_sa = sa.dig('child-sas')&.values&.max_by { |csa| csa['uniqueid'] }
      dst_ts = child_sa&.dig('remote-ts')&.map { |ts| ts[/^([^\[]+)/, 1] }

      unless dst_ip && src_ip && dst_ts
        pp sa if options[:debug]
        raise 'destination, source or tunnel subnets unavailable'
      end

      # lookup gw (or use direct route)
      gw = route_get_gw(dst_ip) || dst_ip

      # delete old and add new
      routes = route_list(IPSEC_ROUTE_TABLE)
      dst_ts.each do |ts|
        route_del(ts, IPSEC_ROUTE_TABLE) if routes.any? { |r| r =~ /^#{ts}/ }
        route_add(ts, gw, src_ip, IPSEC_ROUTE_TABLE)
      end
    rescue StandardError => e
      raise "Unable to program IPsec tunnel source route: #{e.message}"
    end

    def src_routes
      route_list(IPSEC_ROUTE_TABLE).map do |r|
        r.split(' ').first
      end.compact
    end

    def del_src_routes
      src_routes.each do |r|
        route_del(r, IPSEC_ROUTE_TABLE)
      end
    end

    def stop_conn(name)
      begin
        vici.terminate('child' => name, 'ike' => name, 'timeout' => 10000)
        vici.terminate('ike' => name, 'timeout' => 10000)
      rescue Vici::CommandExecError => e
        pp e if options[:debug]
        raise unless e.message =~ /no matching/
      end
    rescue Vici::CommandExecError => e
      pp e if options[:debug]
      raise 'Unable to stop IPsec Security Association'
    end
  end
end
