# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'blockbridge/host/control/udev'

module BlockbridgeHost
  # Linux iSCSI control support.
  class LinuxISCSIControl < LinuxControl
    util :iscsiadm, paths: ['/bb/sbin', '/sbin', '/usr/sbin', '/usr/bin'],
         from_package: { rpm: 'iscsi-initiator-utils', dpkg: 'open-iscsi' }
    util :lsscsi, paths: ['/usr/bin'], from_package: 'lsscsi'
    util :scsi_id, paths: ['/sbin', '/usr/lib/udev', '/lib/udev']
    util :mkdir, paths: ['/usr/bin', '/bin']
    util :mknod, paths: ['/usr/bin', '/bin']
    util :bash, paths: ['/bin']
    util :iscsid, run: :system, paths: ['/sbin', '/usr/sbin']

    #########################################################################
    # initialize iscsid by starting it (ignore errors)
    #########################################################################
    def iscsi_init
      # start iscsid
      service('iscsid', 'start')
    rescue
      nil
    end

    #########################################################################
    # call iscsiadm with retry
    #########################################################################
    def iscsiadm_retry(*args, wait: 10)
      retriable max_elapsed_time: wait do
        begin
          iscsiadm(*args)
        rescue
          iscsi_init
          raise
        end
      end
    end

    # generate a chap password appropriate for a Linux host. Suuuper duper long
    # secrets are allowed, unlike in Windows.
    def gen_init_pass
      SecureRandom.base64(64).slice(0, 64)
    end

    #########################################################################
    # create interface. idempotent.
    #########################################################################
    def create_iface(iface)
      # check for existance -- this command will raise if the interface is not
      # defined.
      iscsiadm('-m', 'iface', '-I', iface)
    rescue
      # create new iface definition
      iscsiadm('-m', 'iface', '-I', iface, '-o', 'new')
      iscsiadm('-m', 'iface', '-I', iface, '--op=update',
               '-n', 'iface.net_ifacename', '-v', iface)
    end

    #########################################################################
    # stat interface. check for existence
    #########################################################################
    def stat_iface(iface)
      iscsiadm('-m', iface, '-I', iface)
    end

    #########################################################################
    # configure a new target node
    #########################################################################
    def create_target(name, addr, port = 3260, iface: nil, **extra_options)
      # create target node and set for automatic startup
      portal_addr = addr.ipv6? ? "[#{addr}]" : addr.to_s
      portal = "#{portal_addr}:#{port},1"

      args = [
        '-m', 'node', '-o', 'new',
        '--targetname', name, '--portal', portal,
      ]

      args.concat(['-I', iface]) unless iface.nil?

      iscsiadm(*args)

      # options (other than addr, port and interface) must be set in a follow-up iscsiadm
      # command. additional node options supplied at create time are accepted but
      # *ignored* by iscsiadm.
      set_target_options(name, **extra_options)
    end

    def iscsiadm_update_node(name, args)
      unless args.empty?
        iscsiadm('-m', 'node', '-o', 'update', '--targetname', name, *args)
      end
    end

    def add_update_option(args, key, value)
      args.concat(['-n', key, '-v', value.to_s]) unless value.nil?
    end

    #########################################################################
    # update an existing target node
    #########################################################################
    def set_target_options(name, iface: nil, automatic: nil, session_timeout: nil,
                           login_timeout: nil, login_retries: nil,
                           authmethod: nil, init_user: nil, init_pass: nil,
                           targ_user: nil, targ_pass: nil, transport: nil,
                           header_digest: nil, data_digest: nil,
                           iscsi_parameters: nil)
      args = []

      # automatic re-connect on boot
      case automatic
      when true
        add_update_option(args, 'node.startup', 'automatic')
        add_update_option(args, 'node.conn[0].startup', 'automatic')
      when false
        add_update_option(args, 'node.startup', 'manual')
        add_update_option(args, 'node.conn[0].startup', 'manual')
      end

      # timeout
      add_update_option(args, 'node.session.timeo.replacement_timeout',
                        session_timeout)

      # auth options
      add_update_option(args, 'node.session.auth.authmethod', authmethod)
      add_update_option(args, 'node.session.auth.username', init_user)
      add_update_option(args, 'node.session.auth.password', init_pass)
      add_update_option(args, 'node.session.auth.username_in', targ_user)
      add_update_option(args, 'node.session.auth.password_in', targ_pass)

      # login options
      add_update_option(args, 'node.conn[0].timeo.login_timeout', login_timeout)
      add_update_option(args, 'node.session.initial_login_retry_max', login_retries)

      # transport option
      add_update_option(args, 'iface.transport_name', transport)

      # digest options
      add_update_option(args, 'node.conn[0].iscsi.HeaderDigest', header_digest)
      add_update_option(args, 'node.conn[0].iscsi.DataDigest', data_digest)

      # iscsi parameters
      iscsi_parameters&.each do |key, val|
        add_update_option(args, "node.session.iscsi.#{key}", val)
      end

      iscsiadm_update_node(name, args)
    end

    #########################################################################
    # set node auth
    #########################################################################
    def set_target_auth(name, init_user:, init_pass:,
                        targ_user: nil, targ_pass: nil)

      set_target_options(name, authmethod: 'CHAP', init_user: init_user,
                         init_pass: init_pass, targ_user: targ_user,
                         targ_pass: targ_pass)
    end

    #########################################################################
    # set node timeout
    #########################################################################
    def set_target_timeout(name, timeo_s = 120)
      set_target_options(name, session_timeout: timeo_s)
    end

    #########################################################################
    # target login
    #########################################################################
    def login_target(name, **args)
      iscsiadm_retry('-m', 'node', '--targetname', name, '--login')
    end

    #########################################################################
    # target logout
    #########################################################################
    def logout_target(name, **args)
      iscsiadm('-m', 'node', '--targetname', name, '--logout')
    rescue ExecError => e
      raise e unless e.message =~ /No matching sessions found/
    end

    #########################################################################
    # delete target node
    #########################################################################
    def remove_target(name, wait: 10)
      retriable max_elapsed_time: wait do
        begin
          # remove iscsi configuration; logout first just in case
          logout_target(name)
          iscsiadm('-m', 'node', '-o', 'delete', '--targetname', name)
        rescue ExecError => e
          raise e unless e.message =~ /No records found/
        end
      end
    end

    #########################################################################
    # stat target node. check for existence
    #########################################################################
    def stat_target(name)
      iscsiadm('-m', 'node', '--targetname', name)
    rescue ExecError => e
      raise e unless e.message =~ /No records found/
    end

    #########################################################################
    # lookup scsi id for dev
    #########################################################################
    def lookup_scsi_id(dev, wait: 10)
      retriable max_elapsed_time: wait do
        x = scsi_id('--whitelisted', dev)
        out = x.out.chomp
        fail "Unable to lookup SCSI ID" if out.empty?
	      out
      end
    end

    #########################################################################
    # lookup initiator name for host
    #########################################################################
    def initiatorname
      file = '/etc/iscsi/initiatorname.iscsi'
      if File.exist? file
        # read iqn from disk
        bash('-c', "echo $(<#{file})").out.each_line do |x|
          x.match('InitiatorName=(.+)') do |n|
            return n[1]
          end
        end
      else
        fail "#{file} not found; ensure 'iscsi-initiator-utils' is installed."
      end
      fail "initiator name not found; verify #{file} contents."
    end

    def lookup_target_devinfo(iqn = nil)
      rx = %r{
              ^\[(?<bus>\S+)\]
              \s+
              (?<type>\S+)
              \s+
              (?<iqn>[^,]+)\S+
              \s+
              (
                # device assigned -- dev path and major+minor available.
                (?<dev>/dev/\S+)
                  \s*
                \[  (?<major>\d+) : (?<minor>\d+)  \]
              |
                # device not assigned -- no information available yet.
                -
              )
            }x
      info = []
      lsscsi('-tdk').out.each_line.grep(rx) do |t|
        md = Regexp.last_match
        if iqn.nil? || iqn == md['iqn']
          info.push Hash[md.names.map(&:to_sym).zip(md.captures)]
        end
      end
      info
    end

    def lookup_target_iqns(dev = nil)
      iqns = []
      lookup_target_devinfo.each do |info|
        next unless dev.nil? || info['dev'] == dev
        iqns.push info['iqn']
      end
      iqns
    end

    #########################################################################
    # lookup devs for target name
    #########################################################################
    def lookup_target_devs(name, wait: 30, settle: true, noraise: false)
      LibUdev.settle if settle
      lsscsi
      rx = %r{
            #{name}\S+
            \s+
            (
              # device assigned -- dev path and major+minor available.
              (?<dev>/dev/\S+)
                \s*
              \[  (?<major>\d+) : (?<minor>\d+)  \]
            |
              # device not assigned -- no information available yet.
              -
            )
           }x

      retriable max_elapsed_time: wait do
        x = lsscsi('-tdk')
        devs = x.out.each_line.grep(rx) do |t|
          md = Regexp.last_match
          Hash[md.names.map(&:to_sym).zip(md.captures)]
        end

        if devs.empty? || devs.any? { |ent| ent[:dev].nil? }
          fail 'device has no /dev entry assigned'
        end

        # success; return the array of devs
        devs
      end
    rescue
      raise unless noraise
    end

    #########################################################################
    # wait for all devs to populate
    #########################################################################
    def wait_for_devs(devs, wait: 30)
      retriable max_elapsed_time: wait do
        return true if devs.all? { |md| File.exist?(md[:dev]) }
        fail 'failed to wait for devices'
      end
    end

    #########################################################################
    # make device nodes
    #########################################################################
    def makedevs(devs, prefix: nil, name: nil)
      if prefix
        mkdir('-p', File.join('/', 'dev', prefix))
      end

      devs.each_with_index do |md, i|
        if name
          if devs.length == 1
            md[:dev] = File.join(File.dirname(md[:dev]), name)
          else
            md[:dev] = File.join(File.dirname(md[:dev]), "#{name}.path#{i}")
          end
        end

        if prefix
          # update the dev node path to include the 'prefix'
          md[:dev] = File.join('/', 'dev', prefix, File.basename(md[:dev]))
        end
        rm('-f', md[:dev])
        mknod(md[:dev], 'b', md[:major], md[:minor])
      end
    end
  end
end
