# Copyright (c) 2016, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'tempfile'

module BlockbridgeHost
  class LinuxSetMultipathOptions
    attr_reader :ctl
    attr_reader :conf
    attr_reader :text
    attr_reader :set_options
    attr_reader :defaults
    attr_reader :user_defaults
    attr_reader :blacklist
    attr_reader :blacklist_device
    attr_reader :user_blacklist
    attr_reader :user_blacklist_device

    MULTIPATH_CONF = '/etc/multipath.conf'
    DEFAULTS_RX = /\s*^defaults.*?{(.*?)}/m
    DEFAULTS = {
      user_friendly_names:  'no',
      find_multipaths: 'yes',
      path_grouping_policy: 'multibus',
    }
    BLACKLIST_FULL_RX = /\s*^blacklist.*?{(.*?(.*?device.*?{(.*?)})*.*?)}/m
    BLACKLIST_DEVICE_RX = /.*?device.*?{(.*?)}/m

    class NOOP < StandardError; end

    def initialize(ctl, conf: MULTIPATH_CONF, set_defaults: false,
                   defaults: DEFAULTS,
                   set_blacklist: false, blacklist: nil, blacklist_device: nil)
      @ctl  = ctl
      @conf = conf
      @text = ''
      @set_options = {
        defaults:  set_defaults,
        blacklist: set_blacklist,
      }
      @defaults = []
      @user_defaults = defaults || []

      @blacklist = []
      @blacklist_device = []
      @user_blacklist = blacklist || []
      @user_blacklist_device = blacklist_device || []
    end

    def rewrite
      read
      process
      write
    rescue NOOP
    end

    private

    def set_defaults?
      set_options[:defaults]
    end

    def set_blacklist?
      set_options[:blacklist]
    end

    def filter_defaults(s)
      user_defaults.keys.any? do |k|
        s.start_with? k.to_s
      end
    end

    def has_all_defaults?
      user_defaults.all? do |k, v|
        defaults.include? "#{k.to_s} #{v}"
      end
    end

    def read_defaults
      text.scan(DEFAULTS_RX) do |m|
        def_vals = m.first.strip.split("\n").map { |s| s.strip }
        defaults.concat def_vals
      end
    end

    def reduce_defaults
      defaults.reject! { |s| filter_defaults(s) }
    end

    def has_all_blacklist?
      user_blacklist.all? do |s|
        blacklist.include? s
      end
    end

    def has_all_blacklist_device?
      user_blacklist_device.all? do |s|
        blacklist_device.include? s
      end
    end

    def filter_blacklist(s)
      user_blacklist.any? do |b|
        s == b
      end
    end

    def filter_blacklist_device(s)
      user_blacklist_device.any? do |b|
        s == b
      end
    end

    def read_blacklist
      text.scan(BLACKLIST_FULL_RX) do |m|
        m.first.scan(BLACKLIST_DEVICE_RX) do |d|
          bl_dev_vals = d.first.strip.split("\n").map { |s| s.strip }
          blacklist_device.concat bl_dev_vals
        end
        m.first.gsub!(BLACKLIST_DEVICE_RX, '')
        bl_vals = m.first.strip.split("\n").map { |s| s.strip }
        blacklist.concat bl_vals
      end
    end

    def reduce_blacklist
      blacklist.reject! { |s| filter_blacklist(s) }
      blacklist_device.reject! { |s| filter_blacklist_device(s) }
    end

    def set_defaults
      reduce_defaults
      user_defaults.each do |k, v|
        defaults.push "#{k} #{v}"
      end
    end

    def set_blacklist
      reduce_blacklist
      user_blacklist.each do |b|
        blacklist.push b
      end
      user_blacklist_device.each do |d|
        blacklist_device.push d
      end
    end

    def process_defaults
      text.gsub!(DEFAULTS_RX, '')
      text.concat "\ndefaults {\n"
      defaults.each do |d|
        text.concat "\t#{d}\n"
      end
      text.concat "}\n"
    end

    def process_blacklist
      text.gsub!(BLACKLIST_FULL_RX, '')
      text.concat "\nblacklist {\n"
      blacklist.each do |bl|
        text.concat "\t#{bl}\n"
      end
      unless blacklist_device.empty?
        text.concat "\tdevice {\n"
        blacklist_device.each do |bld|
          text.concat "\t\t#{bld}\n"
        end
        text.concat "\t}\n"
      end
      text.concat "}\n"
    end

    def read_noop?
      (!set_defaults? || has_all_defaults?) &&
      (!set_blacklist? || (has_all_blacklist? && has_all_blacklist_device?))
    end

    def read
      if File.exist? conf
        @text = ctl.cat(conf).out
        read_defaults if set_defaults?
        read_blacklist if set_blacklist?
      end

      raise NOOP if read_noop?

      set_defaults if set_defaults?
      set_blacklist if set_blacklist?
    end

    def process
      process_defaults if set_defaults?
      process_blacklist if set_blacklist?
    end

    def write
      Tempfile.open('multipath.conf') do |f|
        f.write(text)
        ctl.mv(MULTIPATH_CONF, "#{MULTIPATH_CONF}.bak") if File.exist? MULTIPATH_CONF
        ctl.mv(f.path, MULTIPATH_CONF)
      end
    end
  end

  class LinuxMultipathControl < LinuxControl
    util :multipath, paths: ['/sbin'],
          from_package: { rpm: 'device-mapper-multipath', dpkg: 'multipath-tools' }

    # enable multipath
    def enable_multipath_service
      # set defaults
      set_multipath_defaults

      # enable
      modprobe('dm_multipath')
      begin
        chkconfig('--add', multipathd)
        chkconfig('multipathd', 'on')
      rescue
      end
      begin
        systemctl('enable', 'multipathd.service')
        systemctl('start', 'multipathd.service')
      rescue
      end

      # restart to ensure it is running after 'start'
      multipath('-r')
    end

    def restart
      multipath('-r')
    end

    # set multipath defaults
    def set_multipath_defaults
      LinuxSetMultipathOptions.new(self, set_defaults: true).rewrite
    end

    # set multipath blacklist
    def set_multipath_blacklist(blacklist: [], blacklist_device: [])
      LinuxSetMultipathOptions.new(self,
                                   set_blacklist: true,
                                   blacklist: blacklist,
                                   blacklist_device: blacklist_device).rewrite
    end

    # track device wwid for multipath
    def add_wwid(wwid)
      res = multipath('-a', wwid, noraise: true)

      # Ugh. Redhat has a broken multipath tools package
      return if res.exitstatus == 1 && res.out != /wwid.*added/
      fail "multipath add wwid failed" if res.exitstatus != 0
    end

    # lookup active paths for a given device (by wwid)
    def lookup_active_paths(dev)
      path_status(dev).select do |path|
        path[:dm_status] == 'active'
      end
    end

    def start_multipath(devs)
      devs.each do |dev|
        add_wwid(dev)
      end
      multipath('-r')
    end

    def lookup_dev(wwid)
      dev = wwid_dev_path(wwid)
      if exist?(dev)
        st = File.stat(dev)
        {
          dev: dev,
          major: st.dev_major,
          minor: st.dev_minor,
        }
      end
    end

    # wait for a dm device to exist (by wwid)
    def wait_for_dev(devs, wwid, wait: 30)
      retriable max_elapsed_time: wait do
        start_multipath(devs)
        lookup_dev(wwid) or fail "dev not found: #{wwid}"
      end
    end

    private

    # path to multipath device for a given wwid
    def wwid_dev_path(wwid)
      fail 'missing wwid' if wwid.nil?
      File.join('/dev', 'mapper', wwid)
    end

    # return array of multipath path status objects (structured translation of
    # the multipath -ll command output.)
    def path_status(dev = nil)
      # lines describing a path look like:
      # | `- 6:0:0:0 sdb 8:16  active ready  running
      #
      # Field meanings are:
      #
      #  `- host:channel:id:lun devnode major:minor
      #     dm_status_if_known path_status online_status
      #
      mp_rx = /-
               \s+
                (?<host>\d+) :
                (?<channel>\d+) :
                (?<id>\d+) :
                (?<lun>\d+)
               \s+
                (?<dev>\S+)
               \s+
                (?<major>\d+) :
                (?<minor>\d+)
               \s+
                (?<dm_status>\S+)
               \s+
                (?<path_status>\S+)
               \s+
                (?<online_status>\S+)
              /x

      mpath_out = multipath('-ll', dev).out
      wwid = mpath_out[/(^\S+)/,1]
      mpath_out.each_line.grep(mp_rx) do |line|
        md = Regexp.last_match
        Hash[md.names.map(&:to_sym).zip(md.captures)].merge(wwid: wwid)
      end
    end
  end
end
