# frozen_string_literal: true

##############################################################################
# Copyright (c) 2016-2017, Blockbridge Networks LLC.  All rights reserved.
# Proprietary and confidential. Unauthorized copying of this file via any
# medium is strictly prohibited. Use of this source code is subject to the
# terms and conditions found in the LICENSE file.
##############################################################################

require 'blockbridge/host/control/libc'

require 'digest/sha1'
require 'em-synchrony'
require 'ffi'
require 'memoist'
require 'uuidtools'
require 'absolute_time'

class LibUdev
  class Queue
    attr_reader :lib_ctx

    def initialize(lib_ctx = nil)
      @lib_ctx = lib_ctx || LibUdev::Library.udev_new
      @queue_ctx = LibUdev::Library.udev_queue_new(@lib_ctx)
    end

    def active?
      LibUdev::Library.udev_queue_get_udev_is_active(@queue_ctx)
    end

    def empty?
      LibUdev::Library.udev_queue_get_queue_is_empty(@queue_ctx)
    end

    def seqnum_is_finished(num)
      LibUdev::Library.udev_queue_get_seqnum_is_finished(@queue_ctx, num)
    end

    def sequence_is_finished(start_num, end_num)
      LibUdev::Library.udev_queue_get_seqnum_sequence_is_finished(
        @queue_ctx, start_num, end_num)
    end

    def entries
      LibUdev::Library.udev_queue_get_queued_list_entry(@queue_ctx)
    end

    def sysfs_seqnum
      File.read('/sys/kernel/uevent_seqnum').to_i
    end

    def kernel_seqnum
      LibUdev::Library.udev_queue_get_kernel_seqnum(@queue_ctx)
    end

    def udev_seqnum
      LibUdev::Library.udev_queue_get_udev_seqnum(@queue_ctx)
    end
  end

  class Monitor
    attr_reader :lib_ctx
    attr_reader :enabled
    attr_reader :filtered

    def initialize(lib_ctx = nil)
      @lib_ctx = lib_ctx || LibUdev::Library.udev_new
    end

    def add_match_tag(tag)
      r = LibUdev::Library.udev_monitor_filter_add_match_tag(mon_ctx, tag)
      raise "failed to add match tag '#{tag}' to udev monitor" if r != 0
      update_filter
      nil
    end

    def set_receive_buffer_size(size)
      r = LibUdev::Library.udev_monitor_set_receive_buffer_size(mon_ctx, size)
      r == 0
    end

    def add_match_subsystem_devtype(subsystem, devtype = nil)
      r = LibUdev::Library.udev_monitor_filter_add_match_subsystem_devtype(
        mon_ctx, subsystem, devtype)
      raise "failed to filter on subsystem/devtype" if r != 0
      update_filter
      nil
    end

    def enable_receiving
      return if @enable_receiving
      r = LibUdev::Library.udev_monitor_enable_receiving(mon_ctx)
      raise "failed to filter on subsystem/devtype" if r != 0
      @enable_receiving = true
      nil
    end

    def receive_device
      ctx = LibUdev::Library.udev_monitor_receive_device(mon_ctx)
      ctx.null? ? nil : LibUdev::Device.new_from_dev_ctx(ctx)
    end

    def to_io
      IO.for_fd(get_fd)
    end

    def watch(cb = nil, &blk)
      cb ||= blk
      raise "no cb or block specified to udev monitor" if cb.nil?

      handler = Module.new do
        define_method :notify_readable do
          return unless (dev = @monitor.receive_device)
          @cb.call(dev)
        end
      end

      @conn = EM.watch(to_io, handler).tap do |c|
        c.instance_variable_set :@monitor, self
        c.instance_variable_set :@cb, cb
        c.notify_readable = true
        enable_receiving
      end

      @enabled = true
      @conn
    end

    def close
      if @enabled
        @conn&.detach
        @enabled = false
        @conn = nil
        @mon_ctx = nil
        @enabled = false
      end
    end

    private

    def mon_ctx
      @mon_ctx ||= LibUdev::Library.udev_monitor_new_from_netlink @lib_ctx, 'udev'
      raise "failed to allocate udev monitor ctx" if @mon_ctx.null?
      @mon_ctx
    end

    def get_fd
      LibUdev::Library.udev_monitor_get_fd(mon_ctx)
    end

    def update_filter
      r = LibUdev::Library.udev_monitor_filter_update(mon_ctx)
      raise "failed to update udev monitor filter" if r != 0
    end
  end

  class Devices
    include Enumerable

    attr_reader :lib_ctx
    attr_reader :subsystem

    def initialize(lib_ctx, subsystem = nil)
      @lib_ctx = lib_ctx
      @subsystem = subsystem
    end

    def each(&blk)
      enumerate_devs.each(&blk)
    end

    private

    def setup_enum_ctx
      ectx = Library.udev_enumerate_new(lib_ctx)
      if subsystem
        if subsystem.respond_to?(:each)
          subsystem.each do |subsys|
            Library.udev_enumerate_add_match_subsystem(ectx, subsys)
          end
        else
          Library.udev_enumerate_add_match_subsystem(ectx, subsystem)
        end
      end
      ectx
    end

    def enumerate_devs
      enum_ctx = setup_enum_ctx
      Array.new.tap do |devs|
        Library.udev_enumerate_scan_devices(enum_ctx)
        ent = Library.udev_enumerate_get_list_entry(enum_ctx)
        while !ent.null?
          path = Library.udev_list_entry_get_name(ent)
          dev_ctx = Library.udev_device_new_from_syspath(lib_ctx, path)
          unless dev_ctx.null?
            devs << Device.new(path, lib_ctx, dev_ctx)
            ent = Library.udev_list_entry_get_next(ent)
          end
        end
      end
    end
  end

  class Device
    extend Memoist

    class << self
      def new_from_devnum(major, minor, lib_ctx = nil, ftype: 'b')
        lib_ctx ||= LibUdev::Library.udev_new
        devnum = LibC.makedev(major.to_i, minor.to_i)
        dev_ctx = LibUdev::Library.udev_device_new_from_devnum(lib_ctx, ftype.ord,
                                                               devnum)
        return if dev_ctx.null?

        syspath = LibUdev::Library.udev_device_get_syspath(dev_ctx)
        LibUdev::Device.new(syspath, lib_ctx, dev_ctx)
      end

      def new_from_devnode(path, lib_ctx = nil)
        fs = File.stat(path)
        ftype = case fs.ftype
                when 'characterSpecial'
                  'c'
                when 'blockSpecial'
                  'b'
                else
                  raise ArgumentError, "'#{path}' is not a device (type=#{fs.ftype})"
                end
        new_from_devnum(fs.rdev_major, fs.rdev_minor, lib_ctx, ftype: ftype)
      end

      def new_from_syspath(path, lib_ctx = nil)
        lib_ctx ||= LibUdev::Library.udev_new
        dev_ctx = LibUdev::Library.udev_device_new_from_syspath(lib_ctx, path)
        return if dev_ctx.null?
        !dev_ctx.null? && LibUdev::Device.new(path, lib_ctx, dev_ctx)
      end

      def new_from_dev_ctx(dev_ctx)
        return if dev_ctx.null? || dev_ctx.nil?
        syspath = LibUdev::Library.udev_device_get_syspath(dev_ctx)
        lib_ctx = LibUdev::Library.udev_device_get_udev(dev_ctx)
        LibUdev::Device.new(syspath, lib_ctx, dev_ctx)
      end
    end

    attr_reader :syspath
    attr_reader :dev_ctx
    attr_reader :lib_ctx

    def initialize(sysp, lib_ctx = nil, dev_ctx = nil)
      @lib_ctx = lib_ctx || Library.udev_new
      @dev_ctx = dev_ctx || Library.udev_device_new_from_syspath(@lib_ctx, sysp)
      @syspath = sysp
    end

    def action
      LibUdev::Library.udev_device_get_action(dev_ctx)
    end
    memoize :action

    def seqnum
      LibUdev::Library.udev_device_get_seqnum(dev_ctx)
    end
    memoize :seqnum

    def subsystem
      LibUdev::Library.udev_device_get_subsystem(dev_ctx)
    end
    memoize :subsystem

    def devpath
      LibUdev::Library.udev_device_get_devpath(dev_ctx)
    end
    memoize :devpath

    def devtype
      LibUdev::Library.udev_device_get_devtype(dev_ctx)
    end
    memoize :devtype

    def devnode
      LibUdev::Library.udev_device_get_devnode(dev_ctx)
    end
    memoize :devnode

    def sysname
      LibUdev::Library.udev_device_get_sysname(dev_ctx)
    end
    memoize :sysname

    def sysnum
      LibUdev::Library.udev_device_get_sysnum(dev_ctx)
    end
    memoize :sysnum

    def devlinks
      LibUdev::Library.udev_device_get_devlinks_list_entry(dev_ctx).to_a
    end
    memoize :devlinks

    def tags
      LibUdev::Library.udev_device_get_tags_list_entry(dev_ctx).to_a
    end
    memoize :tags

    # retrieve the list of sysfs attributes for this device. ignore the 'uevent'
    # and events* attributes.
    def sysattr
      LibUdev::Library.udev_device_get_sysattr_list_entry(dev_ctx).reject do |a|
        a == 'uevent' || a.start_with?('events')
      end
    end
    memoize :sysattr

    def devnum
      LibUdev::Library.udev_device_get_devnum(dev_ctx)
    end
    memoize :devnum

    def devnum_s
      "#{devnum_major}:#{devnum_minor}"
    end
    memoize :devnum_s

    def devnum_major
      LibC.major(devnum)
    end
    memoize :devnum_major

    def devnum_minor
      LibC.minor(devnum)
    end
    memoize :devnum_minor

    def encode_property_value(val)
      val.encode(Encoding::UTF_8)
    rescue
      "binary-0x#{val.unpack('H*').first}"
    end

    def properties
      Hash.new.tap do |h|
        vals = LibUdev::Library.udev_device_get_properties_list_entry(dev_ctx)
        vals.each { |k, v| h[k.downcase.to_sym] = encode_property_value(v) }
      end
    end
    memoize :properties

    def syspath_join(*p)
      File.join(syspath, *p)
    end

    def generic_dev_path
      syspath_join('device', 'generic')
    end
    memoize :generic_dev_path

    def generic_dev
      return unless Dir.exists? generic_dev_path
      self.class.new_from_syspath generic_dev_path
    end

    def sas_address
      sysfs_attr('device', 'sas_address')
    end
    memoize :sas_address

    def rotational
      sysfs_bool_attr('queue', 'rotational')
    end
    memoize :rotational

    def size
      @size ||= (s = sysfs_attr('size')) && (s.to_i * 512)
    end

    def sysfs_attr(*p)
      Library.udev_device_get_sysattr_value(dev_ctx, File.join(*p))
    end

    def sysfs_bool_attr(*p)
      sysfs_attr(*p) == '1'
    end

    def has_path?(path)
      devnode == path || devlinks.include?(path)
    end

    def bool_property(key)
      case properties[key]
      when nil, false, 0, '0', 'false', 'FALSE', 'no', 'NO'
        false
      else
        true
      end
    end

    def type
      if properties.key? :id_cdrom
        'cdrom'
      else
        properties[:devtype]
      end
    end

    def mounted?
      File.open('/proc/mounts').each do |m|
        fs_name, _ = m.split(' ', 2)
        fs_name = File.realpath(fs_name) rescue nil
        next if fs_name.nil?

        return true if fs_name == devnode
      end
      false
    rescue => e
      puts "failed to check mount status for #{devnode}: #{e.message}"
      false
    end

    def bb_reserved?
      bool_property(:blockbridge_reserved)
    end

    def swap?
      properties[:id_fs_type] == 'swap'
    end

    def lvm_member?
      properties[:id_fs_type] == 'LVM2_member'
    end

    def udisks_hidden?
      (properties.key?(:udisks_presentation_hide) ||
       properties.key?(:udisks_hide))
    end

    def bb_whitelist?
      bool_property(:blockbridge_whitelist) && 'udev'
    end

    def gpt_name_whitelist?
      /^blockbridge/i.match(properties[:id_part_entry_name]) && 'part-name'
    end

    def gpt_type_whitelist?
      (properties[:id_part_entry_type] ==
       Blockbridge::PartitionTable::BB_PART_TYPE_ID) && 'part-type'
    end

    def gpt_whitelist?
      ((properties[:devtype] == 'partition' &&
        properties[:id_part_entry_scheme] == 'gpt') &&
       gpt_name_whitelist? || gpt_type_whitelist?)
    end

    def whitelisted?
      bb_whitelist? || gpt_whitelist?
    end

    def md?
      properties.key? :md_metadata
    end

    def dm?
      properties.key? :dm_uuid
    end

    def holders_path
      syspath_join('holders')
    end
    memoize :holders_path

    def slaves_path
      syspath_join('slaves')
    end
    memoize :slaves_path

    def uevent_path
      syspath_join('uevent')
    end

    def md_uuid
      return unless properties[:md_uuid]
      UUIDTools::UUID.parse_hexdigest(properties[:md_uuid].gsub(':', '')).to_s
    end

    def md_devname
      properties[:md_devname]
    end

    def id_wwn
      properties[:id_wwn_with_extension] || properties[:id_wwn]
    end

    def id_vendor
      properties[:id_vendor]
    end

    def id_model
      properties[:id_model]
    end

    def id_revision
      properties[:id_revision]
    end

    def id_scsi_serial
      properties[:id_scsi_serial]
    end

    def id_serial_short
      properties[:id_serial_short]
    end

    def id_gpt_uuid
      properties[:id_part_entry_uuid]
    end

    def uuid
      properties[:blockbridge_uuid] || id_gpt_uuid || md_uuid
    end

    def dm_uuid
      properties[:dm_uuid]
    end

    def default_label
      md_devname || sysname
    end

    def to_h
      {
        syspath:    syspath,
        devpath:    devpath,
        devtype:    devtype,
        devnode:    devnode,
        sysname:    sysname,
        sysnum:     sysnum,
        devlinks:   devlinks,
        tags:       tags,
        properties: properties,
        rotational: rotational,
      }
    end

    def parent_syspath
      pctx = Library.udev_device_get_parent(dev_ctx)
      return if pctx.null? || pctx.nil?
      LibUdev::Library.udev_device_get_syspath(pctx)
    end
    memoize :parent_syspath
  end

  def self.devices(subsystem = nil, lib_ctx = nil)
    Devices.new(lib_ctx || Library.udev_new, subsystem)
  end

  attr_reader :lib_ctx

  def initialize
    @lib_ctx = Library.udev_new
  end

  def devices(subsystem = nil)
    Devices.new(@lib_ctx, subsystem)
  end

  class Settle < EM::DefaultDeferrable
    attr_reader :res
    attr_reader :error
    attr_reader :dev
    attr_reader :monitor
    attr_reader :queue
    attr_reader :timer
    attr_reader :start_time

    def initialize(timeout: 30, dev: nil, lib_ctx: nil, exit_if_exists: nil, **)
      @lib_ctx = lib_ctx || LibUdev::Library.udev_new
      @queue = LibUdev::Queue.new(@lib_ctx)
      @monitor = LibUdev::Monitor.new(@lib_ctx)
      @timer = nil
      @start_time = nil
      @res = OpenStruct.new
      @error = nil
      @goal_seqnum = 0
      @if_exists = exit_if_exists
      @timeout = timeout
      @dev = dev ? Device::new_from_devnode(dev, @lib_ctx) : nil
    end

    def wait
      @start_time = AbsoluteTime.now
      @res.wait_time = 0
      @res.event_count = 0

      timeout(@timeout)
      callback(&method(:handle_success))
      errback(&method(:handle_failure))

      if settled?
        succeed(false)
      else
        activate_monitor
        trigger_change_on_dev unless dev.nil?
        periodically_check_status
      end

      self
    end

    private

    def settled?
      queue.empty? || file_condition_met?
    end

    def file_condition_met?
      @if_exists && File.exist?(@if_exists)
    end

    def handle_success(waited = false)
      timer&.cancel
      monitor.close
      @res.file_condition_met = file_condition_met?
      @res.queue_empty = queue.empty?
      @res.wait_time = AbsoluteTime.now - start_time if waited
      @res = @res.to_h
      set_deferred_status :succeeded, self
    end

    def handle_failure(e = Timeout::Error.new('timeout waiting for events'))
      timer&.cancel
      monitor.close
      @error = e
      set_deferred_status :failed, self
    end

    def activate_monitor
      monitor.add_match_subsystem_devtype(dev.subsystem, dev.devtype) if dev
      monitor.watch do |ctx|
        @res.event_count = @res.event_count + 1
        succeed(true) if ctx.seqnum >= @goal_seqnum || settled?
      end

      @goal_seqnum = queue.sysfs_seqnum
    end

    def trigger_change_on_dev
      # generate a change event for the device to ensure *any* event arrives
      File.open(@dev.uevent_path, 'w') { |fh| fh.puts 'change' }
    end

    def periodically_check_status
      @timer = EM.add_periodic_timer(0.1) do
        succeed(true) if settled?
      end
    end

  end

  def asettle(**kwargs)
    Settle.new(**kwargs).wait
  end

  def settle(**kwargs)
    ctx = EM::Synchrony.sync(asettle(**kwargs))
    raise ctx.error unless ctx.error.nil?
    ctx.res
  end

  def self.settle(**kwargs)
    LibUdev.new.settle(**kwargs)
  end

  # FFI bindings
  module Library
    extend FFI::Library
    ffi_lib ['udev.so.0.5.1','udev.so.1', 'libudev.so.1']

    class LibCtx < FFI::AutoPointer
      def self.release(ptr)
        Library.udev_unref(ptr)
      end
    end

    class EnumerateCtx < FFI::AutoPointer
      def self.release(ptr)
        Library.udev_enumerate_unref(ptr)
      end
    end

    class DeviceCtx < FFI::AutoPointer
      def self.release(ptr)
        Library.udev_device_unref(ptr)
      end
    end

    class MonitorCtx < FFI::AutoPointer
      def self.release(ptr)
        Library.udev_monitor_unref(ptr)
      end
    end

    class QueueCtx < FFI::AutoPointer
      def self.release(ptr)
        Library.udev_queue_unref(ptr)
      end
    end

    def self.attach_function(*params)
      if params.length == 3
        ruby_name, args, returns = params
        ffi_name = "udev_#{ruby_name.to_s}"
        super(ffi_name, args, returns)
      else
        ffi_name = params[0]
        super
      end
    rescue FFI::NotFoundError
      define_method(ffi_name) do |*params, &blk|
        raise "#{ffi_name} not defined on this platform"
      end
      module_function ffi_name
    end

    def self.enum_function(ruby_name, args, returns)
      attach_function("enumerate_#{ruby_name}", args, returns)
    end

    def self.list_ent_function(ruby_name, args, returns)
      attach_function("list_entry_#{ruby_name}", args, returns)
    end

    def self.dev_function(ruby_name, args, returns)
      attach_function("device_#{ruby_name}", args, returns)
    end

    def self.dev_alloc_function(ruby_name, args)
      dev_function(ruby_name, args, DeviceCtx)
    end

    def self.list_function(lib_name, args, pairs = false)
      raw_name = "#{lib_name}_raw"
      attach_function(raw_name, lib_name, args, :pointer)
      private raw_name

      # wrap the raw udev function with a method which returns an enumerator.
      define_method(lib_name) do |*params, &blk|
        entry = self.send(raw_name, *params)
        if pairs
          list_to_pair_enumerator(entry, &blk)
        else
          list_to_enumerator(entry, &blk)
        end
      end
      module_function lib_name
    end

    def self.dev_list_function(name, args, pairs = false)
      list_function("udev_device_#{name}", args, pairs)
    end

    def self.monitor_function(ruby_name, args, returns)
      attach_function("monitor_#{ruby_name}", args, returns)
    end

    def self.monitor_alloc_function(ruby_name, args)
      monitor_function(ruby_name, args, MonitorCtx)
    end

    def self.queue_function(ruby_name, args, returns)
      attach_function("queue_#{ruby_name}", args, returns)
    end

    def self.queue_alloc_function(ruby_name, args)
      queue_function(ruby_name, args, QueueCtx)
    end

    def self.queue_list_function(name, args, pairs = false)
      list_function("udev_queue_#{name}", args, pairs)
    end

    def self.list_to_pair_enumerator(entry)
      return to_enum(:list_to_pair_enumerator, entry) unless block_given?
      while !entry.null?
        k = unescape_hex(Library.udev_list_entry_get_name(entry))
        v = unescape_hex(Library.udev_list_entry_get_value(entry))
        yield k, v
        entry = udev_list_entry_get_next(entry)
      end
    end

    def self.list_to_enumerator(entry)
      return to_enum(:list_to_enumerator, entry) unless block_given?
      while !entry.null?
        k = unescape_hex(Library.udev_list_entry_get_name(entry))
        yield k
        entry = udev_list_entry_get_next(entry)
      end
    end

    def self.unescape_hex(str)
      str.gsub(/\\x(\h{2})/) do |c|
        Regexp.last_match[1].to_i(16).chr(str.encoding)
      end
    end

    def self.udev_device_get_ancestry(dev_ctx)
      return to_enum(:udev_device_get_ancestry, dev_ctx) unless block_given?

      loop do
        dev_ctx = Library.udev_device_get_parent(dev_ctx)
        break if dev_ctx.null?
        next if Library.udev_device_get_subsystem(dev_ctx).nil?
        yield dev_ctx
      end
    end

    def self.udev_device_get_action(dev_ctx)
      p = udev_device_get_action_raw(dev_ctx)
      p.null? ? nil : p.read_string
    end

    # library management
    attach_function :new, [], LibCtx
    attach_function :ref, [:pointer], :pointer
    attach_function :unref, [:pointer], :void
    attach_function :get_userdata, [:pointer], :pointer
    attach_function :set_userdata, [:pointer, :pointer], :void

    # device enumeration
    enum_function :new, [:pointer], EnumerateCtx
    enum_function :ref, [:pointer], :pointer
    enum_function :unref, [:pointer], :void
    enum_function :add_match_subsystem, [:pointer, :string], :int
    enum_function :add_nomatch_subsystem, [:pointer,:string],:int
    enum_function :add_match_sysattr, [:pointer, :string], :int
    enum_function :add_nomatch_sysattr, [:pointer, :string], :int
    enum_function :add_match_property, [:pointer, :string, :string], :int
#    enum_function :add_match_tag, [:string], :int
#    enum_function :add_match_parent, [:pointer], :int
#    enum_function :add_match_is_initialized, [:pointer], :int
    enum_function :add_match_sysname, [:pointer, :string], :int
    enum_function :add_syspath, [:pointer, :string], :int
    enum_function :scan_devices, [:pointer], :int
    enum_function :scan_subsystems, [:pointer], :int
    enum_function :get_list_entry, [:pointer], :pointer

    # device access
    dev_alloc_function :new_from_syspath, [:pointer, :string]
    dev_alloc_function :new_from_devnum, [:pointer, :char, :uint32]
    dev_alloc_function :new_from_subsystem_sysname, [:pointer, :string, :string]
#    dev_alloc_function :new_from_environment, [:pointer, :string]

    dev_list_function :get_devlinks_list_entry, [:pointer]
    dev_list_function :get_tags_list_entry, [:pointer]
    dev_list_function :get_sysattr_list_entry, [:pointer]
    dev_list_function :get_properties_list_entry, [:pointer], true

    dev_function :ref, [:pointer], :pointer
    dev_function :unref, [:pointer], :void
    dev_function :get_udev, [:pointer], :pointer
    dev_function :get_parent, [:pointer], :pointer
    dev_function(:get_parent_with_subsystem_devtype,
                 [:pointer, :string, :string], :pointer)
    dev_function :get_devpath, [:pointer], :string
    dev_function :get_subsystem, [:pointer], :string
    dev_function :get_devtype, [:pointer], :string
    dev_function :get_syspath, [:pointer], :string
    dev_function :get_sysname, [:pointer], :string
    dev_function :get_sysnum, [:pointer], :string
    dev_function :get_devnode, [:pointer], :string
#    dev_function :get_is_initialized, [:pointer], :bool
    dev_function :get_property_value, [:pointer, :string], :string
    dev_function :get_driver, [:pointer], :string
    dev_function :get_devnum, [:pointer], :ulong_long
    dev_function :get_sysattr_value, [:pointer, :string], :string
    dev_function :get_seqnum, [:pointer], :ulong_long
#    dev_function :get_usec_since_initialized, [:pointer], :ulong_long
#    dev_function :has_tag, [:pointer, :string], :bool
    attach_function :udev_device_get_action_raw, :udev_device_get_action,
                    [:pointer], :pointer

    # list traversal
    list_ent_function :get_next, [:pointer], :pointer
    list_ent_function :get_by_name, [:pointer, :string], :pointer
    list_ent_function :get_name, [:pointer], :string
    list_ent_function :get_value, [:pointer], :string

    # event monitoring
    monitor_alloc_function :new_from_netlink, [:pointer, :string]
    monitor_function :enable_receiving, [:pointer], :int
    monitor_function :get_fd, [:pointer], :int
    monitor_function :receive_device, [:pointer], DeviceCtx
    monitor_function(:filter_add_match_subsystem_devtype,
                     [:pointer, :string, :string], :int)
    monitor_function :filter_add_match_tag, [:pointer, :string], :int
    monitor_function :filter_update, [:pointer], :int
    monitor_function :filter_remove, [:pointer], :int
    monitor_function :set_receive_buffer_size, [:pointer, :int], :int

    # queue access
    queue_alloc_function :new, [:pointer]
    queue_function :ref, [:pointer], :void
    queue_function :unref, [:pointer], :void
    queue_function :get_udev, [:pointer], :pointer
    queue_function :get_udev_is_active, [:pointer], :bool
    queue_function :get_queue_is_empty, [:pointer], :bool
    queue_function :get_seqnum_is_finished, [:pointer, :ulong_long], :bool
    queue_function :get_seqnum_sequence_is_finished, [:pointer, :ulong_long, :ulong_long], :bool
    queue_function :get_kernel_seqnum, [:pointer], :ulong_long
    queue_function :get_udev_seqnum, [:pointer], :ulong_long
    queue_list_function :get_queued_list_entry, [:pointer], true
  end
end
