# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'blockbridge/host/control/generic'
require 'blockbridge/host/control/windows_iscsi'
#require 'blockbridge/host/control/windows_multipath'

module BlockbridgeHost
  class WindowsControl < GenericControl
    def initialize(**kwargs)
      @options ||= kwargs

      if @options[:winrm_endpoint]
        @executor ||= RemoteWindowsExecutor.new(**@options)
      else
        @executor ||= WindowsExecutor.new(**@options)
      end
      super
    end

    def linux?
      false
    end

    def windows?
      true
    end

    def verify_remotes(target, remotes)
      target_ifnames = target.portals.map { |p| p.interface }
      return target_ifnames unless remotes
      remotes.each do |ifa|
        unless remote_ifdevs.include?(ifa)
          fail "Remote interface '#{ifa} not found."
        end
      end
      remotes
    end

    def verify_locals(locals)
      unless locals
        return local_ipaddrs { |ifa| ifa.addr.afamily == Socket::AF_INET }
      end

      unk_ips = (locals - local_ipaddrs)
      return locals if unk_ips.empty?

      fail "Local IP address '#{unk_ips.first}' not found."
    end

    # configure the Windows initiator to connect to a remote blockbridge target.
    def target_attach(target, **conn_config)
      # verify locals selection, if any. in windows, we only support binding to
      # local interfaces by IP address.
      conn_config[:locals] = verify_locals(conn_config[:locals])

      # determine the path(s) to use to access the target
      paths = compute_target_paths(target, **conn_config)

      # configure the first (possibly only) path by creating target
      # configuration.
      path0 = paths.first

      debug_puts "iscsi: creating target #{path0[:remote].addr}, #{target.iqn}"
      iscsi.create_target(target.iqn, path0[:remote].addr, path0[:remote].port,
                          iface: path0[:local])

      # set auth on target
      iscsi.set_target_auth(target.iqn, **conn_config[:chap])

      if conn_config[:multipath]
        # set timeout for multipath
        iscsi.set_target_timeout(target.iqn, 10)
      end

      # attempt to login to the target
      debug_puts "iscsi: logging in to #{target.iqn}"
      iscsi.login_target(target.iqn)

      # create additional connections as needed for multipathing
      paths[1..-1].each do |path|
        iscsi.add_target_connection(target.iqn,
                                    path[:remote].addr,
                                    path[:remote].port,
                                    iface: path[:local],
                                    **conn_config[:chap])
      end

      # lookup local devs
      devs = iscsi.lookup_target_devs(target.iqn)
      if devs.length != 1
        fail "expected 1 device for #{target.iqn}, but found #{devs.length}"
      end

      dev = devs.first

      {
        dev: dev,
        dev_path: dev[:dev],
        host_dev_label: "#{hostname}~#{dev[:dev]}",
        paths: paths,
        multipath: conn_config[:multipath]
      }
    rescue
      # something went wrong -- try to tear everything down.
      iscsi.remove_target(target.iqn) rescue nil
      raise
    end

    def lookup_target_paths(targ_name, dev_path)
      iscsi.lookup_dev_connections(targ_name, dev_path)
    end

    def device_tune(dev, settings)
    end

    def device_settle
    end

    def iscsi
      @iscsi ||= WindowsISCSIControl.new(**@options).tap do |ctl|
        # before executing the first iscsicli command, ensure the MSiSCSI
        # service is running.
        ctl.enable_msiscsi_service
      end
    end

    def mpath
      @mpath ||= WindowsMultipathControl.new(**@options)
    end

    def os_desc
      "Windows"
    end

    def hostname
      @hostname ||= execute('hostname').out.strip
    end
  end
end
