# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'ipaddr'
require 'wmi-lite'

module BlockbridgeHost
  class WindowsISCSIControl < GenericControl
    util :iscsicli
    util :reg

    def enable_msiscsi_service
      x = execute('sc', 'config', 'MSiSCSI', 'start=', 'auto',
                  privileged: true, noraise: true)
      unless x.success?
        #fail 'Failed to set MSiSCSI service to automatic startup'
      end
      x = execute('net', 'start', 'MSiSCSI', privileged: true, noraise: true)
      unless x.exitstatus == 0 || x.exitstatus == 2
        #fail 'Failed to start the MSiSCSI service'
      end
    end

    #########################################################################
    # create interface. idempotent.
    #########################################################################
    def create_iface(iface)
      fail "not implemented"
    end

    #########################################################################
    # stat interface. check for existence
    #########################################################################
    def stat_iface(iface)
      fail "not implemented"
    end

    #########################################################################
    # configure a target
    #########################################################################
    def create_target(name, host, port = 3260, iface: nil)
      iin = initiator_instance_name
      src_port = src_port_for_addr(iface)

      iscsicli('AddTarget', name, '*', host, port.to_s, '*', 'T', '*', '*',
               '*', '*', '*', '*', '*', '*', '0', '0')

      iscsicli('PersistentLoginTarget', name, 'T', '*', '*', iin, src_port,
               '*', '*', '*', '*', '*', '*', '*', '*', '*', '*', '*', '*')
    end


    # generate a chap password appropriate for a Windows host (limit of 16
    # characters.)
    def gen_init_pass
      SecureRandom.base64(16).slice(0, 16)
    end

    def hex_encode(str)
      '0x' + str.unpack('H*').first
    end

    #########################################################################
    # set target auth
    #########################################################################
    def set_target_auth(name, init_user:, init_pass:, **kwargs)
      iscsicli('AddTarget', name, '*', '*', '*', '0x4', 'T', '*', '*',
               '*', '*', '*', '*', init_user, hex_encode(init_pass), '1', '0')
    end

    #########################################################################
    # set target timeout
    #########################################################################
    def set_target_timeout(name, timeo_s = 120)

    end

    # Logs in to a previously defined iSCSI target.
    def login_target(name, iface: nil)
      src_port = src_port_for_addr(iface)
      iin = initiator_instance_name
      iscsicli('LoginTarget', name, 'T', '*', '*', iin, src_port, '*', '*',
               '*', '*', '*', '*', '*', '*', '*', '*', '*', '*')
    end

    def lookup_dev_connections(name, dev)
      sl = sessions(name)
      session = sl.find do |s|
        s[:devices].find do |d|
          "#{d[:device_type]} #{d[:device_number]}" == dev
        end
      end
      return [] if session.nil?
      return session[:connections]
    end

    # add additional connection to an existing target session
    def add_target_connection(name, dst_ip, dst_port, init_user:, init_pass:,
                              targ_user:, targ_pass:, iface: nil)
      src_port = src_port_for_addr(iface)

      # find sessions by target iqn
      sl = sessions(name)

      fail "No sessions for target #{name}" if sl.empty?
      fail ">1 session for target #{name}" if sl.length > 1

      sess = sl.first
      iin = initiator_instance_name

      # add portal
      iscsicli('AddTargetPortal', dst_ip, dst_port, iin, src_port,
               '*', '*', '*', '*', '*', '*', '*', init_user,
               hex_encode(init_pass), '1')

      # add connection
      iscsicli('AddConnection', sess[:session_id], iin, src_port, dst_ip,
               dst_port, '*', '*', '*', '*', '*', '*', '*',
               init_user, hex_encode(init_pass), '1', '*')
    end

    #########################################################################
    # target logout
    #########################################################################
    def logout_target(name)
      sessions(name).each { |si| iscsicli('LogoutTarget', si[:session_id]) }
    end

    #########################################################################
    # delete target node
    #########################################################################
    def remove_target(name)
      if (t = stat_target(name))
        # check for a corresponding persistent targets and remove any with a
        # matching iqn.
        persistent_targets(name).each do |pt|
          iscsicli('RemovePersistentTarget', pt[:initiator_name], name,
                   pt[:port_number], pt[:host], pt[:port])
        end

        iscsicli('RemoveTarget', name)
      end
    end

    def initiator_instance_name
      'Root\ISCSIPRT\0000_0'
    end

    def persistent_targets(name = nil)
      xd = iscsicli('ListPersistentTargets')

      cur_target = nil
      chunked = xd.out.each_line.chunk do |line|
        case line
        when /Target Name\s+:\s*(\S+)$/ then cur_target = $~[1]
        when /^\s*$/ then cur_target = nil
        else cur_target
        end
      end

      pts = chunked.map { |(_, lines)| kv_lines_to_h(lines) }

      # optionally filter to a single target iqn
      name ? pts.select { |pt| pt[:target_name] == name } : pts
    end

    def targets
      xd = iscsicli('ListTargets')

      sliced = xd.out.each_line.slice_before(/(^Targets List:|^\S)/)
      ts = sliced.find { |s| s.first =~ /^Targets List/ }

      # strip leading & trailing whitespace from each target iqn
      ts[1..-1].map(&:strip)
    end

    def sessions(name = nil)
      xd = iscsicli('SessionList')

      sliced = xd.out.each_line.slice_before(/^Session Id\s+:\s*(.+)$/)
      sl = sliced.select do |s|
        # ignore output which doesn't correspond to a sequence of session lines.
        s.first =~ /^Session Id/
      end.map do |lines|
        parse_session(lines)
      end

      # optionally filter to a single target iqn
      name ? sl.select { |s| s[:target_name] == name } : sl
    end

    #########################################################################
    # stat target node. check for existence
    #########################################################################
    def stat_target(name)
      xd = iscsicli('TargetInfo', name, 'Manually Configured:')
      return if xd.out =~ /The target name is not found/

      values = kv_lines_to_h(xd.out.each_line)
      Hash.new.tap do |h|
        h[:name] = name
        if values[:address_and_socket]
          h[:host], h[:port] = values[:address_and_socket].split(' ')
        end
      end
    end

    #########################################################################
    # lookup scsi id for dev
    #########################################################################
    def lookup_scsi_id(dev, wait: 10)
      nil
    end

    #########################################################################
    # lookup initiator name for host
    #########################################################################
    def initiatorname
      ns = WmiLite::Wmi.new('root\wmi')
      things = ns.instances_of('MSiSCSIInitiator_MethodClass')
      fail "Can't determine initiator IQN" if things.empty?
      things.first['iscsinodename']
    rescue WmiLite::WmiException => e
      fail "Can't determine initiator IQN. Please check that you are running as Administrator."
    end

    #########################################################################
    # lookup devs for target name
    #########################################################################
    def lookup_target_devs(name, wait: 30, settle: false, noraise: false)
      retriable max_elapsed_time: wait do
        tgt_sess = sessions(name)
        fail "no sessions for target" if tgt_sess.empty?
        fail "waiting for devs" if tgt_sess.any? { |s| s[:devices].empty? }

        # concatenate all devices from all sessions and convert to generic
        # hash-based description.
        devs = tgt_sess.flat_map { |s| s[:devices] }
        devs.map do |dev|
          if dev[:device_number] == '-1' || dev[:device_type] == "Unknown"
            fail "dev not ready"
          end
          {
            dev:   "#{dev[:device_type]} #{dev[:device_number]}",
            major: dev[:device_number],
            minor: 0
          }
        end
      end
    rescue
      raise unless noraise
    end

    def lookup_target_devinfo(iqn = nil)
      lookup_target_devs(iqn, noraise: true) || []
    end

    #########################################################################
    # wait for all devs to populate
    #########################################################################
    def wait_for_devs(devs, wait: 30)
      true
    end

    #########################################################################
    # make device nodes
    #########################################################################
    def makedevs(devs, wait: 30, prefix: nil)
      fail "Creating device files is not required on Windows"
    end

    private

    def src_port_for_addr(src_addr = nil)
      return '*' unless src_addr
      info = local_ports.find { |p| p[:ipv4_addr] == src_addr }
      fail "Can't determine port for local address #{src_addr}" unless info
      info[:number]
    end

    def local_ports
      ns = WmiLite::Wmi.new('root\wmi')

      info = ns.instances_of('MSiSCSI_PortalInfoClass')
      info.flat_map do |o|
        pi = o['portalinformation'].map { |o| WmiLite::Wmi::Instance.new(o) }
        pi.map do |info|
          ip = WmiLite::Wmi::Instance.new(info['ipaddr'])
          ipv4_raw = ip['ipv4address']
          ipv4_str = [0, 8, 16, 24].map { |s| (ipv4_raw >> s) & 0xff }.join('.')
          ipv6_str = ip['ipv6address'].pack('C*').unpack('n*').
                     map { |p| "%x" % p }.join(':')

          {
            number: info['port'],
            ipv4_addr: ipv4_str,
            ipv6_addr: ipv6_str,
          }
        end
      end
    rescue WmiLite::WmiException => e
      fail "Can't determine local IP portals. Please check that you are running as Administrator."
    end

    def parse_kv_line(line)
      if (md = /(?<key>(?:\w|\s)+)\s*:\s*(?<val>.*)$/.match(line))
        key = md['key'].downcase.strip.gsub(/\s/, '_')
        val = md['val']

        # special handling for some common value formats.
        val = case val
              when /^\d+$/ then val.to_i
              when '(null)' then nil
              when '<is established>' then true
              when '<Any Port>' then '*'
              else val
              end

        [ key.to_sym, val ]
      end
    end

    def kv_lines_to_h(lines)
      Hash[ lines.map { |line| parse_kv_line(line) }.compact ].tap do |h|
        if h[:address_and_socket]
          h[:host], h[:port] = h[:address_and_socket].split(' ')
        end
      end
    end

    def parse_connections(lines)
      lines.slice_before(/Connection Id\s+:/).map do |lines|
        kv_lines_to_h(lines)
      end
    end

    def parse_devices(lines)
      lines.slice_before(/Device Type:/).map do |lines|
        kv_lines_to_h(lines)
      end
    end

    def parse_session(lines)
      # chunk by information source
      section = 'session'
      groups = lines.chunk do |line|
        if (md = /^\s*(Connections|Devices):$/.match(line))
          section = md[1].downcase and next nil
        end
        section
      end

      groups.inject({connections:[], devices: []}) do |memo, (sect, lines)|
        lines.delete("\n")
        case sect
        when 'session'     then memo.merge!(kv_lines_to_h(lines))
        when 'connections' then memo[:connections] = parse_connections(lines)
        when 'devices'     then memo[:devices] = parse_devices(lines)
        end
        memo
      end
    end
  end
end
