# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'shellwords'
require 'posix/spawn'
require 'ostruct'
require 'open3'

module BlockbridgeHost
  def self.create_executor(for_os: BlockbridgeHost.detect_os, remote: nil)
    case host_os
    when :linux, :unix, :macosx
      UnixExecutor.new
    when :windows
      WindowsExecutor.new
    else
      fail "Don't know how to run commands on host os '#{host_os}'."
    end
  end

  class ExecError < StandardError
    attr_reader :exec_ctx

    def initialize(exec_ctx)
      @exec_ctx = exec_ctx
    end

    def exitstatus
      exec_ctx.exitstatus
    end

    def stdout
      exec_ctx.out
    end

    def stderr
      exec_ctx.err
    end

    def message
      exec_ctx.exception&.message || super
    end
  end

  # Abstraction for running commands on the local host operating system.
  class GenericExecutor
    # Platform independent bundle of execution status.
    ExecResults = Struct.new(:command, :options, :out, :err, :exitstatus) do
      attr_accessor :start_time, :end_time
      attr_accessor :exception

      def duration
        end_time - start_time if start_time && end_time
      end

      def success?
        exitstatus == 0
      end

      def to_s
        parts = ["cmd='#{command.join(' ')}'"]
        parts.push("opts=#{options.inspect}") unless options.empty?
        parts.push("dur=#{duration}s")
        parts.push(success? ? 'success' : "failed [sts=#{exitstatus}]")
        parts.push("exception=#{exception}") if exception
        parts.join('; ')
      end
    end

    attr_reader :last_execution
    attr_accessor :options

    def initialize(**kwargs)
      @options = kwargs
    end

    # Execute a single command and capture the output and exit status.
    def execute(*command, **kwargs)
      if command.length == 1 && command.first.respond_to?(:to_ary)
        command = command.first
      end

      start_time = Time.now
      command.flatten!
      command.compact!
      command.map!(&:to_s)

      begin
        @last_execution = xd =
        if kwargs[:privileged] &&
           (!user_privileged? || kwargs[:privileged] == :force)
          run_elevated(command, **kwargs)
        else
          run_plain(command, **kwargs)
        end
      rescue => e
        xd = ExecResults.new(command, kwargs, nil, nil, nil)
        xd.exception = e
      end

      xd.start_time = start_time
      xd.end_time = Time.now

      if option(kwargs, :debug)
        STDERR.puts "# HOST EXEC: #{xd}"
        if option(kwargs, :verbose)
          xd.out && xd.out.each_line { |l| STDERR.puts "| O: #{l.chomp}" }
          xd.err && xd.err.each_line { |l| STDERR.puts "| E: #{l.chomp}" }
        end
      end

      return xd if xd.success? || option(kwargs, :noraise)

      raise ExecError.new(xd), xd.err ? xd.err.chomp : nil
    end

    # Run an array of commands, collecting all output and status. Each array
    # element should be a hash containing :args (required -- an array of command
    # arguments) and any additional (optional) command-specific options passed
    # through to the #run method) option keys.
    #
    # Additional keyword arguments passed to #run_many are passed through to
    # #run, after applying any per-command options.
    #
    # Finally, the 'noraise' option will cause a failed command execution to
    # *not* halt execution of subsequent commands in the array.
    def execute_batch(cmd_specs = [], **kwargs)
      cmd_specs.map do |cmd_spec|
        cmd_spec = { command: cmd_spec } if cmd_spec.is_a? Array
        kwargs.merge(cmd_spec)
      end.tap do |results|
        results.each do |run_spec|
          opts = run_spec.dup.tap { |o| o.delete :command }
          run_spec[:child] = execute(*run_spec[:command], **opts)
        end
      end
    end

    private

    def option(cmd_opts, opt)
      cmd_opts.fetch(opt, options[opt])
    end

    def user_privileged?
      nil
    end

    def elevate_cmd_default
      fail "Don't know how to elevate privileges."
    end

    def elevate_cmd_env
      nil
    end

    def elevate_cmd(o)
      @elevate_cmd ||= (
        option(o, :elevate_command) || elevate_cmd_env || elevate_cmd_default)
    end

    def elevate_command(command, cmd_options)
      return command if (elevate_prefix = elevate_cmd(cmd_options)).nil?
      case command
      when Array
        [ elevate_prefix ].concat(command)
      when String
        [ elevate_prefix, command ]
      else
        raise "Unsupported command class: #{command.inspect}"
      end
    end

    def run_elevated(command, **cmd_options)
      command = elevate_command(command, cmd_options)
      run_plain(command, cmd_options)
    end

    def run_plain(command, **cmd_options)
      if option(cmd_options, :run) == :system
        run_system(command, **cmd_options)
      else
        spawn(command, {})
      end
    end

    def spawn(command, options)
      res = POSIX::Spawn::Child.new(*command, options)
      ExecResults.new(command, options, res.out, res.err, res.status.exitstatus)
    end

    def run_system(command, options)
      res_out, res_err, sts = Open3.capture3(*command, options[:spawn_options] || {})
      ExecResults.new(command, options, res_out, res_err, sts)
    end
  end

  # Class for executing commands on a local unix machine.
  class UnixExecutor < GenericExecutor
    def user_privileged?
      Process.euid == 0
    end

    def elevate_cmd_default
      if (env_value = ENV['BLOCKBRIDGE_SUDO_COMMAND'])
        env_value unless env_value.empty?
      else
        '/usr/bin/sudo'
      end
    end
  end

  # Class for executing commands on a local windows machine.
  class WindowsExecutor < GenericExecutor
    def user_privileged?
      @user_privileged ||=
        run_plain(['net sessions >nul 2>&1']).success?
    end

    def elevate_cmd_default
      # can't do it? seems possibly difficult to determine & achieve.
      fail 'This command requires Administrator privileges.'
    end
  end

  # Class for executing commands on a remote windows machine via WinRM (Windows
  # Remote Management)
  class RemoteWindowsExecutor < WindowsExecutor
    attr_reader :winrm

    def initialize(**kwargs)
      require 'winrm'

      # XXX: the defaults here are insecure and only suitable for
      # debug/testing. (Due to UNENCRYPTED PLAINTEXT.)
      @winrm_endpoint    = kwargs[:winrm_endpoint]
      @winrm_auth_scheme = kwargs.fetch(:winrm_auth_scheme, :plaintext)
      @winrm_auth_params = kwargs.fetch(:winrm_auth_params, {
                                          user: kwargs[:winrm_auth_user],
                                          pass: kwargs[:winrm_auth_pass],
                                          disable_sspi: true
                                        })

      @winrm = WinRM::WinRMWebService.new(@winrm_endpoint,
                                          @winrm_auth_scheme,
                                          @winrm_auth_params)
      super
    end

    def spawn(command, options)
      res = winrm.cmd(command.join(' '))
      ExecResults.new(command, options, res.stdout, res.stderr, res[:exitcode])
    end
  end
end
