# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

describe BlockbridgeHost::UnixExecutor do
  describe '#execute' do
    subject { Object.new.extend described_class }

    let(:command)      { %w(ls -1 /) }
    let(:mock_out)     { '_XXX_STDOUT_XXX_' }
    let(:mock_err)     { '_XXX_STDERR_XXX_' }
    let(:sudo_command) { [ '/usr/bin/sudo', *command ] }

    let(:child_double) do
      status = instance_double(Process::Status, exitstatus: 0, success?: true)
      instance_double(POSIX::Spawn::Child, err: mock_err, out: mock_out,
                      status: status)
    end

    before do
      allow(POSIX::Spawn::Child).to receive(:new).and_return(child_double)
    end

    it 'captures standard output' do
      expect(subject.execute(command).out).to match(mock_out)
    end

    it 'captures standard err' do
      expect(subject.execute(command).err).to match(mock_err)
    end

    it 'supports bare string (via shell) execution' do
      expect(subject).to receive(:spawn).with(['ls -l'], {}).
                          and_return(child_double)
      subject.execute('ls -l')
    end

    it 'supports flattened argv execution' do
      expect(subject).to receive(:spawn).with(['ls', '-l'], {}).
                          and_return(child_double)
      subject.execute('ls', '-l')
    end

    it 'supports argv-array execution' do
      expect(subject).to receive(:spawn).with(['ls', '-l'], {}).
                          and_return(child_double)
      subject.execute(['ls', '-l'])
    end

    context 'when running as a non-privileged user' do
      before do
        allow(subject).to receive(:user_privileged?) { false }
      end

      it 'doesn\'t use sudo for plain execution' do
        expect(subject).to receive(:spawn).with(command, {}).
                            and_return(child_double)
        subject.execute(command, privileged: false)
      end

      it 'uses sudo to elevate privilege' do
        expect(subject).to receive(:spawn).with(sudo_command, {}).
                            and_return(child_double)
        subject.execute(command, privileged: true)
      end

      it 'uses sudo when forced' do
        expect(subject).to receive(:spawn).with(sudo_command, {}).
                            and_return(child_double)
        subject.execute(command, privileged: :force)
      end
    end

    context 'when running as a privileged user' do
      before do
        allow(subject).to receive(:user_privileged?) { true }
      end

      it 'doesn\'t use sudo for plain execution' do
        expect(subject).to receive(:spawn).with(command, {}).
                            and_return(child_double)
        subject.execute(command, privileged: false)
      end

      it 'doesn\'t use sudo to elevate privilege' do
        expect(subject).to receive(:spawn).with(command, {}).
                            and_return(child_double)
        subject.execute(command, privileged: true)
      end

      it 'uses sudo when forced' do
        expect(subject).to receive(:spawn).with(sudo_command, {}).
                            and_return(child_double)
        subject.execute(command, privileged: :force)
      end
    end
  end

  describe '#execute_batch' do
    subject { Object.new.extend described_class }

    let(:command_specs) do
      [
        { command:
            [ 'ruby', '-e', 'STDOUT.print("OUTONLY")' ] },
        { command:
            [ 'ruby', '-e', 'STDERR.print("ERRONLY")' ] },
        { command:
            [ 'ruby', '-e', 'STDOUT.print("OUT"); STDERR.print("ERR")' ] },
      ]
    end

    let(:results) do
      subject.execute_batch(command_specs)
    end

    it 'returns an equal length result array' do
      expect(results.length).to eq command_specs.length
    end

    it 'has the correct result data' do
      expect(results[0][:child].out).to eq "OUTONLY"
      expect(results[0][:child].err).to be_empty
      expect(results[1][:child].out).to be_empty
      expect(results[1][:child].err).to eq "ERRONLY"
      expect(results[2][:child].out).to eq "OUT"
      expect(results[2][:child].err).to eq "ERR"
    end

    it 'supports bare argv arrays' do
      res = subject.execute_batch([['ls']])
      expect(res[0][:child].status.success?)
    end

    it 'raises an exception on failure' do
      commands = [ { command: [ 'ls', '-?' ] } ]
      expect { subject.execute_batch(commands) }.to raise_error(RuntimeError)
    end

    it 'does not raise if :nofail is true' do
      commands = [ { command: [ 'ls', '-?' ], nofail: true } ]
      expect { subject.execute_batch(commands) }.to_not raise_error
    end
  end
end
