# frozen_string_literal: true

##############################################################################
# Copyright (c) 2020, Blockbridge Networks LLC.  All rights reserved.
# Proprietary and confidential. Unauthorized copying of this file via any
# medium is strictly prohibited. Use of this source code is subject to the
# terms and conditions found in the LICENSE file.
##############################################################################

require 'absolute_time'
require 'pathname'
require 'multi_json'
require 'uuidtools'
require 'yaml'

# Simple wrapper class to monitor and cache a file's contents. This allows for
# simple monitoring of configuration files, and keeps the messy caching/retry
# logic away from the rest of the application.
module Blockbridge
  class FileData
    BUILTIN_READERS = {
      uuid: ->(io) { UUIDTools::UUID.parse(io.read.strip) },
      json: ->(io) { MultiJson.load(io, symbolize_keys: false) },
      yaml: ->(io) { YAML.safe_load(io) }
    }.freeze

    def initialize(path, recheck_after: 1, dirty_reads: true, reader: nil, &blk)
      @path = path
      @recheck_after = recheck_after
      @dirty_reads = dirty_reads

      @value = nil
      @reader = blk || choose_reader(reader)
      @checked = 0
      @last_mtime = nil
      @read_cnt = 0
      @last_err = nil
    end

    def fresh?
      @checked + @recheck_after >= AbsoluteTime.now
    end

    def fetch(dfl = (dfl_not_passed = true; nil), &blk)
      if dfl_not_passed && blk.nil?
        value(raise_errors: true)
      elsif (v = value).nil?
        block_given? ? yield(self) : dfl
      else
        v
      end
    end

    def current_mtime
      File.stat(@path).mtime
    rescue StandardError
      nil
    end

    def file_changed?
      @last_mtime != current_mtime
    end

    def value(raise_errors: false, force: false)
      now = AbsoluteTime.now
      return @value if !force && fresh?

      mtime = current_mtime
      return @value if !force && (mtime == @last_mtime)

      read_current_value
    rescue StandardError => e
      if @last_err != e.to_s
        @last_err = e.to_s
        warn "failed to load data from #{@path}: #{@last_err}"
      end

      raise if raise_errors && (@value.nil? || !@dirty_reads)

      if @dirty_reads
        @value
      else
        @value = nil
      end
    ensure
      @checked = now
      @last_mtime = mtime
    end

    private

    def choose_reader(arg)
      arg.is_a?(Symbol) ? BUILTIN_READERS.fetch(arg) : arg
    end

    def read_current_value
      old_value = @value
      value = read_file_contents

      @read_cnt += 1
      @last_err = nil
      @value = value.freeze
    end

    def read_file_contents
      File.open(@path) do |fh|
        @reader.nil? ? fh.read.chomp : @reader.call(fh)
      end
    end
  end

  # Wrapper class around Blockbridge::FileData for monitoring a collection of
  # files. Handles the allocation/storage of FileData instances, and presents an
  # extremely stripped down hash-like API.
  class FileDataCacher
    DEFAULT_CLASSIFIER = lambda do |path|
      if path == '/proc/sys/kernel/random/boot_id'
        :uuid
      else
        case Pathname.new(path).extname
        when '.uuid' then :uuid
        when '.json' then :json
        when '.yaml' then :yaml
        end
      end
    end

    class << self
      def singleton
        @@singleton ||= self.new
      end

      def entry_for(*args)
        singleton.entry_for(*args)
      end

      def get(path)
        singleton.get(path)
      end

      def [](path)
        singleton[path]
      end

      def fetch(*args, &blk)
        singleton.fetch(*args, &blk)
      end
    end

    def initialize(entry_opts: {}, classifier: DEFAULT_CLASSIFIER)
      @data = {}
      @entry_opts = entry_opts
      @classifier = classifier
    end

    def entry_for(path)
      @data[path] ||= new_entry(path)
    end

    def get(path)
      entry_for(path).value
    end

    def [](path)
      get(path)
    end

    def fetch(path, *rest, &blk)
      entry_for(path).fetch(*rest, &blk)
    end

    def reader_for(path)
      @classifier&.call(path)
    end

    def new_entry(path)
      ::Blockbridge::FileData.new(path, reader: reader_for(path), **@entry_opts)
    end
  end
end
