# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'shellwords'

module Blockbridge
  module Util
    module Filesystem
      class DebugFs
        SUPPORTED_VERSIONS = [ '1.41.12', '1.42' ]
        LOCATIONS = [ '/sbin/debugfs' ]

        # base class instance methods & attrs
        attr_reader :block_device, :mount_point

        def initialize(dev, mnt_pt)
          @block_device = dev
          @mount_point = mnt_pt
        end

        def map_matchdata(md, &blk)
          mapped = {}
          md.names.each do |name|
            key = name.to_sym
            val = md[name]
            if /^\d+$/.match(val)
              val = val.to_i
            end

            if block_given?
              key, val = blk.call(key, val)
            end

            mapped[key] = val
          end
          mapped
        end

        # translate an absolute or relative path to a mountpoint-relative
        # path.
        def mount_rel_path(path)
          path = File.expand_path(path)
          unless path.start_with? mount_point
            raise "path '#{path} doesn't reside on filesystem mounted at #{mount_point}"
          end

          path[mount_point.length .. -1]
        end

        def bin_path
          if @bin_path.nil?
            path = LOCATIONS.find { |p| File.exists?(p) }
            raise "can't find debugfs binary" if path.nil?
            @bin_path = path
          end
          @bin_path
        end

        def check_version
          ver = version

          if @version_checked.nil?
            if SUPPORTED_VERSIONS.include?(ver)
              # all good...
            else
              STDERR.puts "WARNING: using unsupported debugfs version '#{ver}' to examine device '#{block_device}'."
            end
            @version_checked = true
          end
        end

        def run_cmd(*cmd)
          check_version

          cmd = [ bin_path, '-R', *cmd, block_device, :err => '/dev/null' ]

          output = nil
          if File.stat(block_device).readable?
            IO::popen(cmd) do |io|
              output = io.read
            end
          else
            Blockbridge::Util.popen_root(cmd) do |io|
              output = io.read
            end
          end

          unless $?.success?
            raise "debugfs command '#{cmd.join(' ')}' failed with exit status #{$?}"
          end
          if output.nil?
            raise "debugfs command '#{cmd.join(' ')}' generated no output"
          end
          output
        end

        def safe_path(path)
          Shellwords.escape(mount_rel_path(path))
        end

        def version
          if @version.nil?
            output = nil
            IO.popen([ bin_path, '-V', :err=>[:child, :out] ]) do |io|
              output = io.read
            end
            raise 'failed to determine debugfs version' if output.nil?

            md = /debugfs\s+(\S+)\s/.match(output)
            raise 'failed to find version in debugfs -V output' if md.nil?

            @version = md[1]
          end
          @version
        end

        GROUP_RX = %r{
          ^\s Group \s+ (?<num>\d+):                                # start of section & group number
          \s block \s bitmap \s at \s (?<blk_bmap>\d+),             # location of block bitmap
          \s inode \s bitmap \s at \s (?<inode_bmap>\d+),           # location of inode bitmap
          \s inode \s table \s at \s (?<inode_tab>\d+).             # location of inodes
          \s+ (?<free_blks>\d+) \s free \s blocks?,                 # free blocks in group
          \s (?<free_inodes>\d+) \s free \s inodes?,                # free inodes in group
          \s (?<used_dirs>\d+) \s used \s (?:directory|directories) # used directories in group
          (?:,\s (?<unused_inodes>\d+) \s unused \s inodes?)?.      # unused inodes (optional)
          (?:
            \s+ \[ (?<bits>[^\]]+?) \]$                             # extra bits (optional)
          )?
        }xm

        ATTR_RX = %r{
          ^(?<key>\w[^:]+?):
          \s+(?<val>.*)$
        }x

        def stats
          info = {}
          output = run_cmd('stats')

          matches = []
          output.scan(GROUP_RX) { matches << $~ } # accumulate named captures

          info[:groups] = matches.map do |md|
            # turn the inuse value into a boolean
            mapped = map_matchdata(md)
            mapped[:inuse] = !mapped[:inuse].nil?
            mapped[:bits] = mapped[:bits].split(/, ?/) if !mapped[:bits].nil?
            mapped
          end

          md = output.scan(ATTR_RX)
          info[:attrs] = md.inject({}) do |memo, attr|
            k, v = attr
            k = k.downcase.gsub(' ', '_').to_sym
            if /^\d+$/.match(v)
              v = v.to_i
            end

            # squash/translate values
            case k
            when :last_mounted_on
              v = nil if v !~ /^\//
            when :filesystem_volume_name
              v = nil if v == '<none>'
            when :filesystem_features
              v = v.split(' ')
            end

            memo[k] = v
            memo
          end

          Hashie::Mash.new(info)
        end

        # Inode: 1   Type: bad type    Mode:  0000   Flags: 0x0
        # Generation: 0    Version: 0x00000000
        # User:     0   Group:     0   Size: 0
        # File ACL: 0    Directory ACL: 0
        # Links: 0   Blockcount: 0
        # Fragment:  Address: 0    Number: 0    Size: 0
        # ctime: 0x4b718a53 -- Tue Feb  9 11:16:19 2010
        # atime: 0x4b718a53 -- Tue Feb  9 11:16:19 2010
        # mtime: 0x4b718a53 -- Tue Feb  9 11:16:19 2010
        # Size of extra inode fields: 0
        # BLOCKS:

        STAT_RX = %r{
          ^
          Inode: \s+          (?<inode>\d+)          \s+
          Type: \s+           (?<type>.+?)           \s+
          Mode: \s+           (?<mode>\d+)           \s+
          Flags: \s+          (?<flags>0x\h+)        \s*.
          Generation: \s+     (?<gen>\d+)            \s+
          Version: \s+        (?<ver>0x\h+(?::\h+)?) \s*.
          User: \s+           (?<user>\d+)           \s+
          Group: \s+          (?<group>\d+)          \s+
          Size: \s+           (?<size>\d+)           \s*.
          File.ACL: \s+       (?<facl>\d+)           \s+
          Directory.ACL: \s+  (?<dacl>\d+)           \s*.
          Links: \s+          (?<links>\d+)          \s+
          Blockcount: \s+     (?<blkcnt>\d+)         \s*.
          .*

          \n (?<allocmeth>EXTENTS|BLOCKS): \n
          (?:(?<extents>[^\n]*)\n?)
        }xm

        BLOCK_RX = %r{
          \(
            (?<v1>\d+)
            (?:- (?<v2>\d+) )?
          \)
          \s* : \s*
          (?<p1>\d+)
          (?:- (?<p2>\d+) )?
        }xm

        def stat(file_path)
          path = safe_path(file_path)

          output = run_cmd("stat \"#{path}\"")

          md = STAT_RX.match(output)
          if md.nil?
            raise RuntimeError, "failed to gather extent data for #{path}"
          end

          info = map_matchdata(md)
          info[:allocmeth].downcase!

          blocks = []
          info[:extents].scan(BLOCK_RX) { blocks << $~ }

          # map blocks to start + num
          info[:extents] = blocks.map do |bmd|
            block = map_matchdata(bmd)
            pstart = block[:p1]
            if block[:p2].nil?
              pnum = 1
            else
              pnum = block[:p2] - block[:p1] + 1
            end

            Hash[ a: pstart, n: pnum ]
          end

          Hashie::Mash.new(info)
        end

        IMAP_RX = %r{
          ^
          Inode \s+ (?<inode>\d+)
            \s is \s part \s of \s block \s group \s
          (?<group>\d+).
            \s+ located \s at \s block \s
          (?<block>\d+),
            \s offset \s
          (?<offset>0x\h+)
        }xm

        def imap(file_path)
          path = safe_path(file_path)

          output = run_cmd("imap \"#{path}\"")

          md = IMAP_RX.match(output)
          return if md.nil?

          Hashie::Mash.new(map_matchdata(md))
        end

        ICHECK_RX = %r{
          Block\s*Inode\snumber.      # header
          (?<block>\d+)\s*(?<inode>\d+)$
        }xm

        def icheck(lba)
          fs_blk = lba / (fs_stats[:attrs][:block_size] / 512)

          md = ICHECK_RX.match(run_cmd("icheck #{fs_blk}"))
          return if md.nil?

          md['inode'].to_i
        end

        NCHECK_RX = %r{
          Inode\s*Pathname.      # header
          (?<inode>\d+)\s*(?<path>.*?)$
        }xm

        def ncheck(inode)
          md = NCHECK_RX.match(run_cmd("ncheck #{inode}"))
          return if md.nil?

          md['path']
        end
      end

    end
  end
end
