# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'pp'
require "bundler"
require "rake"
require "rake/tasklib"
require 'zlib'
require 'archive/tar/minitar'
require 'erb'

include Archive::Tar

module Blockbridge; end

module Blockbridge::Util

  class BuildTask < ::Rake::TaskLib
    attr_accessor :name
    attr_accessor :options

    def initialize(name = :tarbuild, options = {})
      @name = name
      @options = options

      yield self if block_given?

      desc "Remove previously built package files"
      task(:clean => (options[:deps] || [])) do
        clean_build
      end

      desc "Build tar package"
      task(:tarbuild => [:rubybuild]) do
        tar_build
      end

      desc "Build rpm package"
      task(:rpmbuild => [:tarbuild]) do
        rpm_build
      end

      desc "Build debian package"
      task(:debbuild => [:tarbuild]) do
        deb_build
      end

      desc "Build windows zip package"
      task(:winbuild => [:tarbuild]) do
        unless File.exist?(bintarpath)
          raise "!!! failed to build WINDOWS: bin tar (#{bintarpath}) not found. tarbuild not run first?"
        end

        bdir = File.join(builddir, "#{bindir}-win")
        zfile = "#{bdir}.zip"
        pkgfile = File.join('pkg', File.basename(zfile))

        if File.exist? pkgfile
          puts "output file #{pkgfile} exists, skipping windows packaging"
        else
          puts "staging windows package in #{bdir}"

          FileUtils.rm_rf bdir, :secure => true
          FileUtils.mkdir_p("#{bdir}/resources")

          render_template('bb-wrapper.cmd.erb', File.join(bdir, 'bb.cmd'), binding)
          render_template('bb-wrapper.cmd.erb', File.join(bdir, 'bb-shell.cmd'), binding)

          tgz = Zlib::GzipReader.new(File.open(bintarpath, 'rb'))
          Minitar.unpack(tgz, File.join(bdir, 'resources'))

          # sanity check
          ex_ver = "== blockbridge-cli version #{package_version}-#{build_num}"
          out = %x[#{bdir}/bb version]
          unless $?.success?
            raise "failed to sanity-check bb wrapper script execution"
          end
          ver_line = out.lines.first
          raise "output has no lines" if ver_line.nil?
          if ver_line.chomp != ex_ver
            raise "version mismatch, expected '#{ex_ver}', got: '#{ver_line}'"
          end

          # create zip file
          puts "creating archive '#{zfile}'"
          IO.popen(['winrar', 'a', '-afzip', '-r', '-ep1', zfile, bdir]) { }

          # copy archive to pkg dir
          FileUtils.cp zfile, File.join('pkg', File.basename(zfile))

          # clean up
          FileUtils.rm_rf bdir, :secure => true
          FileUtils.rm zfile
        end
      end

      task(:rubybuild => (options[:deps] || [])) do
        ruby_build
      end
    end

    private

    def render_template(template, output, scope)
      tmpl = File.read(template)
      erb = ERB.new(tmpl, 0, "<>")
      File.open(output, "w") do |f|
        f.puts erb.result(scope)
      end
    end

    def comp_dir
      # determine components dir
      @comp_dir ||= options[:component_dir] || ENV['BUILD_COMPONENTS'] || "/dist/components"
    end

    def os_platform
      return @os_platform if @os_platform

      if RUBY_PLATFORM =~ /mingw32/
        return @os_platform = 'windows'
      end

      platform = nil
      IO.popen(["/usr/bin/lsb_release", "-is"]) {|io|
        platform = io.read.chomp
      }
      if !platform
        raise "!!! unable to determine OS platform"
      end

      @os_platform = platform
    end

    def os_release
      return @os_release if @os_release

      if RUBY_PLATFORM =~ /mingw32/
        return @os_release = 'generic'
      end

      release = nil
      IO.popen(["/usr/bin/lsb_release", "-rs"]) {|io|
        release = io.read.chomp
      }
      if !release
        raise "!!! unable to determine OS release"
      end

      @os_release = release
    end

    def os_arch
      return @os_arch if @os_arch

      @os_arch = RbConfig::CONFIG['host_cpu']
    end

    def os_dist
      @os_dist ||=
        begin
          dist = nil
          IO.popen(['/usr/lib/rpm/redhat/dist.sh']) { |io|
            dist = io.read.chomp
          }
          if !dist
            raise "!!! unable to determine rpm dist"
          end
          dist
        end
    end

    def gemspec
      return @gemspec if @gemspec

      gemspec_file =
        if options.has_key? :gemspec
          options[:gemspec]
        elsif options.has_key? :package
          options[:package] + ".gemspec"
        else
          gemspecs = Dir.glob("*.gemspec")
          raise("!!! no gemspec found") if gemspecs.length == 0
          raise("!!! multiple gemspecs found: #{gemspecs.join(' ')}") if gemspecs.length > 1
          gemspecs.first
        end

      gem = Gem::Specification.load(gemspec_file)
      raise "!!! failed to load specification from '#{gemspec_file}'" if gem.nil?

      @gemspec = gem
    end

    def package_name
      @package_name ||= options[:package] || gemspec.name
    end

    def package_version
      @package_version ||= options[:version] || ENV['BUILD_VERSION'] || gemspec.version
    end

    def package_files
      @package_files ||= options[:files] || gemspec.files
    end

    def build_num
      @build_num ||= options[:build_number] || ENV['BUILD_NUMBER'] || "1"
    end

    def ruby_version
      @ruby_version ||= options[:ruby_version] || ENV['RUBY_VERSION'] || gemspec.required_ruby_version
    end

    def bb_ruby_version
      @bb_ruby_version ||= "= #{ENV['RUBY_VERSION']}" || ruby_version
    end

    def bb_ruby_release
      unless ENV['RUBY_RPM_RELEASE'].to_s.empty?
        "-#{ENV['RUBY_RPM_RELEASE']}#{os_dist}"
      else
        ""
      end
    end

    def ruby_platform
      @ruby_platform ||= options[:ruby_platform] || RUBY_PLATFORM
    end

    def with_emb_ruby
      @with_emb_ruby ||=
        begin
          case ENV['with_emb_ruby']
          when '1', 'yes', 'true'
            true
          else
            false
          end
        end
    end

    def srcdir
      @srcdir ||= "#{package_name}-#{package_version}"
    end

    def builddir
      @builddir ||= ENV['BUILDDIR'] || File.join(Dir.pwd, 'build')
    end

    def srcdirpath
      @srcdirpath ||= File.join(builddir, srcdir)
    end

    def bindir
      @bindir ||= "#{srcdir}-#{build_num}"
    end

    def bindirpath
      @bindirpath ||= File.join(builddir, bindir)
    end

    def bintar
      if with_emb_ruby
        @bintar ||= "#{bindir}-embedded.tar.gz"
      else
        @bintar ||= "#{bindir}.tar.gz"
      end
    end

    def bintarpath
      @bintarpath ||= "pkg/#{bintar}"
    end

    def bintaremb
      @bintaremb ||= "#{bindir}-embedded.tar.gz"
    end

    def bintarpathemb
      @bintarpathemb ||= "pkg/#{bintaremb}"
    end

    def debname
      @debname ||= "#{package_name}-#{package_version}-#{build_num}.deb"
    end

    def debpkg
      @debpkg ||= "pkg/#{debname}"
    end

    def rpmname
      @rpmname ||= "#{package_name}-#{package_version}-#{build_num}.#{os_arch}.rpm"
    end

    def rpmpkg
      @rpmpkg ||= "pkg/#{rpmname}"
    end

    def get_best_ruby(ver)
      choices = []

      req_ver = ver.to_s.sub(/-?p/, '.')

      matches = Dir.glob("pkg/ruby-*-#{os_platform}-#{os_release}-#{ruby_platform}.tar.gz")
      matches.each do |m|
        md = /^ruby-([^-]*)(?:-p(\d+))?-#{os_platform}-#{os_release}-#{ruby_platform}\.tar\.gz$/.match(File.basename(m))
        raise "failed to extract ruby version from file '#{m}'" if md.nil?

        major = md[1]
        patch = md[2] || "0"
        this_ver = "#{major}.#{patch}"

        if Gem::Dependency.new('', req_ver).match?('', this_ver)
          choices.push({ f: m, m: md[1], p: md[2], v: Gem::Version.new(this_ver)})
        end
      end

      return nil if choices.length == 0

      # pick the highest version that satisfies the request
      best = choices.max_by { |ent| ent[:v] }

      unless File.exist? best[:f]
        raise "something weird happened while looking for ruby binaries"
      end

      best[:f]
    end

    #######################################################
    ### CLEAN
    #######################################################
    def clean_build
      FileUtils.rm_f bintarpath, :verbose => true
      FileUtils.rm_f debpkg, :verbose => true
      FileUtils.rm_f rpmpkg, :verbose => true
    end

    #######################################################
    ### BUILD TAR
    #######################################################
    def tar_build
      if File.exist?(bintarpath)
        puts "output file #{bintarpath} exists, not re-making."
        return
      end

      #############################
      ##### PREPARE BINARY ARCHIVE
      #############################

      if with_emb_ruby
        # ensure a precompiled ruby is available for our platform
        rubybin = get_best_ruby(ruby_version)

        if rubybin.nil?
          raise "!!! failed to find precompiled ruby (for '#{ruby_version}' on #{ruby_platform} for #{os_platform}-#{os_release})"
        end
      end

      FileUtils.rm_rf srcdirpath, :secure => true
      FileUtils.mkdir_p(srcdirpath)

      # copy package files
      package_files.each do |file|
        FileUtils.mkdir_p File.join(srcdirpath, File.dirname(file))
        FileUtils.cp file, File.join(srcdirpath, file)
      end

      # copy vendor cache if it exists
      if ENV['BUNDLE_PATH'] && File.exist?(ENV['BUNDLE_PATH'])
        FileUtils.cp_r ENV['BUNDLE_PATH'], srcdirpath
      elsif File.exist? 'bundle'
        FileUtils.cp_r "bundle", srcdirpath
      end

      if with_emb_ruby
        # include a precompiled binary ruby
        FileUtils.cp rubybin, File.join(srcdirpath, 'ruby-emb.tar.gz')
      end

      Bundler.with_clean_env do
        # build binaries from source
        if with_emb_ruby
          system('bash', '-x', 'build/build-bin.sh', srcdirpath, bindirpath, 'embedded')
        else
          system('bash', '-x', 'build/build-bin.sh', srcdirpath, bindirpath)
        end

        unless $?.success?
          raise "!!! failed to build binary paths, exited with status #{$?.exitstatus}"
        end

        # remove .git directories from tar build
        Dir.glob(File.join(bindirpath, "**", ".git")) do |path|
          FileUtils.rm_rf srcdirpath, :secure => true
        end

        # make tar from binaries
        FileUtils.mkdir_p(File.dirname(bintarpath))

        tgz = Zlib::GzipWriter.new(File.open(bintarpath, 'wb'))
        Dir.chdir(bindirpath) { Minitar.pack('.', tgz) }
      end

      # clean
      unless options[:noclean]
        FileUtils.rm_rf srcdirpath, :secure => true
        FileUtils.rm_rf bindirpath, :secure => true
      end
    end

    def rpm_spec_file_cwd
      Dir["*.spec"].each do |spec|
        return spec
      end
      nil
    end

    def rpm_spec_file
      @rpm_spec_file ||= options[:rpm_spec_file] || ENV['rpm_spec_file'] || rpm_spec_file_cwd
    end

    def rpm_buildnum_var
      @buildnum_var ||= options[:rpm_buildnum_var] || 'BB_BUILD_NUMBER'
    end

    def rpm_buildver_var
      @buildver_var ||= options[:rpm_version_var] || 'BB_VERSION'
    end

    def rpm_buildsrc_var
      @buildsrc_var ||= options[:rpm_source_var] || 'BB_SOURCE'
    end

    def rpm_buildrubyver_var
      @buildrubyver_var ||= options[:rpm_rubyversion_var] || 'BB_RUBY_VERSION'
    end

    def rpm_builddir
      @rpm_builddir ||= options[:rpm_builddir] || File.join(Dir.pwd, 'RPMBUILD')
    end

    #######################################################
    ### BUILD RPM
    #######################################################
    def rpm_build
      unless File.exist?(bintarpath)
        raise "!!! failed to build RPM: bin tar (#{bintarpath}) not found. tarbuild not run first?"
      end

      unless rpm_spec_file
        raise "No spec file specified"
      end

      puts "Using #{rpm_spec_file}"

      bpkg = "pkg/#{rpmname}"

      FileUtils.mkdir_p("#{rpm_builddir}/SOURCES")
      FileUtils.cp bintarpath, "#{rpm_builddir}/SOURCES", :verbose => true

      rpmcmd =
        [
         'rpmbuild',
         '-v',
         '--define', "_topdir #{rpm_builddir}",
         '--define', "#{rpm_buildnum_var} #{build_num}",
         '--define', "#{rpm_buildver_var} #{package_version}",
         '--define', "#{rpm_buildsrc_var} #{bintarpath}",
         '--define', "#{rpm_buildrubyver_var} #{bb_ruby_version}#{bb_ruby_release}",
         '-bb',      "#{rpm_spec_file}",
        ]

      Bundler.with_clean_env do
        system(*rpmcmd)
        unless $?.success?
          raise "!!! failed to build RPM package. rpmbuild exited with status #{$?.exitstatus}"
        end
      end

      # copy the RPM(s) out to the 'pkg' directory
      Dir.glob(File.join(rpm_builddir, 'RPMS', '**', '*.rpm')).each do |r|
        FileUtils.mkdir_p(File.dirname(r))
        FileUtils.cp r, File.join('pkg', File.basename(r)), :verbose => true
      end

      # clean
      unless options[:noclean]
        # purge RPM build directory hierarchy
        FileUtils.rm_rf rpm_builddir, :secure => true
      end
    end

    #######################################################
    ### BUILD DEB
    #######################################################
    def deb_build
      unless File.exist?(bintarpathemb)
        raise "!!! failed to build DEB: bin tar (#{bintarpathemb}) not found. tarbuild not run first?"
      end

      bpkg = "pkg/#{debname}"
      if File.exist?(bpkg)
        puts "output file #{bpkg} exists, not re-making."
        return
      end

      bver = "#{package_version}-#{build_num}"
      bdir = "build/#{package_name}-#{bver}"

      FileUtils.rm_rf bdir, :secure => true
      FileUtils.mkdir_p(bdir)
      system('tar', 'xf', bintarpathemb, '-C', bdir)

      # copy control files
      FileUtils.cp_r File.join(Dir.pwd, 'DEBIAN'), bdir

      # update version
      open(File.join(bdir, 'DEBIAN', 'control'), 'a') { |f|
        f.puts "Version: #{bver}"
      }

      # build package
      system('dpkg-deb', '-b', bdir, File.join(Dir.pwd, 'pkg'))
      unless $?.success?
        raise "!!! failed to build DEB package. dpkg-deb exited with status #{$?.exitstatus}"
      end

      # copy the DEB(s) out to the 'pkg' directory
      Dir.glob(File.join('build', '*.deb')).each do |d|
        FileUtils.mkdir_p(File.dirname(d))
        FileUtils.cp d, File.join('pkg', File.basename(d)), :verbose => true
      end

      # clean
      unless options[:noclean]
        FileUtils.rm_rf bdir, :secure => true
      end
    end

    #######################################################
    ### BUILD WINDOWS
    #######################################################
    def win_build
    end

    #######################################################
    ### ruby build embedded
    #######################################################
    def ruby_build
      return unless with_emb_ruby

      build_emb_sh = "build/build-emb-ruby.sh"
      if !File.exist?(build_emb_sh)
        raise "!!! failed to find ruby embedded ruby script: #{build_emb_sh}. in right directory?"
      end

      # version
      build_emb_ver = Gem::Requirement.parse(ruby_version).last.version
      build_emb_dir = "ruby-#{build_emb_ver}-#{os_platform}-#{os_release}-#{ruby_platform}"
      build_emb_tar = "#{build_emb_dir}.tar.gz"

      build_emb_pkg = "pkg/#{build_emb_tar}"

      # remake check
      if File.exist?(build_emb_pkg)
        puts "output file #{build_emb_pkg} exists, not re-making."
        return
      end

      # ruby build
      system(build_emb_sh, build_emb_ver, build_emb_dir, "prebuilt")
      unless $?.success?
        raise "!!! failed to build embedded ruby. exited with status #{$?.exitstatus}"
      end

      # copy to package
      FileUtils.mkdir_p(File.dirname(build_emb_pkg))
      FileUtils.cp build_emb_tar, build_emb_pkg, :verbose => true

      # clean build area
      unless options[:noclean]
        FileUtils.rm_rf build_emb_dir, :secure => true
        FileUtils.rm    build_emb_tar
      end
    end
  end
end
