# frozen_string_literal: true
##############################################################################
# Copyright (c) 2017, Blockbridge Networks LLC.  All rights reserved.
# Proprietary and confidential. Unauthorized copying of this file via any
# medium is strictly prohibited. Use of this source code is subject to the
# terms and conditions found in the LICENSE file.
##############################################################################

require 'fileutils'
require 'open3'

module RootFileUtils
  def ruby_path
    @ruby_path ||= ENV['BB_RUBY_WRAPPER'] || RbConfig.ruby
  end

  SUDO_HELPERS = <<-EOF
    def decode(s)
      Marshal.load([s].pack('H*'))
    end
  EOF

  SUDO_FILE_READ_RB = <<-EOF
    #{SUDO_HELPERS}

    filename = decode(ENV['RFU_FILENAME'])
    begin
      File.open(filename, 'r') do |f|
        loop do
          break if (buf = f.gets(nil, 4096)).nil?
          STDOUT.write(buf)
        end
      end
    rescue Errno::ENOENT
    end
  EOF

  SUDO_FILE_WRITE_RB = <<-EOF
    #{SUDO_HELPERS}

    def exec_op
      filename = decode(ENV['RFU_FILENAME'])
      written = 0
      File.open(filename, 'w') do |f|
        loop do
          break if (buf = STDIN.gets(nil, 4096)).nil?
          written += f.write(buf)
        end
      end
      [ true, written ]
    rescue => e
      [ false, e ]
    end
    STDOUT.write Marshal.dump(exec_op).unpack('H*').first
  EOF

  SUDO_OP_RB = <<-EOF
    #{SUDO_HELPERS}

    def exec_op
      klass = Object.const_get(decode(ENV['RFU_OP_CLASS']))
      method = decode(ENV['RFU_OP_METHOD'])
      args = decode(ENV['RFU_OP_ARGS'])
      res = klass.send(method, *args)
      [ true, res ]
    rescue => e
      [ false, e ]
    end
    STDOUT.write Marshal.dump(exec_op).unpack('H*').first
  EOF

  def encode_env(**kw)
    Hash.new.tap do |h|
      kw.each_pair do |k, v|
        k = "RFU_#{k.to_s.upcase}"
        h[k] = Marshal.dump(v).unpack('H*').first
      end
    end
  end

  def process_op_res(raw_res)
    ok, res = Marshal.load([raw_res].pack('H*'))
    raise res unless ok
    res
  end

  def running_as_root?
    Process.euid == 0
  end

  def direct_exec_simple_op(klass, method, *args, requires: nil, **extra_env)
    k = Object.const_get(klass)
    k.send(method, *args)
  end

  def exec_simple_op(klass, method, *args, requires: nil, **extra_env)
    if running_as_root?
      return direct_exec_simple_op(klass, method, *args, requires: requires, **extra_env)
    end

    env = encode_env(op_class: klass, op_method: method, op_args: args, **extra_env)
    args = [ env, 'sudo', '-E', ruby_path ]
    unless requires.nil?
      requires.each { |f| args.concat(['-r', f]) }
    end
    args.concat(['-e', SUDO_OP_RB])
    Open3.popen2(*args) { |i, out| i.close; process_op_res(out.read) }
  end

  def file_exist_as_root(path)
    exec_simple_op(:File, :exist?, path.to_s)
  end

  def dir_exist_as_root(path)
    exec_simple_op(:Dir, :exist?, path.to_s)
  end

  def mkpath_as_root(path)
    exec_simple_op(:FileUtils, :mkpath, path.to_s, requires: ['fileutils'])
  end

  def chown_as_root(user, group, list)
    exec_simple_op(:FileUtils, :chown, user, group, list, requires: ['fileutils'])
  end

  def chmod_as_root(perms, list)
    exec_simple_op(:FileUtils, :chmod, perms, list, requires: ['fileutils'])
  end

  def mv_as_root(from, to)
    exec_simple_op(:FileUtils, :mv, from.to_s, to.to_s, requires: ['fileutils'])
  end

  def mkdir_as_root(path)
    exec_simple_op(:Dir, :mkdir, path.to_s)
  end

  def rmdir_as_root(path)
    exec_simple_op(:Dir, :rmdir, path.to_s)
  end

  def rm_f_as_root(list)
    exec_simple_op(:FileUtils, :rm_f, list, requires: ['fileutils'])
  end

  def unlink_as_root(path)
    exec_simple_op(:File, :unlink, path.to_s)
  end

  def symlink_as_root(old_name, new_name)
    exec_simple_op(:File, :symlink, old_name.to_s, new_name.to_s)
  end

  def direct_file_read(filename)
    File.open(filename, 'r') do |fh|
      if block_given?
        yield fh
      else
        fh.read
      end
    end
  end

  # read data from a file as root; returns nil if the file does not exist.
  def read_as_root(filename)
    return direct_file_read(filename) if running_as_root?

    env = encode_env(filename: filename.to_s)
    Open3.popen2(env, 'sudo', '-E', ruby_path, '-e', SUDO_FILE_READ_RB) do |_in, out|
      if block_given?
        yield out
      else
        out.read
      end
    end

  rescue Errno::ENOENT
    nil
  end

  def direct_file_write(filename, data = '')
    File.open(filename, 'w') do |fh|
      if block_given?
        yield fh
      else
        fh.write(data)
      end
    end
  end

  # write data to a file as root; create the file if it doesn't exist.
  def write_as_root(filename, data = '')
    return direct_file_write(filename, data) if running_as_root?

    env = encode_env(filename: filename.to_s)
    output = Open3.popen2(env, 'sudo', '-E', ruby_path, '-e', SUDO_FILE_WRITE_RB) do |i, o|
      if block_given?
        yield i, o
      else
        i.write(data)
        i.close
        o.read
      end
    end
    process_op_res(output)
  end
end
