# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'naturally'
require 'blockbridge/util/color'
require 'paint'

module Blockbridge
  module Util
    class Tabular
      class Column
        attr_accessor :id
        attr_accessor :label
        attr_accessor :min
        attr_accessor :max
        attr_accessor :lpad
        attr_accessor :rpad
        attr_accessor :dat_max
        attr_accessor :width
        attr_reader   :align
        attr_reader   :color
        attr_reader   :no_trunc

        def initialize(opts = {}, idx = nil)
          if opts.respond_to? :to_str
            opts = { label: opts }
          end

          @idx      = idx
          @label    = opts.fetch(:label, '')
          @id       = opts[:id]
          @max      = opts[:max]
          @lpad     = opts[:lpad]
          @rpad     = opts[:rpad]
          @color    = opts[:color]
          @min      = opts.fetch(:min, Blockbridge::Util::Color.length(@label))
          @align    = opts.fetch(:align, :left)

          # set the no-trunc default -- for column 0, which is often the
          # 'identity' column, default to NOT truncating it.
          @no_trunc = opts.fetch(:no_trunc, idx == 0)
        end

        def align=(val)
          val = val.to_sym
          unless [:left, :right, :center].include? val
            raise ArgumentError.new("alignment value '#{val}' not supported")
          end
          @align = val
        end

        def color=(val)
          val = Array(val) unless val.nil?
          @color = val
        end

        def extra_width
          no_trunc ? 0 : [ width - min, 0].max
        end
      end

      attr_accessor :trunc_str
      attr_accessor :max_width
      attr_accessor :title
      attr_reader   :rows
      attr_accessor :sort_cols
      attr_accessor :hide

      def initialize(opts = {}, &block)
        @trunc_str = opts[:trunc_str] || " \xc2\xbb"
        @max_width = opts[:max_width] || 80
        @title     = opts[:title]
        @opts      = opts
        @rows      = []
        @max_cols  = 0
        @sort_cols = opts[:sort_cols]
        @hide      = opts[:hide]

        cols(*opts.fetch(:cols, []))
      end

      def col(spec)
        if spec.is_a? Hash
          # find by... id?
          id_sym = spec[:id].to_sym
          cfg = @cols.find { |c| c.id == id_sym }
          raise ArgumentError.new("no column with id #{id_sym}") if cfg.nil?
          cfg
        else
          @cols[spec] ||= Column.new
        end
      end

      def cols(*cfgs)
        @cols ||= []
        cfgs.each_with_index do |c, idx|
          @cols[idx] = Column.new(c, idx)
        end
        @cols
      end

      def <<(row)
        @max_cols = [ @max_cols, row.length ].max
        @rows << row
      end

      def add_row(*vals)
        self.<<(vals)
      end

      def col_idx(cspec)
        cidx = case cspec
               when Integer
                 cspec
               when String
                 @cols.find_index { |c| c.label == cspec }
               when Symbol
                 @cols.find_index { |c| c.id == cspec }
               else
                 raise ArgumentError.new("column spec has unsupported class: #{cspec.class}")
               end

        raise ArgumentError.new("can't find column '#{cspec}'") if cidx.nil?
        cidx
      end

      def sort_rows
        return if @sort_cols.nil?

        # convert sort spec into index array
        sort_idx = []

        if @sort_cols.is_a? Array
          sort_idx = @sort_cols.map { |cspec| col_idx(cspec) }
        else
          sort_idx.push col_idx(@sort_cols)
        end

        # now sort the data
        @rows.sort_by! do |o|
          sort_idx.map do |cidx|
            # use Naturally to wrap each value -- this sorts strings that have
            # embedded numbers in a way that makes more sense to humans. See
            # https://github.com/dogweather/naturally for more information.
            Naturally.normalize(o[cidx])
          end
        end

        sort_idx.each_with_index do |cidx, order|
          @cols[cidx].label += " [#{order + 1}]"
        end
      end

      def render
        return '' if @hide
        render_array.join("\n") + "\n"
      end

      def render_array
        @max_cols = [@max_cols, @cols.length].max

        # fill in missing column configs
        for nidx in 0...@max_cols do
          @cols[nidx] ||= Column.new
        end

        # possibly sort data
        sort_rows

        # scan by column, looking for max length value
        tab_width = 0
        @cols.each_with_index do |c, idx|
          max_row = @rows.max_by do |r|
            Blockbridge::Util::Color.length(r[idx].to_s)
          end
          lab_len = Blockbridge::Util::Color.length(c.label)

          if max_row
            dat_max = Blockbridge::Util::Color.length(max_row[idx].to_s)
          else
            dat_max = 0
          end

          c.dat_max = [lab_len, dat_max].max

          width = c.min
          if width < c.dat_max
            width = c.dat_max
          end
          if c.max && width > c.max
            width = c.max
          end
          c.width = width

          # set rpad/lpad defaults
          c.lpad ||= 0
          c.rpad ||= (idx == @cols.length - 1) ? 0 : 2

          tab_width += c.width + c.lpad + c.rpad
        end

        while tab_width > @max_width
          most_over = @cols.max_by(&:extra_width)

          adj = [ most_over.extra_width, tab_width - @max_width ].min

          break if adj == 0 # nothing more we can do...

          most_over.width -= adj
          tab_width -= adj
        end

        rows = []

        unless @opts[:header] == false
          # header & separator
          rows << render_row(@cols.map { |c| c.label })
          rows << render_row(@cols.map { |c| '-' * c.width })
        end

        # data
        @rows.each do |r|
          rows << render_row(r)
        end

        rows
      end

      protected

      def render_row(data)
        out = []

        @cols.each_with_index do |c, idx|
          if data[idx].nil?
            val = @opts[:nil_str] || "-"
          else
            val = data[idx].to_s
          end

          len = Blockbridge::Util::Color.length(val)

          if len > c.width
            # truncate, if needed
            val  = Blockbridge::Util::Color.truncate(val, c.width - @trunc_str.length)
            val += @trunc_str
            len  = c.width
          end

          if c.align == :center && len < c.width
            # center the text by appending 1/2 of the whitespace to the right
            # side.
            val = val + " " * ((c.width - len) / 2)
          end

          case c.align
          when :left
            val = val + " " * (c.width - len)
          when :right
            val = " " * (c.width - len) + val
          when :center
            pad_len = c.width - len
            if pad_len.odd?
              val = " " * (pad_len / 2) + val + " " * (pad_len / 2) + 1
            else
              val = " " * (pad_len / 2) + val + " " * (pad_len / 2) + 1
            end
          end

          out << " " * c.lpad + val + " " * c.rpad
        end

        out.join('')
      end

    end
  end
end
