# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

module Blockbridge::CLI::ApiOption
  def map_param(path)
    t = api_param_map[path] || api_param_map["/" + path]
    if t
      aliases = t.map { |s| Paint[s, :blue] }.join(', ')
      "Option #{aliases}"
    else
      "API parameter '#{Paint[path, :blue]}'"
    end
  end

  def api_param_map
    @combined_param_map ||= Hash.new.tap do |h|
      h.merge! instance_hash(:@api_param_map)
      h.merge! self.class.instance_hash(:@api_param_map)
    end
  end

  protected

  def apiopt(switches, type, desc, **opts, &block)
    api_param_or_option(:option, switches, type, desc, **opts, &block)
  end

  def apiparam(type, desc, **opts, &block)
    api_param_or_option(:parameter, nil, type, desc, **opts, &block)
  end

  def api_param_or_option(style, switches, type, desc, **opts, &block)
    # be sure to account for the options resulting from the current option
    # group. internally the option-specific options hash is layered on top of
    # the current option group' hash... but... that happens internally to the
    # Clamp option instance. Essentially, we mirror the same hash merge here to
    # be consistent with what's done inside Clamp.
    opts = current_option_group.merge(opts)

    param_method = (opts[:param_group] || 'api_params').to_sym
    param_ivar = "@#{param_method}".to_sym
    param_map_ivar = :@api_param_map

    # define an instance accessor method for fetching/building the
    # api_params.
    method_defined?(param_method) || define_method(param_method) do
      inst_param_spec = instance_hash(param_ivar)
      class_param_spec = self.class.instance_hash(param_ivar)

      # start with class_param_spec -- default values will be in
      # there. overlay any user-specified params.
      param_spec = class_param_spec.merge(inst_param_spec)

      # enumerate each set option and splice it into the params hash.
      param_hash = {}
      param_spec.each_pair do |_name, (opt_inst, copts)|
        cur_h = param_hash

        # calculate the ordered list of hash keys
        names = param_names(opt_inst, copts)

        names.each_with_index do |name, idx|
          name = copts[:stringify_keys] ? name.to_s : name.to_sym

          if idx == names.length - 1
            param_val = copts.fetch(
              :param_value, self.send(opt_inst.read_method)
            )

            if (cfn = copts[:param_coerce])
              cfn = method("parse_#{cfn}") if cfn.is_a? Symbol
              param_val = cfn.call(param_val)
            end

            if copts[:param_list_merge]
              if param_val.length > 0
                cur_h[name] ||= {}
                param_val.each { |pvi| cur_h[name].deep_merge!(pvi) }
              end
            elsif copts[:param_list_concat] && cur_h.key?(name)
              cur_h[name].concat(param_val)
            else
              cur_h[name] = param_val
            end
          else
            cur_h = cur_h[name] ||= {}
          end
        end
      end

      param_hash
    end

    # declare the option ref before the hook lambda, so that by the time the
    # hook gets executed, we can reach back in to determine the attribute
    # name. (with opt_inst.attribute_name)
    opt_inst = nil

    hook = lambda do |val|
      # proxy to validation block, if specified
      if block.nil?
        val
      else
        val = instance_exec(val, &block)
      end

      # keep map of specefied command parameters
      instance_hash_store(param_ivar, opt_inst.attribute_name,
                          [ opt_inst, opts ])

      sw = style == :option ? opt_inst.switches : opt_inst.name

      # here, in the command instance itself, keep a map of actually
      # specified api params
      instance_hash_store(param_map_ivar, opts[:param_path], sw)

      val
    end

    case style
    when :option
      # define the option with clamp
      opt_inst = option(switches, type, desc, **opts, &hook)
    when :parameter
      # define the parameter with clamp
      opt_inst = parameter(type, desc, **opts, &hook)
    end
    if opts.has_key? :default
      # defaults are set in the class instance, as the instance of the
      # command object has not been created yet.
      instance_hash_store(param_ivar, opt_inst.attribute_name,
                          [ opt_inst, opts ])
    end

    if opts.fetch(:param_map, true)
      # set up the default map for all defined apiopts in the class
      # instance. if/when the params are actually used, THAT map is set up in
      # the command instance object.
      instance_hash_store(param_map_ivar,
                          api_param_path(opt_inst, opts),
                          style == :option ? opt_inst.switches : opt_inst.name)
    end

    opt_inst
  end

  def apiopt_map(path, switches)
    instance_hash_store(:@api_param_map, path, switches)
  end

  def instance_hash(ivar)
    if instance_variable_defined? ivar
      instance_variable_get(ivar)
    else
      instance_variable_set(ivar, {})
    end
  end

  def instance_hash_store(ivar, key, val)
    instance_hash(ivar)[key] ||= val
  end

  def api_param_path(opt_inst, opts)
    opts[:param_path] ||= opts.fetch(:param_path_prefix, '/') +
                          param_names(opt_inst, opts).join('/')
  end

  def param_names(opt_inst, opts)
    names = opts[:param_name] || opt_inst.attribute_name
    names = [names] unless names.respond_to? :each
    names = [opts[:param_name_prefix], *names] if opts.key?(:param_name_prefix)
    names
  end
end
