# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

# load blockbridge-api gem
require "blockbridge/api"

require "fileutils"
require "pp"
require "cgi"
require "netrc"
require "highline/import"
require "moneta"
require "socket"

module Blockbridge::CLI
  class Auth
    attr_reader :opts
    attr_reader :client_opts
    attr_reader :credentials

    def initialize(opts = {}, client_opts = {})
      @opts = opts
      @client_opts = client_opts
      if opts[:force_reauth]
        delete_credentials
      end
    end

    def noninteractive?
      opts[:noninteractive]
    end

    def puts_i(*args)
      puts(*args) unless noninteractive?
    end

    def print_i(*args)
      print(*args) unless noninteractive?
    end

    def api
      @api ||= Blockbridge::Api::Client.new_oauth(api_key, default_api_params,
                                                  client_opts)
    end

    def login
      delete_credentials
      get_credentials
    end

    def logout
      delete_credentials
    end

    # just a stub; will raise if not authenticated
    def check
      api.status.authorization
    end

    def api_url
      @opts[:api_url]
    end

    def reauthorize
      @credentials = ask_for_and_save_credentials
    end

    def api_user    # :nodoc:
      get_credentials[0]
    end

    def api_key     # :nodoc:
      get_credentials[1]
    end

    def info
      api.status.authorization
    end

    def get_api_key(user, password)
      # look up an api key using a username & password. currently, this is
      # merely a passthrough to session_create, and the so-called API key is a
      # session auth token.
      extra_headers = {}

      if su_user = opts[:su_user]
        # use basic auth to request a token for another user. this only works if
        # the su-user has permission to do so.
        extra_headers['Authorization'] =
          'Basic ' + Base64.strict_encode64("#{user}:#{password}")
        user     = su_user
        password = nil
      end

      options = default_api_params.tap do |h|
        h[:default_headers].update(extra_headers)
      end
      authapi = Blockbridge::Api::Client.new_basic(nil, options, client_opts)

      create_args = Hash.new.tap do |h|
        h[:grant_type]    = 'password'
        h[:username]      = user
        h[:password]      = password unless password.nil?
        h[:client_id]     = nil
        h[:client_secret] = nil
        h[:method]        = opts[:method]
        h[:expires_in]    = opts[:expires_in]
        h[:idp]           = opts[:idp] if opts.key?(:idp)
        h[:project]       = opts[:project] if opts.key?(:project)
      end

      begin
        authz = authapi.oauth2_token.create(create_args)
      rescue Blockbridge::Api::ForbiddenError => e
        if su_user
          raise RuntimeError, "You are not authorized to su to #{Paint[su_user, :bold]}"
        else
          raise
        end
      rescue Blockbridge::Api::UnauthorizedError => e
        if e.otp_required?
          extra_headers['X-Blockbridge-OTP'] = read_otp

          connect_params = default_api_params.tap do |h|
            h[:default_headers].update(extra_headers)
          end

          authapi.connect_basic(nil, connect_params)
          retry
        end

        # re-raise generic unauthorized failure
        raise
      end

      puts_i "Authenticated; token expires in #{authz.expires_in} seconds."
      authz.access_token
    end

    def read_otp
      if opts.key? :otp
        puts_i "Using supplied OTP."
        opts[:otp]
      else
        if noninteractive?
          raise RuntimeError, "authentication requires OTP to continue"
        end
        ask("Enter your 2-factor-auth OTP: ")
      end
    end

    def get_credentials    # :nodoc:
      @credentials ||= (read_credentials || ask_for_and_save_credentials)
    end

    def delete_credentials
      if netrc
        netrc.delete(api_url)
        netrc.save
      end
      @api, @client, @credentials = nil, nil, nil
    end

    def netrc_path
      default = ENV['BLOCKBRIDGE_NETRC_PATH'] || Netrc.default_path
      encrypted = default + ".gpg"
      if File.exists?(encrypted)
        encrypted
      else
        default
      end
    end

    def netrc
      return if opts[:disable_netrc]

      @netrc ||= begin
        File.exists?(netrc_path) && Netrc.read(netrc_path)
      rescue => error
        if error.message =~ /^Permission bits for/
          perm = File.stat(netrc_path).mode & 0777
          abort("Permissions #{perm} for '#{netrc_path}' are too open. You should run `chmod 0600 #{netrc_path}` so that your credentials are NOT accessible by others.")
        else
          raise error
        end
      end
    end

    def read_credentials
      if ENV['BLOCKBRIDGE_API_KEY']
        ['', ENV['BLOCKBRIDGE_API_KEY']]
      else
        # read netrc credentials if they exist
        netrc[api_url] if netrc
      end
    end

    def write_credentials
      return if opts[:disable_netrc]
      FileUtils.mkdir_p(File.dirname(netrc_path))
      FileUtils.touch(netrc_path)
      FileUtils.chmod(0600, netrc_path)
      netrc[api_url] = credentials
      netrc.save
    end

    def exchange_token_plus_otp(token)
      client_params = default_api_params.tap do |h|
        h[:default_headers]['X-Blockbridge-OTP'] = read_otp
      end

      # use temporary api handle to exchange [token, otp] for a new access
      # token.
      tapi = Blockbridge::Api::Client.new_oauth(token, client_params, client_opts)
      authz = tapi.oauth2_token.create(method: opts[:method],
                                       expires_in: opts[:expires_in])

      print_i "Authenticated as #{Paint[authz.user.user.name, :bold]}"
      puts_i  "; token expires in #{authz.expires_in} seconds."

      return [authz.user.user.name, authz.access_token]
    end

    def ask_for_credentials
      if opts[:method]
        puts_i "Authenticating to #{api_url} using method '#{opts[:method]}'"
      else
        puts_i "Authenticating to #{api_url}"
      end
      puts_i

      if (user = opts[:user]).nil?
        if noninteractive?
          raise RuntimeError, "missing user parameter in non-interactive mode"
        end
        user = ask("Enter user or access token: ")
      end

      raise Blockbridge::CLI::RuntimeError, "empty username specified" if user.empty?

      if (su_user = opts[:su_user]) && su_user.empty?
        raise Blockbridge::CLI::RuntimeError, "empty su username specified"
      end

      if user.length > 16 && user[0,2] =~ /^\d\//
        # auto-detect persistent auth access token
        print_i "Detected access token; checking token status: "

        begin
          tapi = Blockbridge::Api::Client.new_oauth(user, default_api_params, client_opts)
          az = tapi.status.authorization
        rescue Blockbridge::Api::UnauthorizedError => e
          raise unless e.otp_required?
          puts_i Paint['OTP-REQUIRED', :bold] + '.'
          puts_i
          return exchange_token_plus_otp(user)
        rescue => e
          puts_i Paint['INVALID', :red] + '.'
          puts_i
          raise
        end

        puts_i Paint['VALID', :green] + '.'

        if su_user
          puts_i "Acquiring access token for #{Paint[su_user, :bold]}."
          return [az.user.login, get_api_key(user, 'x-oauth-basic')]
        else
          # if not using the token to su, just store it for persistent use
          return [az.user.login, user]
        end
      end

      if (pass = opts[:password]).nil?
        if noninteractive?
          raise RuntimeError, "missing password parameter in " \
                              "non-interactive mode"
        end
        pass = ask("Password for #{Paint[user, :bold]}: ") { |q| q.echo = false }
      else
        puts_i "Using supplied password for user #{Paint[user, :bold]}."
      end

      if su_user
        puts_i "Acquiring access token for #{Paint[su_user, :bold]}."
      end

      [user, get_api_key(user, pass)]
    end

    def ask_for_and_save_credentials
      begin
        @credentials = ask_for_credentials
        write_credentials
        check
      rescue Exception => e
        delete_credentials
        raise e
      end
      @credentials
    end

    protected

    def task_ctx_str
      return unless (ctx = opts[:task_ctx])&.any?
      ctx.map { |(k, v)| "#{k}=#{v}" }.join(',')
    end

    def default_api_params
      {
        :url                   => api_url,
        :user                  => '',
        :default_headers       => {
          'User-Agent'         => Blockbridge::CLI.user_agent,
          'X-Blockbridge-SU'   => ENV['BLOCKBRIDGE_API_SU'],
          'X-Blockbridge-Task' => task_ctx_str,
        },
      }
    end
  end
end
