# frozen_string_literal: true

# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

# include each sub-command class
require "delegate"
require 'blockbridge/cli/command/base'

module Blockbridge::CLI
  class AutoloadedCmd < Delegator
    def initialize(name, cmd_file)
      @klass_name = name
      @path = "blockbridge/cli/command/#{cmd_file}"
    end

    def __getobj__
      Kernel.require(@path)
      Blockbridge::CLI::Command.const_get(@klass_name, true)
    end
  end

  class RootCommand < Blockbridge::CLI::Command::Base
    def parse(argv)
      # scan argv for invalid utf8; best to fail here, rather than
      # later with some generic ArgumentError exception.
      argv.each_with_index do |arg, idx|
        unless arg.valid_encoding?
          signal_usage_error "command argument at position #{idx + 1} " \
                             'contains invalid UTF-8'
        end
      end

      # checks out, continue processing
      super
    end

    option_group 'Tenant', roles: %w[tenant operator] do
      subcommand 'auth', 'authentication (login, logout, status)',
                 AutoloadedCmd.new(:Auth, 'auth'),
                 level: :basic
      subcommand 'authorization', 'manage persistent authorizations',
                 AutoloadedCmd.new(:Authorization, 'authorization'),
                 level: :basic
      subcommand 'backup', 'manage object-store based backups',
                 AutoloadedCmd.new(:Backup, 'backup'),
                 level: :advanced
      subcommand 'batch', 'perform batch object operations',
                 AutoloadedCmd.new(:Batch, 'batch'),
                 level: :expert
      subcommand 'ca', 'manage the embedded certificate authority',
                 AutoloadedCmd.new(:CA, 'ca'),
                 level: :advanced
      subcommand 'catalog', 'manage and query the catalog',
                 AutoloadedCmd.new(:Catalog, 'catalog'),
                 level: :advanced
      subcommand 'disk', 'manage virtual disks',
                 AutoloadedCmd.new(:Disk, 'disk'),
                 level: :basic
      subcommand 'disk-set', 'perform operations on a disk-set',
                 AutoloadedCmd.new(:DiskSet, 'disk_set'),
                 level: :dev
      subcommand 'host', 'collection of host integration tools',
                 AutoloadedCmd.new(:Host, 'host'),
                 level: :basic
      subcommand 'obj-store', 'manage object stores',
                 AutoloadedCmd.new(:ObjStore, 'obj_store'),
                 level: :advanced
      subcommand 'profile', 'manage initiator profiles',
                 AutoloadedCmd.new(:Profile, 'profile'),
                 level: :basic
      subcommand 'rule', 'manage disk rules',
                 AutoloadedCmd.new(:Rule, 'rule'),
                 level: :dev
      subcommand 'snapshot', 'manage snapshots',
                 AutoloadedCmd.new(:Snapshot, 'snapshot'),
                 level: :basic
      subcommand 'target', 'manage targets',
                 AutoloadedCmd.new(:Target, 'target'),
                 level: :basic
      subcommand 'user', 'manage users',
                 AutoloadedCmd.new(:User, 'user'),
                 level: :basic
      subcommand 'vss', 'manage virtual storage services',
                 AutoloadedCmd.new(:Vss, 'vss'),
                 level: :basic
      subcommand 'xmd', 'manage xmd records',
                 AutoloadedCmd.new(:Xmd, 'xmd'),
                 level: :expert
    end

    option_group 'Operator', roles: %w[operator] do
      subcommand 'account', 'manage accounts',
                 AutoloadedCmd.new(:Account, 'account'),
                 level: :basic
      subcommand 'datastore', 'manage datastores',
                 AutoloadedCmd.new(:Datastore, 'datastore'),
                 level: :basic
      subcommand 'dev', 'manage storage devices',
                 AutoloadedCmd.new(:Dev, 'dev'),
                 level: :basic
      subcommand 'net', 'manage network interfaces',
                 AutoloadedCmd.new(:Net, 'net'),
                 level: :basic
      subcommand 'node', 'manage nodes',
                 AutoloadedCmd.new(:Node, 'node'),
                 level: :basic
      subcommand 'site', 'manage sites',
                 AutoloadedCmd.new(:Site, 'site'),
                 level: :basic
      subcommand 'global', 'manage global configuration parameters',
                 AutoloadedCmd.new(:Global, 'global'),
                 level: :basic
    end

    option_group 'Miscellaneous', roles: %w[tenant operator] do
      subcommand 'config', 'get and set configuration options',
                 AutoloadedCmd.new(:Config, 'config'),
                 level: :advanced
      subcommand 'diags', 'diagnostic tools',
                 AutoloadedCmd.new(:Diags, 'diags'),
                 level: :expert
      subcommand 'version', 'show software version(s)',
                 AutoloadedCmd.new(:Version, 'version'),
                 level: :basic
    end
  end
end
