# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

module Blockbridge::CLI::Command
  class Account < Base
    self.default_subcommand = 'list'

    class AccountCreate < Base
      apiopt('--name', 'NAME', "specify new account name", :required => true)
      apiopt('--template', 'TEMPLATE', "specify account template")
      option('--password', :flag, 'prompt to specify initial account password')
      option('--password-file', 'PATH', 'read password from specified file')

      mixin :info_extraction

      def exec
        create_params = api_params.tap do |h|
          h[:password] = prompt_for_password if password? && !password_file
          h[:password] = password_from_file(password_file) if password_file
        end
        account = api.account.create(create_params)
        process_obj_created account
      end
    end
    subcommand 'create', 'create a new account', AccountCreate

    mixin :account
    mixin :info_extraction

    subcommand ['list', 'ls'], 'enumerate accounts' do
      mixin :info_filtering
      mixin :sort_by

      def exec
        accounts = obj_list(:account)
        process_results accounts, view: 'account_list'
      end
    end

    subcommand 'info', 'show detailed account information' do
      mixin :info_filtering
      mixin :info_extraction

      option ['--usage'], :flag, "include account resource usage"

      def exec
        accounts = obj_list(:account)
        process_results accounts, show_usage: usage?
      end
    end

    subcommand 'update', 'update basic account parameters' do
      apiopt('--size-reserve-limit', 'LIMIT', "account wide storage reserve limit (in bytes)", {
             :param_name  => %w(limits size_reserve_total)
            }) do |v|
              parse_bytes(v)
            end
      apiopt('--no-size-reserve-limit', :flag, 'clear existing size reserve limit',
             :param_value => nil,
             :param_name  => %w(limits size_reserve_total))
      apiopt('--vss-num-limit', 'NUM', "account wide limit on the number of VSS", {
             :param_name  => %w(limits vss_num)
           }) do |v|
             Integer(v)
           end
      apiopt('--no-vss-num-limit', :flag, 'clear existing vss num limit',
             :param_value => nil,
             :param_name  => %w(limits vss_num))
      apiopt('--vss-size-reserve-max', 'MAX', "per vss maximum reserved storage (in bytes)", {
             :param_name => %w(provisioning size_reserve_max)
             }) do |v|
               parse_bytes(v)
             end
      apiopt('--no-vss-size-reserve-max', :flag, 'clear exsting per vss size reserve max',
             :param_value => nil,
             :param_name => %w(provisioning size_reserve_max))
      apiopt('--vss-iops-reserve-max', 'MAX', "per vss maximum reserved iops", {
             :param_name => %w(provisioning iops_reserve_max)
             }) do |v|
               Integer(v)
             end
      option('--no-vss-iops-reserve-max', :flag, 'clear exsting per vss iops reserve max',
             :param_value => nil,
             :param_name => %w(provisioning iops_reserve_max))
      apiopt('--tag-include', 'TAGS', 'provisioning tags always included for the account (may be used multiple times)',
             :multivalued => true,
             :param_name => %w(provisioning tags_include))
      apiopt('--clear-include-tags', :flag, 'clear existing include tags',
             :param_value => [],
             :param_name => %w(provisioning tags_include))
      apiopt('--tag-exclude', 'TAGS', 'provisioning tags always excluded for the account (may be used multiple times)',
             :multivalued => true,
             :param_name => %w(provisioning tags_exclude))
      apiopt('--clear-exclude-tags', :flag, 'clear existing exclude tags',
             :param_value => [],
             :param_name => %w(provisioning tags_exclude))
      apiopt('--tag-permit', 'TAGS', 'provisioning tags permitted for the account (may be used multiple times)',
             :multivalued => true,
             :param_name => %w(provisioning tags_permit))
      apiopt('--no-tags-permitted', :flag, 'permit no provisioning tags for the account',
             :param_value => [],
             :param_name => %w(provisioning tags_permit))
      apiopt('--all-tags-permitted', :flag, 'permit all provisioning tags for the account',
             :param_value => nil,
             :param_name => %w(provisioning tags_permit))

      def exec
        account = obj_list(:account, restrict: :only_one)
        account = api.account.update(account.id, api_params)
        process_obj_updated account
      end
    end

    subcommand 'remove', 'remove an existing account' do
      def exec
        account = obj_list(:account, restrict: :only_one)

        api.account.remove(account.id)
        process_obj_removed account
      end
    end

  end

end
