# Copyright (c) 2015-2017, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

module Blockbridge::CLI::Command
  class Auth < Base
    subcommand 'login', 'log in with your Blockbridge credentials', self, level: :basic do
      apiopt('--user', 'LOGIN', 'specify username/token to use')
      apiopt('--su', 'LOGIN', 'specify username to assume the identity of',
             param_name: :su_user)

      option_group nil, level: :advanced do
        option '--password-file', 'PATH', 'read password from specified file'
        apiopt '--otp', 'OTP', 'specify OTP to use, if requested by server'
        apiopt('--method', 'METHOD', 'use alternate auth method. (e.g., keystone)') do |v|
          enum_check(v, %w(keystone))
        end
        apiopt('--expires-in', 'SECONDS', 'set session expiration time (default server assigned)') do |v|
          Integer(v)
        end
        option('--noninteractive', :flag, 'silence output and print token on success')
        option('--disable-netrc', :flag, 'disable reading and writing ~/.netrc')
        option('--idp', 'IDP', 'specify alternate identity provider name')
        option('--project', 'NAME', 'specify idp-specific project name')
      end

      def exec
        auth_params = api_params.tap do |h|
          h[:idp] = idp if idp
          h[:project] = project if project
          h[:password] = password_from_file(password_file) if password_file
          h[:noninteractive] = noninteractive?
          h[:disable_netrc] = disable_netrc?
        end

        begin
          info = auth(auth_params).login
          if noninteractive?
            puts info.last
            return
          end
        rescue Blockbridge::Api::UnauthorizedError
          if noninteractive?
            runtime_error "Authentication failed"
          end
          if auth_params.key? :password
            # trap exception so the main handler doesn't keep retrying auth with
            # an invalid password.
            runtime_error "Authentication failed with pre-specified password"
          end
          raise
        end
        process_results api.status.authorization
      end
    end

    subcommand 'logout', 'clear stored authentication credentials', self, level: :basic do
      def exec
        auth.logout
      end
    end

    subcommand 'token', 'display current access token', self, level: :basic do
      def exec
        puts auth.api_key
      end
    end

    subcommand 'whoami', 'display Blockbridge username', self, level: :basic do
      def exec
        puts auth.info.user.login
      end
    end

    subcommand 'info', 'display authentication information', self, level: :basic do
      mixin :info_extraction

      def exec
        process_results [api.status.authorization]
      end
    end
  end
end
