# frozen_string_literal: true

# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'set'
require 'active_support/core_ext/file/atomic'
require 'blockbridge/host'
require 'blockbridge/util/tabular'
require 'blockbridge/api'
require 'blockbridge/cli/auth'
require 'blockbridge/cli/api_option'
require 'blockbridge/cli/attr_config'
require 'digest/sha1'
require 'netrc'
require 'uuidtools'

module Blockbridge::CLI::Command
  def self.find(argv)
    argv.inject(Blockbridge::CLI::RootCommand) do |klass, arg|
      sc = klass.recognised_subcommands.find { |c| c.names.include? arg }
      if sc.nil?
        raise ArgumentError, "can't find command class for '#{argv.join(' ')}'"\
                             " (at '#{arg}')"
      end
      sc.subcommand_class
    end
  end

  class Base < Clamp::Command
    class << self
      # wrap Clamp options, parameters and subcommands:
      # 1) options and parameters default to level :basic
      # 2) subcommands default to level :advanced

      alias clamp_option option
      alias clamp_parameter parameter
      alias clamp_subcommand subcommand

      def option(switches, type, desc, **opts, &blk)
        opts[:level] ||= current_option_group.fetch(:level, :basic)
        clamp_option(switches, type, desc, **opts, &blk)
      end

      def parameter(type, desc, **opts, &blk)
        opts[:level] ||= current_option_group.fetch(:level, :basic)
        clamp_parameter(type, desc, **opts, &blk)
      end

      def subcommand(name, desc, cls = self, level: :advanced, **opts, &blk)
        clamp_subcommand(name, desc, cls, level: level, **opts, &blk)
      end
    end

    def execute
      # wrap Clamp's 'execute' method to ensure any raised exceptions include
      # the final resolved command instance object. this allows any state that
      # was built up (e.g., while constructing api parameters) to be accessed
      # in the exception handling code. most importantly, this makes
      # generating pretty validation errors more-or-less automatic. (assuming
      # proper 'apiopt' params are defined)
      this_command_instance = self

      begin
        self.exec
      rescue StandardError, Interrupt => e
        # add a command_instance attribute to whatever excetion was thrown,
        # and set it to the previously saved value of 'self'

        e.class.class_eval { attr_accessor :command_instance }
        e.command_instance = this_command_instance

        # and re-raise the same exception
        raise
      end

      # return self on success, too.
      self
    end

    attr_reader :results

    include Blockbridge::CLI::Helpers
    include Blockbridge::CLI::Parsing
    include Blockbridge::CLI::Attach

    # include and extend with ApiOption so all the utilities are available in
    # the class and the instance. sheesh, this seems complicated.
    include Blockbridge::CLI::ApiOption
    extend  Blockbridge::CLI::ApiOption
    extend  Blockbridge::CLI::AttrConfig

    option_group 'Global' do
      # global options applicable to all commands
      option(['-h', '--help'], :flag,
             'print help (use --verbose to show hidden options)',
             level: :basic) do
        request_help
      end
      option('--verbose', :flag, 'enable detailed output', level: :basic) do |_|
        Blockbridge::CLI.verbose = true
      end
      option('--debug', :flag, 'enable additional debugging output', hidden: true)

      option('--pry', :flag, 'enable pry debugger', level: :dev) do |v|
        require 'pry-byebug'
        require 'pry-rescue'
        require 'pry-stack_explorer'
        binding.pry
      end

      option(['-R', '--raw'], :flag, 'enable raw json output', level: :advanced)
      option('--fail-on-empty', :flag, 'fail on empty result set',
             hidden: true,
             level: :advanced)
      option('--show-trash', :flag, 'show trashed objects', level: :advanced)

      option(['--[no-]color'], :flag,
             'enable colorized output (default: autodetect)', hidden: true) do |v|
        Paint.mode = (v ? 256 : 0)
      end

      option(['-H', '--api-host'], 'HOST', 'specify api endpoint host',
             environment_variable: 'BLOCKBRIDGE_API_HOST', hidden: true)
      option('--[no-]ssl', :flag, 'use ssl/tls', default: true, hidden: true)

      def api_scheme
        ssl? ? 'https' : 'http'
      end

      option('--[no-]ssl-verify-peer', :flag, 'verify peer certificate',
             hidden: true, default: true,
             environment_variable: 'BLOCKBRIDGE_SSL_VERIFY_PEER') do |val|
        @ssl_verify_peer = parse_boolean(val, true)
      end

      option('-k', :flag, 'disable peer certificate verification',
             attribute_name: :no_ssl_verify_peer, hidden: true) do |_|
        @ssl_verify_peer = false
      end

      attr_config :ssl_verify_peer?, 'api.ssl_verify_peer'

      option('--ssl-ca-path', 'DIR', 'specify path to CA certificate directory',
             hidden: true)
      option('--ssl-ca-file', 'FILE', 'specify path to CA certificate file',
             hidden: true)

      # options for tweaking the blockbridge task context
      option_group nil, level: :expert do
        option('--task-timeout', 'SECONDS', 'task timeout, in seconds', level: :advanced)
        option('--task-id', 'UUID', 'explicit task uuid', hidden: true, level: :advanced)
        option('--task-ext-auth', 'SPEC', 'external auth specification',
               hidden: true, level: :expert)
      end

      option('--api-url', 'URL', 'specify api endpoint url',
             attribute_name: :raw_api_url,
             environment_variable: 'BLOCKBRIDGE_API_URL', hidden: true)

      def default_raw_api_url
        api_host && "#{api_scheme}://#{api_host}/api"
      end
      attr_config(:raw_api_url, 'api.hosts') do |hosts|
        if api_host
          hosts[api_host].url if hosts.key?(api_host)
        elsif (dfl = hosts.values.find { |v| v.default })
          dfl.url
        end
      end
      def api_url
        raw_api_url && URI.parse(raw_api_url)
      end

      # these stub options are not actually used -- they exist to consume the
      # options if specified via the command line. The actual option/env/config
      # support for these two parameters is in the CLI::Main class (in main.rb)
      option '--level', 'LEVEL', 'specify cli user skill level', level: :dev
      option '--role', 'ROLE', 'specify cli user role', level: :dev

      ## configuration/cache directory options
      option('--home', 'DIR', 'specify config and cache directory',
             attribute_name: :home_dir,
             environment_variable: 'BLOCKBRIDGE_HOME',
             default: File.join(Dir.home, '.blockbridge-cli'),
             hidden: true)
      option('--config', 'FILE',
             'specify config file location (default: [HOME]/config.yml)',
             attribute_name: :config_file_name,
             environment_variable: 'BLOCKBRIDGE_CONFIG',
             hidden: true)
      option('--no-config', :flag, "don't read config file on startup",
             attribute_name: :no_config, hidden: true)

      def default_config_file_name
        File.join(home_dir, 'config.yml')
      end

      def config_file
        no_config? ? '' : config_file_name
      end

      option('--cache', 'DIR',
             'specify cache directory (default: [HOME]/cache)',
             attribute_name: :cache_dir, hidden: true)
      option('--no-cache', :flag, 'disable caching',
             attribute_name: :cache_dir, hidden: true) { |_| nil }

      def default_cache_dir
        File.join(home_dir, 'cache')
      end

      option('--error-format', 'FMT',
             'spectify output error format (human, machine)',
             environment_variable: 'BLOCKBRIDGE_ERROR_FORMAT',
             default: 'human', hidden: true, level: :advanced) do |v|
        enum_check(v, %w[machine human json])
      end
    end

    # predefined reusable options for scoping operations
    mixin :vss do |opts = {}|
      opt_desc = opts.fetch(:desc, 'specify vss by label or serial number')
      opt_flgs = opts.fetch(:flags, {})
      opt_inst = option(['-v', '--vss'], 'VSS', opt_desc, level: :basic, **opt_flgs) do |v|
        signal_usage_error "option '-v/--vss' requires an argument" if v.nil?
        v
      end
      apiopt_map(opts[:param_path] || '/vss_id', opt_inst.switches)
    end
    mixin :disk do |opts = {}|
      obj_name = opts.fetch(:obj_name, 'disk')
      opt_desc = opts.fetch(:desc, "specify #{obj_name} by label or serial number")
      opt_inst = option(['-d', '--disk'], 'DISK', opt_desc, level: :basic) do |v|
        signal_usage_error "option '-d/--disk' requires an argument" if v.nil?
        v
      end
      apiopt_map(opts[:param_path] || '/vdisk_id', opt_inst.switches)
    end
    mixin :rule do |opts = {}|
      option ['-r', '--rule'], 'RULE', 'specify rule by label or serial number', level: :basic do |v|
        signal_usage_error "option '-r/--rule' requires an argument" if v.nil?
        v
      end
    end
    mixin :target do |opts = {}|
      opt_inst = option(['-t', '--target'], 'TARGET',
                        'specify target by label or serial number', level: :basic) do |v|
        signal_usage_error "option '-t/--target' requires an argument" if v.nil?
        v
      end
      apiopt_map(opts[:param_path] || '/target_id', opt_inst.switches)
    end
    mixin :profile do |opts = {}|
      opt_inst = option(['-p', '--profile'], 'PROFILE',
                        'specify profile by label or serial number', level: :basic) do |v|
        signal_usage_error "option '-p/--profile' requires an argument" if v.nil?
        v
      end
      apiopt_map(opts[:param_path] || '/profile_id', opt_inst.switches)
    end
    mixin :snapshot do |opts = {}|
      opt_desc = opts.fetch(:desc, 'specify snapshot by label or serial number')
      opt_inst = option(['-s', '--snapshot'], 'SNAPSHOT', opt_desc, level: :basic) do |v|
        signal_usage_error "option '-s/--snapshot' requires an argument" if v.nil?
        v
      end
      apiopt_map(opts[:param_path] || '/snapshot_id', opt_inst.switches)
    end
    mixin :datastore do |opts = {}|
      opt_desc = opts.fetch(:desc, 'specify datastore by label or serial number')
      opt_inst = option(['-D', '--datastore'], 'DATASTORE', opt_desc, level: :basic) do |v|
        signal_usage_error "option '-D/--datastore' requires an argument" if v.nil?
        v
      end
      apiopt_map(opts[:param_path] || '/sds_id', opt_inst.switches)
    end
    mixin :site do |opts = {}|
      opt_inst = option('--site', 'SITE',
                        'specify site by label or serial number') do |v|
        signal_usage_error "option '--site' requires an argument" if v.nil?
        v
      end
      apiopt_map(opts[:param_path] || '/site_id', opt_inst.switches)
    end
    mixin :node do |opts = {}|
      opt_inst = option('--node', 'NODE',
                        'specify node by label or serial number') do |v|
        signal_usage_error "option '--node' requires an argument" if v.nil?
        v
      end
      apiopt_map(opts[:param_path] || '/node_id', opt_inst.switches)
    end
    mixin :dev do |opts = {}|
      opt_inst = option('--dev', 'DEVICE',
                        'specify device by label or serial number') do |v|
        signal_usage_error "option '--dev' requires an argument" if v.nil?
        v
      end
      apiopt_map(opts[:param_path] || '/dev_id', opt_inst.switches)
    end
    mixin :net do |opts = {}|
      opt_inst = option('--net', 'INTERFACE',
                        'specify network interface by label or serial number') do |v|
        signal_usage_error "option '--net' requires an argument" if v.nil?
        v
      end
      apiopt_map(opts[:param_path] || '/net_id', opt_inst.switches)
    end
    mixin :account do |opts = {}|
      opt_inst = option('--account', 'ACCOUNT',
                        'specify account by name or serial number') do |v|
        signal_usage_error "option '--account' requires an argument" if v.nil?
        v
      end
      apiopt_map(opts[:param_path] || '/account_id', opt_inst.switches)
    end
    mixin :user do |opts = {}|
      opt_inst = option('--user', 'USER',
                        'specify user by login or serial number') do |v|
        signal_usage_error "option '--user' requires an argument" if v.nil?
        v
      end
      apiopt_map(opts[:param_path] || '/user_id', opt_inst.switches)
    end
    mixin :authz do |opts = {}|
      opt_inst = option('--authz', 'AUTHZ',
                        'specify authorization by serial number') do |v|
        signal_usage_error "option '--authz' requires an argument" if v.nil?
        v
      end
      apiopt_map(opts[:param_path] || '/authz_id', opt_inst.switches)
    end
    mixin :obj_store do |opts = {}|
      opt_desc = opts.fetch(:desc, 'specify obj-store by label or serial number')
      opt_inst = option(['-o', '--obj-store'], 'OBJ-STORE', opt_desc) do |v|
        if v.nil?
          signal_usage_error "option '-o/--obj-store' requires an argument"
        end
        v
      end
      apiopt_map(opts[:param_path] || '/obj_store_id', opt_inst.switches)
    end
    mixin :backup_id do |opts = {}|
      opt_inst = option(['-b', '--backup-id'], 'BACKUP-ID',
                        'specify backup by ID') do |v|
        if v.nil?
          signal_usage_error "option '-b/--backup-id' requires an argument"
        end
        v
      end
      apiopt_map(opts[:param_path] || '/backup_id', opt_inst.switches)
    end

    mixin :uuid do |opts = {}|
      apiopt('--uuid', 'UUID', 'specify explicit object UUID') do |v|
        signal_usage_error "option '--uuid' requires an argument" if v.nil?
        v && UUIDTools::UUID.parse(v)
      end
    end

    mixin :identity_param do |opts = {}|
      name = opts.fetch(:obj_name, 'object')
      desc = opts.fetch(:param_desc, "#{name} identity (label, id, serial, or uuid)")
      attr_name = opts.fetch(:attribute_name, :identity_param)
      parameter("[#{name.upcase}]", desc, attribute_name: attr_name)
    end

    mixin :sort_by do
      option('--sort-by', 'COLS', 'comma-separated list of sort columns',
             attribute_name: :user_sort_by) do |v|
        if v.length == 0
          []
        else
          v.split(',')
        end
      end
    end

    mixin :info_filtering do
      option_group 'Formatting', hidden: true, level: :advanced do
        option('--filter',
               'PATH',
               'filter output through json path before display',
               multivalued: true) do |v|
          signal_usage_error "option '--filter' requires an argument" if v.nil?
          @filtering = true
          v
        end

        option('--limit', 'COUNT', 'limit number of returned objects') do |v|
          signal_usage_error "option '--limit' requires an argument" if v.nil?
          @limiting = true
          Integer(v)
        end
      end
    end

    mixin :info_extraction do
      option_group 'Formatting', hidden: true, level: :advanced do
        option '--tabular', :flag, 'format extracted data as a table'
        option '--flatten', :flag, 'flatten nested arrays into extracted rows',
               level: :expert
        option('--col-sep', 'SEP', 'set column separator for output csv',
               default: ',')
        option '--extract-format', 'STR', 'format string for extracted fields'
        option(['--extract', '-X'],
               'PATH',
               'json path to extract from each object',
               multivalued: true) do |v|
          if v.nil?
            signal_usage_error "option '-X/--extract' requires an argument"
          end
          @extracting = true
          v
        end
        option '--process', 'SCRIPT', 'evaluate output data with ruby code',
               level: :expert
      end
    end

    def extracting?
      @extracting
    end

    def filtering?
      @filtering
    end

    def limiting?
      @limiting
    end

    mixin :label_notes do |opts|
      obj_name = opts.delete(:obj_name) || 'object'
      apiopt ['-l', '--label'], 'LABEL', "specify #{obj_name} label", level: :basic
      option_group nil, opts do
        apiopt('--no-label',
               :flag,
               "specify explicitly undefined #{obj_name} label",
               param_name: :label,
               param_value: nil,
               map_param: false,
               level: :advanced)
        apiopt '--notes', 'NOTES', "specify #{obj_name} notes", level: :basic
      end
    end

    mixin :catalog_query_params do
      option_group 'Querying' do
        apiopt(['-c', '--capacity'], 'CAPACITY', 'requested service capacity',
               {
                 param_group: :query_params
               }) do |c|
          parse_bytes(c)
        end
        apiopt('--type', 'TYPE', 'requested service type', {
                 param_group: :query_params,
                 param_name: :template
               })
        apiopt('--iops', 'IOPS', 'requested service IOPS', {
                 param_group: :query_params
               }) do |val|
          Integer(val)
        end
        apiopt('--include', 'TAG',
               'restrict search by including tag(s) (may be used multiple times)', {
                 multivalued: true,
                 attribute_name: :include_tags,
                 param_group: :query_params,
                 param_name: %w[tags include]
               })
        apiopt('--exclude', 'TAG',
               'restrict search by excluding tag(s) (may be used multiple times)', {
                 multivalued: true,
                 attribute_name: :exclude_tags,
                 param_group: :query_params,
                 param_name: %w[tags exclude]
               })
        option_group nil, hidden: true do
          apiopt('--scheduling-algorithm', 'ALG', 'scheduling algorithm', {
                   param_group: :query_params
                 })
        end
      end
    end

    mixin :xref do
      apiopt '--xref', 'REF', 'external reference key', level: :advanced
    end

    mixin :xmd_refs_create do
      apiopt '--xmd-ref', 'REF', 'xmd record reference key',
             multivalued: true,
             attribute_name: :xmd_refs,
             level: :expert
    end

    mixin :xmd_refs_update do
      option('--clear-xmd-refs', :flag, 'clear xmd refs',
             param_name: 'xmd_refs', level: :expert) do
        []
      end

      apiopt('--xmd-ref', 'REF', 'xmd record reference key',
             multivalued: true, attribute_name: :xmd_refs,
             level: :expert) do |v|
        v
      end
    end

    def set_xmd_refs(params, ex_refs = nil)
      return if !clear_xmd_refs? && xmd_refs.empty?

      ex_refs = yield params if block_given?

      params[:xmd_refs] = clear_xmd_refs? ? [] : (ex_refs || [])
      params[:xmd_refs].push(*xmd_refs)
      params[:xmd_refs].uniq!
      params[:xmd_refs]
    end

    mixin :tags_create do |opts = {}|
      apiopt(['--tag', '-T'], 'TAGS', 'add tag (may be used multiple times)',
             multivalued: true,
             param_name: opts.fetch(:param_name, ['tags']),
             level: :advanced)
    end

    mixin :tags_update do |opts = {}|
      option('--clear-tags', :flag, 'clear existing tags (may be combined with -T)',
             level: :advanced)
      apiopt(['--tag', '-T'], 'TAGS', 'add tag (may be used multiple times)',
             multivalued: true,
             param_name: opts.fetch(:param_name, ['tags']),
             level: :advanced)
      option(['--untag'], 'TAGS', 'remove tag (may be used multiple times)',
             multivalued: true,
             level: :advanced)
    end

    def set_tags(params = nil, ex_tags = nil)
      params.delete(:tags) unless params.nil?
      ex_tags ||= yield params if block_given?
      ex_tags ||= []

      ex_tags  = Set.new(ex_tags)
      add_tags = Set.new(tag_list.flat_map { |t| t.split(/,\s*/) })
      del_tags = Set.new(untag_list.flat_map { |t| t.split(/,\s*/) })

      new_tags  = clear_tags? ? Set.new : ex_tags
      new_tags -= del_tags
      new_tags += add_tags

      return if new_tags == ex_tags

      if params
        params[:tags] = new_tags.to_a
      else
        new_tags.to_a
      end
    end

    def parse_tag_query(tq)
      tq = tq.split(/\s+/) if tq.respond_to? :to_str

      {}.tap do |p|
        h = p[:tags] = { include: [], exclude: [] }

        tq.each do |ts|
          tt = ts[0]
          tv = ts[1..-1]

          next if tv.nil? || tv.empty?

          case tt
          when '+'
            h[:include].push(tv)
          when '-'
            h[:exclude].push(tv)
          else
            usage_error "'#{ts}' tag prefix must be '+' or '-'"
          end
        end
      end
    end

    def enum_check(val, options, fmt = nil, secret_opts: nil)
      all_options = [options, secret_opts].compact.flatten
      return val if all_options.include?(val)

      msg = (fmt || "must be one of: #{options.join(', ')}")
      raise ArgumentError, msg
    end

    def enum_map(val, map, fmt = nil)
      unless !val.nil? && map.key?(val.to_sym)
        msg = (fmt || "must be one of: #{map.keys.join(', ')}")
        raise ArgumentError, msg
      end
      map[val.to_sym]
    end

    def friendly_type(ot)
      if @friendly_types.nil?
        @friendly_types = {}
        OBJ_TYPE_INFO.each_pair do |k, v|
          @friendly_types[v[:apitype]] = k.to_s
        end
      end
      @friendly_types[ot.to_sym]
    end

    def display(*args)
      puts(*args)
    end

    def verbose(*args)
      display(args) if verbose?
    end

    def render_extraction(data, opts)
      extras = {
        extract_list:    extract_list,
        extract_format:  extract_format,
        extract_flatten: flatten?,
        col_sep:         col_sep,
      }

      opts.merge! extras
      data = [ data ] unless data.is_a? Array

      # create view container and render the appropriate template
      view = Blockbridge::CLI::View.new
      tmpl = tabular? ? :table_extraction : :csv_extraction

      view.render(tmpl, data, opts)
    end

    def render(data, opts = {})
      if extracting?
        render_extraction(data, opts)
      elsif raw?
        JSON.pretty_generate(data)
      else
        # try to automatically determine a template name
        name = (opts[:view] || @invocation_path.sub(/^bb[^ ]* /, '')).to_s.gsub(/[ -]/, '_')

        begin
          script = respond_to?(:process) && self.send(:process)
          if script
            Blockbridge::CLI::View.new.render_script(script, data, opts)
          else
            Blockbridge::CLI::View.new.render(name, data, opts)
          end
        rescue Blockbridge::CLI::RuntimeError
          raise
        rescue => e
          if filtering?
            s = "failed to render output -- perhaps your filter got in the way.\n"
            if verbose?
              s << "here's the raw filtered JSON:\n\n"
            end
          else
            raise e if debug?

            s = "failed to render template #{name}: #{e.message}"
            s << "\n\nException backtrace:\n"
            s << e.backtrace.join("\n")

            if verbose?
              s << "\n\nRaw JSON follows:\n\n"
            end
          end
          if verbose?
            s << JSON.pretty_generate(data)
          end
        end
      end
    end

    def process_results(data = nil, opts = {})
      if filtering?
        filter_list.each do |f|
          data = JsonPath.new(f).on(data)
        end
      end

      if limiting?
        if !limit.nil? && !data&.empty?
          data = data.take(limit)
        end
      end

      # save results for interactive/scripted usage
      @results = data

      if (data.nil? || data.empty?) && (fail_on_empty? rescue false)
        usage_error("no results found")
      end

      opts[:verbose] ||= verbose?
      opts[:cmd]       = self
      opts[:api_cb]    = lambda { api }

      if sort = user_sort_by rescue nil
        opts[:sort_cols] = sort
      end

      output = render(data, opts)
      display(output) unless output.nil? || output.empty?
    end

    def process_obj_created(obj, opts = {})
      process_results([obj], opts.merge(view: :obj_created))
    end

    def process_obj_updated(obj, opts = {})
      process_results([obj], opts.merge(view: :obj_updated))
    end

    def process_obj_removed(obj, opts = {})
      process_results([obj], opts.merge(view: :obj_removed))
    end

    def process_action_complete(data, opts = {})
      process_results(data, opts.merge(view: :action_complete))
    end

    def setup_default_api_host
      hosts = config.api.hosts
      return unless hosts.empty? && !api_url

      label = ask('Enter a default management host: ') do |q|
        q.validate = ->(v) { !v.empty? }
        q.responses[:not_valid] = 'Enter a non-empty string.'
      end

      case label
      when /^https?:\/\//
        url = URI.parse(label)
      else
        url = URI.parse("https://#{label}/api")
      end

      config.api.hosts = { label => { url: url.to_s, default: true } }

      # set as current, too
      self.raw_api_url = url.to_s
    end

    def setup_default_config
      return if config.disabled?
      config.global.uuid ||= UUIDTools::UUID.random_create.to_s
      setup_default_api_host
      config.save
    end

    def task_ctx
      Hash.new.tap do |h|
        h[:timeout] = task_timeout if task_timeout
        h[:id] = task_id if task_id
        h[:ext_auth] = task_ext_auth if task_ext_auth
      end
    end

    def default_api_opts
      Hash.new.tap do |h|
        h[:force_reauth] = context[:force_reauth]
        h[:api_url]      = api_url.to_s
        h[:ssl_ca_path]  = ssl_ca_path
        h[:ssl_ca_file]  = ssl_ca_file || api_cert_path
        h[:task_ctx]     = task_ctx
      end
    end

    class VerifyContext
      attr_reader :preverify_ok
      attr_reader :uri
      attr_reader :chain
      attr_reader :cert
      attr_reader :error
      attr_reader :error_string
      attr_reader :error_depth

      def initialize(uri, preverify_ok, store_ctx)
        @uri = uri
        @preverify_ok = preverify_ok
        @chain = store_ctx.chain
        @cert = store_ctx.current_cert
        @error = store_ctx.error
        @error_string = store_ctx.error_string
        @error_depth = store_ctx.error_depth
      end

      def serial
        @serial ||= cert.serial.to_s + " (0x#{cert.serial.to_s(16).downcase})"
      end

      def subject
        @subject ||= cert.subject.to_s(OpenSSL::X509::Name::ONELINE)
      end

      def issuer
        @issuer ||= cert.issuer.to_s(OpenSSL::X509::Name::ONELINE)
      end

      def sha1_digest
        @sha1_digest ||= Digest::SHA1.hexdigest(cert.to_der)
      end

      def sha256_digest
        @sha256_digest ||= Digest::SHA256.hexdigest(cert.to_der)
      end

      def hex_pp(digest)
        digest.scan(/../).join(':')
      end

      def cert_id_match?
        @cert_id_match ||= OpenSSL::SSL.verify_certificate_identity(cert, uri.hostname)
      end

      def ok?
        preverify_ok && cert_id_match?
      end

      def error_detail
        if !cert_id_match?
          "hostname '#{uri.hostname}' does not match the server certificate"
        elsif !preverify_ok
          "#{error_string} (#{error})"
        end
      end

      def to_text
        io = StringIO.new

        if !ok?
          io.puts "Peer verification failed: #{error_detail}"
        else
          io.puts "Peer verification ok."
        end

        io.puts
        io.puts "  API:         #{uri}"
        io.puts "  Serial:      #{serial}"
        io.puts "  Subject:     #{subject}"
        io.puts "  Issuer:      #{issuer}"
        io.puts "  Fingerprints:"
        io.puts "    SHA1:      #{hex_pp(sha1_digest)}"

        d1, d2 = sha256_digest.scan(/.{32}/)
        io.puts "    SHA256:    #{hex_pp(d1)}"
        io.puts "               #{hex_pp(d2)}"
        io.string
      end
    end

    class PeerVerifier
      attr_reader :uri
      attr_reader :cert_dir
      attr_reader :verify_peer

      def initialize(uri, cert_dir, verify_peer)
        @uri = uri
        @cert_dir = cert_dir
        @verify_peer = verify_peer
      end

      def options
        Hash.new.tap do |h|
          if uri.scheme == 'https'
            h[:ssl_verify_peer] = verify_peer
            h[:ssl_verify_callback] = method(:verify_callback)
          end
        end
      end

      def verify_callback(pv, vctx)
        return true if pv

        if vctx.current_cert.to_der == vctx.chain[0].to_der
          offer_whitelisting VerifyContext.new(uri, pv, vctx)
        else
          # fake trust for intermediate certs in the chain.
          true
        end

      rescue => e
        puts "exception: #{e.message}"
        puts e.backtrace
      end

      def cert_path
        @cert_path ||=
          begin
            dir = "#{uri.host}:#{uri.port}"
            dir = URI.encode(dir, ':') if BlockbridgeHost.detect_os == :windows
            File.join(cert_dir, dir, 'peer.crt')
          end
      end

      def pinned_cert
        if File.exist?(cert_path)
          OpenSSL::X509::Certificate.new(File.read(cert_path))
        end
      end

      def offer_whitelisting(vctx)
        if pcert = pinned_cert
          return true if pcert.to_der == vctx.cert.to_der
        end

        puts vctx.to_text
        puts

        return false unless vctx.cert_id_match?

        wl = ask "Manually trust this certificate for #{uri.hostname}? [y/N] "
        return false unless wl == 'y'

        FileUtils.mkpath(File.dirname(cert_path))
        File.atomic_write(cert_path) do |f|
          f.write vctx.cert.to_s
        end

        puts
        puts "Peer certificate registered for #{uri}"
        puts

        true

      rescue => e
        puts "Failed to whitelist peer: #{e.message}"
        raise if debug?
        false
      end
    end

    def peer_verifier
      @peer_verifier ||= PeerVerifier.new(api_url, cert_dir, ssl_verify_peer?)
    end

    def auth(extra_opts = {})
      return @auth if defined?(@auth)

      config # ensure config has been loaded
      setup_default_config

      if api_url.nil?
        runtime_error 'No API host specified; use -H to specify an explicit '  \
                      'host or configure a default using `bb config api host`' \
                      ' commands.'
      end

      Blockbridge::Api::Client.defaults.update(peer_verifier.options)
      @auth = Blockbridge::CLI::Auth.new(default_api_opts.merge(extra_opts))
    end

    def api_cert_path
      url_cert_path(api_url)
    end

    def cert_dir
      File.join(home_dir, 'certs.d')
    end

    def host_cert_dir(url)
      dir = "#{url.host}:#{url.port}"
      dir = URI.encode(dir, ':') if BlockbridgeHost.detect_os == :windows
      File.join(cert_dir, dir)
    end

    def url_cert_path(url)
      dir = "#{url.host}:#{url.port}"
      dir = URI.encode(dir, ':') if BlockbridgeHost.detect_os == :windows
      File.join(cert_dir, dir, 'peer.crt')
    end

    def cert_for_url(url)
      return unless File.exist?(path = url_cert_path(url))

      OpenSSL::X509::Certificate.new(File.read(path))
    end

    def url_whitelisted?(url)
      File.exist? url_cert_path(url)
    end

    def api
      config # ensure config has been loaded
      auth.api
    end

    def config
      @config ||= Blockbridge::CLI::Config.new(config_file)
    end

    def diags_dir
      @diags_dir ||= home_dir && File.join(home_dir, 'diags')
    end

    # configure and inspect the host operating system
    def host_control
      opts = { debug: debug?, verbose: verbose? }
      @host_control ||= BlockbridgeHost.control(**opts)
    end
  end
end
