# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

module Blockbridge::CLI::Command
  class Config < Base
    def config_support!
      usage_error "Configuration support disabled" if config.disabled?
    end

    subcommand 'clear', 'clear all local configuration and cached certificates' do
      def exec
        config_support!

        # clear all configuration except for the global.uuid and global.role
        # values, if present.
        save_uuid = config.global.uuid
        save_role = config.global.role
        config.clear
        config.global.uuid = save_uuid unless save_uuid.nil?
        config.global.role = save_role unless save_role.nil?
        config.save

        # remove all cached ca certificates
        FileUtils.rm_rf cert_dir, secure: true
      end
    end

    subcommand 'role', 'specify cli user role', self, level: :expert do
      option '--clear', :flag, 'reset user role configuration'

      parameter '[ROLE]', 'user role (tenant, operator)' do |v|
        enum_check(v, %w[tenant operator])
      end

      def exec
        config_support!
        if clear?
          config.global.delete('role')
          config.save
        elsif role
          config.global.role = role
          config.save
        end
        puts "global.role = #{config.global.role}"
      end
    end

    subcommand 'level', 'specify cli user level', self, level: :expert do
      option '--clear', :flag, 'reset user level configuration'

      parameter '[LEVEL]', 'user skill level (basic, advanced, expert)' do |v|
        enum_check(v, %w[basic advanced expert dev])
      end

      def exec
        config_support!
        if clear?
          config.global.delete('level')
          config.save
        elsif level
          config.global.level = level
          config.save
        end
        puts "global.level = #{config.global.level}"
      end
    end

    subcommand 'host', 'configure command line host details' do
      subcommand 'device', 'configure device settings' do
        self.default_subcommand = 'list'

        subcommand 'set', 'set device settings' do
          option '--queue-depth', 'QUEUE-DEPTH', 'set device setting queue depth' do |v|
            Integer(v)
          end
          option '--nomerges', 'NOMERGES', 'set device queue setting nomerges' do |v|
            parse_boolean(v) ? 1 : 0
          end
          option '--rq-affinity', 'RQ-AFFINITY', 'set device queue setting rq affinity' do |v|
            Integer(v)
          end
          option '--scheduler', 'SCHEDULER', 'set device queue setting scheduler (noop, none, deadline, mq-deadline, bfq, cfq)' do |v|
            enum_check(v, %w(noop deadline mq-deadline bfq cfq none))
          end
          option '--rotational', 'ROTATIONAL', 'set device queue setting rotational' do |v|
            parse_boolean(v) ? 1 : 0
          end
          option '--add-random', 'ADD-RANDOM', 'set device queue setting add_random' do |v|
            Integer(v)
          end
          option '--nr-requests', 'NR-REQUESTS', 'set device queue setting nr_requests' do |v|
            Integer(v)
          end

          def exec
            config_support!
            device = {
              'queue_depth' => queue_depth,
              'nomerges'    => nomerges,
              'rq_affinity' => rq_affinity,
              'scheduler'   => scheduler,
              'rotational'  => rotational,
              'add_random'  => add_random,
              'nr_requests' => nr_requests,
            }.reject { |_, v| v.nil? }

            signal_usage_error "specify a device setting to set" if device.keys.empty?

            config.host.device.merge!(device)
            config.save
            process_results config.host.device, view: :config_host_device_list
          end
        end

        subcommand 'unset', 'set device settings' do
          option '--queue-depth', :flag, 'unset device setting queue depth'
          option '--nomerges', :flag, 'unset device queue setting nomerges'
          option '--rq-affinity', :flag, 'unset device queue setting rq affinity'
          option '--scheduler', :flag, 'unset device queue setting scheduler'
          option '--rotational', :flag, 'unset device queue setting rotational'
          option '--add-random', :flag, 'unset device queue setting add_random'
          option '--nr-requests', :flag, 'unset device queue setting nr_requests'

          def exec
            config_support!
            device = {
              'queue_depth' => queue_depth?,
              'nomerges'    => nomerges?,
              'rq_affinity' => rq_affinity?,
              'scheduler'   => scheduler?,
              'rotational'  => rotational?,
              'add_random'  => add_random?,
              'nr_requests' => nr_requests?,
            }.reject { |_, v| v.nil? }

            signal_usage_error "specify a device setting to unset" if device.keys.empty?

            device.each_pair do |k,v|
              config.host.device.delete(k)
            end
            config.save
            process_results config.host.device, view: :config_host_device_list
          end
        end

        subcommand 'clear', 'clear all device settings' do
          def exec
            config_support!
            config.host.device.each_pair do |k,v|
              config.host.device.delete(k)
            end
            config.save
            process_results config.host.device, view: :config_host_device_list
          end
        end

        subcommand ['list', 'ls'], 'show configured device settings' do
          def exec
            process_results config.host.device, view: 'config_host_device_list'
          end
        end
      end
    end

    subcommand 'api', 'configure command line api details' do
      subcommand 'ssl-verify-peer', 'enable or disable peer verification' do
        parameter 'MODE', 'verification mode (enabled, disabled)' do |v|
          enum_map v, enabled: true, disabled: false
        end
        def exec
          config_support!
          config.api.ssl_verify_peer = mode
          config.save
        end
      end

      subcommand 'host', 'configure and display api host information' do
        self.default_subcommand = 'list'

        subcommand ['list', 'ls'], 'list all configured api hosts' do
          mixin :info_extraction

          def exec
            config_support!
            process_results config.api.hosts, view: 'config_api_host_list'
          end
        end

        subcommand 'add', 'add a new api host' do
          parameter 'HOST', 'hostname, hostname:port or url of api endpoint'
          option '--label', 'LABEL', 'specify explicit label',
                 attribute_name: :exp_label
          option '--default', :flag, 'set host as default'

          def exec
            config_support!

            case host
            when %r{^https?://}
              url = URI.parse host
            else
              url = URI.parse "https://#{host}/api"
            end

            label = exp_label || url.hostname
            hosts = config.api.hosts
            runtime_error "Host #{label} already configured" if hosts.key? label

            hosts[label] = { url: url.to_s }

            if default?
              hosts.each_value { |v| v.delete('default') }
              hosts[label].default = true
            end

            config.save
          end
        end

        subcommand 'update', 'update host configuration' do
          parameter 'LABEL', 'host label'
          option '--url', 'URL', 'set host url'
          option '--default', :flag, 'set host as default'

          def exec
            config_support!

            hosts = config.api.hosts
            unless hosts.key? label
              runtime_error "Host #{label} not configured"
            end

            hosts[label].tap do |host|
              if default?
                hosts.each_value { |v| v.delete('default') }
                host.default = true
              end
              host.url = url if url
            end

            config.save
          end
        end

        subcommand 'remove', 'remove host configuration' do
          parameter 'LABEL', 'host label'

          def exec
            config_support!

            hosts = config.api.hosts
            unless hosts.delete(label)
              runtime_error "Host #{label} not configured"
            end

            config.save
          end
        end
      end
    end
  end
end
