# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

module Blockbridge::CLI::Command
  class DatastoreBase < Base
    def avail_devs(node_id, arg = :dev, sds_id = nil)
      devs = obj_list(:dev, arg_map: { dev: arg })
      devs.reject! { |d| d.node_id != node_id }
      devs.select! do |d|
        d.sds_id.nil? || (sds_id && sds_id == d.sds_id)
      end
      devs
    end

    def pool_dev(sys_ds, arg = :dev)
      devs = avail_devs(sys_ds.node_id, arg, sys_ds.id)
      dev = only_one("device", devs, "storage")

      unless sys_ds.pool.devs.any? { |d| d.dev_id == dev.id }
        # device not in pool
        runtime_error "Device #{dev} is not in #{sys_ds} pool."
      end

      dev
    end

    mixin :pool_rebalance do |opts = {}|
      option_group 'Rebalancing', opts do
        apiopt('--rebalance', 'FLAG',
               'enable or disable rebalancing (enable or disable)',
               param_name: %w(rebalance enabled),
               param_coerce: :boolean)
        apiopt('--rebalance-threshold', 'PERCENT',
               'out-of-balance threshold percent',
               param_name: %w(rebalance threshold),
               param_coerce: :percent)
        apiopt('--rebalance-start-time', 'TIME',
               'time of day to start rebalancing',
               param_name: %w(rebalance start_time),
               param_coerce: :time_of_day)
        apiopt('--rebalance-end-time', 'TIME',
               'time of day to stop rebalancing',
               param_name: %w(rebalance end_time),
               param_coerce: :time_of_day)
        apiopt('--rebalance-bw-limit', 'BYTES/s',
               "rebalance bandwidth limit (bytes or 'none' for no limit)",
               param_name: %w(rebalance bandwidth_limit),
               param_coerce: :bw_limit)
        apiopt('--rebalance-evaluate', 'MODE',
               'rebalance evaluation mode (once, continuous)',
               param_name: %w(rebalance evaluate)) do |v|
          enum_check v, %w(once continuous)
        end
      end
    end

    mixin :compression do |opts = {}|
      option_group 'Compression', opts do
        apiopt('--compression', 'FLAG',
               'enable or disable compression (enable or disable)',
               param_name: %w(compression enabled),
               param_coerce: :boolean)
        apiopt('--compression-codec', 'CODEC',
               'compression codec',
               param_name: %w(compression codec)) do |v|
                 enum_check v, %w(none economy adaptive_economy adaptive_performance performance)
               end
        apiopt('--compression-iops-limit', 'IOPS',
               'limit sustained I/O operations per second',
               param_name: %w(compression iops_limit)) do |v|
                 (val = Integer(v)) == 0 ? nil : val
               end
        apiopt('--compression-activity-monitor', 'MINUTES',
               'prevent compression of active data (in minutes)',
               param_name: %w(compression activity_monitor),
               param_coerce: :integer)
        apiopt('--compression-cycle-detection', 'MINUTES',
               'recurring compression of active data (in minutes)',
               param_name: %w(compression cycle_detection),
               param_coerce: :integer)
      end
    end
  end

  class Datastore < DatastoreBase
    self.default_subcommand = 'list'

    mixin :node, :datastore
    mixin :info_extraction

    subcommand ['list', 'ls'], 'enumerate system datastores' do
      mixin :info_filtering
      mixin :sort_by

      def exec
        datastores = obj_list(:datastore)
        process_results datastores, view: 'datastore_list'
      end
    end

    subcommand 'info', 'show detailed system datastore information' do
      mixin :info_filtering
      mixin :info_extraction

      def exec
        datastores = obj_list(:datastore)
        process_results datastores
      end
    end

    class DatastoreCreate < DatastoreBase
      mixin :node
      mixin :label_notes, obj_name: 'datastore'
      mixin :uuid

      apiopt('--dev', 'DEVICE',
             'specify device for both metadata storage and data journal',
             param_name: %w(meta dev_id),
             param_group: 'aux_params')
      apiopt('--meta-dev', 'DEVICE', 'specify metadata storage device',
             param_name: %w(meta dev_id),
             param_group: 'aux_params')
      apiopt('--meta-size', 'SIZE', 'specify metadata db size', required: true,
             param_name: %w(meta size),
             param_group: 'aux_params') do |v|
        parse_bytes(v)
      end
      apiopt('--meta-cache-size', 'SIZE',
             'specify metadata cache size (% of size or bytes)',
             param_name: %w(meta cache_size),
             param_group: 'aux_params')

      apiopt('--journal-dev', 'DEVICE', 'specify data journal device',
             param_name: %w(data log_dev_id),
             param_group: 'aux_params')
      apiopt('--journal-size', 'SIZE', 'specify data journal size',
             required: true,
             param_name: %w(data log_size),
             param_group: 'aux_params') do |v|
        parse_bytes(v)
      end

      apiopt('--encryption', 'MODE',
             'set default block encryption mode (aes128-xts, aes256-gcm)',
             param_name: 'blk_encode_default') do |v|
        enum_map(v,
                 :"aes128-xts" => 'aes128-xts-4k',
                 :"aes256-gcm" => 'aes256-gcm-4k')
      end

      apiopt('--[no-]sequential-write-direct', :flag, 'enable sequential write direct to pool',
             :param_name => %w(data sequential_write_direct))
      apiopt('--write-merge-timeout', 'TIMEOUT', 'write merge timeout (in microseconds)',
             :param_name => %w(data write_merge_timeout)) do |v| Integer(v) end
      apiopt('--segment-size', 'SIZE', 'segment size (128KiB, 144KiB, 256KiB)',
             :param_name => %w(data segment_size)) do |v|
        enum_check v, %w(128KiB 144KiB 256KiB)
      end

      apiopt('--[no-]data-log-compress', :flag, 'enable data journal compression',
             :param_name => %w(data log_compress))

      apiopt('--[no-]data-log-bypass', :flag, 'enable data journal bypass (XTS-only)',
             :param_name => %w(data log_bypass))

      apiopt('--size-reservable', 'BYTES', 'size reservable',
             :param_name => 'size_reservable') do |v| parse_bytes(v) end
      apiopt('--size-reserve-hard-threshold', 'BYTES', 'hard reserve threshold',
             :param_name => 'size_reserve_hard_threshold') do |v| parse_bytes(v) end
      apiopt('--size-reserve-soft-threshold', 'BYTES', 'soft reserve threshold',
             :param_name => 'size_reserve_soft_threshold') do |v| parse_bytes(v) end

      apiopt('--cplx', 'ID', 'processor complex assignment', param_name: :cplx_id) do |v|
        Integer(v)
      end
      apiopt('--networks', 'NETS', 'comma-separated network interface bindings') do |v|
        v.split(/,\s+/).map { |s| obj_list(:net, value: s, restrict: :only_one).id }
      end

      mixin :pool_rebalance
      mixin :compression
      mixin :tags_create, param_name: %w(product tags)
      mixin :info_extraction

      def exec
        sn = only_one_storage_node

        if dev
          devs = avail_devs(sn.id)
          mdev = jdev = only_one("device", devs, "storage")
        else
          unless meta_dev && journal_dev
            usage_error "you must specify devices for metadata and journal storage"
          end

          devs = avail_devs(sn.id, :meta_dev)
          mdev = only_one("device", devs, "meta")

          devs = avail_devs(sn.id, :journal_dev)
          jdev = only_one("device", devs, "journal")
        end

        unless meta_cache_size.nil?
          @meta_cache_size = Integer(disk_quota_size(meta_cache_size, meta_size, 1))
        end

        create_params = {
          node_id: sn.id,
          meta: {
            dev_id: mdev.id,
            size: meta_size,
            cache_size: meta_cache_size,
          },
          data: {
            log_dev_id: jdev.id,
            log_size: journal_size,
          }
        }.deep_merge(api_params)

        sys_ds = api.sys_ds.create(create_params)
        process_obj_created sys_ds
      end
    end
    subcommand 'create', 'create a new system datastore', DatastoreCreate

    subcommand 'update', 'update existing system datastore' do
      mixin :label_notes, obj_name: 'datastore', hidden: false

      # provisioning params
      apiopt('--[no-]vss-provisioning', :flag, 'permit automatic vss provisioning')
      apiopt('--iops-rating', 'IOPS', 'iops rating') do |v| Integer(v) end
      apiopt('--[no-]iops-scheduling', :flag, 'enforce iops reservations and limits')
      apiopt('--iops-io-size', 'BYTES', 'IOPS accounting I/O size') do |v| Integer(v) end
      apiopt('--size-reservable', 'BYTES', 'size reservable',
             :param_name => 'size_reservable') do |v| parse_bytes(v) end
      apiopt('--size-reserve-hard-threshold', 'BYTES', 'hard reserve threshold',
             :param_name => 'size_reserve_hard_threshold') do |v| parse_bytes(v) end
      apiopt('--size-reserve-soft-threshold', 'BYTES', 'soft reserve threshold',
             :param_name => 'size_reserve_soft_threshold') do |v| parse_bytes(v) end
      apiopt('--iops-reservable', 'PERCENT', 'percent iops reservable',
             :param_name => 'iops_reserve_pct') do |v| parse_percent(v) end

      apiopt('--[no-]label-tag', :flag, 'use label as tag')

      apiopt('--[no-]sequential-write-direct', :flag, 'enable sequential write direct to pool',
             :param_name => %w(data sequential_write_direct))

      apiopt('--write-merge-timeout', 'TIMEOUT', 'write merge timeout (in microseconds)',
             :param_name => %w(data write_merge_timeout)) do |v| Integer(v) end

      apiopt('--[no-]data-log-compress', :flag, 'enable data journal compression',
             :param_name => %w(data log_compress))

      apiopt('--[no-]data-log-bypass', :flag, 'enable data journal bypass (XTS-only)',
             :param_name => %w(data log_bypass))

      apiopt('--cplx', 'ID', 'processor complex assignment', param_name: 'cplx_id') do |v|
        Integer(v)
      end
      apiopt('--no-cplx', :flag, 'unbind datastore from processor complex',
             param_name: 'cplx_id', param_value: nil)
      apiopt('--networks', 'NETS', 'comma-separated network interface bindings') do |v|
        v.split(/,\s+/).map { |s| obj_list(:net, value: s, restrict: :only_one).id }
      end
      apiopt('--all-networks', :flag, 'bind to all dataplane network interfaces',
             param_name: 'networks', param_value: nil)

      mixin :pool_rebalance, param_name_prefix: 'pool'
      mixin :compression
      mixin :tags_update, :param_name => %w(product tags)

      def exec
        sys_ds = obj_list(:datastore, restrict: :only_one)

        update_params = api_params.tap do |p|
          if new_tags = set_tags(nil, sys_ds.product.tags)
            p[:product] = { tags: new_tags }
          else
            p.delete(:product)
          end
        end

        sys_ds = api.sys_ds.update(sys_ds.id, update_params)
        process_obj_updated sys_ds
      end
    end

    subcommand 'remove', 'remove a system datastore' do
      def exec
        sys_ds = obj_list(:datastore, restrict: :only_one)

        _rsp = api.sys_ds.remove(sys_ds.id)
        process_obj_removed sys_ds
      end
    end

    subcommand 'stats', 'display datastore statistics' do
      apiopt('--start', 'TIME', 'specify query range start time',
             param_name: 'tstart')
      apiopt('--end', 'TIME', 'specify query range end time',
             param_name: 'tend')
      apiopt('--daily', :flag, 'specify daily data granularity',
             param_name: 'granularity', param_value: 'day')
      apiopt('--hourly', :flag, 'specify hourly data granularity',
             param_name: 'granularity', param_value: 'hour')
      apiopt('--stat', 'NAME',
             'limit results to one or more statistic (comma separated)',
             param_name: 'stats')
      apiopt('--detail', :flag, 'display detailed tabular data',
             param_name: 'with_points', param_value: true)
      option('--list', :flag, 'enumerate available stats')

      def exec
        sys_ds = obj_list(:datastore, restrict: :only_one)
        usage = api.sys_ds.query_usage(sys_ds.id, **api_params)

        if list?
          process_results({ obj: sys_ds, stats: usage }, view: 'datastore_stats_list')
        else
          process_results(obj: sys_ds, stats: usage)
        end
      end
    end

    subcommand 'pool', 'manage datastore pool devices' do

      subcommand 'add', 'add a datastore pool device' do
        option '--dev', 'DEVICE', 'specify device to add'

        def exec
          sys_ds = obj_list(:datastore, restrict: :only_one)
          pool_devs = sys_ds.pool.devs

          devs = avail_devs(sys_ds.node_id, :dev, sys_ds.id)
          dev = only_one("device", devs, "storage")

          # is dev already in pool?
          if pool_devs.any? { |d| d.dev_id == dev.id }
            runtime_error "Device #{dev} already in #{sys_ds} pool."
          end

          rsp = api.sys_ds.pool_add(sys_ds.id, dev_id: dev.id)
          process_action_complete(sys_ds,
            desc: "Added device #{dev} to storage pool for #{sys_ds}")
        end
      end

      subcommand 'del', 'remove a datastore pool device' do
        option '--dev', 'DEVICE', 'specify device to remove'

        def exec
          sys_ds = obj_list(:datastore, restrict: :only_one)
          pool_devs = sys_ds.pool.devs

          devs = avail_devs(sys_ds.node_id, :dev, sys_ds.id)
          dev = only_one("device", devs, "storage")

          unless pool_devs.any? { |d| d.dev_id == dev.id }
            # device not in pool
            runtime_error "Device #{dev} is not in #{sys_ds} pool."
          end

          rsp = api.sys_ds.pool_delete(sys_ds.id, dev_id: dev.id)
          process_action_complete(sys_ds,
            desc: "Removed device #{dev} from storage pool for #{sys_ds}")
        end
      end

      subcommand 'destage', 'destage a datastore pool device' do
        option '--dev', 'DEVICE', 'specify device to destage'

        def exec
          sys_ds = obj_list(:datastore, restrict: :only_one)
          pool_devs = sys_ds.pool.devs

          devs = avail_devs(sys_ds.node_id, :dev, sys_ds.id)
          dev = only_one("device", devs, "storage")

          unless pool_devs.any? { |d| d.dev_id == dev.id }
            # device not in pool
            runtime_error "Device #{dev} is not in #{sys_ds} pool."
          end

          rsp = api.sys_ds.pool_destage(sys_ds.id, dev_id: dev.id)
          process_action_complete(sys_ds,
            desc: "Initiated destage for #{dev} in #{sys_ds}")
        end
      end

      subcommand 'update', 'update datastore pool configuration' do
        apiopt('--strategy', 'STRATEGY', "change the pool allocation strategy\n" \
               "(weighted, ranked, ranked_weighted, " \
               "balance_usage_pct, balance_avail)") do |v|
          enum_map(v,
                   weighted: 'weighted',
                   ranked: 'ranked',
                   ranked_weighted: 'ranked_weighted',
                   balance_usage_pct: 'balanced_unused_percentage',
                   balance_avail: 'balanced_unused')
        end

        # optional, if specified will update a dev's pool configuration
        option '--dev', 'DEVICE', 'update device parameters'
        apiopt('--rank', 'RANK', 'update device rank',
               :param_group => :dev_params) do |v| Integer(v) end
        apiopt('--weight', 'WEIGHT', 'update device weight',
               :param_group => :dev_params) do |v| Integer(v) end
        apiopt('--[no-]allocation', :flag, 'allocation permitted',
               :param_group => :dev_params)

        def exec
          sys_ds = obj_list(:datastore, restrict: :only_one)

          update_params = {
            pool: sys_ds.pool
          }

          update_params[:pool].merge!(api_params)

          if dev
            ex_dev = pool_dev(sys_ds)

            # find existing entry
            pool_ent = update_params[:pool][:devs].find do |ent|
              ent.dev_id == ex_dev.id
            end

            # update with supplied params
            pool_ent.merge!(dev_params || {})
          end

          sys_ds = api.sys_ds.update(sys_ds.id, update_params)
          process_obj_updated(sys_ds,
            desc: "Updated datastore pool configuration for #{sys_ds}")
        end
      end

    end

  end
end
