# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

module Blockbridge::CLI::Command
  ##########################################################################
  # storage device configuration
  ##########################################################################
  class Dev < Base
    self.default_subcommand = 'list'

    mixin :site
    mixin :node
    mixin :dev
    mixin :info_extraction

    class DevDiscover < Base
      option '--inuse', :flag, 'include in-use devices'
      mixin :info_extraction
      mixin :site
      mixin :node

      def exec
        node = only_one_storage_node
        process_results api.node.discover_devices(node.id, {})
      end
    end
    subcommand 'discover', 'enumerate attached storage devices', DevDiscover

    subcommand ['list', 'ls'], 'enumerate configured storage devices' do
      mixin :info_filtering
      mixin :sort_by

      def exec
        devs = obj_list(:os_dev, arg_map: { os_dev: 'dev' })
        process_results devs, view: 'dev_list'
      end
    end

    subcommand 'info', 'show detailed storage device information' do
      mixin :info_filtering
      mixin :info_extraction

      def exec
        devs = obj_list(:os_dev, arg_map: { os_dev: 'dev' })
        process_results devs
      end
    end

    class CreateDev < Base
      mixin :label_notes, obj_name: 'device'
      mixin :uuid
      option '--path', 'PATH', 'device path', :required => true
      option '--size', 'SIZE', 'specify raw device size (default auto-detect)' do |v|
        parse_bytes(v)
      end
      option '--preserve-path', :flag, 'use exact supplied path (default auto-selects best)'

      mixin :site
      mixin :node
      mixin :info_extraction

      def exec
        node = only_one_storage_node
        devs = api.node.discover_devices(node.id, {})
        devs.reject! { |dev| dev.inuse }

        # find path in discovered devices
        devs.select! do |dev|
          dev.devlinks.include?(path) || dev.path == path || dev.syspath == path
        end

        # ensure only one dev was selected
        if devs.empty? && !preserve_path?
          runtime_error "Failed to find available device at path: #{path}"
        end
        if devs.length > 1
          runtime_error "Found more than one device at path: #{path}"
        end

        dev = devs.first

        if preserve_path?
          # use exact path
          bp = path

          if dev.nil? && size.nil?
            runtime_error "Unable to detect device size; use --size to specify one."
          end
        else
          bp = dev.preferred_path
        end

        create_params = {
          node_id: node.id,
          path: bp,
          raw_size: size || dev.raw_size,
        }.merge(api_params)

        if !create_params[:label] && dev && dev.name
          create_params[:label] = dev.name
        end

        os_dev = api.os_dev.create(create_params)
        process_obj_created os_dev
      end
    end
    subcommand 'create', 'create storage device configuration', CreateDev

    subcommand 'update', 'update storage device parameters' do
      mixin :label_notes, obj_name: 'device', hidden: false
      apiopt '--path', 'PATH', 'change device path'

      def exec
        dev = obj_list(:os_dev, restrict: :only_one,
                       arg_map: { os_dev: 'dev' })

        process_obj_updated api.os_dev.update(dev.id, api_params)
      end
    end

    subcommand 'remove', 'remove configuration for a storage device' do

      def exec
        dev = obj_list(:os_dev, restrict: :only_one,
                       arg_map: { os_dev: 'dev' })

        _rsp = api.os_dev.remove(dev.id)

        process_obj_removed dev
      end
    end

  end
end
