# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

module Blockbridge::CLI::Command
  AdapterDiags = lambda do
    subcommand 'adapter', 'api adapter debugging utilities', self, level: :dev do
      subcommand 'info', 'dump adapter status' do
        mixin :info_extraction
        def exec
          data = api.status.info
          process_results data
        end
      end

      subcommand 'authorization', 'dump api authorization state', self, level: :dev do
        mixin :info_extraction
        def exec
          data = api.status.authorization
          process_results data
        end
      end

      subcommand 'debug', 'dump api adapter debug state', self, level: :dev do
        mixin :info_extraction

        apiopt('--objects', :flag, 'include object statistics')

        def exec
          data = api.status.debug
          process_results data
        end
      end

      subcommand 'platform', 'dump api adapter os platform', self, level: :dev do
        mixin :info_extraction
        def exec
          data = api.status.platform
          process_results data
        end
      end

      subcommand 'ps', 'dump api adapter os process listing', self, level: :dev do
        apiopt('--uid',   'ID', 'limit to uid')
        apiopt('--user',  'NAME', 'limit to username')
        apiopt('--gid',   'ID', 'limit to uid')
        apiopt('--group', 'NAME', 'limit to username')
        apiopt('--state', 'STATE', 'limit to state')
        apiopt('--cmd',   'NAME', 'limit to command name',
               :param_name => 'name')

        option('--cmdline', :flag, 'show full command line')

        mixin :info_extraction
        mixin :info_filtering
        mixin :sort_by
        def exec
          data = api.status.process_list(api_params)
          process_results data
        end
      end
    end
  end

  class Diags < Base
    def diags_data
      runtime_error 'Diagnostics directory unspecified' unless diags_dir

      Dir.glob(File.join(diags_dir, 'diags.*')).map do |file|
        c = MultiJson.load(File.read(file), symbolize_keys: true) rescue nil

        unless debug?
          # hide common diags unless --debug is enabled
          next if c.nil?

          case (c[:exception] && c[:exception][:class])
          when 'Clamp::UsageError', 'Interrupt', 'Blockbridge::CLI::LookupError'
            next
          end
        end

        { file: file, contents: c }
      end.compact
    rescue StandardError
      []
    end

    subcommand ['list', 'ls'], 'list saved diagnostics information', self, level: :expert do
      mixin :info_extraction
      mixin :info_filtering
      mixin :sort_by
      def exec
        process_results diags_data, view: 'diags_list'
      end
    end

    subcommand 'info', 'show detail about saved diagnostics information', self, level: :expert do
      mixin :info_extraction
      mixin :info_filtering

      option ['--num', '-n'], 'NUM', 'show NUM diags info', default: 1
      option '--all', :flag, 'show all diags'

      def exec
        d = diags_data
        d = d.last(num.to_i) unless all?
        process_results d
      end
    end

    subcommand 'clear', 'clear saved diagnostics information', self, level: :expert do
      def exec
        diags_data.each do |f|
          File.unlink(f.file)
        end
      end
    end

    instance_exec(&AdapterDiags) if ENV.key? 'BLOCKBRIDGE_DIAGS'
  end
end
