# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'blockbridge/cli/command/disk_replication'

module Blockbridge::CLI::Command
  class Disk < Base
    include Blockbridge::CLI::Formatters

    self.default_subcommand = 'list'

    mixin :vdisk_create_common do
      option_group nil, hidden: true do
        apiopt('--multi-map', 'VAL', 'allow disk to be multi-mapped') do |v| parse_boolean(v) end
        apiopt('--read-only', 'VAL', 'configure read-only device', level: :expert) do |v| parse_boolean(v) end
        apiopt('--removable', 'VAL', 'configure removable device emulation', level: :expert) do |v| parse_boolean(v) end

        mixin :tags_create
        mixin :xmd_refs_create
        mixin :xref
      end
      mixin :info_extraction
    end

    mixin :vdisk_compression do
      option_group nil, hidden: true do
        apiopt('--compression', 'CODEC', 'set disk compression codec') do |v|
          enum_check(v, %w[none inherit economy adaptive_economy
                           adaptive_performance performance])
        end
      end
    end

    mixin :vdisk_create_data_sizing do
      option_group 'Quota', hidden: true do
        option '--size-reserve', 'RESERVE', 'size reservation (in bytes or % capacity)'
        option '--size-limit', 'LIMIT', 'size limit (in bytes or % capacity)'
        option '--size-notify', 'NOTIFY', 'size notification threshold (in bytes or % capacity)'
        option '--iops-limit', 'IOPS', 'iops limit' do |s| Integer(s) end
      end

      option_group :nil, :hidden => true do
        apiopt('--readahead', 'VAL', 'size readahead (in bytes) or "none"') do |v| parse_readahead(v) end
        apiopt('--data-cache', 'VAL', 'configure data cache (default: true)') do |v| parse_boolean(v) end
        apiopt('--write-once', 'VAL', 'configure write-once device', level: :expert) do |v| parse_boolean(v) end
      end
    end

    mixin :vdisk_external do
      option_group 'External' do
        apiopt('--target-host', 'HOST', 'external target host IP address', :param_group => :ext_params)
        apiopt('--target-port', 'PORT', 'external target port', default: 3260, :param_group => :ext_params) do |s| Integer(s) end
        apiopt('--target-name', 'IQN', 'external target IQN', :param_group => :ext_params)
        apiopt('--target-lun', 'LUN', 'external target LUN', :param_group => :ext_params) do |v|
          Integer(v)
        end
        apiopt('--initiator-login', 'LOGIN', 'initiator auth user', :param_group => :ext_params)
        apiopt('--initiator-pass', 'PASS', 'initiator auth password', :param_group => :ext_params)
        apiopt('--initiator-name', 'IQN', 'initiator IQN', :param_group => :ext_params)
      end

      option_group 'Mutual Auth' do
        apiopt('--mutual-auth', :flag, 'require mutual authentication', :param_group => :ext_params)
        apiopt('--target-login', 'LOGIN', 'target auth user', :param_group => :ext_params)
        apiopt('--target-pass', 'PASS', 'target auth password', :param_group => :ext_params)
      end

      option_group 'Transport Security' do
        apiopt('--transport-mode', 'MODE', 'transport security mode', {
                 :default => 'insecure',
                 :param_name => %w(transport mode),
                 :param_group => :ext_params,
               }) do |val|
                enum_check(val, %w(insecure tls))
        end
      end
    end

    class DiskCreateExternal < Base
      mixin :vss
      mixin :label_notes, obj_name: 'external disk'
      mixin :vdisk_external
      option_group 'Attachment', :hidden => true do
        mixin :snapshot, desc: 'select snapshot to attach by label or serial number'
        mixin :disk, desc: 'select disk to attach by label or serial number'
        option '--clone', :flag, 'attach clone of disk '
      end
      mixin :vdisk_create_common

      def exec
        if !snapshot && !disk && !target_name
          usage_error "Target name not specified."
        end

        if (snapshot || disk) && target_name
          usage_error "Only one of snapshot, disk or target-name can be specified"
        end

        if snapshot
          snap = obj_list(:snapshot, restrict: :only_one, arg_map: { vss: nil })
        end

        if disk && !snapshot
          disk = obj_list(:disk, restrict: :only_one, arg_map: { :vss => nil })
        end

        # select vss
        vss = obj_list(:vss, restrict: :only_one)

        create_params = api_params.merge(vss_id: vss.id, external: true)
        if snapshot 
          create_params.merge!(snapshot_id: snap.id)
        elsif disk
          create_params.merge!(vdisk_id: disk.id)
          create_params.merge!(clone: true) if clone?
        else
          create_params.merge!(ext_params)
        end

        created = api.vdisk.create(create_params)
        if disk 
          desc = "Attached #{created} to disk #{disk}"
        elsif snapshot
          desc = "Attached #{created} to snapshot #{snapshot}"
        else
          desc = "Attached #{created} to external #{target_name} at #{target_host}"
        end

        process_obj_created created, desc: desc
      end
    end

    class DiskCreateClone < Base
      mixin :vss, desc: 'specify destination vss for clone by label or serial number'

      option_group 'Capacity', :hidden => true do
        apiopt(['-c', '--capacity'], 'CAPACITY', 'specify larger capacity for clone') do |c|
          parse_bytes(c)
        end
      end

      mixin :label_notes, obj_name: 'clone'

      option_group 'Source basis' do
        mixin :snapshot, desc: 'select snapshot as basis by label or serial number'
        mixin :disk, desc: 'select disk as basis by label or serial number'
        mixin :obj_store, desc: 'select object store as basis by label or serial number'
        mixin :backup_id, desc: 'select backup to restore from object store'
      end

      option_group 'Thicken' do
        apiopt('--thicken', :flag, 'pro-actively copy all data from source basis')
        apiopt('--bandwidth-limit', 'LIMIT', 'bandwidth limit (in bytes/s) for thicken operation') do |v|
          parse_bytes(v)
        end
      end

      mixin :vdisk_create_common
      mixin :vdisk_create_data_sizing
      mixin :vdisk_compression

      def exec
        # select basis
        if !snapshot && !disk && !obj_store && !backup_id
          usage_error "One of snapshot, disk or object store must be specified as basis"
        end

        if snapshot
          snap      = obj_list(:snapshot,
                               restrict: :only_one,
                               arg_map: { vss: nil })
          capacity  = snap.capacity if capacity.nil?
          snap_disk = api.vdisk.info(snap.vdisk_id)
          cs_ratio  = snap_disk.cap_size_ratio
          @vss      = snap.vss_id if vss.nil?
        end

        if disk and !snapshot
          disk     = obj_list(:disk,
                              restrict: :only_one,
                              arg_map: { :vss => nil })
          capacity = disk.capacity if capacity.nil?
          cs_ratio = disk.cap_size_ratio
          @vss     = disk.vss_id if vss.nil?
        end

        if obj_store || backup_id
          if snapshot || disk
            usage_error "Only one of either --disk/--snapshot or --obj-store can be specified to clone from"
          end

          unless backup_id
            usage_error "No backup id specified to clone from #{obj_store}"
          end

          ostore = obj_list(:obj_store,
                            restrict: :only_one,
                            arg_map: { :vss => nil })
          @thicken = true if @thicken.nil?
        end

        # select vss
        vss = obj_list(:vss, restrict: :only_one)

        if !bandwidth_limit.nil? && !thicken?
          usage_error "Bandwidth limit only valid if thickening"
        end

        cs_ratio = vss.cap_size_ratio if cs_ratio.nil?

        # now that disk capacity is known, compute size values which may be
        # expressed as % of capacity.
        self.size_reserve = disk_quota_size(size_reserve, capacity, cs_ratio)
        self.size_limit   = disk_quota_size(size_limit, capacity, cs_ratio)
        self.size_notify  = disk_quota_size(size_notify, capacity, cs_ratio)

        # build create params
        create_params = api_params.merge(vss_id: vss.id)
        create_params[:quota] = update_quota_params
        if snapshot
          create_params.merge!(snapshot_id: snap.id)
        elsif disk
          create_params.merge!(vdisk_id: disk.id)
        elsif backup_id
          create_params.merge!(obj_store_id: ostore.id)
          create_params.merge!(backup_id: backup_id)
        end
        create_params.merge!(thicken: true) if thicken?
        create_params.merge!(bandwidth: bandwidth_limit) unless bandwidth_limit.nil?

        created = api.vdisk.create(create_params)
        if snapshot
          desc = "Created clone #{created} from snapshot #{snapshot} of #{snap_disk}"
        elsif disk
          desc = "Created clone #{created} from disk #{disk}"
        elsif backup_id
          desc = "Created clone #{created} from object store #{ostore} backup-id #{backup_id}"
        end

        process_obj_created created, desc: desc
      end
    end

    class DiskCreate < Base
      self.optional_subcommands = true

      subcommand 'clone', 'clone a vdisk', DiskCreateClone
      subcommand 'external', 'create an external vdisk', DiskCreateExternal

      mixin :vss

      apiopt(['-c', '--capacity'], 'CAPACITY', 'requested disk capacity') do |c|
        parse_bytes(c)
      end

      mixin :label_notes, obj_name: 'disk'
      mixin :vdisk_create_common
      mixin :vdisk_create_data_sizing
      mixin :vdisk_compression

      def exec
        # select vss
        vss = obj_list(:vss, restrict: :only_one)

        # now that disk capacity is known, compute size values which may be
        # expressed as % of capacity.
        cs_ratio = vss.cap_size_ratio if cs_ratio.nil?
        self.size_reserve = disk_quota_size(size_reserve, capacity, cs_ratio)
        self.size_limit   = disk_quota_size(size_limit, capacity, cs_ratio)
        self.size_notify  = disk_quota_size(size_notify, capacity, cs_ratio)

        # build create params
        create_params = api_params.merge(vss_id: vss.id)
        create_params[:quota] = update_quota_params

        process_obj_created api.vdisk.create(create_params)
      end
    end

    subcommand 'create', 'create a new virtual disk', DiskCreate, level: :basic

    mixin :vss, flags: { level: :expert }
    mixin :disk, :param_path => '/id'
    mixin :info_extraction

    subcommand ['list', 'ls'], 'enumerate virtual disks', self, level: :basic do
      mixin :info_filtering
      mixin :sort_by

      def exec
        disks = obj_list(:disk)
        process_results disks, view: 'disk_list'
      end
    end

    subcommand 'info', 'show detailed virtual disk information', self, level: :basic do
      mixin :info_filtering
      mixin :info_extraction

      def exec
        disks = obj_list(:disk)
        process_results disks
      end
    end

    class DiskUpdateExternal < Base
      mixin :label_notes, obj_name: 'disk', hidden: false
      apiopt('--multi-map', 'VAL', 'allow disk to be multi-mapped') do |v| parse_boolean(v) end
      apiopt('--read-only', 'VAL', 'configure read-only device') do |v| parse_boolean(v) end
      apiopt('--write-once', 'VAL', 'configure write-once device') do |v| parse_boolean(v) end
      apiopt('--data-cache', 'VAL', 'configure data cache (default: true)') do |v| parse_boolean(v) end
      apiopt('--removable', 'VAL', 'configure removable device emulation') do |v| parse_boolean(v) end
      apiopt('--readahead', 'VAL', 'size readahead (in bytes) or "none"') do |v| parse_readahead(v) end

      mixin :tags_update
      mixin :xmd_refs_update

    end

    subcommand 'update', 'update basic disk parameters', self, level: :basic do
      self.optional_subcommands = true

      mixin :disk
      mixin :label_notes, obj_name: 'disk', hidden: false

      option_group nil, level: :advanced do
        apiopt('--multi-map', 'VAL', 'allow disk to be multi-mapped') do |v| parse_boolean(v) end
        apiopt('--read-only', 'VAL', 'configure read-only device') do |v| parse_boolean(v) end
      end

      option_group nil, level: :expert do
        apiopt('--write-once', 'VAL', 'configure write-once device', level: :expert) do |v| parse_boolean(v) end
        apiopt('--data-cache', 'VAL', 'configure data cache (default: true)') do |v| parse_boolean(v) end
        apiopt('--removable', 'VAL', 'configure removable device emulation') do |v| parse_boolean(v) end
        apiopt('--readahead', 'VAL', 'size readahead (in bytes) or "none"') do |v| parse_readahead(v) end
      end

      mixin :tags_update
      mixin :xmd_refs_update
      mixin :vdisk_compression

      # update external
      subcommand 'external', 'update external disk parameters' do
        mixin :vdisk_external

        def exec
          disk = obj_list(:disk, restrict: :only_one)
          if disk.type != 'external'
            runtime_error "#{disk} is not an external disk; an external disk must be specified"
          end

          update_params = api_params

          # assign updated xmd refs
          set_xmd_refs(update_params, disk.xmd_refs)

          # update tags
          set_tags(update_params, disk.tags)
          update_params

          update_params.merge!(ext_params)

          disk = api.vdisk.update(disk.id, update_params)
          process_obj_updated disk
        end
      end

      # update non-external
      def exec
        disk = obj_list(:disk, restrict: :only_one)

        update_params = api_params

        # assign updated xmd refs
        set_xmd_refs(update_params, disk.xmd_refs)

        # update tags
        set_tags(update_params, disk.tags)

        disk = api.vdisk.update(disk.id, update_params)
        process_obj_updated disk
      end
    end

    subcommand 'quota', 'manage disk quota' do
      subcommand 'update', 'update disk quota' do
        option '--size-reserve', 'RESERVE', 'size reservation (in bytes or % capacity)'
        option '--size-limit', 'LIMIT', 'size limit (in bytes or % capacity)'
        option '--no-size-limit', :flag, 'remove size limit'
        option '--size-notify', 'NOTIFY', 'size notification threshold (in bytes or % capacity)'
        option '--no-size-notify', :flag, 'remove size notification'
        option '--iops-limit', 'IOPS', 'iops limit' do |s| Integer(s) end
        option '--no-iops-limit', :flag, 'remove iops limit'

        def exec
          disk = obj_list(:disk, restrict: :only_one)

          # compute size values which may be expressed as % of capacity.
          self.size_reserve = disk_quota_size(size_reserve, disk.capacity, disk.cap_size_ratio)
          self.size_limit   = disk_quota_size(size_limit, disk.capacity, disk.cap_size_ratio)
          self.size_notify  = disk_quota_size(size_notify, disk.capacity, disk.cap_size_ratio)

          quota_params = update_quota_params(disk.quota)

          api.vdisk.quota_update(disk.id, quota_params)

          # re-read new disk
          disk = api.vdisk.info(disk.id)
          process_obj_updated disk, desc: "Updated quota for disk #{disk}"
        end
      end
    end

    subcommand 'resize', 'alter the capacity of an existing disk', self, level: :basic do
      option ['-c', '--capacity'], 'CAPACITY', 'requested disk capacity', :required => true do |c|
        parse_bytes(c)
      end
      option '--confirm', :flag, 'confirm dangerous operation'

      def exec
        params = {
          capacity: capacity,
          warning_confirm: confirm?
        }

        disk = obj_list(:disk, restrict: :only_one)
        rsp = api.vdisk.resize(disk.id, params)

        # re-read disk data
        desc = "Resized disk #{disk}: #{format_bytes(disk.capacity)} --> #{format_bytes(capacity)}"
        disk = api.vdisk.info(disk.id)
        process_obj_updated disk, desc: desc
      end
    end

    subcommand 'thicken', 'thicken disk from basis' do
      def exec
        disk = obj_list(:disk, restrict: :only_one)

        rsp = api.vdisk.thicken(disk.id)
        disk = api.vdisk.info(disk.id)
        if disk.task&.op == 'thicken'
          desc = "Thickening disk #{disk}."
        else
          desc = "#{disk} is thickened."
        end
        process_action_complete disk, desc: desc
      end
    end

    subcommand 'compress', 'compress a disk' do
      def exec
        disk = obj_list(:disk, restrict: :only_one)

        rsp = api.vdisk.compress(disk.id)
        disk = api.vdisk.info(disk.id)
        if disk.task&.op == 'compress'
          desc = "Compressing disk #{disk}."
        else
          desc = "#{disk} is now compressed."
        end
        process_action_complete disk, desc: desc
      end
    end

    subcommand 'backup', 'initiate a backup' do
      mixin :obj_store
      mixin :snapshot

      apiopt('--label', 'LABEL', 'optional backup label', level: :basic)
      option '--take-snapshot', :flag, 'take new transient snapshot for the backup'

      def exec
        obj_store = obj_list(:obj_store, restrict: :only_one)
        disk = obj_list(:disk, restrict: :only_one)
        unless take_snapshot?
          snapshot = obj_list(:snapshot, restrict: :only_one)
        end

        backup_params = {}.tap do |h|
          h[:obj_store_id] = obj_store.id
          h[:snapshot_id] = snapshot ? snapshot.id : nil
        end.merge(api_params)

        rsp = api.vdisk.backup(disk.id, backup_params)
        desc = case rsp['status']
               when 'success'
                 'Completed'
               when 'running'
                 'Initiated'
               else
                 "[task-status=#{rsp['status']}]"
               end

        process_action_complete({ disk: disk, rsp: rsp },
                                desc: "#{desc} backup of #{disk} with "\
                                      "backup-id #{rsp[:backup_id]}")
      end
    end

    subcommand 'remove', 'remove a virtual disk', self, level: :basic do
      option('--batch', :flag, 'allow multiple disks to be removed', level: :advanced)
      option('--force', :flag, 'force a disk to be removed while still attached')

      def exec
        batch_remove :disk do |disk|
          attached, _ = disk_get_attached(vdisk: disk)
          if !attached.empty? && !force?
            usage_error "Disk #{disk} cannot be removed while attached. Detach first or use '--force' to force remove."
          end
          not_found_ok { api.vdisk.remove(disk.id) }
          process_obj_removed disk
        end
      end
    end

    subcommand 'load', 'load removable media', self, level: :expert do
      mixin :profile

      option '--clone', :flag, 'load transient clone', :default => false
      apiopt '--label', 'LABEL', 'transient clone label (default auto-selects label)', level: :basic

      def exec
        profile = obj_list(:profile, restrict: :only_one)
        disk = obj_list(:disk, restrict: :only_one)

        if clone?
          clone_params = api_params.merge(profile_id: profile.id)
          clone = api.vdisk.load_cloned_media(disk.id, clone_params)
          process_obj_created clone, desc: "Loaded transient clone of disk #{disk}: #{clone}"
        else
          rsp = api.vdisk.load_media(disk.id, profile_id: profile.id)
          process_action_complete disk, desc: "Loaded removable disk #{disk}"
        end
      end
    end

    subcommand 'eject', 'eject removable media', self, level: :expert do
      def exec
        disk = obj_list(:disk, restrict: :only_one)

        api.vdisk.eject_media(disk.id)
        process_action_complete disk, desc: "Ejected removable media for disk #{disk}"
      end
    end

    subcommand 'lock', 'lock disk encryption key vault' do
      apiopt '--passphrase', 'PASS', 'specify encryption key vault passphrase (default is to prompt)'

      def exec
        disk = obj_list(:disk, restrict: :only_one)

        lock_params = api_params.tap do |h|
          h[:passphrase] ||= prompt_for_password "passphrase", confirm: true
        end

        rsp = api.vdisk.lock(disk.id, lock_params)
        process_action_complete disk, desc: "Locked disk #{disk}"
      end
    end

    subcommand 'unlock', 'unlock disk encryption key vault' do
      apiopt '--passphrase', 'PASS', 'specify encryption key vault passphrase (default is to prompt)'

      def exec
        disk = obj_list(:disk, restrict: :only_one)

        unlock_params = api_params.tap do |h|
          h[:passphrase] ||= prompt_for_password "passphrase"
        end

        rsp = api.vdisk.unlock(disk.id, unlock_params)
        process_action_complete disk, desc: "Unlocked disk #{disk}"
      end
    end

    subcommand 'replication', 'manage replication', Blockbridge::CLI::Command::DiskReplication

    subcommand 'barrier', 'manage disk I/O barrier', self, level: :expert do
      subcommand 'create', 'create a time-limited disk I/O barrier', self, level: :expert do
        apiopt '--expires', 'EXPIRES', 'relative expiration time (in ms)', :required => true do |s| Integer(s) end
        apiopt '--uuid', 'UUID', 'barrier unique ID', :required => true

        def exec
          disk = obj_list(:disk, restrict: :only_one)

          create_params = api_params

          rsp = api.vdisk.barrier_create(disk.id, create_params)
          process_action_complete disk, desc: "I/O barrier active on disk #{disk}"
        end
      end

      subcommand 'remove', 'remove a disk I/O barrier', self, level: :expert do
        option '--uuid', 'UUID', 'barrier unique ID', :required => true

        def exec
          disk = obj_list(:disk, restrict: :only_one)
          rsp = api.vdisk.barrier_remove(disk.id, uuid)
          if rsp.barrier == "expired"
            runtime_error "I/O barrier expired on disk #{disk}"
          end
          process_action_complete disk, desc: "I/O barrier removed from disk #{disk}"
        rescue Blockbridge::Api::NotFoundError, Excon::Errors::NotFound, Excon::Errors::Gone
          lookup_error "No disk found that matches '#{disk}'."
        rescue Blockbridge::Api::ConflictError
          lookup_error "Barrier exists, but not with uuid '#{uuid}'."
        end
      end
    end
  end
end
