# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'blockbridge/cli/helpers/replication'

module Blockbridge::CLI::Command
  ##########################################################################
  # REPLICATION
  ##########################################################################

  class DiskReplication < Base
    include Blockbridge::CLI::Formatters # blech
    include Blockbridge::CLI::Helpers::Replication

    option '--master-vss', 'VSS', 'specify master vss by label or serial number'
    option [ '--master', '-M' ], 'DISK', 'specify master disk by label or serial number'
    option '--slave-vss', 'VSS', 'specify slave vss by label or serial number'
    option [ '--slave', '-S' ], 'DISK', 'specify slave disk by label or serial number'

    mixin :info_extraction

    # SLAVE configuration -- adding and removing a master disk to a slave.
    subcommand 'master', 'add and remove a replication master' do
      subcommand 'add', 'add a replication master' do
        self.description = <<-EOS
          Assign a master disk to a slave. This only affects the slave-side
          configuration.  After assigning the master disk to the slave, use
          the 'disk replication slave add' command to create the other half of
          the replication relationship.

          See also: 'disk replication join'
        EOS

        def exec
          slave_disk = lookup_slave
          master_disk = lookup_master

          if slave_disk.replication.master
            if slave_disk.replication.master.disk_id == master_disk.id
              success "Master disk #{master_disk} already assigned to slave #{slave_disk}."
            end

            runtime_error "Disk #{slave_disk} already has a configured replication master."
          end

          rsp = api.vdisk.add_replication_master(slave_disk.id, vdisk_id: master_disk.id)
          process_action_complete([master_disk, slave_disk],
            desc: "Assigned master disk #{master_disk} to slave disk #{slave_disk}")
        end
      end

      subcommand 'remove', 'remove a configured replication master' do
        self.description = <<-EOS
          Remove master disk assignment from a slave. This only affects the
          slave-side configuration. To remove the master-side configuration, use
          the 'disk replication slave remove' command.

          See also: 'disk replication split'
        EOS

        def exec
          slave_disk = lookup_slave

          if slave_disk.replication.master.nil?
            runtime_error "Disk #{slave_disk} has no configured replication master."
          end

          # first, examine the current master assignment.
          master_id = slave_disk.replication.master.disk_id

          # look up the master disk...
          master_desc = master_sn = nil

          begin
            ex_master = api.vdisk.info(master_id)

            # master disk exists
            master_sn = ex_master.serial
            master_desc = ex_master
          rescue Blockbridge::Api::NotFoundError, Excon::Errors::NotFound, Excon::Errors::Gone
            # master disk deleted
            master_desc = master_sn = slave_disk.replication.master.disk_serial
          end

          if master_specified?
            # lookup & verify slave-side master configuration
            _cur_master = slave_master_cfg(slave_disk, master_desc)
          end

          rsp = api.vdisk.remove_replication_master(slave_disk.id, master_id)
          process_action_complete([ex_master, slave_disk].compact,
            desc: "Removed master #{master_desc} from slave #{slave_disk}")
        end
      end
    end

    # MASTER configuration -- adding, configuring and removing a slave from
    # a master.
    subcommand 'slave', 'add and remove replication slaves' do
      subcommand 'add', 'add a replication slave' do
        def exec
          master_disk = lookup_master
          slave_disk = lookup_slave

          slaves = master_disk.replication.slaves

          if slaves.length > 0
            # currently limit to single slave per master, until the GUI can
            # support more complex configurations.
            if slaves.length > 1
              runtime_error "Disk #{master_disk} has more than one slave. Oops."
            end

            cur_slave = slaves.first
            if cur_slave.disk_id == slave_disk.id
              unless display_raw
                runtime_success "Slave disk #{slave_disk} already assigned " \
                                "to master #{master_disk}."
              end
              return # already assigned, no more work to do.
            end

            runtime_error "Disk #{master_disk} already has a configured replication slave."
          end

          # add slave configuration to master disk
          rsp = api.vdisk.add_replication_slave(master_disk.id, vdisk_id: slave_disk.id)
          process_action_complete([master_disk, slave_disk],
            desc: "Assigned slave #{slave_disk} to master #{master_disk}")
        end
      end

      subcommand 'remove', 'remove a configured replication slave' do
        def exec
          master_disk = lookup_master

          slaves = master_disk.replication.slaves

          if slaves.length == 0
            runtime_error "Disk #{master_disk} has no configured replication slave."
          end

          if slaves.length > 1
            runtime_error "Disk #{master_disk} has more than one slave. Oops."
          end

          slave_id = slaves.first.disk_id

          #################################################################
          # look up the existing slave disk information. if the object has
          # been deleted, all we have is the ID and serial number.
          #################################################################
          slave_sn = nil
          slave_desc = nil

          begin
            ex_slave = api.vdisk.info(slave_id)

            # slave disk exists
            slave_sn = ex_slave.serial
            slave_desc = ex_slave
          rescue Blockbridge::Api::NotFoundError, Excon::Errors::NotFound, Excon::Errors::Gone
            # slave disk removed, all we have left is the s/n
            slave_desc = slave_sn = slaves.first.disk_serial
          end

          if slave_specified?
            # look up & verify master-side slave configuration
            _slave_cfg = master_slave_cfg(master_disk, slave_desc)
          end

          # remove slave configuration from the master disk
          rsp = api.vdisk.remove_replication_slave(master_disk.id, slave_id)
          process_action_complete([master_disk, ex_slave].compact,
            desc: "Removed slave #{slave_desc} from master #{master_disk}")
        end
      end

    end # end replication slave

    subcommand 'update', 'modify replication parameters', self, level: :basic do
      option_group nil, level: :basic do
        apiopt('--mode', 'MODE', "replication mode (sync, async, ordered)",
               :param_name => %w(config mode)) do |val|
          enum_map(val, sync: 'sync', async: 'async', ordered: 'async-ordered')
        end

        apiopt('--rebuild-method', 'METHOD', 'replication rebuild method (repair, sync)',
               :param_name => %w(config recovery)) do |val|
          enum_map(val, repair: 'drl', sync: 'none')
        end

        apiopt('--command-lag', 'NUM', 'async replication command lag (# of commands)',
               :param_name => %w(config max_lag_cmds)) do |val|
          Integer(val)
        end

        apiopt('--transfer-lag', 'NUM', 'async replication transfer lag (bytes)',
               :param_name => %w(config max_lag_bytes)) do |val|
          parse_bytes(val)
        end

        apiopt('--rebuild-bandwidth', 'BYTES', 'replication rebuild bandwith limit (bytes/s)',
               :param_name => %w(config rebuild_bandwidth)) do |val|
          parse_bytes(val)
        end

        apiopt('--timeout', 'SECONDS', 'replication I/O timeout (seconds)',
               :param_name => %w(config timeout)) do |val|
          (Float(val) * 1000).to_i
        end

        apiopt('--similarity-detection', 'METHOD', 'hashing method (sha1, none)',
               :param_name => %w(config rebuild_hash)) do |val|
          enum_map(val, sha1: true, none: false)
        end
      end

      def exec
        # get existing configuration
        master_disk = lookup_master(existing: true)

        # ensure a slave is configured
        cur_slave = master_slave_cfg(master_disk)

        # merge existing config with supplied options
        update_params = Hashie::Mash.new({ config: cur_slave.config })
        update_params.deep_merge!(api_params)

        if update_params[:config][:mode] == 'sync'
          # for synchronous replication, command lag and transfer lag are
          # not permitted.
          if command_lag
            usage_error "Command lag (--comand-lag) only makes sense for asynchronous replication."
          end
          if transfer_lag
            usage_error "Transfer lag (--transfer-lag) only makes sense for asynchronous replication."
          end

          # these will be left-over if switching from async to sync
          update_params[:config].delete :max_lag_cmds
          update_params[:config].delete :max_lag_bytes
        end

        rsp = api.vdisk.update_replication_slave(master_disk.id, cur_slave.disk_id,
                                                 update_params)
        process_action_complete([master_disk, cur_slave],
          desc: "Updated master #{master_disk} replication " \
                "parameters for slave #{cur_slave.disk_serial}")
      end
    end

    subcommand 'enable', 'enable replication' do
      def exec
        # get existing configuration
        master_disk = lookup_master(existing: true)

        # ensure a slave is configured
        cur_slave = master_slave_cfg(master_disk)

        case cur_slave.state
        when 'enabled'
          success "Replication from master #{master_disk} to slave " \
                  "#{cur_slave.disk_serial} already in state 'enabled'."
        when 'suspended'
          runtime_error "Replication from master #{master_disk} to slave "   \
                        "#{cur_slave.disk_serial} in state 'suspended' -- to resume" \
                        " replication, use the 'resume' command."
        end

        rsp = api.vdisk.update_replication_slave(master_disk.id, cur_slave.disk_id,
                                                 state: 'enable')
        process_action_complete([master_disk, cur_slave],
          desc: "Enabled replication from master #{master_disk} to " \
                "slave #{cur_slave.disk_serial}")

      end
    end # replication enable

    subcommand 'disable', 'disable replication' do
      def exec
        # get existing configuration
        master_disk = lookup_master(existing: true)

        # ensure a slave is configured
        cur_slave = master_slave_cfg(master_disk)

        if cur_slave.state == 'disabled'
          success "Replication from master #{master_disk} to slave " \
                  "#{cur_slave.disk_serial} already in state 'disabled'."
        end

        rsp = api.vdisk.update_replication_slave(master_disk.id, cur_slave.disk_id,
                                                 state: 'disable')
        process_action_complete([master_disk, cur_slave],
          desc: "Disabled replication from master #{master_disk} to " \
                "slave #{cur_slave.disk_serial}. (previous state = "  \
                "#{cur_slave.state})")
      end
    end # replication disable

    subcommand 'suspend', 'suspend replication' do
      def exec
        # get existing configuration
        master_disk = lookup_master(existing: true)

        # ensure a slave is configured
        cur_slave = master_slave_cfg(master_disk)

        case cur_slave.state
        when 'suspended'
          success "Replication from master #{master_disk} to slave " \
                  "#{cur_slave.disk_serial} already in state 'suspended'."
        when 'enabled'
          # proceed...
        else
          runtime_error "Replication from master #{master_disk} to slave " \
                        "#{cur_slave.disk_serial} is not in state 'enabled'."      \
                        " (current state = #{cur_slave.state})"
        end

        rsp = api.vdisk.update_replication_slave(master_disk.id, cur_slave.disk_id,
                                                 state: 'suspend')
        process_action_complete([master_disk, cur_slave],
          desc: "Suspended replication from master #{master_disk} to " \
                "slave #{cur_slave.disk_serial}. (previous state = "   \
                "#{cur_slave.state})")
      end
    end # replication suspend

    subcommand 'resume', 'resume replication' do
      def exec
        # get existing configuration
        master_disk = lookup_master(existing: true)

        # ensure a slave is configured
        cur_slave = master_slave_cfg(master_disk)

        case cur_slave.state
        when 'enabled'
          success "Replication from master #{master_disk} to slave " \
                  "#{cur_slave.disk_serial} already in state 'enabled'."
        when 'suspended'
          # proceed
        else
          runtime_error "Replication from master #{master_disk} to slave " \
                        "#{cur_slave.disk_serial} is not in state 'suspended'."    \
                        " (current state = #{cur_slave.state})"
        end

        rsp = api.vdisk.update_replication_slave(master_disk.id, cur_slave.disk_id,
                                                 state: 'resume')
        process_action_complete([master_disk, cur_slave],
          desc: "Resumed replication from master #{master_disk} to "  \
                "slave #{cur_slave.disk_serial}. (previous state = "  \
                "#{cur_slave.state})")
      end
    end # replication suspend

    subcommand 'join', 'set up master and slave for replication' do
      self.description = <<-EOS
        Configure both master and slave disks to be replication peers with
        each other. On successful configuration, optionally enable replication
        (if the --enable flag is used).

        This command requires both master and slave disks to be specified.
      EOS

      option '--enable', :flag, 'enable replication'

      def exec
        unless master_specified? && slave_specified?
          usage_error "Join requires master and slave disks to be specified."
        end

        # get master & slave disks
        master_disk = lookup_master
        slave_disk = lookup_slave

        if slave_disk.replication.master
          master_sn = slave_disk.replication.master.disk_serial
          runtime_error "Slave disk #{slave_disk} already has " \
                        "#{master_sn} configured as a master."
        end

        if master_disk.replication.slaves.length > 0
          slave_sn = master_disk.replication.slaves.first.disk_serial
          runtime_error "Master disk #{master_disk} already has " \
                        "#{slave_sn} configured as a slave."
        end

        attach_params = {
          master_id: master_disk.id,
          slave_id: slave_disk.id
        }

        rsp = api.vdisk.attach_replication_peers(attach_params)

        if enable?
          rsp = api.vdisk.update_replication_slave(master_disk.id, slave_disk.id,
                                                   state: 'enable')
        end

        process_action_complete([master_disk, slave_disk],
          desc: "Configured replication for master #{master_disk} and " \
                "slave #{slave_disk}")
      end
    end # replication join

    subcommand 'split', 'unconfigure master and slave replication (best-effort)' do
      self.description = <<-EOS
        Detach and remove configuration for both replication peers.
        This operation is best-effort; examine each disk on completion to
        ensure the desired effect has occurred.

        If only the slave is specified, its currently assigned master (if any)
        will be detached. If only a master is specified, its currently assigned
        slave will be detached. If both peers are specified, the configuration
        must match exactly or the split will not proceed.
      EOS

      def exec
        unless master_specified? || slave_specified?
          usage_error "Split requires master and/or slave disks"
        end

        # get master & slave disks
        master_disk = lookup_master if master_specified?
        slave_disk = lookup_slave if slave_specified?

        if master_disk
          master_id = master_disk.id
          master_desc = master_disk
        else
          # lookup slave's current master.
          cur_master = slave_master_cfg(slave_disk)
          master_id = cur_master.disk_id
          master_desc = cur_master.disk_serial
        end

        if slave_disk
          slave_id = slave_disk.id
          slave_desc = slave_disk
        else
          # look up master's current slave.
          cur_slave = master_slave_cfg(master_disk)
          slave_id = cur_slave.disk_id
          slave_desc = cur_slave.disk_serial
        end

        detach_params = {
          master_id: master_id,
          slave_id: slave_id
        }

        rsp = api.vdisk.detach_replication_peers(detach_params)
        process_action_complete([master_disk, slave_disk].compact,
          desc: "Unconfigured replication for master #{master_desc} and " \
                "slave #{slave_desc}")
      end
    end # replication split

    subcommand 'rebuild', 'rebuild a halted replication slave' do
      def exec
        master_disk = lookup_master(existing: true)
        cur_slave = master_slave_cfg(master_disk)

        rebuild_params = {
          master_id: master_disk.id,
          slave_id: cur_slave.disk_id
        }

        rsp = api.vdisk.rebuild_replication_slave(rebuild_params)

        process_action_complete([master_disk, cur_slave],
          desc: "Initiated replication rebuild for master #{master_disk} and " \
                "slave #{cur_slave.disk_serial}.")
      end
    end # replication rebuild

    subcommand 'status', 'show current replication status' do
      mixin :info_filtering
      mixin :info_extraction

      def exec
        # lookup current configuration
        master_disks = lookup_masters
        add_replication_status(master_disks)

        # sort by vss/disk labels
        master_disks.sort_by do |d|
          [ d.replication_status.master.vss.label, d.label ]
        end

        process_results master_disks
      end
    end

  end
end
