# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

module Blockbridge::CLI::Command
  class DiskSet < Base
    self.default_subcommand = 'list'

    def get_disk_set
      if label && ref
        usage_error "Please specify only one of '--ref' and '--label'"
      end

      xmds = []
      if ref
        xmd = api.xmd.info(ref)
        unless xmd.tags.include? 'disk-set'
          xmd = nil
        else
          xmds = [ xmd ]
        end
      else
        xmds = api.xmd.list(api_params).select { |x| x.tags.include? 'disk-set' }
      end

      only_one(:xmd, xmds, "disk-set")
    end

    def disk_set_get_disks(set_ref)
      vdisks = []
      set_ref.rdeps.select { |e|
        e.type == 'vdisk'
      }.map { |e|
        not_found_ok { api.vdisk.info(e.id) }
      }.compact.sort { |a, b|
        a.vss_id <=> b.vss_id
      }.each_with_index { |vdisk, idx|
        attached = vdisk.xmd.select { |xmd|
          xmd.tags.include?('host-attach') && xmd.data.attach.data.host
        }.map { |xmd|
          "#{xmd.data.attach.data.host}:#{xmd.data.attach.data.dev}"
        }

        vdisk = {
          vdisk:     vdisk,
          idx:       idx,
          attached:  attached,
        }
        vdisks.push vdisk
      }
      vdisks
    end

    subcommand 'create', 'create disk-set' do
      apiopt('--label', 'LABEL', 'disk-set label', :required => true)
      option '--num', 'NUM', 'number of disks to create', {
             :required => true,
        } do |val|
          Integer(val)
        end
      apiopt(['-c', '--capacity'], 'CAPACITY', 'capacity of each disk created disk', {
             :required => true,
             :param_group => :query_params,
        }) do |c|
          parse_bytes(c)
        end
      apiopt('--iops', 'IOPS', 'reserve iops for each disk', {
             :param_group => :query_params,
        })  do |val|
          Integer(val)
        end
      option '--tag', 'TAG', 'require disk placement diversified among specified tags (may be used multiple times)',
             :multivalued => true
      option '--unique-nodes', :flag, 'require unique node for each disk'
      option '--unique-tags', :flag, 'require unique tag for each disk'

      def disk_set_provision(dryrun: false, ref:)
        exclude = []

        # provision disks up to the num requested
        num.times { |n|
          # search catalog for match. loop to reset excludes on failure
          while true do
            params = {
              tags: {
                include: tag_list - exclude,
                exclude: exclude,
              }
            }.merge!(query_params)

            # search for product
            search = api.product.search_create(params)

            # validate search results
            if search.length > 0
              product = search[0]
              print "disk:#{n} placed on node #{product.node}"

              # add successfully queried tag to exclude list
              if tag_list.length == 0
                exclude.push product.node
              elsif t = tag_list.find { |t| product.tags.include? t }
                print" with tag '#{t}'."
                exclude.push t
              end
              puts

              # found product, now provision it
              break
            else
              if unique_nodes? or unique_tags?
                runtime_error "Unable to place disk:#{n} #{unique_nodes? ? 'on unique node' : 'with unique tag'}."
              end
              if exclude.length == 0
                runtime_error "No resources found to create disk:#{n}."
              else
                exclude = []
              end
            end
          end

          # dryrun only does query. skip provisioning if dryrun
          next if dryrun

          # set provision params with token
          params = {
            token: product.token,
            vss: {
              xmd_refs:         [ ref ],
              xref:             ref,
              label:            "#{label}:disk:#{n}",
            },
            disk: {
              create:   true,
              xmd_refs: [ ref ],
              xref:     ref,
              label:    nil,
              tags:     product.tags,
            },
          }

          # provision
          vss = api.vss.create(params)
        }
      end

      def disk_set_group_create(ref:)
        # set xmd create params
        xmd_params = {
          ref: ref,
          xref: ref,
          label: label,
          tags: [ 'disk-set' ],
          data: {
            group: {
              _schema:  "group",
              _publish: false,
              _type:    "disk-set",
              data: {
                label:     label,
                disks_num: num,
              },
            },
          },
        }

        # create xmd for disk-set group
        xmd = api.xmd.create(xmd_params)
      end

      def exec
        set_ref = "disk-set-#{::SecureRandom.uuid}"

        xmds = api.xmd.list(api_params).select { |x| x.tags.include? 'disk-set' }
        unless xmds.length == 0
          signal_usage_error "Disk-set already found with label #{label}"
        end

        if unique_tags? && tag_list.empty?
          signal_usage_error "unique tags flagged as required but no tags specified"
        end

        begin
          puts "Creating disk-set '#{label}' with #{num} disks..."

          # create disk-set group
          # grouping done by generated uuid
          group = disk_set_group_create(ref: set_ref)

          # now provision the disks into the disk-set
          disk_set_provision(ref: set_ref)

          # done
          process_action_complete group, {desc: "Disk-set '#{label}' created."}
        rescue
          api.objects.remove_by_xref(set_ref, scope: "xmd,vss,vdisk")
          raise
        end
      end
    end

    subcommand 'remove', 'remove disk-set' do
      option '--ref', 'REF', 'disk-set ref'
      apiopt '--label', 'LABEL', 'disk-set label'

      def exec
        set = get_disk_set

        api.objects.remove_by_xref(set.ref, scope: "xmd,vss,vdisk")

        process_action_complete set, desc: "Disk-set '#{set.label}' removed."
      end
    end

    subcommand 'attach', 'attach disk-set disks to local host' do
      option '--ref', 'REF', 'disk-set ref'
      apiopt '--label', 'LABEL', 'disk-set label'
      option '--multipath', :flag, 'use multiple paths to connect to target'
      option('--local-interface', 'LOCAL',
             'use specified local interfaces for multipath',
             :multivalued => true) do |v|
        @local_interfaces_set = true
        v
      end
      option('--remote-interface', 'REMOTE',
             'use specified remote interfaces for multipath',
             :multivalued => true) do |v|
        @remote_interfaces_set = true
        v
      end
      option '--[no-]local-bind', :flag,
             '(multipath) bind to local interfaces',
             default: true
      option '--mknod', :flag, 'make device node after attach'
      option '--mknod-prefix', 'MKNOD-PREFIX', 'device node prefix for mknod'

      def local_intf
        return unless @local_interfaces_set
        @local_interfaces.flat_map { |v| l.split(/,\s*/) }
      end

      def remote_intf
        return unless @remote_interfaces_set
        @remote_interfaces.flat_map { |v| l.split(/,\s*/) }
      end

      def exec
        remotes = remote_intf
        locals = local_intf

        if !local_bind? && !multipath?
          signal_usage_error "--no-local-bind option only valid for --multipath"
        end

        unless multipath?
          if locals && locals.length > 1
            signal_usage_error "multiple local interfaces specified when " \
                               "not multipathing"
          end
          if remotes && remotes.length > 1
            signal_usage_error "multiple remote interfaces specified when " \
                               "not multipathing"
          end
        end

        set = get_disk_set
        set.vdisks_in_set = disk_set_get_disks(set)

        # each disk in disk-set, attach it
        set.vdisks_in_set.select do |d|
          d.attached.length == 0
        end.each do |d|
          # attach disk to host
          vss = api.vss.info(d.vdisk.vss_id)
          ref = "host-attach-#{SecureRandom.uuid}"
          params = {
            vdisk:   d.vdisk,
            vss:     vss,
            ref:     ref,
            locals:  locals,
            remotes: remotes,
          }

          process_results(disk_attach(params), view: :disk_attach)
        end

        puts "All disks in disk-set '#{set.label}' attached."
      end
    end

    subcommand 'detach', 'detach disk-set disks from local host' do
      option '--ref', 'REF', 'disk-set ref'
      apiopt '--label', 'LABEL', 'disk-set label'
      option '--force', :flag, 'force detach of remote target'

      def exec
        set = get_disk_set
        set.vdisks_in_set = disk_set_get_disks(set)

        # each disk in disk-set, detach it
        set.vdisks_in_set.select { |d|
          d.attached.length > 0
        }.each { |d|
          vss = api.vss.info(d.vdisk.vss_id)
          attached, dangling = disk_get_attached(vdisk: d.vdisk, vss: vss)

          unless attached.length > 0
            if dangling.length > 0
              # remove ref from disk
              params = {
                xmd_refs: d.vdisk.xmd_refs - dangling,
              }
              api.vdisk.update(d.vdisk.id, params)
            end
          end

          # detach each attached host
          attached.each { |xmd|
            params = {
              vdisk: d.vdisk,
              vss:   vss,
              xmd:   xmd,
            }
            disk_detach(params)
          }
        }

        puts "All disks in disk-set '#{set.label}' detached."
      end
    end

    subcommand ['list', 'ls'], 'enumerate disk-sets' do
      apiopt '--label', 'LABEL', 'disk-set label'

      mixin :info_filtering
      mixin :sort_by

      def exec
        sets = api.xmd.list(api_params).select { |x|
          x.tags.include? 'disk-set'
        }.each { |x|
          x.vdisks_in_set = disk_set_get_disks(x)
        }
        process_results sets, view: 'disk_set_list'
      end
    end

    subcommand 'info', 'show detailed disk-set information' do
      apiopt '--label', 'LABEL', 'disk-set label'

      mixin :info_filtering
      mixin :info_extraction

      def exec
        sets = api.xmd.list(api_params).select { |x|
          x.tags.include? 'disk-set'
        }.each { |x|
          x.vdisks_in_set = disk_set_get_disks(x)
        }
        process_results sets
      end
    end
  end
end
