# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

module Blockbridge::CLI::Command
  class Host < Base
    subcommand 'attach', 'attach disk to host operating system', level: :basic do
      mixin :disk
      mixin :vss
      option('--repair', :flag, 'repair disk attached to this host',
             default: false)
      option('--persist', :flag, 'persist attachment across reboot',
             default: false)

      option_group 'Metadata', level: :advanced do
        option '--ref', 'REF', 'supply reference ID to attachment'
        option '--notes', 'NOTES', 'supply notes to attachment',
               default: 'generated by host attach'
        option '--name', 'NAME', 'supply attachment name'
        option(['--tag', '-T'], 'TAGS', 'add tag (may be used multiple times)',
             :multivalued => true,
             :attribute_name => :tags)
      end

      option_group 'Policy', level: :advanced do
        option('--secure-token', :flag, 'connect with secure access token',
               default: false, level: :advanced)
        option_group nil, :hidden => true do
          option('--passphrase-file', 'FILE', 'read token passphrase from file')
          option('--restrict-ip', :flag, 'restrict access to this host\'s IPs',
                 default: false)
          option('--read-only', :flag, 'attach for read-only access',
                 default: false)
          option('--write-only', :flag, 'attach for write-only access',
                 default: false)
          option('--read-write', :flag, 'attach for read-write access',
                 default: true)
          option('--multi-writer', :flag, 'allow multiple writers with write access',
                 default: false)
        end
      end

      option_group 'Clone', level: :advanced do
        option('--clone', :flag, 'create a temporary clone',
             default: false)
        option_group nil, :hidden => true do
          option('--snapshot', 'SNAPSHOT', 'use specific snapshot as basis')
          option('--snapshot-tag', 'SNAPSHOT-TAG', 'select basis by snapshot tag')
          option('--snapshot-latest', :flag, 'select most recent snapshot as basis')
        end
      end

      option_group 'Transport' do
        option('--tls', :flag, 'use TLS transport encryption',
               default: false)
        option('--ipsec', :flag, 'use IPsec transport encryption',
               default: false)
        option('--compression', 'ALGO', 'enable TLS compression (zlib, rle)', level: :advanced) do |v|
          tls_cfg[:compression] = enum_check(v, %w[zlib rle])
        end
        option('--tls-debug', 'LEVEL', 'set TLS debug log level (default: 5)', level: :advanced) do |v|
          tls_cfg[:debug] = Integer(v)
        end
        option_group nil, :hidden => true do
          option('--ipsec-hostname', 'HOSTNAME',
                 'connect to specified hostname for IPsec')
        end
      end

      option_group 'Network' do
        option('--multipath', :flag, 'use multiple paths to connect to target',
               default: false)
        option(['--ipv6', '-6'], :flag, 'use IPv6 (default: IPv4)') do |v|
          @af = Socket::AF_INET6
        end
        option_group nil, :hidden => true do
          option('--local-interface', 'LOCAL',
                 '(multipath) use specified local interfaces only',
                 :attribute_name => :local_interfaces,
                 :multivalued => true) do |v|
            @local_interfaces_set = true
            v
          end
          option '--[no-]local-bind', :flag,
                 '(multipath) bind to local interfaces (forces traffic to use specific local interfaces)',
                 default: false
          option('--remote-interface', 'REMOTE',
                 'use specified remote interfaces only',
                 :attribute_name => :remote_interfaces,
                 :multivalued => true) do |v|
            @remote_interfaces_set = true
            v
          end
        end
      end

      option_group 'System', :hidden => true do
        option '--mknod', :flag, 'make device node after attach'
        option '--mknod-prefix', 'MKNOD-PREFIX', 'device node prefix for mknod'
        option '--mknod-name', 'MKNOD-NAME', 'device name for mknod'
      end

      option_group nil, :hidden => true do
        option('--use-group', 'GROUP-REF', 'attach to existing group',
               :attribute_name => :group) do |g|
          signal_usage_error "--use-group requires an argument" if g.nil?
          g
        end
      end

      mixin :info_filtering
      mixin :info_extraction

      def tls_cfg
        @tls_cfg ||= {}
      end

      def local_intf
        return unless @local_interfaces_set
        local_interfaces.flat_map { |v| v.split(/,\s*/) }
      end

      def remote_intf
        return unless @remote_interfaces_set
        remote_interfaces.flat_map { |v| v.split(/,\s*/) }
      end

      def repair_dangling?(dangling)
        return if not repair?
        return !dangling.empty?
      end

      def repair_attached?(attached)
        return if not repair?
        return !attached.empty?
      end

      def af
        @af || Socket::AF_INET
      end

      def exec
        vdisk   = obj_list(:disk, restrict: :only_one)
        vss     = api.vss.info(vdisk.vss_id)
        @ref    = ref || "host-attach-#{SecureRandom.uuid}"
        remotes = remote_intf
        locals  = local_intf

        if read_only? && write_only?
          signal_usage_error 'both --read-only and --write-only cannot be specified together'
        end

        if read_only? && multi_writer?
          signal_usage_error '--multi-writer not valid for read-only access mode'
        end

        if ipsec? && tls?
          signal_usage_error 'only one of --tls or --ipsec can be specified at the same time'
        end

        if persist? && (ipsec? || tls?)
          signal_usage_error '--persist is not possible with --tls or --ipsec'
        end

        if (snapshot || snapshot_tag || snapshot_latest?) && !clone?
          signal_usage_error 'snapshot options not valid without --clone'
        end

        if local_bind? && !multipath?
          signal_usage_error "--local-bind option only valid for --multipath"
        end

        if mknod_prefix && !mknod?
          signal_usage_error "--mknod-prefix option only valid if " \
                             "--mknod specified"
        end

        if group && vdisk.xmd.fragments_for(schema: 'group', ref: group).empty?
          runtime_error "group specified but disk is not currently " \
                        "in that group"
        end

        unless multipath?
          if locals && locals.length > 1
            signal_usage_error "multiple local interfaces specified when " \
                               "not multipathing"
          end
          if remotes && remotes.length > 1
            signal_usage_error "multiple remote interfaces specified when " \
                               "not multipathing"
          end
        end

        # snapshot
        if clone? && (snapshot || snapshot_latest? || snapshot_tag)
          if snapshot_latest?
            snap = api.snapshot.list(vdisk.id).sort_by { |s| s.timestamp }.reverse.first
          elsif snapshot_tag
            snap = api.snapshot.list(vdisk.id).select { |s| s.tags.include? snapshot_tag }.sort_by { |s| s.timestamp }.reverse.first
          else
            snap = obj_list(:snapshot, restrict: :only_one)
          end
          if !snap
            signal_usage_error "snapshot not found from given parameters"
          end
        end

        # repair if specified and required
        attached, dangling = disk_get_attached(vdisk: vdisk, vss: vss)
        if repair_dangling?(dangling)
          # remove dangling ref from disk
          api.vdisk.update(vdisk.id, xmd_refs: vdisk.xmd_refs - dangling)
        end
        if repair_attached?(attached)
          repair = attached.select do |xmd|
            host_control.hostname == disk_attach_host(xmd)
          end
          if repair.empty?
            usage_error "Disk is not attached to this host " \
                        "and cannot be repaired. First, detach the disk."
          end

          repair.each do |xmd|
            # repair attached disk to host
            params = {
              vdisk:   vdisk,
              vss:     vss,
              xmd:     xmd,
              af:      af,
              locals:  locals,
              remotes: remotes,
              group:   group,
              tls_cfg: tls_cfg,
            }

            process_results(disk_attach_wrap(:repair, params), view: :disk_attach)
          end
        else
          begin
            # attach disk to host
            params = {
              vdisk:    vdisk,
              vss:      vss,
              ref:      ref,
              snapshot: snap,
              af:       af,
              locals:   locals,
              remotes:  remotes,
              hostname: ipsec_hostname,
              group:    group,
              mode:     attach_mode,
              tls_cfg:  tls_cfg
            }

            process_results(disk_attach_wrap(:attach, params), view: :disk_attach)
          rescue => e
            unless e.message.include? 'Disk is already attached'
              disk_attach_group_remove(ref: ref, vdisk: vdisk)
            end
            raise
          end
        end
      end
    end

    subcommand 'detach', 'detach disk from host operating system', level: :basic do
      mixin :disk
      mixin :vss
      option_group 'Detach by' do
        option '--ref', 'REF', 'detach by attachment group reference', level: :advanced
        option '--uuid', 'UUID', 'detach by disk UUID'
        option '--path', 'PATH', 'detach by device path'
        option '--name', 'NAME', 'detach by attachment name'
        option '--iqn',  'IQN',  'detach by target IQN'
      end
      #option '--unmount', :flag, 'unmount disk before detach'
      option '--force', :flag, 'force detach of remote target', default: false

      def vss_label
        vss.label || vss.serial || vss.uuid
      end

      def vdisk_label
        vdisk.label || vdisk.serial || vdisk.uuid
      end

      def lookup_vdisk
        not_found_ok { api.vdisk.info(attach_data(@xmd, 'vdisk')) } ||
          not_found_ok { api.vdisk.info(attach_data(@xmd, 'uuid')) }
      end

      def lookup_vss
        not_found_ok { api.vss.info(attach_data(@xmd, 'vss')) }
      end

      def attachment_error
        force_msg = ", use --force to clean up a stranded host session."
        if ref
          "#{ref} did not match a reference for an attachment"
        elsif uuid
          "#{uuid} did not match a UUID for an attachment".concat force_msg
        elsif path
          "#{path} did not match a path for an attachment".concat force_msg
        elsif name
          "#{name} did not match a name for an attachment".concat force_msg
        elsif iqn
          "#{iqn} did not match a target IQN for an attachment".concat force_msg
        else
          "is invalid"
        end
      end

      attr_reader :vdisk
      attr_reader :vss
      attr_reader :xmd

      def exec
        attached = []
        dangling = []
        if ref || uuid || path || name || iqn
          attached, dangling = disk_get_attached(ref: ref,
                                                 uuid: uuid,
                                                 path: path,
                                                 name: name,
                                                 iqn: iqn)
          if (@xmd = attached.first).nil? && !force?
            usage_error attachment_error
          end
          @vdisk = lookup_vdisk
          @vss   = lookup_vss
        else
          @vdisk = obj_list(:disk, restrict: :only_one_attached, desc: "attached")
          @vss   = api.vss.info(vdisk.vss_id)
          attached, dangling = disk_get_attached(vdisk: vdisk, vss: vss)
        end

        if attached.empty?
          if !dangling.empty?
            # remove ref from disk
            api.vdisk.update(vdisk.id, xmd_refs: vdisk.xmd_refs - dangling)
          end

          unless force?
            usage_error "Disk #{disk_detach_label(vss, vdisk, ref)} is not attached."
          end

          disk_detach(vdisk: vdisk, vss: vss, xmd: xmd,
                      ref: ref, uuid: uuid, path: path, name: name, iqn: iqn)
        else
          # detach each attached host
          attached.select { |xmd|
            force? || disk_attach_host(xmd) == host_control.hostname
          }.each { |xmd|
            disk_detach(vdisk: vdisk, vss: vss, xmd: xmd)
          }
        end
      end
    end

    subcommand 'info', 'show detailed host information', level: :basic do
      option '--all-hosts', :flag, 'display attachment information for all hosts (default: localhost)'
      option_group 'Select info' do
        option ['-d', '--disk'], 'DISK', 'show info only for matching disk by label or serial number'
        option ['-v', '--vss'], 'VSS', 'show info only for matching vss by label or serial number'
        option '--ref', 'REF', 'show info only for matching attachment ref'
        option '--uuid', 'UUID', 'show info only for matching disk UUID'
        option '--path', 'PATH', 'show info only for matching device path'
        option '--name', 'NAME', 'show info only for matching attachment name'
        option '--iqn',  'IQN',  'show info only for matching target IQN'
      end
      mixin :info_filtering
      mixin :info_extraction

      attr_reader :vdisk
      attr_reader :vss

      def exec
        host_info_wrap
      end

      def host_info_wrap
        EM.synchrony do
          begin
            host_control.device_settle
            host_info_exec
          ensure
            EM.stop
          end
        end
      end

      attr_reader :schema

      def xmd_match_schema(otype, objects, desc)
        objects = objects.select { |o| o.data.values.any? { |v| v._schema == schema } }
      end

      def host_info_exec
        host_name = host_control.hostname

        if disk || vss
          @vdisk = obj_list(:disk, restrict: :only_one_attached, desc: "attached")
          @vss   = api.vss.info(vdisk.vss_id)
        end

        # find attachments
        if disk || vss || ref || uuid || path || name || iqn
          attached = disk_get_attached(vdisk: vdisk,
                                       vss: vss,
                                       ref: ref,
                                       uuid: uuid,
                                       path: path,
                                       name: name,
                                       iqn: iqn).first
        else
          xmd_filter = {
            'data/attach/_schema': 'host-attach'
          }
          attached = obj_list(:xmd, custom_filter: xmd_filter).select { |xmd|
            next false unless all_hosts? || attach_data(xmd, 'host') == host_name
            true
          }
        end

        localhost = nil
        remote_hosts = {}
        attached.each do |xmd|
          attachment = attachment_info(xmd)
          if attachment[:local]
            host = localhost ||= {
              attached: [],
              local:    true,
            }
          else
            host = remote_hosts[attach_data(xmd, 'host')] ||= {
              host_name: attach_data(xmd, 'host'),
              attached:  [],
              local:     false,
            }
          end
          host[:attached].push attachment
        end

        results = ([ localhost ] + remote_hosts.values).compact.flat_map { |host|
          host[:attached]
        }
        process_results results
      end
    end
  end
end
