# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

module Blockbridge::CLI::Command
  class Net < Base
    mixin :net_services do
      apiopt('--[no-]application-service', :flag, 'enable/disable application service',
             :param_name => %w(service application enabled))
      apiopt('--[no-]management-service', :flag, 'enable/disable management service',
             :param_name => %w(service management enabled))
      apiopt('--[no-]storage-service', :flag, 'enable/disable storage service',
             :param_name => %w(service storage enabled))
      apiopt('--[no-]storage-service-shared', :flag, 'enable/disable shared storage service use',
             :param_name => %w(service storage shared))
      apiopt('--[no-]replication-service', :flag, 'enable/disable replication service',
             :param_name => %w(service replication enabled))
      apiopt('--[no-]infrastructure-service', :flag, 'enable/disable infrastructure service',
             :param_name => %w(service infrastructure enabled))
      apiopt('--[no-]external-service', :flag, 'enable/disable external service',
             :param_name => %w(service external enabled))

      apiopt('--all-services', :flag, 'enable all services', :param_name => 'service') do |v|
        {
          application: { enabled: true },
          management: { enabled: true },
          storage: { enabled: true },
          replication: { enabled: true },
          infrastructure: { enabled: true },
          external: { enabled: true },
        }
      end
      apiopt('--no-services', :flag, 'disable all services', :param_name => 'service') do |v|
        {
          application: { enabled: false },
          management: { enabled: false },
          storage: { enabled: false },
          replication: { enabled: false },
          infrastructure: { enabled: false },
          external: { enabled: false },
        }
      end
    end

    self.default_subcommand = 'list'

    mixin :net
    mixin :node
    mixin :info_extraction

    subcommand ['list', 'ls'], 'enumerate nets' do
      mixin :info_filtering
      mixin :sort_by

      def exec
        nets = obj_list(:net)
        process_results nets, view: 'net_list'
      end
    end

    subcommand 'info', 'show detailed net information' do
      mixin :info_filtering
      mixin :info_extraction

      def exec
        nets = obj_list(:net)
        process_results nets
      end
    end

    class NetDiscover < Base
      option '--inuse', :flag, 'include in-use devices'
      mixin :info_filtering
      mixin :info_extraction
      mixin :node
      mixin :site

      def exec
        node = obj_list(:node, restrict: :only_one)
        nets = api.node.discover_interfaces(node.id)

        # format nets for display
        process_results nets, show_inuse: inuse?
      end
    end
    subcommand 'discover', 'discover network interfaces', NetDiscover

    class NetCreate < Base
      mixin :label_notes, obj_name: 'net'
      mixin :uuid

      apiopt('--interface', 'NAME', 'interface name', :required => true)
      apiopt('--hostname', 'NAME', 'host name')
      apiopt('--nat-addr', 'ADDR', 'NAT address')
      apiopt('--enable', :flag, 'enable interface (default)',
             :param_name => 'enabled') do |v|
        true
      end
      apiopt('--disable', :flag, 'disable interface',
             :param_name => 'enabled') do |v|
        false
      end
      apiopt('--portals', 'IP', 'specify allowed portal addresses',
             :multivalued => true)
      apiopt('--cplx', 'ID', 'bind interface to a storage complex',
             param_name: 'cplx_id') do |v|
        Integer(v)
      end
      apiopt('--shared', :flag, 'share interface between all complexes',
             param_name: 'cplx_id', param_value: nil)
      mixin :net_services
      mixin :node
      mixin :info_extraction

      def exec
        node = obj_list(:node, restrict: :only_one)

        create_params = { node_id: node.id }.merge(api_params)

        net = api.net.create(create_params)
        process_obj_created net
      end
    end
    subcommand 'create', 'create network interface', NetCreate

    subcommand 'update', 'update existing net' do
      mixin :label_notes, obj_name: 'net', hidden: false

      apiopt('--interface', 'NAME', 'update interface name')
      apiopt('--hostname', 'NAME', 'update hostname')
      apiopt('--no-hostname', :flag, 'clear hostname setting',
             :param_name => 'hostname',
             :param_value => nil)
      apiopt('--nat-addr', 'ADDR', 'update NAT address')
      apiopt('--no-nat-addr', :flag, 'clear NAT address setting',
             :param_name => 'nat_addr',
             :param_value => nil)
      apiopt('--enable', :flag, 'enable interface',
             :param_name => 'enabled') do |v|
        true
      end
      apiopt('--disable', :flag, 'disable interface',
             :param_name => 'enabled') do |v|
        false
      end
      apiopt('--portals', 'IP', 'specify allowed portal addresses',
             :multivalued => true,
             :param_name => 'portals')
      apiopt('--all-portals', :flag, 'allow all portal addresses',
             :param_name => 'portals',
             :param_value => nil)
      apiopt('--cplx', 'ID', 'bind interface to a storage complex',
             param_name: 'cplx_id') do |v|
        Integer(v)
      end
      apiopt('--shared', :flag, 'share interface between all complexes',
             param_name: 'cplx_id', param_value: nil)
      mixin :net_services
      mixin :node

      def exec
        net = obj_list(:net, restrict: :only_one)
        net = api.net.update(net.id, api_params)

        process_obj_updated net
      end
    end

    subcommand 'remove', 'remove existing net' do
      def exec
        net = obj_list(:net, restrict: :only_one)
        rsp = api.net.remove(net.id)

        process_obj_removed net
      end
    end

  end
end
