# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'ipaddr'

module Blockbridge::CLI::Command
  class Node < Base
    self.default_subcommand = 'list'

    mixin :info_extraction

    subcommand ['list', 'ls'], 'enumerate nodes' do
      mixin :site
      mixin :node
      mixin :info_filtering
      mixin :sort_by
      mixin :identity_param, obj_name: 'node'

      option '--storage', :flag, 'only show storage nodes' do |v|
        @filter_type = 'storage'
      end
      option '--management', :flag, 'only show management nodes' do |v|
        @filter_type = 'management'
      end
      option '--disk', :flag, 'only show disk nodes' do |v|
        @filter_type = 'disk'
      end
      option '--hidden', :flag, 'include hidden nodes' do |v|
        custom_filter['$visibility'] = 'any'
      end

      def custom_filter
        @custom_filter ||= {}
      end

      def exec
        nodes = obj_list(:node, custom_filter: custom_filter)
        nodes.select! { |n| n.type == @filter_type } if @filter_type
        process_results nodes, view: 'node_list'
      end
    end

    subcommand 'info', 'show detailed node information' do
      mixin :site
      mixin :node
      mixin :info_filtering
      mixin :info_extraction
      mixin :identity_param, obj_name: 'node'

      option '--storage', :flag, 'only show storage nodes' do |v|
        @filter_type = 'storage'
      end
      option '--management', :flag, 'only show management nodes' do |v|
        @filter_type = 'management'
      end
      option '--disk', :flag, 'only show disk nodes' do |v|
        @filter_type = 'disk'
      end
      option '--enclosures', :flag, 'include enclosure data' do |v|
        custom_filter['with_enclosures'] = 'true'
      end
      option '--hidden', :flag, 'include hidden nodes' do |v|
        custom_filter['$visibility'] = 'any'
      end

      def custom_filter
        @custom_filter ||= {}
      end

      def exec
        nodes = obj_list(:node, custom_filter: custom_filter)
        nodes.select! { |n| n.type == @filter_type } if @filter_type
        process_results nodes
      end
    end

    class NodeCreate < Base
      mixin :site

      apiopt '--label', 'LABEL', 'specify node label'
      mixin :uuid
      apiopt '--notes', 'NOTES', 'specify node notes'
      apiopt('--cluster', :flag, 'node is an HA-cluster (default is solo)',
             :param_name => :mode) do |val|
        "cluster"
      end
      apiopt('--type', 'TYPE', 'specify node type (storage, disk)',
             default: 'storage') do |val|
        enum_check(val, %w(storage disk))
      end
      apiopt('--license', 'LICENSE', 'specify license for node')
      mixin :tags_create

      mixin :info_extraction

      def exec
        site = obj_list(:site, restrict: :only_one)

        node = api.node.create(api_params.merge(site_id: site.id))
        process_obj_created node
      end
    end
    subcommand 'create', 'create a new node object', NodeCreate

    subcommand 'params', 'update node parameters' do
      subcommand 'openstack', 'update openstack parameters' do
        self.default_subcommand = 'info'

        subcommand 'info', 'info on openstack parameters' do
          mixin :info_extraction
          def exec
            node = only_one_management_node
            process_results(node, view: :params_idp)
          end
        end

        subcommand 'update', 'update openstack parameters' do
          option_group 'Identity Provider' do
            apiopt('--[no-]idp', :flag, 'enable or disable identity provider',
                  :param_name => %w(openstack_idp enabled))
            apiopt('--idp-url', 'URL', 'update the identity provider url',
                  :param_name => %w(openstack_idp url))
            apiopt('--idp-label', 'LABEL', 'update the identity provider label',
                  :param_name => %w(openstack_idp label))
            apiopt('--idp-user', 'USER', 'update the identity provider administrative user',
                  :param_name => %w(openstack_idp auth user name))
            apiopt('--idp-password', 'PASSWORD', 'update the identity provider administrative password',
                  :param_name => %w(openstack_idp auth user password))
            apiopt('--idp-project', 'PROJECT', 'update the identity provider administrative project',
                  :param_name => %w(openstack_idp auth project name))
            apiopt('--idp-domain', 'DOMAIN', 'update the identity provider domain',
                  :param_name => %w(openstack_idp auth domain name))
          end

          option_group 'Target Portal Filter' do
            apiopt('--no-portal-filter', :flag, 'clear openstack target portal filter',
                   :param_value => nil,
                   :param_name  => %w(openstack_portal_filter))
            apiopt('--portal-filter', 'NETWORK/MASK', 'update the openstack target portal filter in network mask CIDR notation (eg: 192.168.1.0/24)',
                   :param_name => %w(openstack_portal_filter)) do |val|
              addr = IPAddr.new(val)
              mask = addr.ipv4? ? '0.0.0.0' : '0:0:0:0:0:0:0:0'
              val.match(/\/([0-9]+)$/) { |cidr|
                mask = IPAddr.new(addr.ipv4? ? '255.255.255.255': 'FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF').mask(cidr[1]).to_s
              }
              {
                ip:   addr.to_s,
                mask: mask,
              }
            end
          end

          def exec
            node = only_one_management_node
            node = api.node.update(node.id, api_params)
            process_obj_updated node
          end
        end
      end
    end

    mixin :disk_nodes_update do |opts = {}|
      option('--clear-disk-nodes', :flag, 'clear existing disk nodes')
      apiopt(['--disk-node'], 'NODES', 'add disk node (may be used multiple times)',
             :multivalued => true,
             :param_name => opts.fetch(:param_name, ['disk_nodes']))
      option(['--remove-disk-node'], 'TAGS', 'remove disk node (may be used multiple times)',
             :multivalued => true)
    end

    def set_disk_nodes(params = nil, ex_nodes = nil)
      params.delete(:disk_nodes) unless params.nil?
      ex_nodes ||= yield params if block_given?
      ex_nodes ||= []

      ex_nodes  = Set.new(ex_nodes)
      add_nodes = Set.new(disk_node_list.flat_map { |t| t.split(/,\s*/) })
      del_nodes = Set.new(remove_disk_node_list.flat_map { |t| t.split(/,\s*/) })

      new_nodes  = clear_disk_nodes? ? Set.new : ex_nodes
      new_nodes -= del_nodes
      new_nodes += add_nodes

      return if new_nodes == ex_nodes

      if params
        params[:disk_nodes] = new_nodes.to_a
      else
        new_nodes.to_a
      end
    end

    subcommand 'update', 'update existing node' do
      mixin :site
      mixin :node
      mixin :label_notes, obj_name: 'node', hidden: false
      option_group 'Tags' do
        mixin :tags_update
        mixin :disk_nodes_update
      end
      mixin :identity_param, obj_name: 'node'

      apiopt('--visibility-ctl', 'MODE',
             'override node visibility (default, hide, show)') do |v|
        enum_check(v, %w(default hide show))
      end

      option_group nil, hidden: true do
        apiopt('--license', 'LICENSE', 'specify license for node')
        apiopt('--accept-eula', :flag, 'accept end user license agreement',
               :param_name => %w(eula accepted))
        apiopt('--clear-accepted-eula', :flag, 'clear previously accepted eula',
               :param_name => %w(eula accepted), :param_value => false)
        apiopt('--pub-ipaddr', 'ADDR', 'specify node public ip address')
        apiopt('--pub-ipaddr-desc', 'ADDR', 'specify node public address description')
        apiopt('--prv-ipaddr', 'ADDR', 'specify node private ip address')
        apiopt('--prv-ipaddr-desc', 'ADDR', 'specify node private address description')
      end

      def exec
        update_params = api_params
        node = obj_list(:node, restrict: :only_one)
        set_tags(update_params, node.tags)
        set_disk_nodes(update_params, node.disk_nodes)
        node = api.node.update(node.id, update_params)
        process_obj_updated node
      end
    end

    subcommand 'registration', 'allow or deny node registration' do
      mixin :site
      mixin :node
      mixin :identity_param, obj_name: 'node'

      option '--allow', :flag, 'allow node registration'
      option '--deny', :flag, 'deny node registration'

      def exec
        if allow?.nil? && deny?.nil?
          signal_usage_error "specify one of --allow or --deny"
        end
        node = obj_list(:node, restrict: :only_one)
        params = {
          registration: {
            allow: allow? ? true : false
          }
        }
        node = api.node.update(node.id, params)
        process_obj_updated node
      end
    end

    subcommand 'remove', 'remove existing node' do
      mixin :site
      mixin :node
      mixin :identity_param, obj_name: 'node'

      option('--batch', :flag, 'allow multiple nodes to be removed')

      def exec
        batch_remove :node
      end
    end

  end
end
