# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'ipaddr'
require 'deep_merge'

module Blockbridge::CLI::Command
  class Profile < Base
    self.default_subcommand = 'list'

    class ProfileCreate < Base
      mixin :vss

      mixin :label_notes, obj_name: 'profile'

      apiopt '--initiator-login', 'USER',
             'CHAP username (defaults to profile serial)'
      apiopt '--initiator-pass', 'PASS',
             'CHAP password'
      apiopt '--initiator', 'IQN',
             'permit initiator IQN (may be specified multiple times)',
             multivalued: true

      option_group 'Mutual Authentication', level: :advanced do
        apiopt('--mutual-auth', 'MODE',
               "mutual authentication mode (disabled, optional, required)",
               default: 'disabled') do |val|
          enum_check(val, %w(disabled optional required))
        end

        apiopt '--target-login', 'USER', 'CHAP user for mutual authentication'
        apiopt '--target-pass', 'PASS', 'CHAP password for mutual authentication'
      end

      option_group 'Transport', level: :expert do
        apiopt('--transport-mode', 'MODE',
               "transport security mode\n(insecure, tls, ikev1, ikev2)", {
                 :default => 'insecure',
                 :param_name => %w(transport mode)
               }) do |val|
          enum_check(val, %w(insecure tls ikev1 ikev2))
        end

        apiopt('--encryption-policy', 'POLICY',
               "transport encryption policy\n(disallowed,optional,required)", {
                 :default => 'disallowed',
                 :param_name => %w(transport encryption_policy)
               }) do |val|
          enum_check(val, %w(disallowed optional required))
        end

        # IKEv1 options
        apiopt('--ikev1-psk', 'PSK', "IPsec pre-shared key (IKEv1)", {
                 :param_group => :ikev1_params,
                 :param_name => %w(transport ikev1 psk)
               })

        # IKEv2 options
        apiopt('--ikev2-mode', 'MODE', "IPsec mode (IKEv2) (tunnel, transport)", {
                 :default => 'tunnel',
                 :param_group => :ikev2_params,
                 :param_name => %w(transport ikev2 mode)
               }) do |val|
          enum_check(val, %w(tunnel transport))
        end

        apiopt('--ikev2-auth', 'MODE', "IPsec auth mode (IKEv2) (mschapv2)", {
                 :default => 'mschapv2',
                 :param_group => :ikev2_params,
                 :param_name => %w(transport ikev2 auth)
               }) do |val|
          enum_check(val, %w(mschapv2))
        end

        apiopt('--ikev2-user', 'USER', "username for mschapv2 (IKEv2)", {
                 :param_group => :ikev2_params,
                 :param_name => %w(transport ikev2 mschapv2_user)
               })

        apiopt('--ikev2-pass', 'PASS', "password for mschapv2 (IKEv2)", {
                 :param_group => :ikev2_params,
                 :param_name => %w(transport ikev2 mschapv2_pass)
               })

        # IKEv2 tunnel mode params
        apiopt('--ikev2-tunnel-subnet', 'NET',
               "tunnel subnet (IKEv2 tunnel mode only)\n(172_30, 172_31, 10_1)", {
                 :default => '172_30',
                 :param_group => :ikev2_tunnel_params,
                 :param_name => %w(transport ikev2 subnet)
               }) do |val|
          enum_check(val, %w(172_31 172_30 10_1))
        end

        # TLS options
        apiopt('--tls-portal-addr', 'ADDR', "portal IP address", {
                 :default => '127.0.0.1',
                 :param_group => :tls_params,
                 :param_name => %w(transport tls portal_ipaddr)
               })
        apiopt('--tls-portal-port', 'PORT', "portal IP port", {
                 :default => '3260',
                 :param_group => :tls_params,
                 :param_name => %w(transport tls portal_ipport)
               }) do |val|
          Integer(val)
        end
      end

      option_group 'Policy', level: :advanced do
        apiopt('--disk-commands', 'COMMANDS', 'disk commands allowed', {
                 param_group: :policy_params,
                 param_name: %w[rules disk commands],
                 level: :expert
               }) do |val|
          enum_check(val, %w[read-write read-only write-only])
        end

        apiopt('--ip-allow', 'IP (CIDR)', "specifcally allow an IP address or range", {
                 :multivalued => true,
                 :param_list_concat => true,
                 :param_group => :policy_params,
                 :param_name  => %w(rules access ip)}) do |val|
          addr = IPAddr.new(val)
          mask = addr.ipv4? ? '0.0.0.0' : '0:0:0:0:0:0:0:0'
          val.match(/\/([0-9]+)$/) { |cidr|
            mask = IPAddr.new(addr.ipv4? ? '255.255.255.255': 'FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF').mask(cidr[1]).to_s
          }
          {
            type: 'allow',
            ip:   addr.to_s,
            mask: mask,
          }
        end

        apiopt('--ip-deny', 'IP (CIDR)', 'specifically deny an IP address or range', {
                 :multivalued => true,
                 :param_list_concat => true,
                 :param_group => :policy_params,
                 :param_name  => %w(rules access ip)}) do |val|
          addr = IPAddr.new(val)
          mask = addr.ipv4? ? '0.0.0.0' : '0:0:0:0:0:0:0:0'
          val.match(/\/([0-9]+)$/) { |cidr|
            mask = IPAddr.new(addr.ipv4? ? '255.255.255.255': 'FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF').mask(cidr[1]).to_s
          }
          {
            type: 'deny',
            ip:   addr,
            mask: mask,
          }
        end
      end

      mixin :xmd_refs_create
      mixin :xref
      mixin :info_extraction

      def exec
        params = api_params
        if vss
          # resolve containing vss
          server = obj_list(:vss, restrict: :only_one)
          params[:vss_id] = server.id
        end

        # add in appropriate transport params
        case params[:transport][:mode]
        when 'insecure'
          params[:transport][:encryption_policy] = 'disallowed'
        when 'ikev1'
          params.deep_merge!(ikev1_params)
        when 'ikev2'
          params.deep_merge!(ikev2_params)
          if params[:transport][:ikev2][:mode] == 'tunnel'
            params.deep_merge!(ikev2_tunnel_params)
          end
        when 'tls'
          params.deep_merge!(tls_params)
        end

        # policy params
        params.deep_merge!(policy_params)

        # submit
        profile = api.initiator_profile.create(params)
        process_obj_created profile
      end
    end
    subcommand 'create', 'create a new initiator profile', ProfileCreate, level: :basic

    mixin :profile, :vss
    mixin :info_extraction

    subcommand ['list', 'ls'], 'enumerate profiles' do
      mixin :info_filtering
      mixin :sort_by

      option '--global', :flag, 'only show global profiles'

      def exec
        profiles = obj_list(:profile)
        profiles.select! { |p| p.vss_id.nil? } if global?
        process_results profiles, view: 'profile_list'
      end
    end

    subcommand 'info', 'show detailed profile information', self, level: :basic do
      mixin :info_filtering
      mixin :info_extraction

      option '--global', :flag, 'only show global profiles'

      def exec
        profiles = obj_list(:profile)
        profiles.select! { |p| p.vss_id.nil? } if global?
        process_results profiles
      end
    end

    subcommand 'update', 'update basic initiator profile parameters', self, level: :basic do
      mixin :label_notes, obj_name: 'profile', hidden: false

      apiopt '--initiator-login', 'USER', 'CHAP username (defaults to profile serial)'
      apiopt '--initiator-pass', 'PASS', 'CHAP password'

      option_group 'Mutual Authentication', level: :advanced do
        apiopt('--mutual-auth', 'MODE', "mutual authentication mode (disabled, optional, required)", :default => 'disabled') do |val|
          enum_check(val, %w(disabled optional required))
        end

        apiopt '--target-login', 'USER', 'CHAP user for mutual authentication'
        apiopt '--target-pass', 'PASS', 'CHAP password for mutual authentication'
      end

      apiopt '--initiator', 'IQN', "permit initiator IQN (may be specified multiple times)", :multivalued => true

      option_group 'Transport', level: :expert do
        apiopt('--transport-mode', 'MODE',
               "transport security mode\n(insecure, tls, ikev1, ikev2)", {
                 :param_name => %w(transport mode)
               }) do |val|
          enum_check(val, %w(insecure tls ikev1 ikev2))
        end

        apiopt('--encryption-policy', 'POLICY',
               "transport encryption policy\n(disallowed,optional,required)", {
                 :param_name => %w(transport encryption_policy)
               }) do |val|
          enum_check(val, %w(disallowed optional required))
        end

        # IKEv1 options
        apiopt('--ikev1-psk', 'PSK', "IPsec pre-shared key (IKEv1)", {
                 :param_group => :ikev1_params,
                 :param_name => %w(transport ikev1 psk)
               })

        # IKEv2 options
        apiopt('--ikev2-mode', 'MODE', "IPsec mode (IKEv2) (tunnel, transport)", {
                 :param_group => :ikev2_params,
                 :param_name => %w(transport ikev2 mode)
               }) do |val|
          enum_check(val, %w(tunnel transport))
        end

        apiopt('--ikev2-auth', 'MODE', "IPsec auth mode (IKEv2) (mschapv2)", {
                 :param_group => :ikev2_params,
                 :param_name => %w(transport ikev2 auth)
               }) do |val|
          enum_check(val, %w(mschapv2))
        end

        apiopt('--ikev2-user', 'USER', "username for mschapv2 (IKEv2)", {
                 :param_group => :ikev2_params,
                 :param_name => %w(transport ikev2 mschapv2_user)
               })

        apiopt('--ikev2-pass', 'PASS', "password for mschapv2 (IKEv2)", {
                 :param_group => :ikev2_params,
                 :param_name => %w(transport ikev2 mschapv2_pass)
               })

        # IKEv2 tunnel mode params
        apiopt('--ikev2-tunnel-subnet', 'NET',
               "tunnel subnet (IKEv2 tunnel mode only)\n(172_30, 172_31, 10_1)", {
                 :param_group => :ikev2_tunnel_params,
                 :param_name => %w(transport ikev2 subnet)
               }) do |val|
          enum_check(val, %w(172_31 172_30 10_1))
        end

        # TLS options
        apiopt('--tls-portal-addr', 'ADDR', "portal IP address", {
                 :param_group => :tls_params,
                 :param_name => %w(transport tls portal_ipaddr)
               })
        apiopt('--tls-portal-port', 'PORT', "portal IP port", {
                 :param_group => :tls_params,
                 :param_name => %w(transport tls portal_ipport)
               }) do |val|
          Integer(val)
        end
      end

      option_group 'Policy', level: :advanced do
        apiopt('--disk-commands', 'COMMANDS', "disk commands allowed", {
                 param_group: :policy_params,
                 param_name: %w(rules disk commands),
                 level: :expert
               }) do |val|
          enum_check(val, %w(read-write read-only write-only))
        end

        apiopt('--ip-allow', 'IP (CIDR)', "specifcally allow an IP address or range",
               multivalued: true,
               param_list_concat: true,
               param_group: :policy_params,
               param_name: %w(rules access ip)) do |val|
          addr = IPAddr.new(val).to_s
          mask = '0.0.0.0'
          val.match(/\/([0-9]+)$/) { |cidr|
            mask = IPAddr.new('255.255.255.255').mask(cidr[1]).to_s
          }
          {
            type: 'allow',
            ip:   addr,
            mask: mask,
          }
        end

        apiopt('--ip-deny', 'IP (CIDR)', "specifcally deny an IP address or range",
               multivalued: true,
               param_list_concat: true,
               param_group: :policy_params,
               param_name: %w(rules access ip)) do |val|
          addr = IPAddr.new(val).to_s
          mask = '0.0.0.0'
          val.match(/\/([0-9]+)$/) { |cidr|
            mask = IPAddr.new('255.255.255.255').mask(cidr[1]).to_s
          }
          {
            type: 'deny',
            ip:   addr,
            mask: mask,
          }
        end
      end

      mixin :xmd_refs_update

      def exec
        params = api_params

        # find existing profile
        profile = obj_list(:profile, restrict: :only_one)

        # update xmd-refs
        set_xmd_refs(params, profile.xmd_refs)

        # determine transport mode (current or new value)
        cur_mode = profile.transport.mode
        new_mode = transport_mode

        if new_mode && cur_mode != new_mode
          if cur_mode == 'insecure'
            # clear policy so it may fall back to per-transport mode default.
            params[:transport][:encryption_policy] = nil
          end

          ####################################################################
          # for the newly-selected transport mode: start with empty config,
          # then merge in user-specified config, then defaults.
          case new_mode
          when 'insecure'
            params[:transport][:encryption_policy] = 'disallowed'
          when 'ikev1'
            params[:transport][:ikev1] ||= {}
            params[:transport][:ikev1].deep_merge!(profile.transport.ikev1 || {})
            params.deep_merge!(ikev1_params)
            params[:transport][:encryption_policy] ||= 'required'
          when 'ikev2'
            params[:transport][:ikev2] ||= {}
            params[:transport][:ikev2].deep_merge!(profile.transport.ikev2 || {})
            params.deep_merge!(ikev2_params)
            params[:transport][:ikev2][:auth] ||= 'mschapv2'
            params[:transport][:ikev2][:mode] ||= 'tunnel'
            if params[:transport][:ikev2][:mode] == 'tunnel'
              params.deep_merge!(ikev2_tunnel_params)
            end
            params[:transport][:encryption_policy] ||= 'required'
          when 'tls'
            params[:transport][:tls] ||= {}
            params[:transport].deep_merge!(profile.transport.tls || {})
            params.deep_merge!(tls_params)
            params[:transport][:encryption_policy] ||= 'optional'
          end
        else
          # no mode change; if there are any user specified values for the
          # current mode, start with current parameters and overlay
          # user-specified values.
          case cur_mode
          when 'ikev1'
            unless ikev1_params.empty?
              params[:transport][:mode] = 'ikev1'
              params[:transport][:ikev1] = profile.transport.ikev1
              params.deep_merge!(ikev1_params)
            end
          when 'ikev2'
            unless ikev2_params.empty? && ikev2_tunnel_params.empty?
              params[:transport][:mode] = 'ikev2'
              params[:transport][:ikev2] = profile.transport.ikev2
              params.deep_merge!(ikev2_params)
              params.deep_merge!(ikev2_tunnel_params)
            end
          when 'tls'
            unless tls_params.empty?
              params[:transport][:mode] = 'tls'
              params[:transport][:tls] = profile.transport.tls
              params.deep_merge!(tls_params)
            end
          end
        end

        # policy params
        params.deep_merge!(policy_params)

        profile = api.initiator_profile.update(profile.id, params)
        process_obj_updated profile
      end
    end

    subcommand 'token', 'initiator secure access token', self, level: :basic do
      subcommand 'create', 'create an initiator token', self, level: :basic do
        option '--passphrase-file', 'PATH', 'read passphrase from specified file'
        def exec
          profile    = obj_list(:profile, restrict: :only_one)
          passphrase = password_from_file_or_prompt(passphrase_file, "passphrase")
          token      = profile_configure_token(profile, passphrase)
          process_obj_updated profile, desc: "The secure access token for #{profile} is: #{token}"
        end
      end
      subcommand 'remove', 'remove an initiator token', self, level: :basic do
        def exec
          profile = obj_list(:profile, restrict: :only_one)
          api.initiator_profile.remove_token(profile.id)
          process_obj_updated profile, desc: "Removed secure access token for #{profile}"
        end
      end
    end

    subcommand 'initiator', 'adjust the permitted initiator list', self, level: :basic do
      subcommand 'add', 'add an initiator to the permitted list' do
        option '--iqn', 'IQN', 'initiator IQN to add', :required => true
        def exec
          profile = obj_list(:profile, restrict: :only_one)

          if profile.iscsi.initiator_list.index(iqn)
            display "#{iqn} already permitted for profile #{profile}"
            return
          end

          params = {
            initiator_list: profile.iscsi.initiator_list << iqn
          }

          profile = api.initiator_profile.update(profile.id, params)
          process_obj_updated profile, desc: "Added #{iqn} to permitted list for profile #{profile}"
        end
      end

      subcommand 'remove', 'remove an initiator from the permitted list', self, level: :basic do
        option '--iqn', 'IQN', 'initiator IQN to remove', :required => true
        def exec
          profile = obj_list(:profile, restrict: :only_one)

          if profile.iscsi.initiator_list.delete(iqn).nil?
            display "#{iqn} not permitted for profile #{profile}"
            return
          end

          params = {
            initiator_list: profile.iscsi.initiator_list
          }

          profile = api.initiator_profile.update(profile.id, params)
          process_obj_updated profile, desc: "Removed #{iqn} from permitted list for profile #{profile}"
        end
      end
    end

    subcommand 'remove', 'remove an existing initiator profile', self, level: :basic do
      def exec
        profile = obj_list(:profile, restrict: :only_one)
        rsp = api.initiator_profile.remove(profile.id)

        process_obj_removed profile
      end
    end

  end
end
