# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require "blockbridge/util"

module Blockbridge::CLI::Command
  class Rule < Base
    self.default_subcommand = 'list'

    class RuleCreate < Base
      mixin :vss, :disk
      mixin :label_notes, obj_name: 'rule'
      mixin :xmd_refs_create
      mixin :xref
      mixin :info_extraction

      def exec
        if vss.nil? && disk.nil?
          signal_usage_error "Must specify a vss and/or disk."
        end

        vss_obj = obj_list(:vss, restrict: :only_one) if vss
        disk_obj = obj_list(:disk, restrict: :only_one) if disk

        if vss_obj && disk_obj && vss_obj.id != disk_obj.vss_id
          signal_usage_error "Disk and rule must reside on the same vss"
        end

        # build create params
        create_params = api_params.tap do |h|
          if disk_obj
            h[:ref_id] = disk_obj.id
            h[:vss_id] = disk_obj.vss_id
          else
            h[:ref_id] = nil # no assigned ref object yet...
            h[:vss_id] = vss_obj.id
          end
        end

        rule = api.rule.create(create_params)

        process_obj_created rule
      end
    end
    subcommand 'create', 'create a new disk rule', RuleCreate, level: :dev

    mixin :vss, :disk, :rule
    mixin :info_extraction

    subcommand ['list', 'ls'], 'enumerate disk rules', self, level: :dev do
      mixin :vss, :disk, :rule
      mixin :info_filtering
      mixin :sort_by

      def exec
        rules = obj_list(:rule)
        process_results rules, view: 'rule_list'
      end
    end

    subcommand 'info', 'show detailed disk rule information', self, level: :dev do
      mixin :info_filtering
      mixin :info_extraction

      def exec
        rules = obj_list(:rule)
        process_results rules
      end
    end

    subcommand 'update', 'update basic rule parameters', self, level: :dev do
      mixin :label_notes, obj_name: 'rule', hidden: false
      mixin :xmd_refs_update

      option(['--no-disk'], :flag, "unassign disk from rule")

      def exec
        rule = obj_list(:rule, restrict: :only_one)

        if disk
          disk_obj = obj_list(:disk, restrict: :only_one)
          if disk_obj.vss_id != rule.vss_id
            signal_usage_error "Disk and rule must reside on the same vss"
          end
        end

        params = api_params.tap do |p|
          set_xmd_refs(p, rule.xmd_refs)

          if no_disk?
            p[:ref_id] = nil
          elsif disk_obj
            p[:ref_id] = disk_obj.id
          end
        end

        rule = api.rule.update(rule.id, params)

        process_obj_updated rule
      end
    end

    subcommand 'remove', 'remove a disk rule', self, level: :dev do
      def exec
        rule = obj_list(:rule, restrict: :only_one)

        rsp = api.rule.remove(rule.id)
        process_obj_removed rule
      end
    end

    subcommand 'fileset-filter', 'manage rule fileset-filter objects', self, level: :dev do

      subcommand 'create', 'create a new fileset filter', self, level: :dev do
        option '--filter', 'LABEL', 'specify filter label', :required => true

        def exec
          rule = obj_list(:rule, restrict: :only_one)

          chk = rule.filters.find { |f| f.label == filter }
          runtime_error "Filter '#{filter}' already exists in rule #{rule}." unless chk.nil?

          # add new filter to to array
          new_filter = {
            label:   filter,
            type:    'logical_block_access',
            access:  'read',
            monitor: false,
            extent_groups: []
          }

          rule = api.rule.update(rule.id, filters: (rule.filters << new_filter))
          process_obj_updated rule
        end
      end

      subcommand 'remove', 'remove existing filset-filter', self, level: :dev do
        option '--filter', 'LABEL', 'specify filter label', :required => true

        def exec
          rule = obj_list(:rule, restrict: :only_one)

          idx = rule.filters.index { |f| f.label == filter }
          runtime_error("Filter '#{filter}' does not exist in rule #{rule}.") if idx.nil?

          if rule.filters[idx].type != 'logical_block_access'
            runtime_error("Filter '#{filter}' in rule #{rule} is not a fileset filter.")
          end

          # remove filter
          rule.filters.delete_at(idx)

          rule = api.rule.update(rule.id, filters: rule.filters)
          process_obj_updated rule, desc: "Removed filter '#{filter}' from rule #{rule}"
        end
      end

      def map_tags(tags, path, type)
        tags.map do |tag|
          if md = /^%(?<char>.)$/.match(tag)
            case md[:char]
            when 't'
              type
            when 'f'
              File.basename(path)
            when 'p'
              path
            else
              signal_usage_error "invalid tag macro character '#{md[:char]}'"
            end
          else
            tag
          end
        end
      end

      def find_filter(rule, type, friendly_type, fl)
        f_list = rule.filters.select do |f|
          f.type == type && (fl.nil? || f.label == fl)
        end

        if f_list.empty?
          if fl.nil?
            runtime_error("Rule #{rule} has no #{friendly_type} filters.")
          else
            runtime_error("Rule #{rule} has no #{friendly_type} filter '#{fl}'.")
          end
        end

        if f_list.length > 1
          if fl.nil?
            runtime_error("Rule #{rule} has more than one #{friendly_type} filter " \
                          "-- specify one using --filter")
          else
            internal_error("Rule #{rule} unexpectedly has more than one " \
                           "#{friendly_type} filter '#{fl}'")
          end
        end

        f_list[0]
      end

      subcommand 'update', 'update basic fileset filter parameters', self, level: :dev do
        option '--filter', 'LABEL', 'specify filter label'
        option '--label', 'LABEL', 'update filter label'
        option('--access', 'MODE', "select command access type(s)\n(read, write, read_write)") do |v|
          enum_check(v, %w(read write read_write))
        end
        option '--[no-]monitor', :flag, 'enable or disable I/O monitoring'
        option '--inode', :flag, 'include inode extent for files (default excluded)'
        option '--no-data', :flag, 'exclude data extents for files (default included)'
        option '--tags', 'TAGS', 'tag spec for files -' do |t| t.split(/\s*,\s*/) end
        option '--offset', 'VAL', 'device offset to use when creating extent groups'
        option('--files', :flag, 'update the fileset extent groups (replace)',
               :attribute_name => :update_extents)
        option('--merge', :flag, 'merge extents instead of replacing',
               :attribute_name => :merge_extents)

        parameter "[FILE] ...", "files to add", :attribute_name => :files

        attr_reader :update_mode

        def update_extents(ff)
          if !Blockbridge::Util.has_feature? :fs_inspection
            runtime_error("Filesystem inspection is not supported on this " \
                          "host -- please contact support for assistance.")
          end

          # optionally replace existing extent groups
          ff.extent_groups = [] unless merge_extents?

          mounts = {}

          files.each do |file|
            path = File.expand_path(file)

            begin
              File.open(path) do |f|
                f.fdatasync
              end

              info = Blockbridge::Util::Filesystem.inspect(path)
            rescue Errno::ENOENT => e
              signal_usage_error "#{e.message}"
            end

            mounts[info.dev] = info.mount_point

            exg = {
              label: path,
              offset: offset || 0,
              extents: [],
            }

            unless no_data?
              data_extents = info.data_extents.map do |ex|
                ex[:t] = map_tags(tags, path, 'data') unless tags.nil?
                ex
              end
              exg[:extents].concat(data_extents)
            end

            if inode?
              inode_extent = info.inode_extent do |ex|
                ex[:t] = map_tags(tags, path, 'inode') unless tags.nil?
                ex
              end
              exg[:extents] << inode_extent
            end

            ff.extent_groups << exg
          end

          if mounts.size > 1
            runtime_error("A fileset may only include files from a " \
                          "single device. (#{mounts.size} devices scanned.)")
          end
        end

        def exec
          rule = obj_list(:rule, restrict: :only_one)

          ff = find_filter(rule, 'logical_block_access', 'fileset', filter)
          filter = ff.label

          ff.label   = label    unless label.nil?
          ff.access  = access   unless access.nil?
          ff.monitor = monitor? unless monitor?.nil?

          if update_extents?
            update_extents(ff)
          else
            if !files.empty?
              usage_error("Found extra arguments -- perhaps " \
                          "you meant to use the '--files' option?")
            end
          end

          rule = api.rule.update(rule.id, filters: rule.filters)
          process_obj_updated(rule,
              desc: "Updated filter '#{filter}' in rule #{rule}")
        end

      end
    end

  end
end
