# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

module Blockbridge::CLI::Command
  class Site < Base
    self.default_subcommand = 'list'

    mixin :site
    mixin :info_extraction

    subcommand ['list', 'ls'], 'enumerate sites' do
      mixin :info_filtering
      mixin :sort_by

      def exec
        sites = obj_list(:site)
        process_results sites, view: 'site_list'
      end
    end

    subcommand 'info', 'show detailed site information' do
      mixin :info_filtering
      mixin :info_extraction

      def exec
        sites = obj_list(:site)
        process_results sites
      end
    end

    class SiteCreate < Base
      apiopt '--label', 'LABEL', 'specify site label'
      apiopt '--notes', 'NOTES', 'specify site notes'
      apiopt '--tier', 'TIER', 'specify datacenter tier (1-4)' do |val|
        enum_map(val, :"1" => 'I', :"2" => 'II', :"3" => 'III', :"4" => 'IV')
      end
      apiopt('--city', 'CITY', 'specify site city',
             param_name: %w(location city))
      apiopt('--state', 'STATE', 'specify site state',
             param_name: %w(location state))
      apiopt('--country', 'COUNTRY', 'specify site country',
             param_name: %w(location country))
      apiopt('--lat', 'LATITUDE', 'specify site latitude',
             param_name: %w(location lat)) do |val| Float(val) end
      apiopt('--long', 'LONGITUDE', 'specify site longitude',
             param_name: %w(location long)) do |val| Float(val) end

      apiopt('--email-type', 'TYPE', 'specify email delivery type (smtp, none)',
             default: 'none', param_name: %w(email type)) do |val|
        enum_check(val, %w(smtp none))
      end

      apiopt('--smtp-addr', 'HOST', 'specify hostname for smtp email delivery',
             param_name: %w(email smtp server_addr))

      apiopt('--smtp-from', 'EMAIL', 'specify from-email-address for smtp email delivery',
             param_name: %w(email smtp mail_from))

      apiopt('--smtp-port', 'PORT', 'specify port for smtp email delivery',
             default: 25, param_name: %w(email smtp server_port)) do |val|
        Integer(val)
      end
      apiopt('--smtp-ssl', 'MODE', "use TLS/SSL for smtp email delivery\n(none, optional, required)",
             default: 'optional', param_name: %w(email smtp server_ssl)) do |val|
        enum_check(val, %w(none optional required))
      end
      apiopt('--[no-]smtp-auth', :flag, 'use smtp authentication',
             default: false, param_name: %w(email smtp server_auth))
      apiopt('--smtp-user', 'USER', 'specify username for smtp authentication',
             param_name: %w(email smtp server_username))
      apiopt('--smtp-pass', 'PASS', 'specify passwordfor smtp authentication',
             param_name: %w(email smtp server_password))

      mixin :info_extraction

      def exec
        create_params = api_params

        if create_params[:email][:type] != 'smtp'
          create_params[:email].delete :smtp
        end

        site = api.site.create(create_params)
        process_obj_created site
      end
    end
    subcommand 'create', 'create a new site object', SiteCreate

    subcommand 'update', 'update existing site' do
      mixin :label_notes, obj_name: 'site', hidden: false
      apiopt '--tier', 'TIER', 'specify datacenter tier (1-4)' do |val|
        enum_map(val, "1" => 'I', "2" => 'II', "3" => 'III', "4" => 'IV')
      end
      apiopt('--city', 'CITY', 'specify site city', param_name: %w(location city))
      apiopt('--state', 'STATE', 'specify site state', param_name: %w(location state))
      apiopt('--country', 'COUNTRY', 'specify site country', param_name: %w(location country))
      apiopt('--lat', 'LATITUDE', 'specify site latitude',
             param_name: %w(location lat)) do |val| Float(val) end
      apiopt('--long', 'LONGITUDE', 'specify site longitude',
             param_name: %w(location long)) do |val| Float(val) end

      apiopt('--email-type', 'TYPE', 'specify email delivery type (smtp, none)',
             param_name: %w(email type)) do |val|
        enum_check(val, %w(smtp none))
      end

      apiopt('--smtp-addr', 'HOST', 'specify hostname for smtp email delivery',
             param_name: %w(email smtp server_addr))

      apiopt('--smtp-from', 'EMAIL', 'specify from-email-address for smtp email delivery',
             param_name: %w(email smtp mail_from))

      apiopt('--smtp-port', 'PORT', 'specify port for smtp email delivery',
             param_name: %w(email smtp server_port)) do |val|
        Integer(val)
      end
      apiopt('--smtp-ssl', 'MODE', "use TLS/SSL for smtp email delivery\n(none, optional, required)",
             param_name: %w(email smtp server_ssl)) do |val|
        enum_check(val, %w(none optional required))
      end
      apiopt('--[no-]smtp-auth', :flag, 'use smtp authentication',
             param_name: %w(email smtp server_auth))
      apiopt('--smtp-user', 'USER', 'specify username for smtp authentication',
             param_name: %w(email smtp server_username))
      apiopt('--smtp-pass', 'PASS', 'specify passwordfor smtp authentication',
             param_name: %w(email smtp server_password))

      def exec
        site = obj_list(:site, restrict: :only_one)

        # site update needs full location specification
        update_params = {
          location: site.location
        }.deep_merge!(api_params)

        site = api.site.update(site.id, update_params)

        process_obj_updated site
      end
    end

    subcommand 'remove', 'remove existing site' do
      option('--batch', :flag, 'allow multiple sites to be removed',
             hidden: true)

      def exec
        batch_remove :site
      end
    end

  end
end
