# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

module Blockbridge::CLI::Command
  class Target < Base
    self.default_subcommand = 'list'

    class TargetCreate < Base
      mixin :vss
      mixin :label_notes, obj_name: 'target'
      option_group nil, :hidden => true do
        apiopt '--iqn-suffix', 'SUFFIX', "override iqn suffix", level: :advanced
        apiopt '--[no-]automap', :flag, "enable/disable auto-mapping of removable disks", level: :expert
        mixin :xmd_refs_create
        mixin :xref
      end
      mixin :info_extraction

      def exec
        vss = obj_list(:vss, restrict: :only_one)

        target = api.target.create(api_params.merge(vss_id: vss.id))
        process_obj_created target
      end
    end
    subcommand 'create', 'create a new iSCSI target', TargetCreate

    mixin :vss, :target
    mixin :info_extraction

    subcommand ['list', 'ls'], 'enumerate targets' do
      mixin :info_filtering
      mixin :sort_by

      def exec
        targets = obj_list(:target)
        process_results targets, view: 'target_list'
      end
    end

    subcommand 'info', 'show detailed target information' do
      mixin :info_filtering
      mixin :info_extraction

      def exec
        targets = obj_list(:target)
        process_results targets
      end
    end

    subcommand 'update', 'update basic target parameters' do
      mixin :label_notes, obj_name: 'target', hidden: false
      option_group nil, :hidden => true do
        apiopt '--iqn-suffix', 'SUFFIX', "override iqn suffix", level: :advanced
        apiopt '--[no-]automap', :flag, "enable/disable auto-mapping of removable disks", level: :expert
        mixin :xmd_refs_update
      end

      def exec
        target = obj_list(:target, restrict: :only_one)

        # build params
        params = api_params

        set_xmd_refs(params, target.xmd_refs)

        target = api.target.update(target.id, params)
        process_obj_updated target
      end
    end

    subcommand 'remove', 'remove an existing target' do
      def exec
        target = obj_list(:target, restrict: :only_one)
        api.target.remove(target.id)
        process_obj_removed target
      end
    end

    subcommand 'lun', 'manage target LUN map' do
      subcommand 'map', 'create or update a target lun map entry' do
        mixin :disk

        apiopt '--lun', 'NUM', "logical unit number" do |s|
          Integer(s)
        end

        def exec
          target = obj_list(:target, restrict: :only_one)
          disk = obj_list(:disk, restrict: :only_one)

          if target.vss_id != disk.vss_id
            signal_usage_error "Disk and target must reside on the same vss"
          end

          if lun
            rsp = api.target.lun_map(target.id, lun, vdisk_id: disk.id)
          else
            rsp = api.target.disk_map(target.id, disk.id, {})
          end

          process_obj_updated target
        end
      end

      subcommand 'unmap', 'remove a lun map entry' do
        mixin :disk
        option '--lun', 'NUM', "logical unit number" do |s|
          Integer(s)
        end

        def exec
          target = obj_list(:target, restrict: :only_one)

          if disk
            disk = obj_list(:disk, restrict: :only_one)

            if target.luns.find { |e| e.vdisk_id == disk.id }
              rsp = api.target.disk_unmap(target.id, disk.id)
            else
              runtime_error "#{disk} is not mapped into #{target}"
            end
          elsif lun
            if target.luns.find { |e| e.lun == lun }
              rsp = api.target.lun_unmap(target.id, lun)
            else
              runtime_error "lun #{lun} is not mapped on #{target}"
            end
          else
            signal_usage_error "unmap must specify disk or LUN"
          end

          process_obj_updated target
        end
      end
    end

    subcommand 'acl', 'manage target ACL' do
      subcommand 'add', 'add a target ACL entry' do
        mixin :profile

        option '--[no-]privacy', :flag, 'require transport security', :default => false
        option '--[no-]discover', :flag, 'permit target discovery', :default => true
        option '--[no-]access', :flag, 'allow access', :default => true

        def exec
          profile = obj_list(:profile, restrict: :only_one)
          target = obj_list(:target, restrict: :only_one)

          if target.acl.find { |ent| ent.profile_id == profile.id }
            display "Profile #{profile} already in ACL for target #{target}"
            return
          end

          # append new acl entry
          target.acl << {
            profile_id: profile.id,
            privacy:    privacy? ? 'yes' : 'no',
            discover:   discover? ? 'yes' : 'no',
            access:     access? ? 'yes' : 'no',
          }

          target = api.target.update(target.id, acl: target.acl)
          process_obj_updated target, desc: "Updated ACL for target #{target}"
        end
      end

      subcommand 'update', 'update a target ACL entry' do
        mixin :profile

        option '--[no-]privacy', :flag, 'require transport security'
        option '--[no-]discover', :flag, 'permit target discovery'
        option '--[no-]access', :flag, 'allow access'

        def exec
          profile = obj_list(:profile, restrict: :only_one)
          target = obj_list(:target, restrict: :only_one)

          if (acl_ent = target.acl.find { |ent| ent.profile_id == profile.id }).nil?
            display "Profile #{profile} not in ACL for target #{target}"
            return
          end

          # update acl entry
          acl_ent.privacy  = (privacy? ? 'yes' : 'no') unless privacy?.nil?
          acl_ent.discover = (discover? ? 'yes' : 'no') unless discover?.nil?
          acl_ent.access   = (access? ? 'yes' : 'no') unless access?.nil?

          target = api.target.update(target.id, acl: target.acl)
          process_obj_updated target, desc: "Updated ACL for target #{target}"
        end
      end

      subcommand 'remove', 'remove an ACL entry' do
        mixin :profile

        def exec
          profile = obj_list(:profile, restrict: :only_one)
          target = obj_list(:target, restrict: :only_one)

          if (target.acl.reject! { |ent| ent.profile_id == profile.id }).nil?
            display "Profile #{profile} not in ACL for target #{target}"
            return
          end

          target = api.target.update(target.id, acl: target.acl)
          process_obj_updated target, desc: "Updated ACL for target #{target}"
        end
      end

      subcommand 'clear', 'remove all entries from the ACL' do
        mixin :profile

        def exec
          target = obj_list(:target, restrict: :only_one)
          target = api.target.update(target.id, acl: [])
          process_obj_updated target, desc: "Cleared ACL for target #{target}"
        end
      end

    end
  end
end
