# Copyright (c) 2015-2017, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'base32'
require 'rqrcode'

module Blockbridge::CLI::Command
  class User < Base
    mixin :info_extraction

    self.default_subcommand = 'list'

    mixin :user_permissions do
      apiopt(['--permit', '--grant', '-P'], 'RIGHT', 'grant specified user right',
             :multivalued => true,
             :param_list_merge => true,
             :param_name => :permissions) do |v|
        scope, right = parse_permission(v)
        { scope => { :rights => { right => true } } }
      end
      apiopt(['--deny', '-D'], 'RIGHT', 'deny specified user right',
             :multivalued => true,
             :param_list_merge => true,
             :param_name => :permissions) do |v|
        scope, right = parse_permission(v)
        { scope => { :rights => { right => false } } }
      end
    end

    class UserCreate < Base
      apiopt('--name', 'NAME', "specify new user name",
             :required => true,
             :param_name => :username)
      option('--password', :flag, "prompt to specify initial user password")
      option('--password-file', 'PATH', 'read password from specified file')
      apiopt('--email', 'ADDR', "specify email address (optional)",
             :default => nil)

      mixin :user_permissions

      option '--account', 'ACCOUNT', "specify account by name or serial number\n(default creates user in current account)"

      mixin :info_extraction

      def exec
        create_params = api_params.tap do |h|
          h[:password] = nil
          h[:password] = prompt_for_password if password? && !password_file
          h[:password] = password_from_file(password_file) if password_file
          if account
            acct = obj_list(:account, restrict: :only_one)
            h[:account_id] = acct.id
          end
        end

        user = api.user_profile.create(create_params)
        process_obj_created user
      end
    end
    subcommand 'create', 'create a new user', UserCreate, level: :basic

    mixin :user
    mixin :account

    subcommand ['list', 'ls'], 'enumerate users', self, level: :basic do
      mixin :info_filtering
      mixin :sort_by

      def exec
        users = obj_list(:user)
        process_results users, view: 'user_list'
      end
    end

    subcommand 'info', 'show detailed user information', self, level: :basic do
      mixin :info_filtering
      mixin :info_extraction

      def exec
        users = obj_list(:user)
        process_results users
      end
    end

    subcommand 'activate', 'activate a user profile', self, level: :basic do
      def exec
        user = obj_list(:user, restrict: :only_one)
        process_obj_updated api.user_profile.update(user.id, activated: true)
      end
    end

    subcommand 'deactivate', 'deactivate a user profile', self, level: :basic do
      def exec
        user = obj_list(:user, restrict: :only_one)
        process_obj_updated api.user_profile.update(user.id, activated: false)
      end
    end

    subcommand 'update', 'update basic user parameters', self, level: :basic do
      option('--password', :flag, "prompt to specify new user password")
      option('--password-file', 'PATH', 'read password from specified file')
      option('--no-password', :flag, "clear user password")
      option('--auth-password', :flag, "prompt to specify your current password to confirm when changing password")
      option('--auth-password-file', 'PATH', 'read auth password from specified file')
      apiopt('--email', 'ADDR', "update email address")
      apiopt('--no-email', :flag, "clear email address",
             :param_name => :email,
             :param_value => nil)

      option('--node-scope', 'VALUE', "set node permission scope "\
                                       "(options: all, none)") do |v|
        enum_check(v, %w(all none))
      end

      apiopt('--[no-]allow-su', :flag, "permit su to this user")

      option_group "Multi-factor authentication" do
        apiopt('--two-factor', 'TYPE', "set two-factor authentication type",
               :param_name => %w(two_factor auth),
               :attribute_name => :mfa)
        option('--qrcode', 'MODE',
               "display key as a QR code (options: ansi, image)\n(default: " \
               "show key for manual entry)") do |v|
          enum_check(v, %w(ansi image))
        end
        apiopt('--otp', 'OTP', "specify OTP for enabling multi-factor auth",
               :param_name => %w(two_factor otp),
               :hidden => true)
        apiopt('--private-id', 'VALUE',
               "specify private ID value (yubikey only)",
               :param_name => %w(two_factor private_id),
               :hidden => true)
        apiopt('--secret', 'VALUE', "specify explicit multi-factor secret",
               :param_name => %w(two_factor secret),
               :hidden => true)
      end

      mixin :user_permissions
      mixin :xmd_refs_update

      def show_qrcode_image(q)
        png = q.as_png(
          resize_gte_to: false,
          resize_exactly_to: false,
          fill: 'white',
          color: 'black',
          size: 120,
          border_modules: 4,
          module_px_size: 6,
          file: nil)
        print_image(png.to_s, 'qr-code')
      end

      def show_google_key(user, key)
        if qrcode
          label = "#{user.login}%20(Blockbridge)"
          qrc = RQRCode::QRCode.new("otpauth://totp/#{label}?secret=#{key}")

          case qrcode
          when 'image'
            show_qrcode_image(qrc)
          when 'ansi'
            puts qrc.as_ansi
            puts
          end

          puts "Scan the above QR code with Google Authenticator."
        else
          puts "Manually enter the key into Google Authenticator:"
          puts "  #{Paint[key,:bold]}"
          puts
        end
      end

      def secret_key
        return SecureRandom.random_bytes(20) if secret.nil?

        Base32.decode(secret)
      rescue => e
        usage_error "Secret key must be base32 encoded (given: '#{secret}')"
      end

      def configure_google_mfa(user, params)
        r = secret_key
        rhex = r.unpack('H*').first
        rb32 = Base32.encode(r)

        show_google_key(user, rb32)

        params[:two_factor][:secret] = rhex
      end

      def configure_yubikey_mfa(user, params)
        params[:two_factor][:private_id] ||= ask("Enter yubikey private ID: ")
        params[:two_factor][:secret] ||= ask("Enter yubikey secret key: ")
      end

      def exec
        user = obj_list(:user, restrict: :only_one)

        if user[:two_factor][:auth] != 'none'
          if mfa && mfa != 'none'
            usage_error "Multi-factor authentication is already configured; " \
                        "disable the current configuration and try again. "   \
                        "(set the multi-factor auth type to 'none')"
          end
        end

        update_params = api_params.tap do |h|
          set_xmd_refs(h, user.xmd_refs)
          h[:new_password] = nil if no_password?
          h[:new_password] = prompt_for_password('new password') if password? && !password_file
          h[:new_password] = password_from_file(password_file) if password_file
          h[:auth_password] = prompt_for_password('your current password') if auth_password? && !auth_password_file
          h[:auth_password] = password_from_file(auth_password_file) if auth_password_file

          unless node_scope.nil?
            perms = h[:permissions] ||= {}
            node_perms = perms[:node] ||= {}

            case node_scope
            when 'all'
              node_perms[:all_realm_scope] = true
            when 'none'
              node_perms[:all_realm_scope] = false
              node_perms[:nodes] = []
            end
          end
        end

        # generate, display key and prompt for OTP

        case mfa
        when 'google'
          configure_google_mfa(user, update_params)
        when 'yubikey'
          configure_yubikey_mfa(user, update_params)
        end

        if mfa && mfa != 'none'
          update_params[:two_factor][:otp] ||= ask("Enter generated OTP: ")
        end

        begin
          user = api.user_profile.update(user.id, update_params)
        rescue Blockbridge::Api::ValidationError => e
          raise unless e.errors&.any? { |o| o[:type] == 'missing' && o[:field] == 'auth_password' }
          update_params.merge!(auth_password: prompt_for_password('your current password'))
          user = api.user_profile.update(user.id, update_params)
        end
        process_obj_updated user
      end
    end

    subcommand 'remove', 'remove an existing user', self, level: :basic do
      def exec
        user = obj_list(:user, restrict: :only_one)
        rsp = api.user_profile.remove(user.id)
        process_obj_removed user
      end
    end

    class UserPermissions < Base
      def exec
        process_results api.user_profile.permissions
      end
    end
    subcommand 'permissions', 'display available user permissions',
               UserPermissions
  end

end
