# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'securerandom'

module Blockbridge::CLI::Command
  class Vss < Base
    self.default_subcommand = 'list'

    class ProvisionVss < Base
      # query parameters
      mixin :catalog_query_params

      option_group 'Provisioning' do
        # vss provisioning parameters
        apiopt(['--label', '-l'], 'LABEL', 'specify service label',
               :param_group => :server_params,
               :param_name => %w(vss label))
        apiopt '--notes', 'NOTES', 'specify service notes',
               :param_group => :server_params,
               :param_name => %w(vss notes)

        option_group nil, :hidden => true do
          apiopt('--xmd-ref', 'REF', 'xmd record reference key',
                 :multivalued => true,
                 :param_group => :server_params,
                 :param_name => %w(vss xmd_refs))
          apiopt('--xref', 'REF', 'external reference key',
                 :param_group => :server_params,
                 :param_name => %w(vss xref))
        end

        # provision with disk (or clone)
        apiopt('--with-disk', :flag, 'create disk with vss',
               :param_group => :disk_params,
               :param_name => %w(disk create))
        apiopt('--disk-label', 'LABEL', 'specify disk label',
               :param_group => :disk_params,
               :param_name => %w(disk label))

        option_group nil, :hidden => true do
          apiopt('--disk-tag', 'TAG', 'specify disk tag (may be used multiple times)',
                 :multivalued => true,
                 :param_group => :disk_params,
                 :param_name => %w(disk tags))
          apiopt('--no-disk-label', :flag, 'specify null disk label',
                 :param_group => :disk_params,
                 :param_value => nil, :param_name => %w(disk label))
          apiopt('--disk-xmd-ref', 'REF', 'xmd record reference key',
                 :multivalued => true,
                 :param_group => :disk_params,
                 :param_name => %w(disk xmd_refs))
          apiopt('--disk-xref', 'REF', 'disk external reference key',
                 :param_group => :disk_params,
                 :param_name => %w(disk xref))
          apiopt('--clone-src', 'SRC', 'create disk as clone of src',
                 :param_group => :disk_params,
                 :param_name => %w(disk clone_src))
          apiopt('--clone-blocks', :flag, 'create disk as clone from block device',
                 :param_group => :disk_params,
                 :param_name => %w(disk clone_blocks))
          apiopt('--backup-id', 'BACKUP-ID', 'if clone from object store, specify backup to clone from',
                 :param_group => :disk_params,
                 :param_name => %w(disk backup_id))
        end

        option_group 'Clone' do
          apiopt('--with-clone', :flag, 'create disk as a clone with vss',
                 :param_group => :disk_params,
                 :param_name => %w(disk create))
          apiopt('--clone-label', 'LABEL', 'specify disk label',
                 :param_group => :disk_params,
                 :param_name => %w(disk label))
          mixin :snapshot
          mixin :disk
          mixin :obj_store
          mixin :backup_id
          option '--blocks', :flag, 'ensure clone from raw block device'
          apiopt('--thicken', :flag, 'make thick clone',
                 :param_group => :disk_params,
                 :param_name => %w(disk thicken))
          apiopt('--bandwidth-limit', 'LIMIT', 'bandwidth limit (in bytes) when restoring from object',
                 :param_group => :disk_params,
                 :param_name => %w(disk bandwidth)) do |v|
            parse_bytes(v)
          end
        end

        # provision with profile
        option_group nil, :hidden => true do
          apiopt('--with-profile', :flag, 'create profile with vss',
                 :param_group => :profile_params,
                 :param_name => %w(profile create))
          apiopt('--profile-label', 'LABEL', 'specify profile label',
                 :param_group => :profile_params,
                 :param_name => %w(profile label))
          apiopt('--no-profile-label', :flag, 'specify null profile label',
                 :param_group => :profile_params,
                 :param_value => nil, :param_name => %w(profile label))
          apiopt('--profile-xmd-ref', 'REF', 'xmd record reference key',
                 :multivalued => true,
                 :param_group => :profile_params,
                 :param_name => %w(profile xmd_refs))
          apiopt('--profile-xref', 'REF', 'profile external reference key',
                 :param_group => :profile_params,
                 :param_name => %w(profile xref))
        end

        # provision with target
        option_group nil, :hidden => true do
          apiopt('--with-target', :flag, 'create target with vss',
                 :param_group => :target_params,
                 :param_name => %w(target create))
          apiopt('--target-label', 'LABEL', 'specify target label',
                 :param_group => :target_params,
                 :param_name => %w(target label))
          apiopt('--no-target-label', :flag, 'specify null target label',
                 :param_group => :target_params,
                 :param_value => nil, :param_name => %w(target label))
          apiopt('--target-xmd-ref', 'REF', 'xmd record reference key',
                 :multivalued => true,
                 :param_group => :target_params,
                 :param_name => %w(target xmd_refs))
          apiopt('--target-xref', 'REF', 'target external reference key',
                 :param_group => :target_params,
                 :param_name => %w(target xref))
        end

        # provision with xmd
        option_group nil, :hidden => true do
          apiopt('--with-xmd', :flag, 'create xmd with vss',
                 :param_group => :xmd_params,
                 :param_name => %w(xmd create))
          apiopt('--xmd-rec-ref', 'REF', 'specify xmd ref',
                 :param_group => :xmd_params,
                 :param_name => %w(xmd ref))
          apiopt('--xmd-label', 'LABEL', 'specify xmd label',
                 :param_group => :xmd_params,
                 :param_name => %w(xmd label))
          apiopt('--no-xmd-label', :flag, 'specify null xmd label',
                 :param_group => :xmd_params,
                 :param_value => nil, :param_name => %w(xmd label))
          apiopt('--xmd-xmd-ref', 'REF', 'xmd record reference key',
                 :multivalued => true,
                 :param_group => :xmd_params,
                 :param_name => %w(xmd xmd_refs))
          apiopt('--xmd-tag', 'TAG', 'specify xmd tag (may be used multiple times)',
                 :multivalued => true,
                 :param_group => :xmd_params,
                 :param_name => %w(xmd tags))
          apiopt('--xmd-xref', 'REF', 'xmd external reference key',
                 :param_group => :xmd_params,
                 :param_name => %w(xmd xref))
        end
      end

      option_group 'Resource', :hidden => true do
        # provision by token
        apiopt(['-T', '--token'], 'TOKEN', 'token from previous catalog query results',
               :param_group => :server_params)
        mixin :datastore
        mixin :account
      end

      option_group 'Grouping', :hidden => true do
        # wrap all created objects into an xmd
        option('--group', :flag, 'create and link xmd grouping object')
        option('--group-label', 'LABEL', 'specify group label')
        option('--group-notes', 'NOTES', 'specify group notes')
        option('--group-ref', 'REF', 'specify group ref')
        option('--group-xref', 'REF', 'specify group xref')
        option('--group-xmd-ref', 'REF', 'xmd record reference key',
               :multivalued => true)
        option('--group-publish', :flag, 'publish the group')
        option('--group-reserve', :flag, 'take a reservation on the group ref during provision')
        option('--group-exists-ok', :flag, 'when creating group, ok if already exists')
        option('--group-tag', 'TAG', 'specify group tag (may be used multiple times)',
               :multivalued => true,
               :attribute_name => :group_tags)
        option('--group-data-json', 'KEY=JSON', 'key=value pair data to insert (parsed json)',
               :multivalued => true, :attribute_name => :group_data) do |v|
          k, v = v.split('=')
          { k => JSON.parse(v) }
        end
      end

      mixin :info_extraction

      parameter("[QUERY-TAGS] ...", "query tags (+tag to include, -tag to exclude)", :attribute_name => :tag_query)

      self.description = <<-EOS
        Provision a new virtual storage service. You may use a product token from
        a previous search or perform the provisioning operation in one step by
        specifying search parameters.
      EOS

      def exec
        params = {}
        qp = nil

        if datastore
          ## manual provisioning

          # provision on explicit datastore -- the api takes a product ID, so
          # lookup the datastore first.

          ds = obj_list(:datastore, restrict: :only_one)

          params = server_params.merge(product_id: ds.product_id)

          if with_disk?
            signal_usage_error "Disk capacity (-c) required" unless capacity
            params[:disk] = { capacity: capacity }
          end

        elsif token
          ## provision a previously-searched-for product
          params = server_params

        else
          signal_usage_error "Capacity (-c) required for catalog query" unless capacity

          qp = query_params.deeper_merge(parse_tag_query(tag_query))

          params = server_params

          # perform query in group reservation
          unless group_reserve?
            ## provision based on query params.
            search = api.product.search_create(qp)
            products = search.is_a?(Array) ? search : search[:products]

            if products.length == 0
              runtime_error "Query returned no results"
            end

            # provision the first result
            params = server_params.merge(token: products[0].token)
          end
        end

        if with_clone?
          clone_params = {}
          unless disk || snapshot || obj_store
            usage_error "One of --disk, --snapshot, or --obj-store is required to specify the clone source"
          end
          if snapshot
            if disk || obj_store
              usage_error "Only one of either --disk, --snapshot or --obj-store can be specified to clone from"
            end
            snap = obj_list(:snapshot, restrict: :only_one)
            if blocks?
              clone_params[:src] = "block"
              clone_params[:block_id] = snap.id
            else
              clone_params[:src] = "snapshot"
              clone_params[:snapshot_id] = snap.id
            end
          end

          if disk
            if snapshot || obj_store
              usage_error "Only one of either --disk, --snapshot or --obj-store can be specified to clone from"
            end
            disk = obj_list(:disk, restrict: :only_one)
            if blocks?
              clone_params[:src] = "block"
              clone_params[:block_id] = disk.id
            else
              clone_params[:src] = "vdisk"
              clone_params[:vdisk_id] = disk.id
            end
          end

          if obj_store || backup_id
            if snapshot || disk
              usage_error "Only one of either --disk, --snapshot or --obj-store can be specified to clone from"
            end

            unless obj_store && backup_id 
              usage_error "Clone from object requires both --obj-store and --backup-id to be specified"
            end

            ostore = obj_list(:obj_store, restrict: :only_one)
            clone_params[:src] = "object"
            clone_params[:obj_store_id] = ostore.id
            clone_params[:backup_id]    = backup_id
          end

          clone_params = { :disk => clone_params }
        end

        # include optional object configuration if specified.
        params.deep_merge!(disk_params)    if with_disk? || with_clone?
        params.deep_merge!(clone_params)   if with_clone?
        params.deep_merge!(profile_params) if with_profile?
        params.deep_merge!(target_params)  if with_target?
        params.deep_merge!(xmd_params)     if with_xmd?
        params.deep_merge!(query: qp)      if group_reserve?

        if group?
          # ensure xmd is created.
          xmd_params = params[:xmd] ||= {}

          xmd_params[:create]     = true
          xmd_params[:exists_ok]  = group_exists_ok? if group_exists_ok?
          xmd_params[:ref]      ||= group_ref || "group-#{::SecureRandom.uuid}"
          xmd_params[:xref]     ||= group_xref if group_xref
          xmd_params[:xmd_refs] ||= []
          xmd_params[:data]     ||= {}
          xmd_params[:tags]     ||= group_tags if group_tags
          xmd_params[:publish]    = group_publish? if group_publish?
          xmd_params[:reservation] = group_reserve? if group_reserve?

          # add group xmd data
          xmd_params[:data][:group] = {
            :_schema  => "group",
            :_publish => false,
            :data => {
              :label => group_label,
              :notes => group_notes,
            }
          }

          # group data
          group_data.each do |group_data|
            xmd_params[:data].deep_merge!(group_data) if group_data
          end

          # add any additional refs
          xmd_params[:xmd_refs].push(*group_xmd_ref_list)

          # for each specified object, link to the xmd
          group_ref = xmd_params[:ref]

          (params[:disk][:xmd_refs] ||= []).push(group_ref) if with_disk?
          (params[:target][:xmd_refs] ||= []).push(group_ref) if with_target?
          (params[:profile][:xmd_refs] ||= []).push(group_ref) if with_profile?
        end

        if account
          acct_obj = obj_list(:account, restrict: :only_one)
          params[:account_id] = acct_obj.id
        end

        vss = api.vss.create(params)
        process_obj_created vss
      end
    end
    subcommand ['provision', 'create'], 'provision a new vss from the catalog', ProvisionVss, level: :advanced

    mixin :vss, :param_path => '/id'
    mixin :info_extraction

    subcommand ['list', 'ls'], 'enumerate virtual storage services', self, level: :basic do
      mixin :info_filtering
      mixin :sort_by

      def exec
        servers = obj_list(:vss)
        process_results servers, view: 'vss_list'
      end
    end

    subcommand 'info', 'show detailed virtual storage service information', self, level: :basic do
      option '--disks', :flag, 'enumerate disks in service'
      option '--targets', :flag, 'enumerate tarets in service'

      mixin :info_filtering
      mixin :info_extraction

      def exec
        servers = obj_list(:vss)

        servers.map! do |s|
          s.disks = api.vdisk.list(vss_id: s.id) if disks?
          s.targets = api.target.list(vss_id: s.id) if targets?
          s
        end

        process_results servers
      end
    end

    subcommand 'update', 'update basic vss parameters', self, level: :basic do
      apiopt ['--label', '-l'], 'LABEL', 'specify new vss label'
      apiopt '--notes', 'NOTES', 'specify new vss notes'
      mixin :xmd_refs_update

      def exec
        params = api_params

        vss = obj_list(:vss, restrict: :only_one)

        set_xmd_refs(params, vss.xmd_refs)

        vss = api.vss.update(vss.id, params)
        process_obj_updated vss
      end
    end

    subcommand 'quota', 'manage vss quota' do
      subcommand 'update', 'update vss quota' do
        apiopt('--size-reserve', 'RESERVE', 'size reservation (in bytes)', {
                 :param_name => %w(size reserve)
               }) do |s|
          parse_bytes(s)
        end
        apiopt('--size-limit', 'LIMIT', 'size limit (in bytes)', {
                 :param_name => %w(size limit)
               }) do |s|
          parse_bytes(s)
        end
        apiopt('--no-size-limit', :flag, 'remove size limit', {
                 :param_name => %w(size limit) }) do |val|
          nil # translate to json 'null'
        end
        apiopt('--size-notify', 'NOTIFY', 'size notification threshold (in bytes)', {
                 :param_name => %w(size notify) }) do |s|
          parse_bytes(s)
        end
        apiopt('--no-size-notify', :flag, 'remove size notification', {
                 :param_name => %w(size notify) }) do |val|
          nil # translate to json 'null'
        end
        apiopt('--no-iops-limit', :flag, 'remove iops limit', {
                 :param_name => %w(iops limit) }) do |v|
          nil # translate to json 'null'
        end
        apiopt('--iops-limit', 'IOPS', 'iops limit', {
               :param_name => %w(iops limit) })  do |s| Integer(s) end
        apiopt('--iops-reserve', 'IOPS', 'iops reservation', {
               :param_name => %w(iops reserve) })  do |s| Integer(s) end

        def exec
          vss = obj_list(:vss, restrict: :only_one)

          quota_params = vss.quota.dup
          quota_params.deep_merge!(api_params)

          api.vss.quota_update(vss.id, quota_params.to_h)
          process_obj_updated vss, desc: "Updated quota for vss #{vss}"
        end
      end
    end

    subcommand 'remove', 'remove a virtual storage service' do
      def exec
        vss = obj_list(:vss, restrict: :only_one)

        api.vss.remove(vss.id)
        process_obj_removed vss
      end
    end
  end
end
