# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require "blockbridge/util"

module Blockbridge::CLI::Command
  class Xmd < Base
    self.default_subcommand = 'list'

    def xmd_match_schema(otype, objects, desc)
      objects = objects.select { |o| o.data.values.any? { |v| v._schema == schema } }
    end

    def only_one_schema(otype, objects, desc)
      objects = xmd_match_schema(otype, objects, desc)
      only_one(otype, objects, desc)
    end

    def xmd_ref
      if ref
        ref
      elsif disk
        disk = obj_list(:disk, restrict: :only_one)
        if schema
          frag = disk.xmd.fragments_for(schema: schema)
          refs = frag.map(&:ref).uniq
          xmds = refs.map{ |r| api.xmd.info(r) }
          rec = only_one(:xmd, xmds)
        else
          rec = only_one(:xmd, disk.xmd)
        end
        rec.ref
      elsif schema
        rec = obj_list(:xmd, restrict: :only_one_schema)
        rec.ref
      elsif vss
        rec = obj_list(:xmd, restrict: :only_one)
        rec.ref
      else
        usage_error "Specify an xmd by ref (--ref), vss (-v, --vss), disk ref (-d, --disk), or schema (--schema)."
      end
    end

    def xmd_objs
      if disk
        disk = obj_list(:disk, restrict: :only_one)
        if schema
          frag = disk.xmd.fragments_for(schema: schema)
          refs = frag.map(&:ref).uniq
          xmds = refs.map{ |r| api.xmd.info(r) }
        else
          xmds = disk.xmd
        end
        one_or_more(:xmd, xmds)
      elsif schema
        xmds = obj_list(:xmd, restrict: :xmd_match_schema)
        one_or_more(:xmd, xmds)
      else
        xmds = obj_list(:xmd)
      end
    end

    mixin :xmd_status do
      apiopt('--sts-val', 'VALUE', 'specify status value (online, degraded, offline, pending)',
             :param_name => %w(ext_status value)) do |v|
        enum_check(v, %w(online degraded offline pending))
      end
      apiopt('--sts-ind', 'INDICATOR', 'specify status indicator (online, degraded, offline)',
             :param_name => %w(ext_status indicator)) do |v|
        enum_check(v, %w(online degraded offline))
      end
      apiopt('--sts-detail', 'STR', 'specify status detail',
             :param_name => %w(ext_status detail)) do |v|
        v.length == 0 ? nil : v
      end
      apiopt('--clear-sts', :flag, 'clear existing external status',
             :param_name => :ext_status,
             :param_value => nil)
    end

    mixin :xmd_data do
      apiopt(['--str'], 'KEY=STR', 'key=value pair to insert (string)',
             :multivalued => true, :param_name => :data, :param_list_merge => true) do |v|
        k, v = v.split('=')
        { k => v }
      end
      apiopt(['--int'], 'KEY=INT', 'key=value pair to insert (integer)',
             :multivalued => true, :param_name => :data, :param_list_merge => true) do |v|
        k, v = v.split('=')
        { k => Integer(v) }
      end
      apiopt(['--float'], 'KEY=FLOAT', 'key=value pair to insert (float)',
             :multivalued => true, :param_name => :data, :param_list_merge => true) do |v|
        k, v = v.split('=')
        { k => Float(v) }
      end
      apiopt(['--bool'], 'KEY=VAL', 'key=value pair to insert (boolean)',
             :multivalued => true, :param_name => :data, :param_list_merge => true) do |v|
        k, v = v.split('=')
        { k => parse_boolean(v) }
      end
      apiopt(['--null'], 'KEY', 'key to set to null',
             :multivalued => true, :param_name => :data, :param_list_merge => true) do |k|
        { k => nil }
      end
      apiopt(['--json'], 'KEY=JSON', 'key=value pair to insert (parsed json)',
             :multivalued => true, :param_name => :data, :param_list_merge => true) do |v|
        k, v = v.split('=')
        { k => JSON.parse(v) }
      end
      apiopt(['--data'], 'KEY=SCHEMA', 'creates nested data block with specified schema type',
             :param_name => :data,
             :param_list_merge => true,
             :multivalued => true) do |v|
        k, v = v.split('=')
        { k => { "_schema" => v, "data" => {} } }
      end
    end

    mixin :vss
    mixin :disk
    mixin :info_extraction
    apiopt(['--schema'], 'SCHEMA', 'schema type of data block')

    class XmdCreate < Base
      option '--ref', 'REF', 'xmd record key', :required => true
      mixin :uuid
      mixin :vss
      mixin :label_notes, obj_name: 'xmd'
      mixin :xmd_data
      mixin :xmd_refs_create
      mixin :tags_create
      mixin :xref
      mixin :xmd_status
      mixin :info_extraction

      apiopt('--publish', :flag, 'publish xmd to the system')

      option_group 'Grouping', :hidden => true do
        option('--group', :flag, 'create xmd as grouping object')
        option('--group-label', 'LABEL', 'specify group label')
        option('--group-notes', 'NOTES', 'specify group notes')
      end

      def exec
        create_params = api_params.merge(ref: ref)

        # create xmd as a grouping object
        if group?
          create_params[:data] ||= {}
          group_label ||= ref

          # add group xmd data
          create_params[:data][:group] = {
            :_schema  => "group",
            :_publish => false,
            :data => {
              :label => group_label,
              :notes => group_notes,
            }
          }
        end

        # optionally bind to a vss
        if vss
          vss = obj_list(:vss, restrict: :only_one)
          create_params[:vss_id] = vss.id
        end

        xmd = api.xmd.create(create_params)
        process_obj_created xmd, desc: "Created xmd ref '#{xmd.ref}'"
      end
    end
    subcommand 'create', 'create a new xmd record', XmdCreate, level: :expert

    subcommand ['list', 'ls'], 'enumerate xmd records', self, level: :expert do
      option '--ref', 'REF', 'xmd record key'
      option '--label', 'LABEL', 'xmd label to search for'
      mixin :info_filtering
      mixin :sort_by

      def exec
        if label
          xmds = api.xmd.list(label: label)
        elsif ref.nil?
          xmds = xmd_objs
        else
          xmds = [ api.xmd.info(ref) ]
        end
        process_results xmds, view: 'xmd_list'
      rescue Blockbridge::Api::NotFoundError, Excon::Errors::NotFound, Excon::Errors::Gone
        lookup_error "No xmd found that matches '#{ref}'."
      end
    end

    subcommand 'info', 'show detailed xmd record information', self, level: :expert do
      option '--ref', 'REF', 'xmd record key'
      mixin :info_filtering
      mixin :info_extraction

      def exec
        if ref.nil?
          xmds = xmd_objs
        else
          xmds = [ api.xmd.info(ref) ]
        end
        process_results xmds
      rescue Blockbridge::Api::NotFoundError, Excon::Errors::NotFound, Excon::Errors::Gone
        lookup_error "No xmd found that matches '#{ref}'."
      end
    end

    subcommand 'patch', 'patch xmd record data', self, level: :expert do
      option '--ref', 'REF', 'xmd record key'

      option '--str', :flag, 'interpret value as string (DEFAULT)' do
        @val_type = :string
      end
      option '--int', :flag, 'interpret value as integer' do
        @val_type = :integer
      end
      option '--float', :flag, 'interpret value as float' do
        @val_type = :float
      end
      option '--bool', :flag, 'interpret value as boolean' do
        @val_type = :boolean
      end
      option '--json', :flag, 'interpret value as json' do
        @val_type = :json
      end
      option '--null', :flag, 'interpret value as null' do
        @val_type = :null
      end
      option '--file-as-array', :flag, 'interpret value as path to file' do
        @val_type = :file_as_array
      end

      apiopt('--noent-action', 'ACTION', "record-not-found error handling.\n[fail,success,create]") do |v|
        enum_check v, %w(fail success create)
      end

      subcommand 'add', 'insert (or replace) value into document', self, level: :expert do
        parameter 'PATH', "JSON Pointer path to location"
        parameter '[VALUE]', "value to add to document"

        def exec
          update_params = {
            mode: 'patch',
            data: [ { op: 'add', path: path, value:
                      parse_value(value, @val_type) } ]
          }

          xmd = api.xmd.update(xmd_ref, update_params)
          process_obj_updated xmd, desc: "Patched xmd ref '#{xmd.ref}'"
        end
      end

      subcommand 'replace', 'replace (must exist) value into document', self, level: :expert do
        parameter 'PATH', "JSON Pointer path to location"
        parameter '[VALUE]', "value to add to document"

        def exec
          update_params = {
            mode: 'patch',
            data: [ { op: 'replace', path: path,
                      value: parse_value(value, @val_type) } ]
          }

          xmd = api.xmd.update(xmd_ref, update_params)
          process_obj_updated xmd, desc: "Patched xmd ref '#{xmd.ref}'"
        end
      end

      subcommand 'remove', 'remove value from document', self, level: :expert do
        parameter 'PATH', "JSON Pointer path to location"

        def exec
          update_params = {
            mode: 'patch',
            data: [ { op: 'remove', path: path } ]
          }

          xmd = api.xmd.update(xmd_ref, update_params)
          process_obj_updated xmd, desc: "Patched xmd ref '#{xmd.ref}'"
        end
      end

      subcommand 'copy', 'copy value from one path to another path', self, level: :expert do
        parameter 'FROM', "JSON Pointer path to source location"
        parameter 'TO', "JSON Pointer path to destination location"

        def exec
          update_params = {
            mode: 'patch',
            data: [ { op: 'copy', from: from, to: to } ]
          }

          xmd = api.xmd.update(xmd_ref, update_params)
          process_obj_updated xmd, desc: "Patched xmd ref '#{xmd.ref}'"
        end
      end

      subcommand 'move', 'move value from one path to another path', self, level: :expert do
        parameter 'FROM', "JSON Pointer path to source location"
        parameter 'TO', "JSON Pointer path to destination location"

        def exec
          update_params = {
            mode: 'patch',
            data: [ { op: 'move', from: from, to: to } ]
          }

          xmd = api.xmd.update(xmd_ref, update_params)
          process_obj_updated xmd, desc: "Patched xmd ref '#{xmd.ref}'"
        end
      end

      subcommand 'test', 'test value in document', self, level: :expert do
        parameter 'PATH', "JSON Pointer path to location"
        parameter '[VALUE]', "value to test for"

        def exec
          update_params = {
            mode: 'patch',
            data: [ { op: 'test', path: path,
                      value: parse_value(value, @val_type) } ]
          }

          xmd = api.xmd.update(xmd_ref, update_params)
          process_obj_updated xmd, desc: "Patched xmd ref '#{xmd.ref}'"
        end
      end
    end

    subcommand 'update', 'update xmd record', self, level: :expert do
      option '--ref', 'REF', 'xmd record key'
      mixin :label_notes, obj_name: 'xmd', hidden: false
      mixin :xmd_data
      apiopt('--replace', :flag, 'replace existing data (default is to merge)',
             :param_name => %w(mode), :default => 'merge') do |val|
        "replace"
      end
      mixin :xmd_refs_update
      apiopt('--noent-action', 'ACTION', "record-not-found error handling.\n[fail,success,create]") do |v|
        enum_check v, %w(fail success create)
      end
      mixin :xmd_status
      mixin :tags_update

      def exec
        update_params = api_params
        ex_rec = nil

        # update refs
        set_xmd_refs(update_params) do
          if (ex_rec ||= api.xmd.info(xmd_ref)).nil?
            usage_error "Failed to find xmd record with ref '#{ref}'"
          end
          ex_rec.xmd_refs
        end

        # update tags
        set_tags(update_params) do
          if (ex_rec ||= api.xmd.info(xmd_ref)).nil?
            usage_error "Failed to find xmd record with ref '#{ref}'"
          end
          ex_rec.tags
        end

        xmd = api.xmd.update(xmd_ref, update_params)
        process_obj_updated xmd, desc: "Updated xmd ref '#{xmd.ref}'"
      end
    end

    subcommand 'remove', 'remove an xmd record', self, level: :expert do
      option '--ref', 'REF', 'xmd record key'
      def exec
        del_ref = xmd_ref

        # look up existing xmd
        xmd = api.xmd.info(del_ref)

        rsp = api.xmd.remove(del_ref)
        process_obj_removed xmd, desc: "Removed xmd ref '#{del_ref}'"
      end
    end

  end
end
