# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'blockbridge/cli/errors'

require 'active_support/core_ext/object/deep_dup'
require 'delegate'
require 'hashdiff'
require 'yaml'

class ConfigDelegator < SimpleDelegator
  def initialize(init_h = nil)
    super init_h || Hash.new
  end

  def to_hash
    @data.each_with_object({}) do |(k, v), h|
      h[k] = v.is_a?(ConfigDelegator) ? v.to_hash : v
    end
  end
end

class ConfigHash < ConfigDelegator
  class << self
    def value_class(klass = nil)
      klass ? @value_class = klass : @value_class || ConfigSection
    end
  end

  def initialize(init_h = Hash.new)
    @data = Hash.new
    init_h.to_hash.each_pair do |k, v|
      @data[k] = wrap_value(v)
    end
    super @data
  end

  def deep_dup
    self.class.new @data.deep_dup
  end

  def []=(key, value)
    @data[key] = wrap_value(value)
  end

  def delete(key)
    @data.delete(key)
  end

  private

  def wrap_value(v)
    klass = self.class.value_class
    if klass.respond_to?(:call)
      klass.call(v)
    elsif klass.is_a?(Class)
      klass.new(v)
    else
      v
    end
  end
end

class ConfigSection < ConfigDelegator
  class << self
    def cfg_sections
      @cfg_sections ||= []
    end

    def cfg_attrs
      @cfg_attrs ||= []
    end

    def cfg_section(name, klass = ConfigSection)
      name = name.to_sym
      raise "cfg_section #{name} already defined" if cfg_sections.include?(name)
      cfg_sections << name
      define_method(name) do
        @data[name.to_s] ||= klass.new
      end
      define_method("#{name}=") do |val|
        raise 'section value must be hash-like' unless val.respond_to? :to_hash
        @data[name.to_s] = klass.new(val)
      end
    end
    def cfg_attr(name, opts = {})
      name = name.to_sym
      raise "cfg_attr #{name} already defined" if cfg_attrs.include?(name)
      cfg_attrs << name
      define_method(name) do
        @data.fetch(name.to_s, opts[:default])
      end
      define_method("#{name}=") do |val|
        @data[name.to_s] = val
      end
      define_method("#{name}_default?") do
        opts.key?(:default) && !@data.key?(name.to_s)
      end
    end
  end

  def initialize(init_h = Hash.new)
    @data = Hash.new
    init_h.to_hash.each_pair do |k, v|
      if section?(k) || attr?(k)
        self.send("#{k}=", v)
      else
        @data[k] = v
      end
    end
    super @data
  end

  def get_path(path)
    path.split('.').inject(self) { |o, p| o && o.send(p) }
  end

  def deep_dup
    self.class.new @data.deep_dup
  end

  def section?(name)
    self.class.cfg_sections.include? name.to_sym
  end

  def attr?(name)
    self.class.cfg_attrs.include? name.to_sym
  end

  def cfg_attrs
    self.class.cfg_attrs
  end

  def [](key)
    send(key) if respond_to?(key)
  end
end

class ApiHostConfig < ConfigSection
  cfg_attr :url
  cfg_attr :default
end

class ApiHostsConfig < ConfigHash
  value_class ->(v) { ApiHostConfig.new(v) }
end

class ApiConfig < ConfigSection
  cfg_section :hosts, ApiHostsConfig
  cfg_attr    :ssl_verify_peer
end

class GlobalConfig < ConfigSection
  cfg_attr :uuid
  cfg_attr :color
  cfg_attr :error_format, default: 'human'
  cfg_attr :role, default: 'tenant'
  cfg_attr :level, default: 'basic'
end

class DevicesConfig < ConfigSection
  cfg_attr :queue_depth, default: 96
  cfg_attr :nomerges,    default: 1
  cfg_attr :rq_affinity, default: 2
  cfg_attr :scheduler,   default: 'noop'
  cfg_attr :rotational,  default: 0
  cfg_attr :add_random,  default: 0
  cfg_attr :nr_requests, default: 128
end

class HostConfig < ConfigSection
  cfg_section :device, DevicesConfig
end

class RootConfig < ConfigSection
  cfg_section :api, ApiConfig
  cfg_section :host, HostConfig
  cfg_section :global, GlobalConfig
end

class Blockbridge::CLI::Config < SimpleDelegator
  class Error < StandardError; end

  attr_reader :filename

  def initialize(filename)
    @filename = filename
    @file = File.basename(filename)
    @dir = File.dirname(filename)
    @data = load unless disabled?
    @pdata = @data.deep_dup
    super @data
  end

  def disabled?
    filename.nil? || filename.empty? || filename == '/dev/null'
  end

  def exist?
    File.exist?(filename)
  end

  def get_path(path)
    path.split('.').inject(@data) { |d, c| d && d.send(c) }
  end

  def save(force = false)
    raise Error, "configuration support is disabled" if disabled?
    return unless force || changed?

    ensure_dir_exists
    write
  rescue Errno::EACCES => e
    raise Error, "permission denied while writing config to '#{filename}'"
  end

  def changed?
    !HashDiff.diff(@pdata, @data).empty?
  end

  private

  def write
    File.open("#{filename}.tmp", 'w') { |f| f.write YAML.dump(@data.to_hash) }
    File.rename("#{filename}.tmp", filename)
    @pdata = @data.deep_dup
  end

  def load
    RootConfig.new YAML.load(File.read(filename))
  rescue Errno::ENOENT
    RootConfig.new # empty configuration
  rescue Errno::EACCES => e
    raise Error, "permission denied while reading config from '#{filename}'"
  end

  def ensure_dir_exists
    FileUtils.mkpath(@dir)
  rescue Errno::EEXIST => e
    raise Error, "non-directory component while creating path to '#{filename}'"
  rescue Errno::EACCES => e
    raise Error, "permission denied while creating path to '#{filename}'"
  end
end
