# frozen_string_literal: true

# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

module Blockbridge::CLI
  class HelpBuilder
    attr_reader :cli_role, :cli_level

    def initialize(opts = {})
      @out = StringIO.new
      @opts = opts
      @cli_role = opts[:role]
      @cli_level = map_level(opts.fetch(:level, :basic))
    end

    def map_level(lvl)
      case lvl.to_sym
      when :basic
        1
      when :advanced
        2
      when :expert
        3
      when :dev
        4
      end
    end

    def string
      @out.string
    end

    def add_usage(invocation_path, usage_descriptions)
      puts 'Usage:'
      usage_descriptions.each do |usage|
        puts "    #{invocation_path} #{usage}".rstrip
      end
    end

    def add_description(description)
      return unless description

      puts ''
      puts description.gsub(/^/, '    ')
    end

    DETAIL_FORMAT = "    %-29s %s\n"

    def add_list(heading, raw_items)
      raw_items.select! do |item|
        if (roles = item.options[:roles])
          next false unless roles.include?(cli_role)
        end

        item_level = map_level(item.options.fetch(:level))
        next false unless cli_level >= item_level

        true
      end

      # optionally group items
      item_groups = raw_items.group_by { |i| i.options[:group] }

      item_groups.each_pair do |name, items|
        visible, hidden = items.partition do |i|
          @opts[:verbose] || !i.hidden?
        end

        # skip empty option group sections
        next if visible.empty?

        print "\n"

        if name
          print "#{name} #{heading.downcase}"
        else
          print heading
        end

        print " (#{hidden.length} hidden)" if hidden.any?
        puts  ':'

        visible.each do |item|
          label, description = item.help

          description.each_line.each_with_index do |line, i|
            puts DETAIL_FORMAT % [i.zero? ? label : '', line]
          end
        end
      end
    end

    private

    def print(*args)
      @out.print(*args)
    end

    def puts(*args)
      @out.puts(*args)
    end

  end

end
