# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'base64'
require 'time'
require 'hashie'
require 'jsonpath'
require 'naturally'
require 'pbkdf2'
require 'digest'
require 'blockbridge/cli/helpers/parsing'
require 'blockbridge/cli/helpers/attach'

module Blockbridge::CLI::Helpers
  OBJ_TYPE_INFO = {
    # storage objects
    disk: {
      scopes: [ :vss, :node, :account ],
      parent: :vss,
      apitype: :vdisk
    },
    snapshot:  {
      scopes: [ :vss, :disk, :node, :account ],
      parent: :disk,
      apitype: :snapshot
    },
    target: {
      scopes: [ :vss, :node, :account ],
      parent: :vss,
      apitype: :target
    },
    profile: {
      scopes: [ :vss, :node, :account ],
      parent: :vss,
      apitype: :initiator_profile,
    },
    rule: {
      scopes: [ :vss, :node, :account ],
      parent: :vss, # or disk? a little weird...
      apitype: :rule,
    },
    vss: {
      scopes: [ :node, :account, :datastore ],
      apitype: :vss,
    },
    product: {
      scopes: [ :node, :datastore ],
      apitype: :product,
    },
    obj_store: {
      scopes: [ :account ],
      apitype: :obj_store,
    },

    # infrastructure objects
    site: {
      scopes: [ ], # top level
      apitype: :site,
    },
    node: {
      scopes: [ :site ],
      apitype: :node,
    },
    datastore: {
      scopes: [ :node, :account ],
      parent: :node,
      apitype: :sys_ds,
    },
    net: {
      scopes: [ :node ],
      parent: :node,
      apitype: :net,
    },
    os_dev: {
      scopes: [ :node ],
      parent: :node,
      apitype: :os_dev,
      id_fields: [ :label, 'device_info/display_name', 'device_info/serial', :path, 'device_info/devname', 'device_info/dsn' ]
    },
    iscsi_dev: {
      scopes: [ :node ],
      parent: :node,
      apitype: :iscsi_dev,
    },
    dev: {
      scopes: [ :node ],
      parent: :node,
      apitype: :os_dev,
      id_fields: [ :label, 'device_info/display_name', 'device_info/serial', :path, 'device_info/devname', 'device_info/dsn' ]
    },

    # account/domain objects
    account: {
      scopes: [ ], # top level
      apitype: :account,
      label: :name,
    },
    user: {
      scopes: [ :account ],
      parent: :account,
      apitype: :user_profile,
      label: :login,
    },

    # misc objects
    task: {
      scopes: [ :disk, :datastore ],
      apitype: :task
    },

    xmd: {
      scopes: [ :vss ],
      apitype: :xmd,
    },

    authz: {
      scopes: [],
      apires: :authorization,
      apitype: :authz,
      label: nil,
    },
  }

  SERIAL_RX   = /^([A-Z]{3})[A-F0-9]{16}$/
  OID_RX      = /^[a-z_]+:(\d+)(:(\d))?/
  OBJ_SPEC_RX = /^(\/?(?<par>[^\/]+)\/)?(?<child>[^\/]+)$/
  UUID_RX     = /^[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-
                  [a-fA-F0-9]{4}-[a-fA-F0-9]{12}$/x

  def scope_val_type(val)
    case val
    when SERIAL_RX
      :serial
    when OID_RX, UUID_RX
      :id
    else
      :label
    end
  end

  def scope_val_filter(val, type_info)
    t = scope_val_type(val)
    f = type_info.fetch(t, t)

    if f.nil?
      signal_usage_error("#{obj_plural(type_info[:authz])} must be specified by id or serial number")
    end

    { f => val }
  end

  def scope_obj_list(otype, filter = {}, opts = {})
    apires = OBJ_TYPE_INFO[otype][:apires] || OBJ_TYPE_INFO[otype][:apitype]

    if cf = opts[:custom_filter]
      filter = cf.merge(filter)
    end

    if filter.keys.length == 1 && (filter.has_key?(:id) || filter.has_key?(:serial))
      # if looking up by ID or serial, do a simple get, and translate the 404
      # status into an empty array.
      begin
        if filter.has_key?(:id)
          [ api.send(apires).info(filter[:id]) ]
        else
          [ api.send(apires).info(filter[:serial]) ]
        end
      rescue Blockbridge::Api::NotFoundError, Excon::Errors::NotFound, Excon::Errors::Gone
        [ ]
      end
    else
      api.send(apires).list(filter)
    end
  end

  def scope_id_fields(otype, type_info, val)
    if type_info[:id_fields]
      type_info[:id_fields].each do |f|
        objects = scope_obj_list otype, { f => val }
        return objects unless objects.empty?
      end
    end
    []
  end

  def obj_list(otype, opts = {}, &blk)
    level = opts[:level] || 0
    arg_map = opts[:arg_map] || {}
    indent = ' ' * level

    puts "#{indent}obj-query[#{level}]: type = #{otype}, opts = #{opts}" if debug?

    val = nil
    oname = opts[:name] || otype
    oname = arg_map[oname] || oname

    if opts.has_key? :value
      val = opts[:value]
    elsif self.respond_to? oname
      val = self.send(oname)
    end

    if val.nil? && self.respond_to?(:identity_param)
      val = self.identity_param
    end

    type_info = OBJ_TYPE_INFO[otype]
    if type_info.nil?
      internal_error("Attempt to look up unknown object type '#{otype}'.")
    end

    selections = {}
    objects = []

    type_info[:scopes].each do |t|
      arg_name = arg_map[t] || t
      unless arg_map.has_key?(t) && arg_map[t].nil?
        selections[t] = self.send(arg_name) if self.respond_to?(arg_name)
      end
    end

    if val
      # lookup object by fields if specified
      objects = scope_id_fields(otype, type_info, val)

      # if not found by specified fields; lookup normal
      if objects.empty?
        if val.start_with? '/'
          # trap inputs starting with a '/' -- perhaps the user has specified a
          # /dev/whatever path as an object name?
          child_val = val
        else
          # split value on '/' and attempt to resolve parent references.
          parts = val.split('/').reject { |p| p.nil? || p == '' }

          if parts.length == 0
            signal_usage_error("Empty string specified for object name.")
          end

          # pop off the most specific portion of the obj specification, then try
          # to resolve the remaining bits as a parent ref.
          child_val = parts.pop

          if scope_val_type(child_val) == :id ||
             scope_val_type(child_val) == :serial
            type_info[:scopes].each do |t|
              selections.delete(t)
            end
          end

          parent_type_info = type_info
          parts.reverse.each do |part|
            break if OBJ_TYPE_INFO[parent_type_info[:parent]].nil?
            selections.delete(parent_type_info[:parent])
            parent_type_info = OBJ_TYPE_INFO[parent_type_info[:parent]]
          end

          if parts.length > 0
            selections[type_info[:parent]] = parts.join('/')
          end
        end

        child_filter = scope_val_filter(child_val, type_info)
        objects = scope_obj_list otype, child_filter, opts
      end
    else
      objects = scope_obj_list otype, {}, opts
    end

    # resolve any applicable scopes
    filters = {}

    selections.each_pair do |t, sval|
      next unless t
      apitype = OBJ_TYPE_INFO[t][:apitype]
      fld = apitype.to_s + '_id'

      if sval.nil?
        # nothing...
      elsif scope_val_type(sval) == :id
        filters[fld] = [ sval ]
      else
        # recursively search for matches
        objs = obj_list t, value: sval, arg_map: arg_map, level: level + 1
        if objs.length == 0
          msg = val ? "No #{obj_singular(otype)} found that matches '#{val}'. " : ''
          msg.concat "No #{t} found that matches '#{sval}'."
          lookup_error msg
        end

        filters[fld] = objs.map { |o| o['id'] }
      end
    end

    puts "#{indent} child_filter: #{child_filter}" if debug?
    puts "#{indent} filters: #{filters}" if debug?

    # filter resulting objects
    objects.select! do |obj|
      filters.all? { |(fld, vals)| vals.include? obj[fld] }
    end

    unless show_trash?
      objects.reject! { |o| o.trashed }
    end

    # if a block is specified, use it as a filter
    objects.select!(&blk) if block_given?

    if objects.empty? && val
      lookup_error "No #{obj_singular(otype)} found that matches '#{val}'." if level == 0
    end

    # process requested restriction
    if opts[:restrict]
      return method(opts[:restrict]).call(otype, objects, opts[:desc])
    end

    puts "#{indent} -> #{objects.length} objects" if debug?

    objects
  end

  def obj_singular(otype)
    case otype
    when :os_dev
      "device"
    when :iscsi_dev
      "iscsi device"
    when :vss
      "virtual storage service"
    when :vss
      "authorization"
    else
      otype.to_s
    end
  end

  def obj_plural(otype)
    case otype
    when :os_dev
      "devices"
    when :iscsi_dev
      "iscsi devices"
    when :vss
      "virtual storage services"
    when :authz
      "authorizations"
    else
      otype.to_s + "s"
    end
  end

  def obj_pluralize(otype, count, with_count = false)
    str = count == 1 ? obj_singular(otype) : obj_plural(otype)
    str = "#{count} #{str}" if with_count
  end

  def only_one(otype, objects, desc = nil)
    if objects.length == 0
      if desc
        usage_error("No matching #{desc} #{obj_plural(otype)} found.")
      else
        lookup_error("No matching #{obj_plural(otype)} found.")
      end
    end

    if objects.length > 1
      labels = Naturally.sort_by(objects, :label).map { |o| o.to_s }
      if desc
        lookup_error("More than one #{desc} #{obj_singular(otype)} found:\n #{labels.join("\n ")}")
      else
        lookup_error("More than one #{obj_singular(otype)} found:\n #{labels.join("\n ")}")
      end
    end
    objects.first # and only
  end

  def one_or_more(otype, objects, desc = nil)
    if objects.length == 0
      if desc
        lookup_error("No matching #{desc} #{obj_plural(otype)} found.")
      else
        lookup_error("No matching #{obj_plural(otype)} found.")
      end
    end
    objects
  end

  def one_or_none(otype, objects, desc = nil)
    if objects.length > 1
      labels = Naturally.sort_by(objects, :label).map { |o| o.to_s }
      if desc
        lookup_error("More than one #{desc} #{obj_singular(otype)} found:\n #{labels.join("\n ")}")
      else
        lookup_error("More than one #{obj_singular(otype)} found:\n #{labels.join("\n ")}")
      end
    end
    objects.first # and only
  end

  def only_one_mn(otype, objects, desc = nil)
    nodes = objects.select do |o|
      o.type == 'management'
    end
    only_one(otype, nodes, desc)
  end

  def only_one_sn(otype, objects, desc = nil)
    nodes = objects.select do |o|
      o.type == 'storage'
    end
    only_one(otype, nodes, desc)
  end

  def only_one_management_node
    obj_list(:node, restrict: :only_one_mn)
  end

  def only_one_storage_node
    obj_list(:node, restrict: :only_one_sn)
  end

  def only_one_storage_device(node_id, arg, desc = nil)
    os_devs = obj_list(:os_dev, arg_map: { os_dev: arg })
    iscsi_devs = obj_list(:iscsi_dev, arg_map: { iscsi_dev: arg })

    devs = os_devs + iscsi_devs

    only_one("storage device", devs, desc)
  end

  def display_raw(o = nil)
    return false unless raw?

    unless o.nil?
      # format input as pretty json
      puts JSON.pretty_generate(o)
    end

    true
  end

  def batch_remove(otype, &blk)
    allow_multi = batch? rescue false

    if allow_multi
      objects = obj_list(otype)
    else
      objects = [ obj_list(otype, restrict: :only_one) ]
    end

    if blk
      cb = blk
    else
      # auto-create basic remove proc
      apitype = OBJ_TYPE_INFO[otype][:apitype]
      rescol = api.public_send(apitype)

      if rescol.nil?
        fail "don't know how to operate on #{obj_plural(otype)}"
      end

      cb = proc do |obj|
        not_found_ok { rescol.remove(obj.id) }
        process_obj_removed obj
      end
    end

    # remove object(s)
    failed = objects.reject do |obj|
      begin
        cb.call(obj)
        true
      rescue
        raise unless allow_multi
        false
      end
    end

    return if failed.empty?

    runtime_error "Failed to remove #{obj_pluralize(otype, failed.length)}:" \
                  "\n#{failed.join("\n ")}"
  end

  def terminal_screen?
    ENV.fetch('TERM', '').start_with?('screen')
  end

  def print_osc
    if terminal_screen?
      print "\033Ptmux;\033\033]"
    else
      print "\033]"
    end
  end

  def print_st
    if terminal_screen?
      print "\a\033\\"
    else
      print "\a"
    end
  end

  def print_image(data, filename = nil, inline = true)
    print_osc
    print "1337;File="
    print "name='#{Base64.strict_encode64(filename)}';" if filename
    print "size=#{data.length};"
    print "inline=#{inline ? '1' : '0'}:"
    print Base64.strict_encode64(data)
    print_st
    print "\n"
  end

  def terminal_width
    cols, _rows = HighLine::SystemExtensions.terminal_size
    cols
  end

  def terminal_height
    _cols, rows = HighLine::SystemExtensions.terminal_size
    rows
  end

  def update_quota_params(cur_params = {})
    quota_opts = cur_params.dup

    # overwrite each element with the user-supplied argument
    if size_reserve
      quota_opts['size'] ||= {}
      quota_opts['size']['reserve'] = size_reserve
    end
    if self.respond_to?(:no_size_limit?) && no_size_limit?
      quota_opts['size'] ||= {}
      quota_opts['size']['limit'] = nil
    elsif size_limit
      quota_opts['size'] ||= {}
      quota_opts['size']['limit'] = size_limit
    end
    if self.respond_to?(:no_size_notify?) && no_size_notify?
      quota_opts['size'] ||= {}
      quota_opts['size']['notify'] = nil
    elsif size_notify
      quota_opts['size'] ||= {}
      quota_opts['size']['notify'] = size_notify
    end
    if self.respond_to?(:no_iops_limit?) && no_iops_limit?
      quota_opts['iops'] ||= {}
      quota_opts['iops']['limit'] = nil
    elsif iops_limit
      quota_opts['iops'] ||= {}
      quota_opts['iops']['limit'] = iops_limit
    end
    quota_opts['workload'] = nil

    quota_opts
  end

  def disk_quota_size(arg, capacity, ratio)
    if arg.nil?
      nil
    elsif md = /^(.+)%$/.match(arg)
      begin
        pct = Float(md[1])
      rescue ArgumentError
        signal_usage_error "Invalid percentage value: #{md[1]}"
      end

      # size as percent of capacity
      pct * (capacity / ratio) / 100
    else
      parse_bytes(arg)
    end
  end

  def prompt_for_password(purpose = "password", confirm: false)
    a = ask("Enter #{purpose}: ") { |q| q.echo = false }
    if confirm
      b = ask("Confirm #{purpose}: ") { |q| q.echo = false }
      if a != b
        runtime_error "Passwords do not match!"
      end
    end
    a
  end

  def password_from_file(password_file, purpose = "password")
    return nil if not password_file
    case password_file
    when /^fd:(\d+)/
      fd_num = $1.to_i
      verbose "# Reading #{purpose} from fd #{fd_num}..."
      IO.open(fd_num) do |f|
        f.readline.chomp
      end
    else
      verbose "# Reading #{purpose} from file #{password_file}..."
      File.open(password_file) do |f|
        f.readline.chomp
      end
    end
  end

  def password_from_file_or_prompt(password_file, purpose = "password")
    return prompt_for_password(purpose) if not password_file
    return password_from_file(password_file, purpose)
  end

  def profile_configure_token(profile, passphrase)
    runtime_error "No profile specified." if not profile
    runtime_error "No passphrase specified." if not passphrase

    # prepare
    prep       = api.initiator_profile.prepare_token(profile.id)

    # encode
    key        = PBKDF2.new do |p|
      p.password      = passphrase
      p.salt          = Base64.strict_decode64(prep.key_salt)
      p.iterations    = prep.key_rounds
      p.key_length    = prep.key_bytes
      p.hash_function = prep.key_alg.gsub(/^pbkdf2-hmac-/i,'')
    end
    pad_bytes  = Base64.strict_decode64(prep.pad).unpack("C*")
    key_bytes  = key.bin_string.unpack("C*")
    akey_bytes = key_bytes.zip(pad_bytes).map { |a,b| a ^ b }
    akey_hash  = Digest::SHA256.digest(akey_bytes.pack("C*"))

    # submit
    params = {
      id:       profile.id,
      tok_hash: Base64.strict_encode64(akey_hash),
      pad:      prep.pad,
    }
    profile = api.initiator_profile.create_token(profile.id, params)

    # token
    id_bytes = Base64.strict_decode64(profile[:access_token][:id]).unpack("C*")
    tok_bin = (id_bytes + akey_bytes).pack("C*")

    Base64.strict_encode64(tok_bin)
  end

  def cert_decode(crt)
    case crt.encoding
    when 'base64'
      # response data is base64 encoded
      Base64.decode64(crt.data)
    else
      # no additional encoding
      crt.data
    end
  end

  def not_found_ok
    begin
      yield if block_given?
    rescue Blockbridge::Api::NotFoundError, Excon::Errors::NotFound, Excon::Errors::Gone
      nil
    end
  end

  def delete_nil_values(h)
    h.delete_if { |k,v| v.nil? }
  end

  def usage_error(msg)
    signal_usage_error(msg)
  end

  def lookup_error(msg)
    raise Blockbridge::CLI::LookupError.new(msg, self)
  end

  def runtime_error(msg)
    raise Blockbridge::CLI::RuntimeError.new(msg, self)
  end

  def success(msg)
    raise Blockbridge::CLI::RuntimeSuccess.new(msg, self)
  end

  def internal_error(msg)
    raise Blockbridge::CLI::InternalError.new(msg, self)
  end

  def exec_script(name, *args)
    paths = [ "bin/#{name}", "/usr/libexec/blockbridge-cli/#{name}" ]
    path = paths.find { |p| File.exist?(p) }
    if path.nil?
      runtime_error "couldn't find script #{name}"
    end

    system(path, *args)
  end

  def debug_puts(*args, &blk)
    return unless debug?
    if blk.nil?
      args.each do |item|
        puts("D: #{item}")
      end
    else
      pfx = ""
      if leader = args.first
        puts("D: #{leader}")
        pfx = "  "
      end
      [blk.call].flatten.each do |item|
        item.each_line do |l|
          puts("D: #{pfx}#{l}")
        end
      end
    end
  end
end
