# Copyright (c) 2018, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'active_support/inflector'
require 'blockbridge/cli/helpers/ipsec'
require 'retriable'

module Blockbridge::CLI
  module Attach
    def ipsec_control
      @ipsec_control ||= IPsecControl.new(debug?)
    end

    def only_one_attached(otype, objects, desc)
      attached = objects.select { |o| o.xmd.schema? 'host-attach' }
      only_one(otype, attached, desc)
    end

    def attach_mode
      if read_only?
        'ro'
      elsif write_only?
        'wo'
      else
        'rw'
      end
    end

    def disk_attach_host(x)
      host = x[:data][:attach][:data][:host]
    end

    def disk_attach_mode(x)
      return x[:mode] if x[:mode]
      return x[:data][:attach][:data][:mode] if x[:data]
    rescue
      return 'unknown'
    end

    def disk_attach_mode_str(x)
      mode = disk_attach_mode(x)
      case mode
      when 'wo'
        'write-only'
      when 'rw'
        'read-write'
      when 'ro'
        'read-only'
      else
        'unknown'
      end
    end

    def disk_attach_group_create(ref:, host:, mode:, vss:, group:)
      # set xmd create params
      xmd_params = {
        ref: ref,
        xref: ref,
        vss_id: vss,
        label: host,
        tags: tags.concat(['host-attach']),
        data: {
          attach: {
            _schema:  'host-attach',
            _publish: true,
            _type:    'host-attach',
            data: {
              label: host,
              host:  host,
              mode:  mode,
            },
          },
        },
      }

      # also create group if not using existing
      unless group
        xmd_params[:data][:group] = {
          _schema: 'group',
          _publish: true,
          data: {
            label: host,
          }
        }
      end

      # create xmd for disk attach group
      xmd = api.xmd.create(xmd_params)
    end

    def disk_attach_group_remove(ref:, vdisk:)
      # update xmd rdeps to unref the removing group
      xmd = not_found_ok { api.xmd.info(ref) }
      if xmd
        xmd.rdeps.select { |dep|
          dep.ref
        }.map { |dep|
          not_found_ok { api.xmd.info(dep.ref) }
        }.compact.select { |xmd|
          xmd.xmd_refs.include?(ref)
        }.each { |xmd|
          api.xmd.update(xmd.ref, xmd_refs: xmd.xmd_refs - [ref])
        }
      end

      # remove all group by xref
      api.objects.remove_by_xref(ref, scope: 'xmd,target,initiator_profile,snapshot,vdisk')

      # update disk to remove ref
      api.vdisk.update(vdisk.id, xmd_refs: vdisk.xmd_refs - [ref]) rescue nil
    end

    def disk_check_attached(vdisk:, vss:, mode:, fail_message: nil)
      vdisk_label = vdisk.label || vdisk.serial
      vss_label   = vss.label || vss.serial

      # check disk is not already attached to a host
      attached = vdisk.xmd.fragments_for(schema: 'host-attach')
      return if attached.empty?

      failed = false
      attached.each do |x|
        if mode != 'ro' && x.mode != 'ro' && !multi_writer?
          puts "Disk #{vss_label}/#{vdisk_label} is attached to host #{x.host} for #{disk_attach_mode_str(x)} access"
          failed = true
        end
      end

      if failed
        runtime_error(fail_message || "Disk is already attached for write access. Use --repair to repair an existing connection.")
      end
    end

    def disk_check_not_mapped(vdisk:, vss:, fail_message: nil)
      vdisk_label = vdisk.label || vdisk.serial
      vss_label   = vss.label || vss.serial

      # multi-map allowed
      return if vdisk.scsi.multi_map

      # check disk is not already mapped into a target
      mapped = obj_list(:target, vss_id: vss.id).select do |t|
        t.luns.find { |l| l.vdisk_id == vdisk.id }
      end.each do |t|
        target_label = t.label || t.serial
        puts "Disk #{vss_label}/#{vdisk_label} is mapped to " \
             "target #{target_label}"
      end

      unless mapped.length == 0
        runtime_error(fail_message ||  "Disk is already mapped in a target and is not configured to allow multi-map.")
      end
    end

    def attach_data(xmd, fld)
      xmd&.dig('data','attach','data',fld)
    end

    def attachment_info(xmd, vdisk = nil, vss = nil)
      multipath = attach_data(xmd, 'multipath') == true
      targ_name = attach_data(xmd, 'targ_name')
      paths_cfg = attach_data(xmd, 'paths_num').to_i
      paths_up  = multipath ? 0 : nil
      devinfo   = host_control.iscsi.lookup_target_devinfo(targ_name)
      path_devs = devinfo.map { |info| info[:dev] }

      if host_control.hostname == attach_data(xmd, 'host')
        local = true
        if multipath
          # choose one path_dev; reports on accessibility of all
          paths = host_control.lookup_target_paths(targ_name, path_devs.first)
          wwid = paths&.first&.dig(:wwid)
          if wwid
            mpath_dev = host_control.mpath.lookup_dev(wwid)&.dig(:dev)
          end
          paths_up = paths.length
        end
      else
        local = false
      end

      unless vdisk
        uuid = attach_data(xmd, 'uuid')
        vdisk = not_found_ok { api.vdisk.info(uuid) } if uuid
      end
      unless vdisk
        vdisk = api.vdisk.list.find do |v|
          !v.xmd.fragments_for(schema: 'host-attach', ref: xmd.ref).empty?
        end
      end
      unless vss
        vss = not_found_ok { api.vss.info(vdisk.vss_id) } if vdisk
      end

      # host attachment
      {
        local:     local,
        localhost: host_control.hostname,
        localinit: host_control.iscsi.initiatorname,
        host:      attach_data(xmd, 'host'),
        ref:       xmd.ref,
        vdisk:     vdisk,
        vss:       vss,
        multipath: multipath,
        mpath_dev: mpath_dev,
        paths_cfg: paths_cfg,
        paths_up:  paths_up,
        paths:     path_devs,
        devs:      path_devs,
        dev:       mpath_dev ? mpath_dev : path_devs.first,
        mode:      disk_attach_mode_str(xmd),
        transport: attach_data(xmd, 'secure') ? attach_data(xmd, 'secure') : 'TCP/IP',
        secure:    attach_data(xmd, 'secure'),
        data:      xmd.data.attach.data,
      }
    end

    def disk_get_attached(vdisk: nil, vss: nil, ref: nil,
                          uuid: nil, path: nil, name: nil, iqn: nil,
                          fail_message: nil)
      attached = []
      dangling = []

      if ref
        attached = [ not_found_ok { api.xmd.info(ref) } ].compact.select { |xmd|
          xmd.tags.include?('host-attach')
        }
        return attached, dangling
      end

      if path
        uuid = host_control.device_scsi_serial(File.realpath(path))
      end

      if uuid
        vdisk = api.vdisk.list(uuid: uuid).first
        unless vdisk
          attached = api.xmd.list.select { |xmd|
            xmd.tags.include?('host-attach') &&
              attach_data(xmd, 'uuid') == uuid
          }
        else
          attached = vdisk.xmd_refs.map { |ref|
            xmd = not_found_ok { api.xmd.info(ref) }
            if not xmd
              dangling.push ref
            end
            xmd
          }.compact.select { |xmd|
            xmd.tags.include?('host-attach')
          }
        end

        return attached, dangling
      end

      if name
        attached = api.xmd.list.select { |xmd|
          xmd.tags.include?('host-attach') &&
            attach_data(xmd, 'name') == name
        }

        return attached, dangling
      end

      if iqn
        attached = api.xmd.list.select { |xmd|
          xmd.tags.include?('host-attach') &&
            attach_data(xmd, 'targ_name') == iqn
        }

        return attached, dangling
      end

      # check disk is attached to a host
      if vdisk
        attached = vdisk.xmd_refs.map { |ref|
          xmd = not_found_ok { api.xmd.info(ref) }
          if not xmd
            dangling.push ref
          end
          xmd
        }.compact.select { |xmd|
          xmd.tags.include?('host-attach')
        }
      end

      return attached, dangling
    end

    def device_tune_paths
      {
        'queue_depth' => 'device',
        'nomerges'    => 'queue',
        'rq_affinity' => 'queue',
        'scheduler'   => 'queue',
        'rotational'  => 'queue',
        'add_random'  => 'queue',
        'nr_requests' => 'queue',
      }
    end

    def device_tune_settings(cfg = {})
      Hash.new.tap do |h|
        device_tune_paths.each_pair do |key, path|
          h[path] ||= {}
          h[path][key] = cfg[key]
        end
      end
    end

    def device_tune_devices(attachment)
      attachment[:multipath_devs] ? attachment[:multipath_devs] : [ attachment[:dev] ]
    end

    # attempt to tune a device's sysfs parameters; in the event of an exception, retry the tuning
    # operation. the default Retriable parameters make 3 attempts over several seconds.
    def device_tune(attachment)
      devs = device_tune_devices(attachment)
      Retriable.retriable(tries: 6) do
        devs.each do |dev|
          host_control.device_tune(dev, device_tune_settings(config.host.device))
        end
      end
    end

    def disk_attach_wrap(type, **kwargs)
      res = nil
      EM.synchrony do
        begin
          case type
          when :attach
            res = disk_attach(**kwargs)
          when :repair
            res = disk_repair(**kwargs)
          else
            raise 'invalid attach type: :attach or :repair'
          end
        ensure
          EM.stop
        end
      end
      res
    end

    def disk_attach(vdisk:, vss:, snapshot: nil, ref:, mode:, hostname: nil,
                    locals: nil, remotes: nil, group: nil, af: nil, tls_cfg: nil)
      host        = host_control.hostname
      init_name   = host_control.iscsi.initiatorname
      vdisk_label = vdisk.label || vdisk.serial
      vss_label   = vss.label || vss.serial

      # clone if needed
      if clone?
        # create or ref snapshot
        if !snapshot
          params = {
            vdisk_id: vdisk.id,
            xref:     ref,
          }
          snapshot = api.snapshot.create(params)
        else
          snapshot = api.snapshot.update(snapshot.id,
            xmd_refs: (snapshot.xmd_refs + [ ref ].compact.uniq))
        end

        # create clone
        params = {
          label:       nil,
          vss_id:      vss.id,
          snapshot_id: snapshot.id,
          xref:        ref,
        }
        vdisk = api.vdisk.create(params)
      end

      # check vdisk is not already attached
      disk_check_attached(vdisk: vdisk, vss: vss, mode: mode)

      # check vdisk is not already mapped
      disk_check_not_mapped(vdisk: vdisk, vss: vss)

      # put disk into group
      vdisk = api.vdisk.update(vdisk.id,
        xmd_refs: (vdisk.xmd_refs + [ group, ref ]).compact.uniq)

      # create attach group
      params = {
        ref:   ref,
        host:  host,
        vss:   vss.id,
        group: group,
        mode:  mode,
      }
      xmd = disk_attach_group_create(params)

      # put group into existing group
      if group
        xmd = api.xmd.info(group)
        api.xmd.update(xmd.ref, xmd_refs: xmd.xmd_refs.push(ref))
      end

      # ip rules
      allow = nil
      if restrict_ip?
        addrs = host_control.local_ipaddrs do |ifa|
          if ifa.addr.ipv4? && (!locals || locals.include?(ifa))
            ifa
          else
            nil
          end
        end
        allow = addrs.map do |a|
          {
            type: 'allow',
            ip:   a,
            mask: '255.255.255.255',
          }
        end
      end

      # disk commands
      disk_commands = nil
      if write_only?
        disk_commands = 'write-only'
      elsif read_only?
        disk_commands = 'read-only'
      end

      # create target
      params = {
        vss_id:     vss.id,
        xref:       ref,
        xmd_refs:   [ group, ref ].compact,
        label:      nil,
        notes:      notes,
      }
      params[:iqn_suffix] = name if name
      target = api.target.create(params)
      targ_name = target.iqn

      # tls setup proxy
      if tls?
        # certificate validation
        # XXX
        #crt = api.ca_certificate.info(cert_format: 'pem')
        #cafile = Tempfile.new('ssl_cafile')
        #cafile.write(cert_decode(crt))
        #cafile.close

        # target path config
        path_config = {
          local_bind: local_bind?,
          multipath:  multipath?,
          locals:     locals,
          remotes:    remotes,
          af:         af,
        }
        paths = host_control.compute_target_paths(target, **path_config)

        ssl_cfg = {
          global: {}.tap do |h|
            h.merge!(tls_cfg) unless tls_cfg.nil?
          end,
          client: {}.tap do |h|
            paths.each do |path|
              h[:connect] = "#{path[:remote].addr}:#{path[:remote].port}"
            end
          end
        }

        ssl_name = "blockbridge.#{target.iqn}"
        #ssl_port = host_control.stunnel.start(ssl_name, ssl_cfg, cafile.path)
        begin
          ssl_port = host_control.stunnel.start(ssl_name, ssl_cfg)
        rescue => e
          raise unless e.message =~ /Incorrect version of SSL protocol/
          runtime_error "TLSv1.2 support not found. Please upgrade stunnel or OS version."
        end
      end

      # profile rules
      rules = nil
      if allow || disk_commands
        rules = Hash.new.tap do |h|
          h[:access] =  { ip: allow } if allow
          h[:disk] =    { commands: disk_commands} if disk_commands
        end
      end

      # create profile
      params = Hash.new.tap do |h|
        h[:vss_id] =          vss.id
        h[:xref] =            ref
        h[:xmd_refs] =        [ group, ref ].compact
        h[:label] =           nil
        h[:initiator_login] = "initiator-#{SecureRandom.uuid}"
        h[:initiator_pass] =  host_control.iscsi.gen_init_pass
        if host_control.linux?
          h[:mutual_auth] =  "required"
          h[:target_login] = "target-#{SecureRandom.uuid}"
          h[:target_pass] =  host_control.iscsi.gen_init_pass
        else
          h[:mutual_auth] = "disabled"
        end
        h[:initiator_list] = [ init_name ]
        if ipsec?
          h[:transport] = {
            mode:              'ikev2',
            encryption_policy: 'required',
            ikev2:             Hash.new.tap do |ikev2|
              ikev2[:mode] =          'tunnel'
              ikev2[:subnet] =        '172_31'
              ikev2[:auth] =          'mschapv2'
              ikev2[:mschapv2_user] = "ipsec-#{SecureRandom.uuid}"
              ikev2[:mschapv2_pass] = SecureRandom.uuid
            end
          }
        elsif tls?
          h[:transport] = {
            mode:              'tls',
            encryption_policy: 'required',
            tls:               Hash.new.tap do |tls|
              tls[:portal_ipaddr] = '127.0.0.1'
              tls[:portal_ipport] = ssl_port
            end
          }
        else
          h[:transport] = { mode: "insecure" }
        end
        if secure_token?
          h[:initiator_chap_mode] = "token"
        end
        h[:notes] = notes
        h[:rules] = rules if rules
      end
      profile = api.initiator_profile.create(params)

      # save away the credentials for later
      init_user  = profile[:iscsi].initiator_login +
                   profile[:iscsi].initiator_login_suffix
      init_pass  = params[:initiator_pass]
      targ_user  = params[:target_login]
      targ_pass  = params[:target_pass]
      ipsec_user = params[:transport][:ikev2][:mschapv2_user] +
                   profile[:iscsi].initiator_login_suffix if ipsec?
      ipsec_pass = params[:transport][:ikev2][:mschapv2_pass] if ipsec?

      # configure secure token
      if secure_token?
        passphrase = password_from_file_or_prompt(passphrase_file, "passphrase")
        init_user  = profile_configure_token(profile, passphrase)
      end

      # update target acl
      target.acl << {
        profile_id: profile.id,
        privacy:    "no",
        discover:   "yes",
        access:     "yes",
      }
      target = api.target.update(target.id, acl: target.acl)

      # map disk into target
      api.target.disk_map(target.id, vdisk.id, {})

      # choose targ host as first available IP
      if target.portals.length == 0 || target.portals[0].ipaddrs.length == 0
        runtime_error "No target portal addresses found for target"
      end

      # require hostname if ipsec
      if ipsec?
        ipsec_portals = target.portals.select do |p|
          p.interface == 'IPsec' && (hostname.nil? || hostname == p.hostname)
        end

        if ipsec_portals.empty?
          runtime_error "No target portals found configured for IPsec. Ensure initiator profile has correct transport."
        end

        ipsec_portal = ipsec_portals.find { |p| hostname.nil? || hostname == p.hostname }

        if ipsec_portal.nil?
          if hostname
            runtime_error "No target portal addresses have a hostname that match requested '#{hostname}'"
          else
            runtime_error "No target portal addresses are configured with a hostname; IPsec requires one. Please contact your administrator."
          end
        end

        ipsec_hostname = ipsec_portal.hostname
        ipsec_port = ipsec_portal.ipaddrs.first&.port

        if ipsec_port.nil?
          runtime_error "Target portal has no iSCSI port defined."
        end

        # configure ipsec
        crt = api.ca_certificate.info(cert_format: 'pem')

        ipsec_control.load_cert(cert_decode(crt))
        ipsec_control.load_cred('EAP', ipsec_user, ipsec_pass)
        #host_control.ipsec.load_cert(cert_decode(crt))
        #host_control.ipsec.load_cred('EAP', ipsec_user, ipsec_pass)

        ipsec_name = "blockbridge.#{target.iqn}"
        ipsec_conn = {
          "#{ipsec_name}" => {
            'version' => 2,
            'vips' => [ '0.0.0.0' ],
            'local_addrs'  => [ '%any' ],
            'remote_addrs' => [ ipsec_hostname ],
            'local' => {
              'auth'   => 'eap-mschapv2',
              'eap_id' => ipsec_user,
            },
            'remote' => {
              'auth' => 'pubkey',
              'id'   => ipsec_hostname,
            },
            'children' => {
              "#{ipsec_name}" => {
                'local_ts'   => [ 'dynamic' ],
                'remote_ts'  => [ "0.0.0.0/0[tcp/#{ipsec_port}]" ],
                'esp_proposals' => [ 'aes128gcm96' ],
                'dpd_action' => 'restart',
              },
            },
          },
        }
        ipsec_control.load_conn(ipsec_conn)
        ipsec_control.start_conn(ipsec_hostname, ipsec_name)
        ipsec_control.add_src_routes(ipsec_name)
        ipsec_control.print_last
      end

      # target connection config
      conn_config = {
        local_bind:   local_bind?,
        multipath:    multipath?,
        ipsec:        ipsec?,
        ssl:          tls?,
        af:           af,
        locals:       locals,
        remotes:      remotes,
        mknod:        mknod?,
        mknod_prefix: mknod_prefix,
        mknod_name:   mknod_name,
        chap: { init_user: init_user, init_pass: init_pass,
                targ_user: targ_user, targ_pass: targ_pass  },
        automatic:    persist?
      }

      attachment = host_control.target_attach(target, **conn_config)

      # settle
      host_control.device_settle

      # Tune device
      device_tune(attachment)

      # add dev to group
      params = {
        mode: 'patch',
        data: [
          { op: 'add',
            path: '/attach/data/label',
            value: host },
          { op: 'add',
            path: '/attach/data/targ_name',
            value: "#{targ_name}" },
          { op: 'add',
            path: '/attach/data/name',
            value: name },
          { op: 'add',
            path: '/attach/data/paths_num',
            value: attachment[:paths].length },
          { op: 'add',
            path: '/attach/data/multipath',
            value: attachment[:multipath] },
          { op: 'add',
            path: '/attach/data/uuid',
            value: vdisk.uuid },
          { op: 'add',
            path: '/attach/data/vss',
            value: vss.id },
          { op: 'add',
            path: '/attach/data/vdisk',
            value: vdisk.id },
          { op: 'add',
            path: '/attach/data/target',
            value: target.id },
          { op: 'add',
            path: '/attach/data/profile',
            value: profile.id },
        ],
      }

      # if attached securely; store secure name
      if ipsec?
        params[:data].push(op: 'add',
                           path: '/attach/data/secure',
                           value: 'IPsec')
        params[:data].push(op: 'add',
                           path: '/attach/data/ipsec',
                           value: ipsec_name)
      elsif tls?
        params[:data].push(op: 'add',
                           path: '/attach/data/secure',
                           value: 'TLSv1.2')
        params[:data].push(op: 'add',
                           path: '/attach/data/ssl',
                           value: ssl_name)
      end

      # if using own group then add label
      if !group
        params[:data].push(op: 'add',
                           path: '/group/data/label',
                           value: host )
      end

      # update xmd
      xmd = api.xmd.update(ref, params)
      attachment_info(xmd, vdisk, vss)
    rescue => e
      unless e.message&.include? 'already attached'
        if ipsec?
          ipsec_control.stop_conn(ipsec_name)
          ipsec_control.unload_conn(ipsec_name)
        elsif tls?
          host_control.stunnel.stop(ssl_name)
        end
      end
      raise
    end

    def disk_attempt_repair(vdisk:, vss:, xmd:)
      # run through a set of queries to determine whether a repair is needed.

      # stat each target
      targets = api.target.list(vss_id: vss.id).find_all do |t|
        t.xmd_refs.include?(xmd.ref)
      end.each do |t|
        # node config
        if !host_control.iscsi.stat_target(t.iqn)
          puts "no iscsi node found for #{t.iqn}" if debug?
          return false
        end

        # sessions
        sessions = 0
        t.sessions.each do |s|
          if s.ini_name == host_control.iscsi.initiatorname
            sessions = sessions + 1
          end
        end

        if sessions == 0
          puts "no sessions found for #{host_control.iscsi.initiatorname}" if debug?
          return false
        end

        # ensure target devs exist and are correct major/minor
        host_control.create_target_devs(t, mknod: mknod?, mknod_prefix: mknod_prefix,
                                        mknod_name: mknod_name, multipath: multipath?)

        # ensure devices are properly tuned
        device_tune(attachment_info(xmd, vdisk, vss))
      end

      true
    end

    def disk_repair(vdisk:, vss:, xmd:, locals: nil, remotes: nil, group: nil,
                    af: nil, tls_cfg: nil)
      host        = host_control.hostname
      init_name   = host_control.iscsi.initiatorname
      vdisk_label = vdisk.label || vdisk.serial
      vss_label   = vss.label || vss.serial

      repaired = disk_attempt_repair(vdisk: vdisk, vss: vss, xmd: xmd)

      if !repaired
        params = {
          vdisk: vdisk,
          vss:   vss,
          xmd:   xmd,
        }
        disk_detach(params)

        vdisk = api.vdisk.info(vdisk.id)

        begin
          # attach disk to host
          params = {
            vdisk:   vdisk,
            vss:     vss,
            ref:     xmd.ref,
            af:      af,
            locals:  locals,
            remotes: remotes,
            group:   group,
            mode:    disk_attach_mode(xmd),
            tls_cfg: tls_cfg
          }

          disk_attach(params)
        rescue
          disk_attach_group_remove(ref: xmd.ref, vdisk: vdisk)
          raise
        end
      else
        attachment_info(xmd, vdisk, vss)
      end
    end

    def disk_detach_label(vss, vdisk, ref = nil)
      label = []
      label.push ref if ref && vdisk.nil?
      label.push vss.label if vss
      label.push vdisk.label if vdisk
      label.join('/')
    end

    def disk_detach(vdisk: nil, vss: nil, xmd: nil,
                    ref: nil, uuid: nil, path: nil, name: nil)
      ref     = ref || xmd&.ref
      host    = attach_data(xmd, 'host')
      secure  = attach_data(xmd, 'secure')
      vss_id  = attach_data(xmd, 'vss')
      targets = []

      if ref
        # a valid ref indicates the backend attachment is still present

        # ensure this host or forced
        if host != host_control.hostname && !force?
          usage_error "Disk is not attached to this host (attached: #{host}) " \
                      "and --force is not specified."
        end

        # each target that is attached, logout/remove session
        targets = api.target.list(vss_id: vss_id).find_all { |t|
          t.xmd_refs.include?(ref)
        }.map { |t| t.iqn }
      else
        # no valid ref; cleanup this host

        # use uuid
        if uuid
          # find devpath from uuid (LibUdev.id_scsi_serial match, then devnode)
          path = host_control.device_from_uuid(uuid)
        end

        # use path
        if path
          # find iqn from /dev/path (lsscsi)
          host_control.iscsi.lookup_target_iqns(path).each do |iqn|
            targets.push iqn
          end
        end

        # use name
        if name
          host_control.iscsi.lookup_target_iqns.each do |iqn|
            targets.push iqn if iqn.end_with? name
          end
        end

        # use iqn
        if name
          host_control.iscsi.lookup_target_iqns.each do |iqn|
            targets.push iqn if iqn == iqn
          end
        end
      end

      # log out of targets
      targets.each do |iqn|
        host_control.iscsi.logout_target(iqn)
        host_control.iscsi.remove_target(iqn)
        puts "logged out session and removed target config for #{iqn}" if debug?

        # remove ipsec and stunnel if needed
        if secure == "IPsec"
          ipsec_name = "blockbridge.#{iqn}"
          ipsec_control.stop_conn(ipsec_name)
          ipsec_control.unload_conn(ipsec_name)
          ipsec_control.del_src_routes
        elsif secure =~ /TLS/
          ssl_name = "blockbridge.#{iqn}"
          host_control.stunnel.stop(ssl_name)
        end
      end

      # remove attach group
      if ref
        disk_attach_group_remove(ref: ref, vdisk: vdisk)

        out = "#{disk_detach_label(vss, vdisk)} detached from host #{host}"
        puts "=" * out.length
        puts out
        puts "=" * out.length
      end
    end
  end
end
