# Copyright (c) 2018, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'multi_json'

class IPsecControl
  attr_reader :ipsec_ctl
  attr_reader :debug
  attr_reader :last_output

  IPSEC_CTL = "bb-ipsecctl"
  SUDO_PATH = '/usr/bin/sudo'
  RUBYCMD   = ENV['BB_RUBY_WRAPPER'] || RbConfig.ruby

  def initialize(debug)
    if File.exist?(path = File.join('exe', IPSEC_CTL))
      @ipsec_ctl = File.expand_path(path)
    elsif File.exist?(path = File.join('', 'usr', 'lib', 'blockbridge-cli', 'bin', IPSEC_CTL))
      @ipsec_ctl = File.expand_path(path)
    else
      raise "Unable to find path to #{IPSEC_CTL} helper"
    end
    @debug = debug
    @last_output = nil
  end

  def command_defs
    [
     'load_cert',
     'load_cred',
     'load_conn',
     'start_conn',
     'get_sa',
     'add_src_routes',
     'del_src_routes',
     'stop_conn',
     'unload_conn',
    ]
  end

  def method_missing(method_sym, *args)
    unless command_defs.include? method_sym.to_s
      raise "unknown IPsec control method #{method_sym.to_s}"
    end
    command(method_sym.to_s, *args)
  end

  def command(cmd, *args)
    cmd_args = [SUDO_PATH, '-E', RUBYCMD, ipsec_ctl, '-c', cmd]
    cmd_args.push '--debug' if debug
    cmd_args.push '--stdin'
    start_time = Time.now
    out = nil
    Open3.popen2e(*cmd_args) do |i, o, wait_thr|
      i.write(args.to_json)
      i.close_write
      @last_output = out = o.read
      exit_status = wait_thr.value
      end_time = Time.now
      if debug
        parts = ["cmd=#{cmd}"]
        parts.push "opts=#{args.to_json}" unless args.compact.empty?
        parts.push "dur=#{end_time - start_time}" if start_time && end_time
        parts.push exit_status.success? ? 'success' : "failed [sts=#{exit_status.exitstatus}]"
        puts "# IPSEC CTL: #{parts.join('; ')}"
      end
      unless exit_status.success?
        raise "failed to update IPsec client configuration. command: #{cmd}, output: #{out}, exit_status: #{exit_status}"
      end
    end
  end

  def print_last
    return if last_output.nil? || last_output.empty?
    puts last_output
  end
end
