# Copyright (c) 2015-2017, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'tod'

module Blockbridge::CLI
  module Parsing
    def parse_permission(str)
      scope, right = str&.split(/[^a-z_]+/, 2)

      if scope.nil? || scope.empty? | right.nil? || right.empty?
        usage_error "Fully specify permission. For example: account.view_events"
      end

      [scope, right]
    end

    BYTE_UNITS = {
      'KiB' => 1024,
      'MiB' => 1024 ** 2,
      'GiB' => 1024 ** 3,
      'TiB' => 1024 ** 4,
      'PiB' => 1024 ** 5,
      'K'   => 1024,
      'k'   => 1000,
      'M'   => 1000 ** 2,
      'G'   => 1000 ** 3,
      'T'   => 1000 ** 4,
      'P'   => 1000 ** 5,
      'kB'  => 1000,
      'MB'  => 1000 ** 2,
      'GB'  => 1000 ** 3,
      'TB'  => 1000 ** 4,
      'PB'  => 1000 ** 5,
      ''    => 1,
    }

    def parse_bytes(str)
      if md = /^([\d.]+)\s*([A-Za-z]*)$/.match(str)
        if BYTE_UNITS.has_key? md[2]
          (md[1].to_f * BYTE_UNITS[md[2]]).to_i
        else
          signal_usage_error "Invalid unit string: '#{md[2]}'"
        end
      else
        signal_usage_error "Invalid byte value: '#{str}'"
      end
    end

    TRUTHY_BOOLS = %w(1 true yes on enable enabled)
    FALSY_BOOLS = %w(0 false no off disable disabled)
    def parse_boolean(val, default_value = false)
      case val
      when nil, '', 'default'
        default_value
      when true
        true
      when false
        false
      when String
        if TRUTHY_BOOLS.include?(val.downcase)
          true
        elsif FALSY_BOOLS.include?(val.downcase)
          false
        else
          signal_usage_error "Invalid boolean value: #{val}"
        end
      else
        signal_usage_error "Invalid boolean type: #{val.class}"
      end
    end

    def parse_bw_limit(str)
      return if ['none', 'disabled'].include?(str)
      n = parse_bytes(str.sub(/\/s$/, ''))
      n == 0 ? nil : n
    end

    def parse_time_of_day(str)
      return if str == 'none'
      Tod::TimeOfDay.parse(str).to_i * 1000
    end

    def parse_percent(str)
      if md = /^(\d+)%?/.match(str)
        md[1].to_i
      else
        usage_error "Invalid percentage value: '#{str}'"
      end
    end

    def parse_integer(str)
      Integer(str)
    end

    def parse_float(str)
      Float(str)
    end

    def parse_json(str)
      JSON.parse(str)
    end

    def parse_null(str)
      nil
    end

    def parse_string(str)
      str.to_s
    end

    def parse_file_as_array(path)
      File.readlines(path).map { |l| l.chomp }
    end

    def parse_value(str, val_type)
      val_type ||= :string

      meth_name = "parse_#{val_type.to_s}"
      if !self.respond_to? meth_name
        signal_usage_error "don't know how to parse value of type '#{val_type.to_s}'"
      end
      self.send(meth_name, str)
    end

    DISABLED_READAHEAD = %w(0 none)
    def parse_readahead(str)
      Hash.new.tap do |h|
        h[:mode] = DISABLED_READAHEAD.include?(str) ? 'none' : 'constant'
        h[:size] = parse_bytes(str) unless h[:mode] == 'none'
      end
    end
  end
end
