# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

module Blockbridge::CLI::Helpers::Replication

  def add_replication_status(disks)
    obj_cache = {}

    # put all disks & vss in the cache
    api.vdisk.list({}).each do |d|
      obj_cache[d[:id]] = d
    end
    api.vss.list({}).each do |v|
      obj_cache[v[:id]] = v
    end

    # translate each disk to its distilled status
    disks.each do |disk|
      if disk.replication.slaves[0]
        disk.replication_status = extract_replication_status(disk, obj_cache)
      end
    end

    disks
  end

  def slave_specified?
    slave || slave_vss
  end

  def only_one_slave(otype, objects, desc)
    slaves = objects.select do |o|
      o.replication&.master != nil # has a master defined
    end
    only_one(otype, slaves, desc)
  end

  def lookup_slave(opts = {})
    restrict = opts[:existing] ? :only_one_slave : :only_one
    obj_list(:disk, restrict: restrict, desc: 'slave',
             arg_map: { disk: :slave, vss: :slave_vss })
  end

  def master_specified?
    master || master_vss
  end

  def only_one_master(otype, objects, desc)
    masters = objects.select do |o|
      o.replication && o.replication.slaves.length > 0 # has a slave defined
    end
    only_one(otype, masters, desc)
  end

  def lookup_master(opts={})
    restrict = opts[:existing] ? :only_one_master : :only_one
    obj_list(:disk, restrict: restrict, desc: 'master',
             arg_map: { disk: :master, vss: :master_vss })
  end

  def one_or_more_master(otype, objects, desc)
    masters = objects.select do |o|
      o.replication && o.replication.slaves.length > 0 # has a slave defined
    end
    one_or_more(otype, masters, desc)
  end

  def lookup_masters
    obj_list(:disk, restrict: :one_or_more_master, desc: 'master',
             arg_map: { disk: :master, vss: :master_vss })
  end

  def lookup_disk(disk_id)
    begin
      api.vdisk.info(disk_id)
    rescue Blockbridge::Api::NotFoundError, Excon::Errors::NotFound, Excon::Errors::Gone
      nil
    end
  end

  # fetch and verify that a master-mode slave is configured
  def master_slave_cfg(master_disk, slave_desc = nil)
    slaves = master_disk.replication.slaves

    if slaves.length == 0
      runtime_error "Disk #{master_disk} has no slave configured."
    end
    if slaves.length > 1
      runtime_error "Disk #{master_disk} has more than 1 slave configured. OOPS."
    end

    # pick the first (and currently only) slave
    cur_slave = slaves.first

    if slave_specified?
      chk_id   = nil
      chk_desc = nil

      if scope_val_type(slave) != :label
        # if matching on oid or serial, don't look up the other object. this
        # permits verifiying slave disk relationships even if the disk has
        # been removed.
        chk_desc = chk_id = slave
      else
        # if specified, ensure the ID matches the current master_id
        chk_disk = lookup_slave
        chk_id   = chk_disk.id
        chk_desc = chk_disk
      end

      matches = case scope_val_type(chk_id)
                when :serial
                  chk_id == cur_slave.disk_serial
                when :id
                  chk_id == cur_slave.disk_id
                else
                  false
                end

      unless matches
        runtime_error "Disk #{master_disk} is configured " \
                      "with #{slave_desc} as slave, not #{chk_desc}."
      end
    end

    cur_slave
  end

  # fetch and verify that a slave-mode master is configured
  def slave_master_cfg(slave_disk, master_desc = nil)
    cur_master = slave_disk.replication.master

    if cur_master.nil?
      runtime_error "Disk #{slave_disk} has no master configured."
    end

    if master_specified?
      chk_id   = nil
      chk_desc = nil

      if scope_val_type(master) != :label
        # if matching on oid or serial, don't look up the other object. this
        # permits verifiying replication relationships even if the peer
        # object has been removed.
        chk_desc = chk_id = master
      else
        # if specified, ensure the ID matches the current configuration
        chk_disk = lookup_master
        chk_id   = chk_disk.id
        chk_desc = chk_disk
      end

      matches = case scope_val_type(chk_id)
                when :serial
                  chk_id == cur_master.disk_serial
                when :id
                  chk_id == cur_master.disk_id
                else
                  false
                end

      unless matches
        runtime_error "Disk #{slave_disk} is configured " \
                      "with #{master_desc} as master, not #{chk_desc}."
      end
    end

    cur_master
  end

  def master_slave_rts(master_disk, slave_cfg)
    return nil if master_disk.replication_rts.nil?

    slave_rts = master_disk.replication_rts.slaves.find do |e|
      e.disk_id == slave_cfg.disk_id
    end

    if slave_rts.nil?
      runtime_error "Disk #{master_disk} missing slave rts entry for " \
      "#{slave_cfg.disk_serial}"
    end

    slave_rts
  end

  def format_master_consistency(consistency_rts, master_desc)
    return unless consistency_rts

    master_id = consistency_rts.master
    _master_sn = consistency_rts.master_serial
    master_time = consistency_rts.time
    master_time = format_time(master_time) if master_time

    return 'n/a' unless master_id
    return 'not consistent' unless master_time

    "#{master_desc} as of #{master_time}"
  end

  def extract_replication_status(master_disk, obj_cache)
    master_vss = obj_cache[master_disk.vss_id]

    # extract replication status as a hash (json formattable) for the
    # given master.
    sts = Hashie::Mash.new({
                             master: {
                               id: master_disk.id,
                               serial: master_disk.serial,
                               label: master_disk.label,
                               desc: "#{master_disk.label} (#{master_disk.serial})",
                               vss: {
                                 id: master_vss.id,
                                 serial: master_vss.serial,
                                 label: master_vss.label,
                                 desc: "#{master_vss.label} (#{master_vss.serial})",
                               }
                             }
                           })

    slave_cfg = master_slave_cfg(master_disk)

    # start gathering slave info
    sts.slave = {
      id: slave_cfg.disk_id,
      serial: slave_cfg.disk_serial,
    }

    # get rts information for slave
    slave_rts = master_slave_rts(master_disk, slave_cfg)

    slave_disk = obj_cache[slave_cfg.disk_id]
    if slave_disk.nil?
      sts.slave.desc = "#{slave_cfg.disk_serial} (deleted)"
      sts.slave.deleted = true
      sts.slave.vss = nil
    else
      slave_master_rts = slave_disk.replication_rts&.master
      slave_vss = obj_cache[slave_disk.vss_id]
      sts.slave.desc = "#{slave_disk.label} (#{slave_disk.serial})"
      sts.slave.label = slave_disk.label
      sts.slave.obj_status = slave_disk.status
      sts.slave.cfg = slave_disk.replication
      sts.slave.rts = slave_disk.replication_rts
      sts.slave.consistency = format_master_consistency(slave_master_rts&.consistency,
                                                        sts.master.desc)
      sts.slave.vss = {
        id: slave_vss.id,
        serial: slave_vss.serial,
        label: slave_vss.label,
        desc: "#{slave_vss.label} (#{slave_vss.serial})",
      }
    end

    # include raw syndrome array
    sts.syndromes = slave_rts.syndromes
    if sts.syndromes.length > 0
      sts.syndromes_str = sts.syndromes.join(", ")
    end

    # include state stat, then further detail
    sts.state = slave_rts.state

    case slave_rts.state
    when 'active'
      # TODO: current stats & history

    when 'halted'
    when 'suspended'
      sts.rebuild_method = slave_rts.repair

    when 'rebuild'
      rebuild = slave_rts.rebuild

      if rebuild.nil?
        sts.rebuild_status = nil
      else
        rsts = Hashie::Mash.new

        # show details
        evaluated = format_bytes(rebuild.evaluated)
        goal = format_bytes(rebuild.goal)
        xfer = format_bytes(rebuild.transfer)
        if rebuild.transfer > 0
          xf_ratio_denom = "%.1f" % (rebuild.evaluated / rebuild.transfer)
          xf_ratio = "1:#{xf_ratio_denom}"
        else
          xf_ratio = nil
        end
        xf_str = "#{xfer} of #{evaluated}"
        xf_str += " (#{xf_ratio})" if xf_ratio

        if rebuild.bandwidth
          eff_bw = (rebuild.evaluated / rebuild.elapsed) * 1000
          eff_bw_str = "#{format_bytes(eff_bw)}/s"
          xfer_bw_str = "#{format_bytes(rebuild.bandwidth)}/s"
          bw_str = "#{xfer_bw_str} (#{eff_bw_str} effective)"
        else
          bw_str = nil
        end

        if rebuild.ettc && rebuild.ettc > 0
          ettc_str = format_duration(rebuild.ettc)
        else
          ettc_str = 'calculating...'
        end

        rsts.mode = rebuild.mode
        rsts.evaluated = evaluated
        rsts.goal = goal
        rsts.progress = rebuild.progress
        rsts.progress_str = "#{evaluated} of #{goal} (#{rebuild.progress}%)"
        rsts.transfer = xfer
        rsts.transfer_ratio = xf_ratio
        rsts.transfer_str = xf_str
        rsts.bandwidth = rebuild.bandwidth
        rsts.bandwidth_str = bw_str
        rsts.started = rebuild.start_time
        rsts.started_str = format_time(rebuild.start_time)
        rsts.elapsed = rebuild.elapsed
        rsts.elapsed_str = format_duration(rebuild.elapsed)
        rsts.completion = rebuild.ettc == 0 ? nil : rebuild.ettc
        rsts.completion_str = ettc_str

        sts.rebuild_status = rsts
      end
    end

    sts
  end

end
