# frozen_string_literal: true

# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

require 'time'
require 'naturally'
require 'securerandom'

if ENV['LOAD_PRY'] == "1"
  require 'pry-byebug'
  require 'pry-rescue'
  require 'pry-stack_explorer'
end

module Blockbridge::CLI
  ExecStatus = Struct.new(:command, :exitstatus, :exception) do
    def success?
      exitstatus == 0
    end
  end

  class Main
    class << self
      attr_accessor :last_argv
      attr_accessor :last_cmd_instance
      attr_accessor :last_exit_status
      attr_accessor :cmd_name
      attr_accessor :argv_override
      attr_accessor :api_param_map
    end

    attr_reader :cmd_instance
    attr_reader :exit_status

    EXIT_SUCCESS  = 0
    EXIT_FAILURE  = 1
    RETRY_AUTH    = :retry_auth
    DIAGS_BACKLOG = 20

    def initialize(argv, stdin = STDIN, stdout = STDOUT,
                   stderr = STDERR, kernel = Kernel)
      if ENV.key? 'HOMEPATH'
        # XXX: trim drive letter (if any) from HOMEPATH and ensure HOMEDRIVE is
        # correct.
        ENV['HOMEPATH'] = ENV['HOMEPATH'].sub(/^[A-Za-z]:/, '')
        ENV['HOMEDRIVE'] = $~[1] if $~
      end

      @stdin = stdin
      @stdout = stdout
      @stderr = stderr
      @kernel = kernel
      @argv = argv

      if (env_dir = ENV['BLOCKBRIDGE_ENV_DIR']) && Dir.exist?(env_dir)
        # inject the contents of each file into the environment before starting
        # execution.
        Dir.glob(File.join(env_dir, '*')).each do |env_file|
          begin
            var_name = 'BLOCKBRIDGE_' + File.basename(env_file).underscore.upcase
            ENV[var_name] = File.read(env_file).chomp unless ENV.key?(var_name)
          rescue StandardError => e
            @stderr.puts 'WARN: failed to read env value from'\
                         " #{env_file} (#{e.message})"
          end
        end
      end
    end

    def cmd_name
      self.class.cmd_name || File.basename($PROGRAM_NAME)
    end

    def argv
      self.class.argv_override || @argv
    end

    def execute!
      # redirect in/out/err streams
      $stderr = @stderr
      $stdin = @stdin
      $stdout = @stdout
      opts = {}

      begin
        exc      = nil
        root_obj = Blockbridge::CLI::RootCommand.new(cmd_name, opts)
        root_obj.parse(argv)

        # assume success if no exception was raised -- also return the
        # command instance object.
        @cmd_instance = root_obj.execute
        @exit_status = 0
      rescue StandardError, Interrupt => e
        # exceptions generated by clamp before 'execute' is called don't get
        # command_instance injected into them. maybe clamp's 'command' is
        # there? if not, there's nothing we can do, but it probably doesn't
        # matter since the command was never executed anyway. (and thus,
        # there probably never was an instance in the first place.)
        @cmd_instance =
          if e.respond_to?(:command_instance) && e.command_instance
            e.command_instance
          elsif e.respond_to?(:command) && e.command
            e.command
          end

        self.class.last_cmd_instance = cmd_instance

        @exit_status = handle_exception(e)

        if exit_status == RETRY_AUTH
          opts[:force_reauth] = true
          retry
        end
      ensure
        # restore in/out/err streams
        $stderr = STDERR
        $stdin  = STDIN
        $stdout = STDOUT
      end

      self.class.last_argv         = argv
      self.class.last_cmd_instance = cmd_instance
      self.class.last_exit_status  = exit_status

      @kernel.exit(exit_status || 0)

      # for interactive/script mode or for unit-testing, return the exec
      # status as an object.
      ExecStatus.new(cmd_instance, exit_status, exc)
    end

    private

    def global_opt(opt, env_key, cfg_key,
                   default: nil, values: nil, secret_values: nil)

      value, src = if opt && (idx = argv.index(opt)) && argv[idx + 1]
                     [argv[idx + 1], :cmdline]
                   elsif env_key && ENV.key?(env_key)
                     [ENV[env_key], :env]
                   elsif cfg_key && (cval = config&.get_path(cfg_key))
                     [cval, :config]
                   elsif default
                     [default, :default]
                   else
                     [nil, :none]
                   end

      return if src == :none
      return value if values.nil? && secret_values.nil?

      valid_values = [values, secret_values].compact.flatten
      return value if valid_values.include?(value)

      puts "value = #{value.inspect}"

      values_str = values.join(', ')
      case src
      when :cmdline
        @stderr.puts("#{opt} option value must be one of: "\
                     "#{values_str} (was: #{value})")
      when :env
        @stderr.puts("#{env_key} env value must be one of: "\
                     "#{values_str} (was: #{value})")
      when :config
        @stderr.puts("#{cfg_key} config value must be one of: "\
                     "#{values_str} (was: #{value})")
      when :default
        @stderr.puts("#{default} default value value must be one of: "\
                     "#{values_str} (was: #{value})")
      end

      exit 1
    end

    def role
      global_opt('--role',
                 'BLOCKBRIDGE_ROLE',
                 'global.role',
                 default: 'tenant',
                 values: %w[tenant operator])
    end

    def level
      global_opt('--level',
                 'BLOCKBRIDGE_LEVEL',
                 'global.level',
                 default: 'basic',
                 values: %w[basic advanced expert],
                 secret_values: %w[dev])
    end

    def error_format
      return cmd_instance.error_format if cmd_instance

      global_opt('--error-format',
                 'BLOCKBRIDGE_ERROR_FORMAT',
                 'global.error_format',
                 default: 'human',
                 values: %w[human machine json])
    end

    def home_dir
      return cmd_instance.home_dir if cmd_instance

      global_opt('--home',
                 'BLOCKBRIDGE_HOME',
                 nil,
                 default: default_home_dir)
    end

    def default_home_dir
      File.join(Dir.home, '.blockbridge-cli') if Dir.home
    end

    def config_file
      return cmd_instance.config_file if cmd_instance
      return if argv.index('--no-config')

      global_opt('--config',
                 'BLOCKBRIDGE_CONFIG',
                 nil,
                 default: default_config_file)
    end

    def default_config_file
      File.join(home_dir, 'config.yml') if home_dir
    end

    def config
      return @config if defined? @config

      @config = config_file && Blockbridge::CLI::Config.new(config_file)
    end

    def verbose?
      argv.include?('--verbose') || cmd_instance&.verbose?
    end

    def debug?
      argv.include?('--debug') || cmd_instance&.debug?
    end

    def diags?
      ENV['BLOCKBRIDGE_DIAGS'].to_i != 0
    end

    def auth_from_env?
      ENV.key? 'BLOCKBRIDGE_API_KEY'
    end

    def filtered_env
      {}.tap do |h|
        ENV.each_pair do |key, val|
          case key
          when /KEY|PASSWORD|TOKEN|SECRET/i
            val = '*REDACTED*'
          when 'LS_COLORS'
            next
          end
          h[key] = val
        end
      end
    end

    def ruby_path
      File.join(RbConfig::CONFIG['bindir'],
                RbConfig::CONFIG['ruby_install_name'])
    end

    def diags_path
      @diags_path ||= home_dir && File.join(home_dir, 'diags')
    end

    def prune_old_diags(ex_diags)
      return unless ex_diags.length > DIAGS_BACKLOG

      (ex_diags.length - DIAGS_BACKLOG).times do |i|
        File.unlink File.join(diags_path, ex_diags[i])
      end
    end

    def save_diags(cur_ex = nil)
      return if diags_path.nil?

      FileUtils.mkpath(diags_path)

      require 'blockbridge/util'

      ex_diags = Dir.entries(diags_path).select { |ent| ent =~ /^diags\.\d+/ }
      ex_diags = Naturally.sort ex_diags
      next_diags = if (last_diags = ex_diags.last)
                     last_diags.sub(/\.json$/, '').next + '.json'
                   else
                     'diags.0.json'
                   end

      prune_old_diags(ex_diags)

      diags_info = {}.tap do |h|
        h[:created_at]    = Time.new.iso8601
        h[:uuid]          = SecureRandom.uuid
        h[:version]       = Blockbridge::CLI.package_version rescue nil
        h[:build_time]    = ENV['BB_BUILD_TIME']
        h[:build_host]    = ENV['BB_BUILD_HOST']
        h[:api_library]   = Blockbridge::Api.version_str rescue nil
        h[:util_library]  = Blockbridge::Util.version_str rescue nil
        h[:ruby_version]  = RUBY_VERSION
        h[:ruby_platform] = RUBY_PLATFORM
        h[:ruby_path]     = ruby_path
        h[:argv]          = argv
        h[:cmd_class]     = cmd_instance.class if cmd_instance
        if cmd_instance && cmd_instance.config
          h[:config] = cmd_instance.config
          h[:param_map] = cmd_instance.api_param_map
          h[:api_url] = cmd_instance.api_url
        end
        if cur_ex
          h[:exception] = {}.tap do |xh|
            xh[:class] = cur_ex.class
            xh[:message] = cur_ex.message
            xh[:backtrace] = cur_ex.backtrace

            if cur_ex.is_a?(Blockbridge::Api::Error)
              xh[:api_detail] = {
                response: cur_ex.data,
                errors: cur_ex.errors
              }
            end
          end
        end
        h[:env] = filtered_env
      end

      File.open(File.join(diags_path, next_diags), 'w') do |fh|
        fh.write MultiJson.dump(diags_info)
      end
    rescue StandardError
      # best effort -- ignore any errors
    end

    def handle_exception(exp)
      opts = {
        cmd: cmd_instance,
        verbose: verbose?,
        debug: debug?,
        diags: diags?,
        home_dir: home_dir,
        role: role,
        level: level
      }

      # format the exception according to the configured error format
      exc_view = Blockbridge::CLI::View.new

      case error_format
      when 'machine'
        # machine output dumps the exception view to stdout
        $stdout.print exc_view.render(:exception_machine, exp, opts)
      when 'json'
        # machine output dumps the exception view to stdout
        $stdout.print exc_view.render(:exception_json, exp, opts)
      else
        # human output dumps the exception view to stderr
        $stderr.print exc_view.render(:exception_human, exp, opts)
      end

      # map to exit status
      case exp
      when Blockbridge::CLI::RuntimeSuccess, Clamp::HelpWanted
        EXIT_SUCCESS
      when Blockbridge::Api::UnauthorizedError
        auth_from_env? ? EXIT_FAILURE : RETRY_AUTH
      else
        # try to save the exception details to the CLI home directory
        save_diags(exp) unless diags?

        # optionally display the exception backtrace
        raise if diags?

        EXIT_FAILURE
      end
    end
  end
end
