# Copyright (c) 2015, Blockbridge Networks LLC.  All rights reserved.
# Use of this source code is governed by a BSD-style license, found
# in the LICENSE file.

if data.length == 0
  header "No datastores found."
  return
end

data.each do |obj|
  fieldset("Datastore: #{obj.label} (#{obj.serial})", heading: true) do
    partial(:obj_basics, obj)
    field "complex", obj[:cplx_id]

    if obj[:size] && obj[:used]
      if obj[:size] == 0
        field "size", "0b"
      else
        avail = obj[:size] - obj[:used]
        rsv = obj.reserve_size - obj.reserve_used

        field "size", obj[:size], cast: :bytes
        field "used", [obj[:used], obj[:size]], cast: :bytes_ratio
        field "available", [avail, obj[:size]], cast: :bytes_ratio
        field "reserved", [obj.reserve_size, obj[:size]], cast: :bytes_ratio

        if obj.reserve_size && obj.reserve_size > 0
          field "reserve used", [obj.reserve_used, obj.reserve_size], cast: :bytes_ratio
        end

        field "compression gain", obj[:compress_gain], cast: :bytes
        field "compression ratio", obj.dig(:metrics, :compression_ratio), cast: :float_ratio
      end
    end

    if obj.blk_encode_default
      md = /(?<cipher>.+)-(?<mode>.+)-(?<size>)/.match(obj.blk_encode_default)
      field "encryption", "#{md['cipher']}-#{md['mode']}" if md
    end
  end

  fieldset("Meta System") do
    field "metadata dev", ref_str(obj.meta.dev_id)
    field "metadata size", obj.meta_rts[:size], cast: :bytes
    field "cache size", obj.meta_rts.cache_size, cast: :bytes

    field_options verbose: true do
      field "page size", obj.meta_rts.page_size
      field "pages total", obj.meta_rts.pages
      field "pages used", obj.meta_rts.pages_used
      field "pages cached", obj.meta_rts.pages_cached
      field "pages dirty", obj.meta_rts.pages_dirty
      field "pages pinned", obj.meta_rts.pages_pinned
      field "pages compressed", obj.meta_rts.pages_compressed
    end
  end

  fieldset("Data Journal") do
    field "journal dev", ref_str(obj.data.log_dev_id)
    field "journal size", obj.data_rts.log_size, cast: :bytes
  end

  fieldset("Provisioning") do
    field "vss provisioning", obj.vss_provisioning, cast: :bool_enabled
    field "reservable", obj.size_reservable, cast: :bytes
    field "hard reserve threshold", obj.size_reserve_hard_threshold, cast: :bytes
    field "soft reserve threshold", obj.size_reserve_soft_threshold, cast: :bytes
    field "tags", obj.product.tags, cast: :cjoin, :empty_str => '-none-'
    field "use label tag", obj.label_tag ? 'yes' : 'no'
  end

  fieldset("Compression", data: obj.compression) do |t, dat|
    field "enabled", dat.enabled, cast: :bool_yn
    field "codec", dat.codec
    field "iops limit", dat.iops_limit, :nil_str => '-none-'
    field "activity monitor", dat.activity_monitor do |v|
      "#{v}m"
    end
    field "cycle detection", dat.cycle_detection do |v|
      "#{v}m"
    end
  end

  if obj.key?(:cplx_id)
    fieldset("Resource Binding") do
      field "processor complex", obj.cplx_id, nil_str: '-unbound-'
    end
  end

  fieldset("Performance") do
    reservable_iops = (obj.iops_rating * obj.iops_reserve_pct / 100).to_i

    if obj.iops_read && obj.iops_write
      iops_usage = obj.iops_read + obj.iops_write
    end

    field "cache size", obj.data_rts.cache_size, cast: :bytes
    field "quality of service", obj.iops_scheduling, cast: :bool_enabled
    field "iops rating", obj.iops_rating
    field "iops io size", obj.iops_io_size, cast: :bytes
    field("reservable", obj.iops_reserve_pct { |v| "#{v}%" })
    field "reserved", [obj.reserve_iops, reservable_iops], cast: :ratio, fmt: "%.1f"
    field "iops usage", [iops_usage, obj.iops_rating], cast: :ratio, fmt: "%.1f"
    field "sequential write direct", obj.data.sequential_write_direct, cast: :bool_enabled
    field "write merge timeout", obj.data.write_merge_timeout do |v|
      if v == 0
        'disabled'
      else
        "#{v}us"
      end
    end

    field "segment size", obj.data.segment_size
    field "journal compression", obj.data.log_compress, cast: :bool_enabled
    field "journal bypass", obj.data.log_bypass, cast: :bool_enabled
  end

  fieldset('Rebalancing', data: obj.pool.rebalance) do |t, dat|
    field "enabled", dat.enabled, cast: :bool_yn
    field "threshold", dat.threshold, cast: :percent
    field "current imbalance", [obj.imbalance, obj.used], cast: :bytes_ratio
    field "start time", dat.start_time, nil_str: '-unset-', cast: :time_of_day
    field "end time", dat.end_time, nil_str: '-unset-', cast: :time_of_day
    field "bandwidth limit", dat.bandwidth_limit, nil_str: '-none-',
          cast: :bandwidth
    field "evaluate", dat.evaluate
  end

  header "Pool Devices"
  if (entries = obj.pool.devs).length > 0
    table do |t|
      t.cols({label: 'dev', lpad: 1}, 'status', 'rank', 'weight', 'allocation', 'usable', 'used')
      entries.each do |e|
        dev = ref_obj(e.dev_id)

        if dev.usable && dev.used
          usable_str = format_bytes(dev.usable)
          used_str   = "#{format_bytes(dev.used)} (%.2f%%)" % (Float(dev.used) / dev.usable * 100)
        end

        t << [ ref_str(e.dev_id), partial(:obj_status, dev.status),
               e.rank, e.weight, e.allocation ? 'permitted' : 'not permitted',
               usable_str, used_str ]
      end
    end
  else
    empty_table " -no devices-"
  end
end
